@@ -4,7 +4,7 @@ import contextlib
 import inspect
 import itertools
 import logging
-from typing import Type
+from typing import Optional, Type
 
 from pytype.abstract import _base
 from pytype.abstract import _classes
@@ -14,12 +14,14 @@ from pytype.abstract import _singletons
 from pytype.abstract import _typing
 from pytype.abstract import abstract_utils
 from pytype.abstract import function
+from pytype.errors import error_types
+from pytype.types import types
 
 log = logging.getLogger(__name__)
 _isinstance = abstract_utils._isinstance  # pylint: disable=protected-access
 
 
-class Function(_instance_base.SimpleValue):
+class Function(_instance_base.SimpleValue, types.Function):
   """Base class for function objects (NativeFunction, InterpreterFunction).
 
   Attributes:
@@ -35,9 +37,12 @@ class Function(_instance_base.SimpleValue):
     self.is_attribute_of_class = False
     self.is_classmethod = False
     self.is_abstract = False
+    self.is_overload = False
     self.is_method = "." in name
+    self.decorators = []
     self.members["func_name"] = self.ctx.convert.build_string(
-        self.ctx.root_node, name)
+        self.ctx.root_node, name
+    )
 
   def property_get(self, callself, is_class=False):
     if self.name == "__new__" or not callself or is_class:
@@ -55,7 +60,7 @@ class Function(_instance_base.SimpleValue):
     if not f:
       # Should not happen but does in some contrived test cases.
       return None
-    for name, v in zip(f.f_code.co_freevars, f.cells):
+    for name, v in zip(f.f_code.freevars, f.cells):
       if v == var:
         return name
     return None
@@ -68,7 +73,7 @@ class Function(_instance_base.SimpleValue):
         # cellvar before it is assigned to in the outer scope.
         name = self._get_cell_variable_name(a)
         assert name is not None, "Closure variable lookup failed."
-        raise function.UndefinedParameterError(name)
+        raise error_types.UndefinedParameterError(name)
     return self._match_args_sequentially(node, args, alias_map, match_all_views)
 
   def _match_args_sequentially(self, node, args, alias_map, match_all_views):
@@ -93,12 +98,24 @@ class Function(_instance_base.SimpleValue):
       # Case 2: Data are entirely Tuple Instances, Unknown or Unsolvable. Make
       # all parameters except self/cls optional.
       # Case 3: Data is anything else. Same as Case 2, but emit a warning.
-      if not (all(isinstance(d, (
-          _instance_base.Instance, _singletons.Unknown, _singletons.Unsolvable))
-                  for d in defaults_var.data) and
-              all(d.full_name == "builtins.tuple"
-                  for d in defaults_var.data
-                  if isinstance(d, _instance_base.Instance))):
+      if not (
+          all(
+              isinstance(
+                  d,
+                  (
+                      _instance_base.Instance,
+                      _singletons.Unknown,
+                      _singletons.Unsolvable,
+                  ),
+              )
+              for d in defaults_var.data
+          )
+          and all(
+              d.full_name == "builtins.tuple"
+              for d in defaults_var.data
+              if isinstance(d, _instance_base.Instance)
+          )
+      ):
         self.ctx.errorlog.bad_function_defaults(self.ctx.vm.frames, self.name)
       # The ambiguous case is handled by the subclass.
       return None
@@ -106,6 +123,9 @@ class Function(_instance_base.SimpleValue):
   def set_function_defaults(self, node, defaults_var):
     raise NotImplementedError(self.__class__.__name__)
 
+  def update_signature_scope(self, cls):
+    return
+
 
 class NativeFunction(Function):
   """An abstract value representing a native function.
@@ -122,16 +142,17 @@ class NativeFunction(Function):
     self.bound_class = lambda callself, underlying: self
 
   def argcount(self, _):
-    return self.func.func_code.co_argcount
+    return self.func.func_code.argcount
 
-  def call(self, node, _, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     sig = None
     if isinstance(self.func.__self__, _classes.CallableClass):
       sig = function.Signature.from_callable(self.func.__self__)
     args = args.simplify(node, self.ctx, match_signature=sig)
     posargs = [u.AssignToNewVariable(node) for u in args.posargs]
-    namedargs = {k: u.AssignToNewVariable(node)
-                 for k, u in args.namedargs.items()}
+    namedargs = {
+        k: u.AssignToNewVariable(node) for k, u in args.namedargs.items()
+    }
     try:
       inspect.signature(self.func).bind(node, *posargs, **namedargs)
     except ValueError as e:
@@ -158,14 +179,14 @@ class NativeFunction(Function):
       if inspect.ismethod(func) and func.__self__ is not None:
         expected_argcount -= 1
       actual_argcount = len(posargs) + len(namedargs)
-      if (actual_argcount > expected_argcount or
-          (not args.starargs and not args.starstarargs)):
+      if actual_argcount > expected_argcount or (
+          not args.starargs and not args.starstarargs
+      ):
         # If we have too many arguments, or starargs and starstarargs are both
         # empty, then we can be certain of a WrongArgCount error.
         argnames = tuple("_" + str(i) for i in range(expected_argcount))
-        sig = function.Signature(
-            self.name, argnames, 0, None, set(), None, {}, {}, {})
-        raise function.WrongArgCount(sig, args, self.ctx)
+        sig = function.Signature.from_param_names(self.name, argnames)
+        raise error_types.WrongArgCount(sig, args, self.ctx)
       assert actual_argcount < expected_argcount
       # Assume that starargs or starstarargs fills in the missing arguments.
       # Instead of guessing where these arguments should go, overwrite all of
@@ -177,15 +198,15 @@ class NativeFunction(Function):
       namedargs = {}
     if "self" in namedargs:
       argnames = tuple(
-          "_" + str(i) for i in range(len(posargs) + len(namedargs)))
-      sig = function.Signature(
-          self.name, argnames, 0, None, set(), None, {}, {}, {})
-      raise function.DuplicateKeyword(sig, args, self.ctx, "self")
+          "_" + str(i) for i in range(len(posargs) + len(namedargs))
+      )
+      sig = function.Signature.from_param_names(self.name, argnames)
+      raise error_types.DuplicateKeyword(sig, args, self.ctx, "self")
     return self.func(node, *posargs, **namedargs)
 
   def get_positional_names(self):
     code = self.func.func_code
-    return list(code.co_varnames[:code.co_argcount])
+    return list(code.varnames[: code.argcount])
 
   def property_get(self, callself, is_class=False):
     return self
@@ -204,11 +225,12 @@ class BoundFunction(_base.BaseValue):
 
     # If the function belongs to `ParameterizedClass`, we will annotate the
     # `self` when do argument matching
-    self.replace_self_annot = None
+    self._self_annot = None
     inst = abstract_utils.get_atomic_value(
-        self._callself, default=self.ctx.convert.unsolvable)
-    if self.underlying.should_replace_self_annot():
-      self.replace_self_annot = abstract_utils.get_generic_type(inst)
+        self._callself, default=self.ctx.convert.unsolvable
+    )
+    if self.underlying.should_set_self_annot():
+      self._self_annot = self._get_self_annot(inst)
     if isinstance(inst, _instance_base.SimpleValue):
       self.alias_map = inst.instance_type_parameters.aliases
     elif isinstance(inst, _typing.TypeParameterInstance):
@@ -216,6 +238,20 @@ class BoundFunction(_base.BaseValue):
     else:
       self.alias_map = None
 
+  def _get_self_annot(self, callself):
+    if isinstance(self.underlying, SignedFunction):
+      self_type = self.underlying.get_self_type_param()
+    else:
+      self_type = None
+    if not self_type:
+      return abstract_utils.get_generic_type(callself)
+    if "classmethod" in self.underlying.decorators:
+      return _classes.ParameterizedClass(
+          self.ctx.convert.type_type, {abstract_utils.T: self_type}, self.ctx
+      )
+    else:
+      return self_type
+
   def argcount(self, node):
     return self.underlying.argcount(node) - 1  # account for self
 
@@ -235,14 +271,24 @@ class BoundFunction(_base.BaseValue):
     if self.argcount(node) >= 0:
       args = args.replace(posargs=(self._callself,) + args.posargs)
     try:
-      if self.replace_self_annot:
-        with self.underlying.set_self_annot(self.replace_self_annot):
-          node, ret = self.underlying.call(node, func, args,
-                                           alias_map=self.alias_map)
+      if self._self_annot:
+        should_set_self_annot = True
       else:
-        node, ret = self.underlying.call(node, func, args,
-                                         alias_map=self.alias_map)
-    except function.InvalidParameters as e:
+        # If a function is recursively calling itself and has set the `self`
+        # annotation for the previous call, we want to clear it for this one.
+        should_set_self_annot = (
+            isinstance(self.underlying, SignedFunction)
+            and self.underlying.has_self_annot
+        )
+      if should_set_self_annot:
+        context = self.underlying.set_self_annot(self._self_annot)
+      else:
+        context = contextlib.nullcontext()
+      with context:
+        node, ret = self.underlying.call(
+            node, func, args, alias_map=self.alias_map
+        )
+    except error_types.InvalidParameters as e:
       if self._callself and self._callself.bindings:
         if "." in e.name:
           # match_args will try to prepend the parent's name to the error name.
@@ -306,6 +352,8 @@ class BoundFunction(_base.BaseValue):
   def get_special_attribute(self, node, name, valself):
     if name == "__self__":
       return self.callself
+    elif name == "__func__":
+      return self.underlying.to_variable(node)
     return super().get_special_attribute(node, name, valself)
 
 
@@ -318,7 +366,7 @@ class BoundInterpreterFunction(BoundFunction):
       yield
 
   def get_first_opcode(self):
-    return self.underlying.code.first_opcode
+    return self.underlying.code.get_first_opcode(skip_noop=True)
 
   @property
   def has_overloads(self):
@@ -340,6 +388,9 @@ class BoundInterpreterFunction(BoundFunction):
     for f in self.underlying.iter_signature_functions():
       yield self.underlying.bound_class(self._callself, f)
 
+  def reset_overloads(self):
+    return self.underlying.reset_overloads()
+
 
 class BoundPyTDFunction(BoundFunction):
   pass
@@ -359,7 +410,8 @@ class ClassMethod(_base.BaseValue):
 
   def call(self, node, func, args, alias_map=None):
     return self.method.call(
-        node, func, args.replace(posargs=(self._callcls,) + args.posargs))
+        node, func, args.replace(posargs=(self._callcls,) + args.posargs)
+    )
 
   def to_bound_function(self):
     return BoundPyTDFunction(self._callcls, self.method)
@@ -405,36 +457,54 @@ class SignedFunction(Function):
   """
 
   def __init__(self, signature, ctx):
+    # We should only instantiate subclasses of SignedFunction
+    assert self.__class__ != SignedFunction
     super().__init__(signature.name, ctx)
     self.signature = signature
     # Track whether we've annotated `self` with `set_self_annot`, since
     # annotating `self` in `__init__` is otherwise illegal.
     self._has_self_annot = False
 
+  @property
+  def has_self_annot(self):
+    return self._has_self_annot
+
   @contextlib.contextmanager
-  def set_self_annot(self, annot_class):
+  def set_self_annot(self, annot_class: Optional[_base.BaseValue]):
     """Set the annotation for `self` in a class."""
     self_name = self.signature.param_names[0]
     old_self = self.signature.annotations.get(self_name)
     old_has_self_annot = self._has_self_annot
-    self.signature.annotations[self_name] = annot_class
-    self._has_self_annot = True
+    if annot_class:
+      self.signature.annotations[self_name] = annot_class
+    elif old_self:
+      del self.signature.annotations[self_name]
+    self._has_self_annot = bool(annot_class)
     try:
       yield
     finally:
       if old_self:
         self.signature.annotations[self_name] = old_self
-      else:
+      elif annot_class:
         del self.signature.annotations[self_name]
       self._has_self_annot = old_has_self_annot
 
+  def get_self_type_param(self):
+    for annot in self.signature.annotations.values():
+      for param in self.ctx.annotation_utils.get_type_parameters(annot):
+        if param.full_name == "typing.Self":
+          return param
+    return None
+
   def argcount(self, _):
     return len(self.signature.param_names)
 
   def get_nondefault_params(self):
-    return ((n, n in self.signature.kwonly_params)
-            for n in self.signature.param_names
-            if n not in self.signature.defaults)
+    return (
+        (n, n in self.signature.kwonly_params)
+        for n in self.signature.param_names
+        if n not in self.signature.defaults
+    )
 
   def match_and_map_args(self, node, args, alias_map):
     """Calls match_args() and _map_args()."""
@@ -457,10 +527,8 @@ class SignedFunction(Function):
       function.FailedFunctionCall: If the caller supplied incorrect arguments.
     """
     # Originate a new variable for each argument and call.
-    posargs = [u.AssignToNewVariable(node)
-               for u in args.posargs]
-    kws = {k: u.AssignToNewVariable(node)
-           for k, u in args.namedargs.items()}
+    posargs = [u.AssignToNewVariable(node) for u in args.posargs]
+    kws = {k: u.AssignToNewVariable(node) for k, u in args.namedargs.items()}
     sig = self.signature
     callargs = {
         name: self.ctx.program.NewVariable(default.data, [], node)
@@ -470,35 +538,34 @@ class SignedFunction(Function):
     posonly_names = set(sig.posonly_params)
     for key in set(positional) - posonly_names:
       if key in kws:
-        raise function.DuplicateKeyword(sig, args, self.ctx, key)
+        raise error_types.DuplicateKeyword(sig, args, self.ctx, key)
     kwnames = set(kws)
     extra_kws = kwnames.difference(sig.param_names + sig.kwonly_params)
     if extra_kws and not sig.kwargs_name:
       if function.has_visible_namedarg(node, args, extra_kws):
-        raise function.WrongKeywordArgs(sig, args, self.ctx, extra_kws)
+        raise error_types.WrongKeywordArgs(sig, args, self.ctx, extra_kws)
     posonly_kws = kwnames & posonly_names
     # If a function has a **kwargs parameter, then keyword arguments with the
     # same name as a positional-only argument are allowed, e.g.:
     #   def f(x, /, **kwargs): ...
     #   f(0, x=1)  # ok
     if posonly_kws and not sig.kwargs_name:
-      raise function.WrongKeywordArgs(sig, args, self.ctx, posonly_kws)
+      raise error_types.WrongKeywordArgs(sig, args, self.ctx, posonly_kws)
     callargs.update(positional)
     callargs.update(kws)
-    for key, kwonly in self.get_nondefault_params():
+    for key, kwonly in itertools.chain(
+        self.get_nondefault_params(), ((key, True) for key in sig.kwonly_params)
+    ):
       if key not in callargs:
         if args.starstarargs or (args.starargs and not kwonly):
           # We assume that because we have *args or **kwargs, we can use these
           # to fill in any parameters we might be missing.
           callargs[key] = self.ctx.new_unsolvable(node)
         else:
-          raise function.MissingParameter(sig, args, self.ctx, key)
-    for key in sig.kwonly_params:
-      if key not in callargs:
-        raise function.MissingParameter(sig, args, self.ctx, key)
+          raise error_types.MissingParameter(sig, args, self.ctx, key)
     if sig.varargs_name:
       varargs_name = sig.varargs_name
-      extraneous = posargs[self.argcount(node):]
+      extraneous = posargs[self.argcount(node) :]
       if args.starargs:
         if extraneous:
           log.warning("Not adding extra params to *%s", varargs_name)
@@ -506,7 +573,7 @@ class SignedFunction(Function):
       else:
         callargs[varargs_name] = self.ctx.convert.build_tuple(node, extraneous)
     elif len(posargs) > self.argcount(node):
-      raise function.WrongArgCount(sig, args, self.ctx)
+      raise error_types.WrongArgCount(sig, args, self.ctx)
     if sig.kwargs_name:
       kwargs_name = sig.kwargs_name
       # Build a **kwargs dictionary out of the extraneous parameters
@@ -557,13 +624,16 @@ class SignedFunction(Function):
         # The annotation is Tuple or Dict, but the passed arg only has to be
         # Iterable or Mapping.
         formal = self.ctx.convert.widen_type(formal)
-      args_to_match.append(function.Arg(name, arg, formal))
+      args_to_match.append(types.Arg(name, arg, formal))
     matcher = self.ctx.matcher(node)
     try:
       matches = matcher.compute_matches(
-          args_to_match, match_all_views, alias_map=alias_map)
-    except matcher.MatchError as e:
-      raise function.WrongArgTypes(self.signature, args, self.ctx, e.bad_type)
+          args_to_match, match_all_views, alias_map=alias_map
+      )
+    except error_types.MatchError as e:
+      raise error_types.WrongArgTypes(
+          self.signature, args, self.ctx, e.bad_type
+      )
     return [m.subst for m in matches]
 
   def get_first_opcode(self):
@@ -578,27 +648,31 @@ class SignedFunction(Function):
 
     Args:
       node: The node where default arguments are being set. Needed if we cannot
-            get a useful value from defaults_var.
+        get a useful value from defaults_var.
       defaults_var: a Variable with a single binding to a tuple of default
-                    values.
+        values.
     """
     defaults = self._extract_defaults(defaults_var)
     if defaults is None:
       defaults = [
           self.ctx.new_unsolvable(node) for _ in self.signature.param_names
       ]
-    defaults = dict(zip(self.signature.param_names[-len(defaults):], defaults))
+    defaults = dict(zip(self.signature.param_names[-len(defaults) :], defaults))
     self.signature.defaults = defaults
 
   def _mutations_generator(self, node, first_arg, substs):
     def generator():
       """Yields mutations."""
-      if (not (self.is_attribute_of_class or self.name == "__new__") or
-          not first_arg or not substs):
+      if (
+          not (self.is_attribute_of_class or self.name == "__new__")
+          or not first_arg
+          or not substs
+      ):
         return
       try:
         inst = abstract_utils.get_atomic_value(
-            first_arg, _instance_base.Instance)
+            first_arg, _instance_base.Instance
+        )
       except abstract_utils.ConversionError:
         return
       if inst.cls.template:
@@ -619,10 +693,15 @@ class SignedFunction(Function):
               else:
                 value.PasteVariable(v, node)
               yield function.Mutation(inst, k, value)
+
     # Optimization: return a generator to avoid iterating over the mutations an
     # extra time.
     return generator
 
+  def update_signature_scope(self, cls):
+    self.signature.excluded_types.update([t.name for t in cls.template])
+    self.signature.add_scope(cls)
+
 
 class SimpleFunction(SignedFunction):
   """An abstract value representing a function with a particular signature.
@@ -636,8 +715,18 @@ class SimpleFunction(SignedFunction):
     self.bound_class = BoundFunction
 
   @classmethod
-  def build(cls, name, param_names, posonly_count, varargs_name, kwonly_params,
-            kwargs_name, defaults, annotations, ctx):
+  def build(
+      cls,
+      name,
+      param_names,
+      posonly_count,
+      varargs_name,
+      kwonly_params,
+      kwargs_name,
+      defaults,
+      annotations,
+      ctx,
+  ):
     """Returns a SimpleFunction.
 
     Args:
@@ -654,24 +743,58 @@ class SimpleFunction(SignedFunction):
     """
     annotations = dict(annotations)
     # Every parameter must have an annotation. Defaults to unsolvable.
-    for n in itertools.chain(param_names, [varargs_name, kwargs_name],
-                             kwonly_params):
+    for n in itertools.chain(
+        param_names, [varargs_name, kwargs_name], kwonly_params
+    ):
       if n and n not in annotations:
         annotations[n] = ctx.convert.unsolvable
     if not isinstance(defaults, dict):
-      defaults = dict(zip(param_names[-len(defaults):], defaults))
-    signature = function.Signature(name, param_names, posonly_count,
-                                   varargs_name, kwonly_params, kwargs_name,
-                                   defaults, annotations)
+      defaults = dict(zip(param_names[-len(defaults) :], defaults))
+    signature = function.Signature(
+        name,
+        param_names,
+        posonly_count,
+        varargs_name,
+        kwonly_params,
+        kwargs_name,
+        defaults,
+        annotations,
+    )
     return cls(signature, ctx)
 
-  def call(self, node, _, args, alias_map=None):
+  def _skip_parameter_matching(self):
+    """Check whether we should skip parameter matching.
+
+    This is use to skip parameter matching for function calls in the context of
+    inference (pyi generation). This is to optimize the case where we don't
+    need to match parameters in cases which the function has explicit type
+    annotations, meaning that we don't need to infer the type.
+
+    Returns:
+      True if we should skip parameter matching.
+    """
+    # We can skip parameter matching if we don't have type any parameters. If we
+    # do, we can't skip it because we need to substitute the type parameters in
+    # the signature's annotations.
+    if self.signature.type_params:
+      return False
+    # We only skip in Infer.
+    if self.ctx.options.analyze_annotated:
+      return False
+
+    return self.signature.has_return_annotation or self.full_name == "__init__"
+
+  def call(self, node, func, args, alias_map=None):
     args = args.simplify(node, self.ctx)
     callargs = self._map_args(node, args)
-    substs = self.match_args(node, args, alias_map)
-    # Substitute type parameters in the signature's annotations.
-    annotations = self.ctx.annotation_utils.sub_annotations(
-        node, self.signature.annotations, substs, instantiate_unbound=False)
+    substs = []
+    annotations = self.signature.annotations
+    if not self._skip_parameter_matching():
+      substs = self.match_args(node, args, alias_map)
+      # Substitute type parameters in the signature's annotations.
+      annotations = self.ctx.annotation_utils.sub_annotations(
+          node, self.signature.annotations, substs, instantiate_unbound=False
+      )
     if self.signature.has_return_annotation:
       ret_type = annotations["return"]
       ret = ret_type.instantiate(node)
@@ -681,6 +804,7 @@ class SimpleFunction(SignedFunction):
       self_arg = ret
     else:
       self_arg = self.signature.get_self_arg(callargs)
-    mutations = self._mutations_generator(node, self_arg, substs)
-    node = abstract_utils.apply_mutations(node, mutations)
+    if not self._skip_parameter_matching():
+      mutations = self._mutations_generator(node, self_arg, substs)
+      node = abstract_utils.apply_mutations(node, mutations)
     return node, ret
