@@ -9,9 +9,11 @@ from pytype.abstract import _instance_base
 from pytype.abstract import abstract_utils
 from pytype.abstract import function
 from pytype.abstract import mixin
+from pytype.errors import error_types
 from pytype.pytd import pytd
 from pytype.typegraph import cfg
 from pytype.typegraph import cfg_utils
+from pytype.types import types
 
 log = logging.getLogger(__name__)
 _make = abstract_utils._make  # pylint: disable=protected-access
@@ -30,13 +32,18 @@ def _get_concrete_sequence_fullhash(seq, seen):
   elif id(seq) in seen:
     return seq.get_default_fullhash()
   seen.add(id(seq))
-  return hash((type(seq),) +
-              tuple(abstract_utils.get_var_fullhash_component(var, seen)
-                    for var in seq.pyval))
+  return hash(
+      (type(seq),)
+      + tuple(
+          abstract_utils.get_var_fullhash_component(var, seen)
+          for var in seq.pyval
+      )
+  )
 
 
 class LazyConcreteDict(
-    _instance_base.SimpleValue, mixin.PythonConstant, mixin.LazyMembers):
+    _instance_base.SimpleValue, mixin.PythonConstant, mixin.LazyMembers
+):
   """Dictionary with lazy values."""
 
   def __init__(self, name, member_map, ctx):
@@ -62,7 +69,7 @@ class ConcreteValue(_instance_base.Instance, mixin.PythonConstant):
     return hash((type(self), id(self.pyval)))
 
 
-class Module(_instance_base.Instance, mixin.LazyMembers):
+class Module(_instance_base.Instance, mixin.LazyMembers, types.Module):
   """Represents an (imported) module."""
 
   def __init__(self, ctx, name, member_map, ast):
@@ -73,6 +80,13 @@ class Module(_instance_base.Instance, mixin.LazyMembers):
 
   def _convert_member(self, name, member, subst=None):
     """Called to convert the items in _member_map to cfg.Variable."""
+    if isinstance(member, pytd.Alias) and isinstance(member.type, pytd.Module):
+      module = self.ctx.vm.import_module(
+          member.type.module_name, member.type.module_name, 0
+      )
+      if not module:
+        raise abstract_utils.ModuleLoadError()
+      return module.to_variable(self.ctx.root_node)
     var = self.ctx.convert.constant_to_var(member)
     for value in var.data:
       # Only do this if this is a class which isn't already part of a module, or
@@ -89,7 +103,7 @@ class Module(_instance_base.Instance, mixin.LazyMembers):
 
   @module.setter
   def module(self, m):
-    assert (m is None or m == self.ast.name), (m, self.ast.name)
+    assert m is None or m == self.ast.name, (m, self.ast.name)
 
   @property
   def full_name(self):
@@ -113,8 +127,9 @@ class Module(_instance_base.Instance, mixin.LazyMembers):
         if len(f.signatures) != 1:
           log.warning("overloaded module-level __getattr__ (in %s)", self.name)
         elif f.signatures[0].return_type != pytd.AnythingType():
-          log.warning("module-level __getattr__ doesn't return Any (in %s)",
-                      self.name)
+          log.warning(
+              "module-level __getattr__ doesn't return Any (in %s)", self.name
+          )
         return True
       else:
         log.warning("__getattr__ in %s is not a function", self.name)
@@ -122,8 +137,9 @@ class Module(_instance_base.Instance, mixin.LazyMembers):
 
   def get_submodule(self, node, name):
     full_name = self.name + "." + name
-    mod = self.ctx.vm.import_module(full_name, full_name,
-                                    0)  # 0: absolute import
+    mod = self.ctx.vm.import_module(
+        full_name, full_name, 0
+    )  # 0: absolute import
     if mod is not None:
       return mod.to_variable(node)
     elif self.has_getattr():
@@ -147,24 +163,15 @@ class Coroutine(_instance_base.Instance):
 
   def __init__(self, ctx, ret_var, node):
     super().__init__(ctx.convert.coroutine_type, ctx)
-    self.merge_instance_type_parameter(node, abstract_utils.T,
-                                       self.ctx.new_unsolvable(node))
-    self.merge_instance_type_parameter(node, abstract_utils.T2,
-                                       self.ctx.new_unsolvable(node))
     self.merge_instance_type_parameter(
-        node, abstract_utils.V, ret_var.AssignToNewVariable(node))
-
-  @classmethod
-  def make(cls, ctx, func, node):
-    """Get return type of coroutine function."""
-    assert func.signature.has_return_annotation
-    ret_val = func.signature.annotations["return"]
-    if func.code.has_coroutine():
-      ret_var = ret_val.instantiate(node)
-    elif func.code.has_iterable_coroutine():
-      ret_var = ret_val.get_formal_type_parameter(
-          abstract_utils.V).instantiate(node)
-    return cls(ctx, ret_var, node)
+        node, abstract_utils.T, self.ctx.new_unsolvable(node)
+    )
+    self.merge_instance_type_parameter(
+        node, abstract_utils.T2, self.ctx.new_unsolvable(node)
+    )
+    self.merge_instance_type_parameter(
+        node, abstract_utils.V, ret_var.AssignToNewVariable(node)
+    )
 
 
 class Iterator(_instance_base.Instance, mixin.HasSlots):
@@ -200,21 +207,25 @@ class BaseGenerator(_instance_base.Instance):
         if self.is_return_allowed:
           type_params.append(abstract_utils.V)
         for param_name in type_params:
-          _, param_var = self.ctx.vm.init_class(
-              node, ret_type.get_formal_type_parameter(param_name))
+          param_var = self.ctx.vm.init_class(
+              node, ret_type.get_formal_type_parameter(param_name)
+          )
           self.merge_instance_type_parameter(node, param_name, param_var)
       else:
         # infer the type parameters based on the collected type information.
         self.merge_instance_type_parameter(
-            node, abstract_utils.T, self.frame.yield_variable)
+            node, abstract_utils.T, self.frame.yield_variable
+        )
         # For T2 type, it can not be decided until the send/asend function is
         # called later on. So set T2 type as ANY so that the type check will
         # not fail when the function is called afterwards.
-        self.merge_instance_type_parameter(node, abstract_utils.T2,
-                                           self.ctx.new_unsolvable(node))
+        self.merge_instance_type_parameter(
+            node, abstract_utils.T2, self.ctx.new_unsolvable(node)
+        )
         if self.is_return_allowed:
           self.merge_instance_type_parameter(
-              node, abstract_utils.V, self.frame.return_variable)
+              node, abstract_utils.V, self.frame.return_variable
+          )
       self.runs += 1
     return node, self.get_instance_type_parameter(abstract_utils.T)
 
@@ -228,8 +239,9 @@ class AsyncGenerator(BaseGenerator):
   """A representation of instances of async generators."""
 
   def __init__(self, async_generator_frame, ctx):
-    super().__init__(ctx.convert.async_generator_type, async_generator_frame,
-                     ctx, False)
+    super().__init__(
+        ctx.convert.async_generator_type, async_generator_frame, ctx, False
+    )
 
 
 class Generator(BaseGenerator):
@@ -263,8 +275,8 @@ class Tuple(_instance_base.Instance, mixin.PythonConstant):
     combined_content = ctx.convert.build_content(content)
     class_params = {
         name: ctx.convert.merge_classes(instance_param.data)
-        for name, instance_param in tuple(enumerate(content)) +
-        ((abstract_utils.T, combined_content),)
+        for name, instance_param in tuple(enumerate(content))
+        + ((abstract_utils.T, combined_content),)
     }
     cls = _make("TupleClass", ctx.convert.tuple_type, class_params, ctx)
     super().__init__(cls, ctx)
@@ -276,8 +288,9 @@ class Tuple(_instance_base.Instance, mixin.PythonConstant):
     self.is_unpacked_function_args = False
 
   def str_of_constant(self, printer):
-    content = ", ".join(" or ".join(_var_map(printer, val))
-                        for val in self.pyval)
+    content = ", ".join(
+        " or ".join(_var_map(printer, val)) for val in self.pyval
+    )
     if self.tuple_length == 1:
       content += ","
     return f"({content})"
@@ -300,8 +313,9 @@ class Tuple(_instance_base.Instance, mixin.PythonConstant):
     if self._is_recursive() or other._is_recursive():
       return self._hash == other._hash
     # Otherwise do an elementwise comparison.
-    return all(e.data == other_e.data
-               for e, other_e in zip(self.pyval, other.pyval))
+    return all(
+        e.data == other_e.data for e, other_e in zip(self.pyval, other.pyval)
+    )
 
   def __hash__(self):
     if self._hash is None:
@@ -309,8 +323,9 @@ class Tuple(_instance_base.Instance, mixin.PythonConstant):
       # tuple containing itself, so we approximate the inner values with their
       # full names.
       approximate_hash = lambda var: tuple(v.full_name for v in var.data)
-      self._hash = hash((self.tuple_length,) +
-                        tuple(approximate_hash(e) for e in self.pyval))
+      self._hash = hash(
+          (self.tuple_length,) + tuple(approximate_hash(e) for e in self.pyval)
+      )
     return self._hash
 
   def get_fullhash(self, seen=None):
@@ -331,8 +346,9 @@ class List(_instance_base.Instance, mixin.HasSlots, mixin.PythonConstant):  # py
     self.set_native_slot("__getslice__", self.getslice_slot)
 
   def str_of_constant(self, printer):
-    return "[%s]" % ", ".join(" or ".join(_var_map(printer, val))
-                              for val in self.pyval)
+    return "[%s]" % ", ".join(
+        " or ".join(_var_map(printer, val)) for val in self.pyval
+    )
 
   def __repr__(self):
     if self.is_concrete:
@@ -385,8 +401,7 @@ class List(_instance_base.Instance, mixin.HasSlots, mixin.PythonConstant):  # py
     If data is an Instance of int, None is returned.
 
     Args:
-      data: The object to extract from. Usually a ConcreteValue or an
-        Instance.
+      data: The object to extract from. Usually a ConcreteValue or an Instance.
 
     Returns:
       The value (an int or None) of the index.
@@ -421,8 +436,9 @@ class List(_instance_base.Instance, mixin.HasSlots, mixin.PythonConstant):  # py
     results = []
     unresolved = False
     if self.is_concrete:
-      for start_val, end_val in cfg_utils.variable_product([start_var,
-                                                            end_var]):
+      for start_val, end_val in cfg_utils.variable_product(
+          [start_var, end_var]
+      ):
         try:
           start = self._get_index(start_val.data)
           end = self._get_index(end_val.data)
@@ -430,7 +446,8 @@ class List(_instance_base.Instance, mixin.HasSlots, mixin.PythonConstant):  # py
           unresolved = True
         else:
           results.append(
-              List(self.pyval[start:end], self.ctx).to_variable(node))
+              List(self.pyval[start:end], self.ctx).to_variable(node)
+          )
     if unresolved or not self.is_concrete:
       results.append(ret)
     return node, self.ctx.join_variables(node, results)
@@ -460,8 +477,10 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
     # self.pyval is only populated for string keys.
     if not self.is_concrete:
       return "{...: ...}"
-    pairs = [f"{name!r}: {' or '.join(_var_map(printer, value))}"
-             for name, value in self.pyval.items()]
+    pairs = [
+        f"{name!r}: {' or '.join(_var_map(printer, value))}"
+        for name, value in self.pyval.items()
+    ]
     return "{" + ", ".join(pairs) + "}"
 
   def __repr__(self):
@@ -480,8 +499,10 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
     elif id(self) in seen:
       return self.get_default_fullhash()
     seen.add(id(self))
-    return hash((type(self),) + abstract_utils.get_dict_fullhash_component(
-        self.pyval, seen=seen))
+    return hash(
+        (type(self),)
+        + abstract_utils.get_dict_fullhash_component(self.pyval, seen=seen)
+    )
 
   def getitem_slot(self, node, name_var):
     """Implements the __getitem__ slot."""
@@ -498,7 +519,7 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
             results.append(self.pyval[name])
           except KeyError as e:
             unresolved = True
-            raise function.DictKeyMissing(name) from e
+            raise error_types.DictKeyMissing(name) from e
     node, ret = self.call_pytd(node, "__getitem__", name_var)
     if unresolved or not self.is_concrete:
       # We *do* know the overall type of the values through the "V" type
@@ -512,7 +533,7 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
     self.merge_instance_type_parameter(node, abstract_utils.V, value_var)
 
   def set_str_item(self, node, name, value_var):
-    name_var = self.ctx.convert.build_string(node, name)
+    name_var = self.ctx.convert.build_nonatomic_string(node)
     self.merge_instance_type_params(node, name_var, value_var)
     if name in self.pyval:
       self.pyval[name].PasteVariable(value_var, node)
@@ -540,7 +561,12 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
   def setitem_slot(self, node, name_var, value_var):
     """Implements the __setitem__ slot."""
     self.setitem(node, name_var, value_var)
-    return self.call_pytd(node, "__setitem__", name_var, value_var)
+    return self.call_pytd(
+        node,
+        "__setitem__",
+        abstract_utils.abstractify_variable(name_var, self.ctx),
+        abstract_utils.abstractify_variable(value_var, self.ctx),
+    )
 
   def setdefault_slot(self, node, name_var, value_var=None):
     if value_var is None:
@@ -580,7 +606,7 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
       try:
         return node, self.pyval.pop(str_key)
       except KeyError as e:
-        raise function.DictKeyMissing(str_key) from e
+        raise error_types.DictKeyMissing(str_key) from e
 
   def _set_params_to_any(self, node):
     self.is_concrete = False
@@ -592,7 +618,7 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
   def _set_params_to_any_on_failure(self, node):
     try:
       yield
-    except function.FailedFunctionCall:
+    except error_types.FailedFunctionCall:
       self._set_params_to_any(node)
       raise
 
@@ -613,15 +639,19 @@ class Dict(_instance_base.Instance, mixin.HasSlots, mixin.PythonDict):
     return node, ret
 
   def update(
-      self, node: cfg.CFGNode,
+      self,
+      node: cfg.CFGNode,
       other_dict: Union["Dict", _Dict[str, cfg.Variable], _base.BaseValue],
-      omit: _Tuple[str, ...] = ()) -> None:
+      omit: _Tuple[str, ...] = (),
+  ) -> None:
     if isinstance(other_dict, (Dict, dict)):
       for key, value in other_dict.items():
         if key not in omit:
           self.set_str_item(node, key, value)
-    if (isinstance(other_dict, _instance_base.Instance) and
-        other_dict.full_name == "builtins.dict"):
+    if (
+        isinstance(other_dict, _instance_base.Instance)
+        and other_dict.full_name == "builtins.dict"
+    ):
       self.is_concrete &= other_dict.is_concrete
       for param in (abstract_utils.K, abstract_utils.V):
         param_value = other_dict.get_instance_type_parameter(param, node)
@@ -674,7 +704,7 @@ class Splat(_base.BaseValue):
     return f"splat({self.iterable.data!r})"
 
 
-class SequenceLength(_base.BaseValue):
+class SequenceLength(_base.BaseValue, mixin.HasSlots):
   """Sequence length for match statements."""
 
   def __init__(self, sequence, ctx):
@@ -688,6 +718,8 @@ class SequenceLength(_base.BaseValue):
         length += 1
     self.length = length
     self.splat = splat
+    mixin.HasSlots.init_mixin(self)
+    self.set_native_slot("__sub__", self.sub_slot)
 
   def __repr__(self):
     splat = "+" if self.splat else ""
@@ -695,3 +727,13 @@ class SequenceLength(_base.BaseValue):
 
   def instantiate(self, node, container=None):
     return self.to_variable(node)
+
+  def sub_slot(self, node, other_var):
+    # We should not get a ConversionError here; this is code generated by the
+    # compiler from a literal sequence in a concrete match statement
+    val = abstract_utils.get_atomic_python_constant(other_var, int)
+    if self.splat:
+      ret = self.ctx.convert.build_int(node)
+    else:
+      ret = self.ctx.convert.constant_to_var(self.length - val, node=node)
+    return node, ret
