@@ -1,10 +1,9 @@
 """Constructs related to type annotations."""
 
-import collections
 import dataclasses
 import logging
 import typing
-from typing import Mapping, Tuple, Type, Union as _Union
+from typing import Mapping, Optional, Set, Tuple, Type, Union as _Union
 
 from pytype import datatypes
 from pytype.abstract import _base
@@ -88,7 +87,10 @@ class AnnotationContainer(AnnotationClass):
     return f"AnnotationContainer({self.name})"
 
   def _sub_annotation(
-      self, annot: _base.BaseValue, subst: Mapping[str, _base.BaseValue]
+      self,
+      annot: _base.BaseValue,
+      subst: Mapping[str, _base.BaseValue],
+      seen: Optional[Set[_base.BaseValue]] = None,
   ) -> _base.BaseValue:
     """Apply type parameter substitutions to an annotation."""
     # This is very similar to annotation_utils.sub_one_annotation, but a couple
@@ -99,21 +101,31 @@ class AnnotationContainer(AnnotationClass):
     # - subst contains the type to be substituted in, not an instance of it.
     #   Again, instantiating the type just to later get the type of the instance
     #   is unnecessary extra work.
+    if seen is None:
+      seen = set()
+    if annot in seen:
+      return annot.ctx.convert.unsolvable
+    seen = seen | {annot}
     if isinstance(annot, TypeParameter):
       if annot.full_name in subst:
         return subst[annot.full_name]
       else:
         return self.ctx.convert.unsolvable
     elif isinstance(annot, mixin.NestedAnnotation):
-      inner_types = [(key, self._sub_annotation(val, subst))
-                     for key, val in annot.get_inner_types()]
+      inner_types = [
+          (key, self._sub_annotation(val, subst, seen))
+          for key, val in annot.get_inner_types()
+      ]
       return annot.replace(inner_types)
     return annot
 
   def _get_value_info(
       self, inner, ellipses, allowed_ellipses=frozenset()
-  ) -> Tuple[Tuple[_Union[int, str], ...], Tuple[_base.BaseValue, ...],
-             Type[_classes.ParameterizedClass]]:
+  ) -> Tuple[
+      Tuple[_Union[int, str], ...],
+      Tuple[_base.BaseValue, ...],
+      Type[_classes.ParameterizedClass],
+  ]:
     """Get information about the container's inner values.
 
     Args:
@@ -127,15 +139,17 @@ class AnnotationContainer(AnnotationClass):
     """
     if self.base_cls.full_name == "typing.Protocol":
       return abstract_utils.build_generic_template(inner, self) + (
-          _classes.ParameterizedClass,)  # pytype: disable=bad-return-type
+          _classes.ParameterizedClass,
+      )  # pytype: disable=bad-return-type
     if isinstance(self.base_cls, _classes.TupleClass):
       template = tuple(range(self.base_cls.tuple_length))
     elif isinstance(self.base_cls, _classes.CallableClass):
       template = tuple(range(self.base_cls.num_args)) + (abstract_utils.RET,)
     else:
       template = tuple(t.name for t in self.base_cls.template)
-    self.ctx.errorlog.invalid_ellipses(self.ctx.vm.frames,
-                                       ellipses - allowed_ellipses, self.name)
+    self.ctx.errorlog.invalid_ellipses(
+        self.ctx.vm.frames, ellipses - allowed_ellipses, self.name
+    )
     last_index = len(inner) - 1
     if last_index and last_index in ellipses and len(inner) > len(template):
       # Even if an ellipsis is not allowed at this position, strip it off so
@@ -183,12 +197,14 @@ class AnnotationContainer(AnnotationClass):
 
   def _validate_inner(self, template, inner, raw_inner):
     """Check that the passed inner values are valid for the given template."""
-    if (isinstance(self.base_cls, _classes.ParameterizedClass) and
-        not abstract_utils.is_generic_protocol(self.base_cls)):
+    if isinstance(
+        self.base_cls, _classes.ParameterizedClass
+    ) and not abstract_utils.is_generic_protocol(self.base_cls):
       # For a generic type alias, we check that the number of typevars in the
       # alias matches the number of raw parameters provided.
       template_length = raw_template_length = len(
-          set(self.ctx.annotation_utils.get_type_parameters(self.base_cls)))
+          set(self.ctx.annotation_utils.get_type_parameters(self.base_cls))
+      )
       inner_length = len(raw_inner)
       base_cls = self.base_cls.base_cls
     else:
@@ -202,7 +218,8 @@ class AnnotationContainer(AnnotationClass):
     if inner_length != template_length:
       if not template:
         self.ctx.errorlog.not_indexable(
-            self.ctx.vm.frames, base_cls.name, generic_warning=True)
+            self.ctx.vm.frames, base_cls.name, generic_warning=True
+        )
       else:
         # Use the unprocessed values of `template` and `inner` so that the error
         # message matches what the user sees.
@@ -211,25 +228,33 @@ class AnnotationContainer(AnnotationClass):
         else:
           error_template = (t.name for t in base_cls.template)
         self.ctx.errorlog.wrong_annotation_parameter_count(
-            self.ctx.vm.frames, self.base_cls, raw_inner, raw_template_length,
-            error_template)
+            self.ctx.vm.frames,
+            self.base_cls,
+            raw_inner,
+            raw_template_length,
+            error_template,
+        )
     else:
       if len(inner) == 1:
-        val, = inner
+        (val,) = inner
         # It's a common mistake to index a container class rather than an
         # instance (e.g., list[0]).
         # We only check the "int" case, since string literals are allowed for
         # late annotations.
-        if (isinstance(val, _instance_base.Instance) and
-            val.cls == self.ctx.convert.int_type):
+        if (
+            isinstance(val, _instance_base.Instance)
+            and val.cls == self.ctx.convert.int_type
+        ):
           # Don't report this error again.
           inner = (self.ctx.convert.unsolvable,)
           self.ctx.errorlog.not_indexable(self.ctx.vm.frames, self.name)
       # Check for a misused Final annotation
       if any(isinstance(val, FinalAnnotation) for val in inner):
         self.ctx.errorlog.invalid_final_type(self.ctx.vm.frames)
-        inner = [val.annotation if isinstance(val, FinalAnnotation) else val
-                 for val in inner]
+        inner = [
+            val.annotation if isinstance(val, FinalAnnotation) else val
+            for val in inner
+        ]
     return inner
 
   def _build_value(self, node, inner, ellipses):
@@ -241,24 +266,28 @@ class AnnotationContainer(AnnotationClass):
       # class_mixin.Class, and (2) we have to postpone error-checking anyway so
       # we might as well postpone the entire evaluation.
       printed_params = []
-      added_imports = collections.defaultdict(set)
+      added_typing_imports = set()
       for i, param in enumerate(inner):
         if i in ellipses:
           printed_params.append("...")
         else:
-          typ = param.get_instance_type(node)
-          annot, imports = pytd_utils.MakeTypeAnnotation(typ)
+          typ = param.to_pytd_type_of_instance(node)
+          annot, typing_imports = pytd_utils.MakeTypeAnnotation(typ)
           printed_params.append(annot)
-          for k, v in imports.items():
-            added_imports[k] |= v
+          added_typing_imports.update(typing_imports)
 
       expr = f"{self.base_cls.expr}[{', '.join(printed_params)}]"
-      annot = LateAnnotation(expr, self.base_cls.stack, self.ctx,
-                             imports=added_imports)
+      annot = LateAnnotation(
+          expr,
+          self.base_cls.stack,
+          self.ctx,
+          typing_imports=added_typing_imports,
+      )
       self.ctx.vm.late_annotations[self.base_cls.expr].append(annot)
       return annot
     template, processed_inner, abstract_class = self._get_value_info(
-        inner, ellipses)
+        inner, ellipses
+    )
     if isinstance(self.base_cls, _classes.ParameterizedClass):
       base_cls = self.base_cls.base_cls
     else:
@@ -268,45 +297,52 @@ class AnnotationContainer(AnnotationClass):
       # usually, the parameterized class inherits the base class's template.
       # Protocol[T, ...] is a shorthand for Protocol, Generic[T, ...].
       template_params = [
-          param.with_module(base_cls.full_name)
-          for param in typing.cast(Tuple[TypeParameter, ...], processed_inner)]
+          param.with_scope(base_cls.full_name)
+          for param in typing.cast(Tuple[TypeParameter, ...], processed_inner)
+      ]
     else:
       template_params = None
     processed_inner = self._validate_inner(template, processed_inner, inner)
     params = {
-        name: (processed_inner[i]
-               if i < len(processed_inner) else self.ctx.convert.unsolvable)
+        name: (
+            processed_inner[i]
+            if i < len(processed_inner)
+            else self.ctx.convert.unsolvable
+        )
         for i, name in enumerate(template)
     }
 
     # Check if the concrete types match the type parameters.
     if base_cls.template:
       processed_params = self.ctx.annotation_utils.convert_class_annotations(
-          node, params)
+          node, params
+      )
       for formal_param in base_cls.template:
         root_node = self.ctx.root_node
         param_value = processed_params[formal_param.name]
-        if (isinstance(formal_param, TypeParameter) and
-            not formal_param.is_generic() and
-            isinstance(param_value, TypeParameter)):
+        if (
+            isinstance(formal_param, TypeParameter)
+            and not formal_param.is_generic()
+            and isinstance(param_value, TypeParameter)
+        ):
           if formal_param.name == param_value.name:
             # We don't need to check if a TypeParameter matches itself.
             continue
           else:
             actual = param_value.instantiate(
-                root_node, container=abstract_utils.DUMMY_CONTAINER)
+                root_node, container=abstract_utils.DUMMY_CONTAINER
+            )
         elif param_value.is_concrete and isinstance(param_value.pyval, str):
-          # TODO(b/261769826): Do we need to collect imports here?
           expr = param_value.pyval
-          annot = LateAnnotation(expr, self.ctx.vm.frames, self.ctx,
-                                 imports=set())
+          annot = LateAnnotation(expr, self.ctx.vm.frames, self.ctx)
           base = expr.split("[", 1)[0]
           self.ctx.vm.late_annotations[base].append(annot)
           actual = annot.instantiate(root_node)
         else:
           actual = param_value.instantiate(root_node)
         match_result = self.ctx.matcher(node).compute_one_match(
-            actual, formal_param)
+            actual, formal_param
+        )
         if not match_result.success:
           if isinstance(param_value, TypeParameter):
             # bad_matches replaces type parameters in the expected type with
@@ -318,15 +354,18 @@ class AnnotationContainer(AnnotationClass):
               expected = dataclasses.replace(match.expected, typ=formal_param)
               bad.append(dataclasses.replace(match, expected=expected))
             if isinstance(formal_param, TypeParameter):
-              details = (f"TypeVars {formal_param.name} and {param_value.name} "
-                         "have incompatible bounds or constraints.")
+              details = (
+                  f"TypeVars {formal_param.name} and {param_value.name} "
+                  "have incompatible bounds or constraints."
+              )
             else:
               details = None
           else:
             bad = match_result.bad_matches
             details = None
           self.ctx.errorlog.bad_concrete_type(
-              self.ctx.vm.frames, root_node, bad, details)
+              self.ctx.vm.frames, root_node, bad, details
+          )
           return self.ctx.convert.unsolvable
 
     try:
@@ -339,89 +378,6 @@ class AnnotationContainer(AnnotationClass):
     return self._call_helper(node, self.base_cls, func, args)
 
 
-class _TypeVariable(_base.BaseValue):
-  """Parameter of a type."""
-
-  formal = True
-
-  _INSTANCE_CLASS = None
-
-  def __init__(self,
-               name,
-               ctx,
-               constraints=(),
-               bound=None,
-               covariant=False,
-               contravariant=False,
-               module=None):
-    super().__init__(name, ctx)
-    # TODO(b/217789659): PEP-612 does not mention constraints, but ParamSpecs
-    # ignore all the extra parameters anyway..
-    self.constraints = constraints
-    self.bound = bound
-    self.covariant = covariant
-    self.contravariant = contravariant
-    self.module = module
-
-  def is_generic(self):
-    return not self.constraints and not self.bound
-
-  def copy(self):
-    return self.__class__(self.name, self.ctx, self.constraints, self.bound,
-                          self.covariant, self.contravariant, self.module)
-
-  def with_module(self, module):
-    res = self.copy()
-    res.module = module
-    return res
-
-  def __eq__(self, other):
-    if isinstance(other, type(self)):
-      return (self.name == other.name and
-              self.constraints == other.constraints and
-              self.bound == other.bound and
-              self.covariant == other.covariant and
-              self.contravariant == other.contravariant and
-              self.module == other.module)
-    return NotImplemented
-
-  def __ne__(self, other):
-    return not self == other
-
-  def __hash__(self):
-    return hash((self.name, self.constraints, self.bound, self.covariant,
-                 self.contravariant))
-
-  def __repr__(self):
-    return ("{!s}({!r}, constraints={!r}, bound={!r}, module={!r})"
-            .format(self.__class__.__name__, self.name, self.constraints,
-                    self.bound, self.module))
-
-  def instantiate(self, node, container=None):
-    var = self.ctx.program.NewVariable()
-    if container and (not isinstance(container, _instance_base.SimpleValue) or
-                      self.full_name in container.all_template_names):
-      instance = self._INSTANCE_CLASS(self, container, self.ctx)  # pylint: disable=not-callable
-      return instance.to_variable(node)
-    else:
-      for c in self.constraints:
-        var.PasteVariable(c.instantiate(node, container))
-      if self.bound:
-        var.PasteVariable(self.bound.instantiate(node, container))
-    if not var.bindings:
-      var.AddBinding(self.ctx.convert.unsolvable, [], node)
-    return var
-
-  def update_official_name(self, name):
-    if self.name != name:
-      message = (f"TypeVar({self.name!r}) must be stored as {self.name!r}, "
-                 f"not {name!r}")
-      self.ctx.errorlog.invalid_typevar(self.ctx.vm.frames, message)
-
-  def call(self, node, func, args, alias_map=None):
-    return node, self.instantiate(node)
-
-
 class _TypeVariableInstance(_base.BaseValue):
   """An instance of a type parameter."""
 
@@ -429,7 +385,11 @@ class _TypeVariableInstance(_base.BaseValue):
     super().__init__(param.name, ctx)
     self.cls = self.param = param
     self.instance = instance
-    self.module = param.module
+    self.scope = param.scope
+
+  @property
+  def full_name(self):
+    return f"{self.scope}.{self.name}" if self.scope else self.name
 
   def call(self, node, func, args, alias_map=None):
     var = self.instance.get_instance_type_parameter(self.name)
@@ -458,6 +418,122 @@ class ParamSpecInstance(_TypeVariableInstance):
   """An instance of a ParamSpec type parameter."""
 
 
+class _TypeVariable(_base.BaseValue):
+  """Parameter of a type."""
+
+  formal = True
+
+  _INSTANCE_CLASS: Type[_TypeVariableInstance] = None
+
+  def __init__(
+      self,
+      name,
+      ctx,
+      constraints=(),
+      bound=None,
+      covariant=False,
+      contravariant=False,
+      scope=None,
+  ):
+    super().__init__(name, ctx)
+    # TODO(b/217789659): PEP-612 does not mention constraints, but ParamSpecs
+    # ignore all the extra parameters anyway..
+    self.constraints = constraints
+    self.bound = bound
+    self.covariant = covariant
+    self.contravariant = contravariant
+    self.scope = scope
+
+  @_base.BaseValue.module.setter
+  def module(self, module):
+    super(_TypeVariable, _TypeVariable).module.fset(self, module)
+    self.scope = module
+
+  @property
+  def full_name(self):
+    return f"{self.scope}.{self.name}" if self.scope else self.name
+
+  def is_generic(self):
+    return not self.constraints and not self.bound
+
+  def copy(self):
+    return self.__class__(
+        self.name,
+        self.ctx,
+        self.constraints,
+        self.bound,
+        self.covariant,
+        self.contravariant,
+        self.scope,
+    )
+
+  def with_scope(self, scope):
+    res = self.copy()
+    res.scope = scope
+    return res
+
+  def __eq__(self, other):
+    if isinstance(other, type(self)):
+      return (
+          self.name == other.name
+          and self.constraints == other.constraints
+          and self.bound == other.bound
+          and self.covariant == other.covariant
+          and self.contravariant == other.contravariant
+          and self.scope == other.scope
+      )
+    return NotImplemented
+
+  def __ne__(self, other):
+    return not self == other
+
+  def __hash__(self):
+    return hash((
+        self.name,
+        self.constraints,
+        self.bound,
+        self.covariant,
+        self.contravariant,
+    ))
+
+  def __repr__(self):
+    return "{!s}({!r}, constraints={!r}, bound={!r}, module={!r})".format(
+        self.__class__.__name__,
+        self.name,
+        self.constraints,
+        self.bound,
+        self.scope,
+    )
+
+  def instantiate(self, node, container=None):
+    var = self.ctx.program.NewVariable()
+    if container and (
+        not isinstance(container, _instance_base.SimpleValue)
+        or self.full_name in container.all_template_names
+    ):
+      instance = self._INSTANCE_CLASS(self, container, self.ctx)  # pylint: disable=not-callable
+      return instance.to_variable(node)
+    else:
+      for c in self.constraints:
+        var.PasteVariable(c.instantiate(node, container))
+      if self.bound:
+        var.PasteVariable(self.bound.instantiate(node, container))
+    if not var.bindings:
+      var.AddBinding(self.ctx.convert.unsolvable, [], node)
+    return var
+
+  def update_official_name(self, name):
+    if self.name != name:
+      message = (
+          f"TypeVar({self.name!r}) must be stored as {self.name!r}, "
+          f"not {name!r}"
+      )
+      self.ctx.errorlog.invalid_typevar(self.ctx.vm.frames, message)
+
+  def call(self, node, func, args, alias_map=None):
+    return node, self.instantiate(node)
+
+
 class TypeParameter(_TypeVariable):
   """Parameter of a type (typing.TypeVar)."""
 
@@ -580,12 +656,16 @@ class Union(_base.BaseValue, mixin.NestedAnnotation, mixin.HasSlots):
     # Check that we are instantiating all the unbound type parameters
     if num_params != len(slice_content):
       self.ctx.errorlog.wrong_annotation_parameter_count(
-          self.ctx.vm.frames, self, [v.data[0] for v in slice_content],
-          num_params)
+          self.ctx.vm.frames,
+          self,
+          [v.data[0] for v in slice_content],
+          num_params,
+      )
       return node, self.ctx.new_unsolvable(node)
     concrete = (
         var.data[0].instantiate(node, container=abstract_utils.DUMMY_CONTAINER)
-        for var in slice_content)
+        for var in slice_content
+    )
     subst = datatypes.AliasingDict()
     for p in params:
       for k in subst:
@@ -617,8 +697,9 @@ class Union(_base.BaseValue, mixin.NestedAnnotation, mixin.HasSlots):
     return function.call_function(self.ctx, node, var, args)
 
   def get_formal_type_parameter(self, t):
-    new_options = [option.get_formal_type_parameter(t)
-                   for option in self.options]
+    new_options = [
+        option.get_formal_type_parameter(t) for option in self.options
+    ]
     return Union(new_options, self.ctx)
 
   def get_inner_types(self):
@@ -647,13 +728,13 @@ class LateAnnotation:
 
   _RESOLVING = object()
 
-  def __init__(self, expr, stack, ctx, *, imports=None):
+  def __init__(self, expr, stack, ctx, *, typing_imports=None):
     self.expr = expr
     self.stack = stack
     self.ctx = ctx
     self.resolved = False
-    # Any new imports the annotation needs while resolving.
-    self._imports = imports or {}
+    # Any new typing imports the annotation needs while resolving.
+    self._typing_imports = typing_imports or set()
     self._type = ctx.convert.unsolvable  # the resolved type of `expr`
     self._unresolved_instances = set()
     self._resolved_instances = {}
@@ -662,9 +743,9 @@ class LateAnnotation:
     # LateAnnotation.__dict__ and self.__dict__. These names are used in
     # __getattribute__ and __setattr__ to determine whether a given get/setattr
     # call should operate on the LateAnnotation itself or its resolved type.
-    self._attribute_names = (
-        set(LateAnnotation.__dict__) |
-        set(super().__getattribute__("__dict__")))
+    self._attribute_names = set(LateAnnotation.__dict__) | set(
+        super().__getattribute__("__dict__")
+    )
 
   def flatten_expr(self):
     """Flattens the expression into a legal variable name if necessary.
@@ -682,7 +763,8 @@ class LateAnnotation:
       # _DOT and _RBAR have no trailing underscore because they precede names
       # that we already prefix an underscore to.
       return "_" + self.expr.replace(".", "_DOT").replace(
-          "[", "_LBAR_").replace("]", "_RBAR").replace(", ", "_COMMA_")
+          "[", "_LBAR_"
+      ).replace("]", "_RBAR").replace(", ", "_COMMA_")
     return self.expr
 
   def unflatten_expr(self):
@@ -691,13 +773,21 @@ class LateAnnotation:
       mod, dot, rest = self.expr.rpartition(".")
       # The [1:] slicing and trailing underscore in _DOT_ are to get rid of
       # leading underscores added when flattening.
-      return mod + dot + rest[1:].replace("_DOT_", ".").replace(
-          "_LBAR_", "[").replace("_RBAR", "]").replace("_COMMA_", ", ")
+      return (
+          mod
+          + dot
+          + rest[1:]
+          .replace("_DOT_", ".")
+          .replace("_LBAR_", "[")
+          .replace("_RBAR", "]")
+          .replace("_COMMA_", ", ")
+      )
     return self.expr
 
   def __repr__(self):
     return "LateAnnotation({!r}, resolved={!r})".format(
-        self.expr, self._type if self.resolved else None)
+        self.expr, self._type if self.resolved else None
+    )
 
   # __hash__ and __eq__ need to be explicitly defined for Python to use them in
   # set/dict comparisons.
@@ -709,15 +799,21 @@ class LateAnnotation:
     return hash(self) == hash(other)
 
   def __getattribute__(self, name):
-    if name == "_attribute_names" or name in self._attribute_names:
-      return super().__getattribute__(name)
-    return self._type.__getattribute__(name)  # pytype: disable=attribute-error
+    # We use super().__getattribute__ directly for attribute access to avoid a
+    # performance penalty from this function recursively calling itself.
+    get = super().__getattribute__
+    if name == "_attribute_names" or name in get("_attribute_names"):
+      return get(name)
+    return get("_type").__getattribute__(name)  # pytype: disable=attribute-error
 
   def __setattr__(self, name, value):
     if not hasattr(self, "_attribute_names") or name in self._attribute_names:
       return super().__setattr__(name, value)
     return self._type.__setattr__(name, value)
 
+  def __contains__(self, name):
+    return self.resolved and name in self._type
+
   def resolve(self, node, f_globals, f_locals):
     """Resolve the late annotation."""
     if self.resolved:
@@ -728,17 +824,19 @@ class LateAnnotation:
     self.resolved = LateAnnotation._RESOLVING
     # Add implicit imports for typing, since we can have late annotations like
     # `set[int]` which get converted to `typing.Set[int]`.
-    if "typing" in self._imports:
+    if self._typing_imports:
       overlay = self.ctx.vm.import_module("typing", "typing", 0)
-      for v in self._imports["typing"]:
+      for v in self._typing_imports:
         if v not in f_globals.members:
           f_globals.members[v] = overlay.get_module(v).load_lazy_attribute(v)
-    var, errorlog = abstract_utils.eval_expr(self.ctx, node, f_globals,
-                                             f_locals, self.expr)
+    var, errorlog = abstract_utils.eval_expr(
+        self.ctx, node, f_globals, f_locals, self.expr
+    )
     if errorlog:
       self.ctx.errorlog.copy_from(errorlog.errors, self.stack)
     self._type = self.ctx.annotation_utils.extract_annotation(
-        node, var, None, self.stack)
+        node, var, None, self.stack
+    )
     if self._type != self.ctx.convert.unsolvable:
       # We may have tried to call __init__ on instances of this annotation.
       # Since the annotation was unresolved at the time, we need to call
