@@ -17,14 +17,16 @@ from pytype.abstract import _pytd_function
 from pytype.abstract import _singletons
 from pytype.abstract import _typing
 from pytype.abstract import class_mixin
+from pytype.abstract import mixin
 
 log = logging.getLogger(__name__)
 
 # For simplicity, we pretend all abstract values are defined in abstract.py.
 BaseValue = _base.BaseValue
 
-# This is technically a mixin, but we use it a lot in isinstance() checks.
+# These are technically mixins, but we use them a lot in isinstance() checks.
 Class = class_mixin.Class
+PythonConstant = mixin.PythonConstant
 
 BuildClass = _classes.BuildClass
 InterpreterClass = _classes.InterpreterClass
@@ -75,6 +77,7 @@ Singleton = _singletons.Singleton
 Empty = _singletons.Empty
 Deleted = _singletons.Deleted
 Unsolvable = _singletons.Unsolvable
+Null = _singletons.Null
 
 AnnotationClass = _typing.AnnotationClass
 AnnotationContainer = _typing.AnnotationContainer
