@@ -5,17 +5,19 @@ import collections
 import dataclasses
 import itertools
 import logging
-from typing import Dict, Optional, Sequence, Tuple
+from typing import Any, Dict, Optional, Tuple, TypeVar, cast
 
 import attrs
-
 from pytype import datatypes
+from pytype import pretty_printer_base as pp
 from pytype.abstract import _base
 from pytype.abstract import abstract_utils
+from pytype.errors import error_types
 from pytype.pytd import pytd
 from pytype.pytd import pytd_utils
 from pytype.typegraph import cfg
 from pytype.typegraph import cfg_utils
+from pytype.types import types
 
 log = logging.getLogger(__name__)
 _isinstance = abstract_utils._isinstance  # pylint: disable=protected-access
@@ -55,22 +57,31 @@ def get_signatures(func):
   else:
     unwrapped = abstract_utils.maybe_unwrap_decorated_function(func)
     if unwrapped:
-      return list(itertools.chain.from_iterable(
-          get_signatures(f) for f in unwrapped.data))
+      return list(
+          itertools.chain.from_iterable(
+              get_signatures(f) for f in unwrapped.data
+          )
+      )
     if _isinstance(func, "Instance"):
       _, call_var = func.ctx.attribute_handler.get_attribute(
-          func.ctx.root_node, func, "__call__",
-          func.to_binding(func.ctx.root_node))
+          func.ctx.root_node,
+          func,
+          "__call__",
+          func.to_binding(func.ctx.root_node),
+      )
       if call_var and len(call_var.data) == 1:
         return get_signatures(call_var.data[0])
     raise NotImplementedError(func.__class__.__name__)
 
 
 def _print(t):
-  return pytd_utils.Print(t.get_instance_type())
+  return pytd_utils.Print(t.to_pytd_type_of_instance())
 
 
-class Signature:
+_SigT = TypeVar("_SigT", bound="Signature")
+
+
+class Signature(types.Signature):
   """Representation of a Python function signature.
 
   Attributes:
@@ -79,9 +90,9 @@ class Signature:
       positional-only parameters and does NOT include keyword-only parameters.
     posonly_count: Number of positional-only parameters. (Python 3.8)
     varargs_name: Name of the varargs parameter. (The "args" in *args)
-    kwonly_params: Tuple of keyword-only parameters. (Python 3)
-      E.g. ("x", "y") for "def f(a, *, x, y=2)". These do NOT appear in
-      param_names. Ordered like in the source file.
+    kwonly_params: Tuple of keyword-only parameters. (Python 3) E.g. ("x", "y")
+      for "def f(a, *, x, y=2)". These do NOT appear in param_names. Ordered
+      like in the source file.
     kwargs_name: Name of the kwargs parameter. (The "kwargs" in **kwargs)
     defaults: Dictionary, name to value, for all parameters with default values.
     annotations: A dictionary of type annotations. (string to type)
@@ -94,15 +105,26 @@ class Signature:
       posonly_count names in param_names).
   """
 
-  def __init__(self, name, param_names, posonly_count, varargs_name,
-               kwonly_params, kwargs_name, defaults, annotations,
-               postprocess_annotations=True):
-    self.name = name
-    self.param_names = param_names
-    self.posonly_count = posonly_count
-    self.varargs_name = varargs_name
-    self.kwonly_params = kwonly_params
-    self.kwargs_name = kwargs_name
+  def __init__(
+      self,
+      name: str,
+      param_names: Tuple[str, ...],
+      posonly_count: int,
+      varargs_name: Optional[str],
+      kwonly_params: Tuple[str, ...],
+      kwargs_name: Optional[str],
+      defaults: Dict[str, cfg.Variable],
+      annotations: Dict[str, _base.BaseValue],
+      postprocess_annotations: bool = True,
+  ) -> None:
+    super().__init__(
+        name,
+        param_names,
+        posonly_count,
+        varargs_name,
+        kwonly_params,
+        kwargs_name,
+    )
     self.defaults = defaults
     self.annotations = annotations
     self.excluded_types = set()
@@ -112,7 +134,8 @@ class Signature:
     self.type_params = set()
     for annot in self.annotations.values():
       self.type_params.update(
-          p.name for p in annot.ctx.annotation_utils.get_type_parameters(annot))
+          p.name for p in annot.ctx.annotation_utils.get_type_parameters(annot)
+      )
 
   @property
   def has_return_annotation(self):
@@ -122,27 +145,33 @@ class Signature:
   def has_param_annotations(self):
     return bool(self.annotations.keys() - {"return"})
 
-  @property
-  def posonly_params(self):
-    return self.param_names[:self.posonly_count]
+  def has_default(self, name):
+    return name in self.defaults
 
-  def add_scope(self, module):
+  def add_scope(self, cls):
     """Add scope for type parameters in annotations."""
     annotations = {}
     for key, val in self.annotations.items():
       annotations[key] = val.ctx.annotation_utils.add_scope(
-          val, self.excluded_types, module)
+          val, self.excluded_types, cls
+      )
     self.annotations = annotations
 
   def _postprocess_annotation(self, name, annotation):
     """Postprocess the given annotation."""
     ctx = annotation.ctx
     if name == self.varargs_name:
-      return _make("ParameterizedClass",
-                   ctx.convert.tuple_type, {abstract_utils.T: annotation}, ctx)
+      return _make(
+          "ParameterizedClass",
+          ctx.convert.tuple_type,
+          {abstract_utils.T: annotation},
+          ctx,
+      )
     elif name == self.kwargs_name:
-      params = {abstract_utils.K: ctx.convert.str_type,
-                abstract_utils.V: annotation}
+      params = {
+          abstract_utils.K: ctx.convert.str_type,
+          abstract_utils.V: annotation,
+      }
       return _make("ParameterizedClass", ctx.convert.dict_type, params, ctx)
     else:
       return annotation
@@ -153,29 +182,120 @@ class Signature:
   def del_annotation(self, name):
     del self.annotations[name]  # Raises KeyError if annotation does not exist.
 
-  def check_type_parameter_count(self, stack):
-    """Check the count of type parameters in function."""
+  def check_type_parameters(self, stack, opcode, is_attribute_of_class):
+    """Check type parameters in function."""
+    if not self.annotations:
+      return
     c = collections.Counter()
-    for annot in self.annotations.values():
-      c.update(annot.ctx.annotation_utils.get_type_parameters(annot))
+    if opcode and opcode.metadata.signature_annotations:
+      src_annots = opcode.metadata.signature_annotations
+    else:
+      src_annots = {}
+    bare_aliases = {}
+    bare_alias_params = {}
+    ctx = next(iter(self.annotations.values())).ctx
+    for name, annot in self.annotations.items():
+      params = ctx.annotation_utils.get_type_parameters(annot)
+      if params and name in src_annots and "[" not in src_annots[name]:
+        # Bare aliases for generic types should not refer to the type param in
+        # their error message.
+        # Record the textual annotation if the source code is not of the form:
+        #   x: Foo[...] or
+        #   x: Foo where Foo is the name of a type param
+        if not any(p.name == src_annots[name] for p in params):
+          for p in params:
+            bare_aliases[p.name] = name
+            bare_alias_params[name] = params
+      c.update(params)
+    bare_alias_errors = set()
     for param, count in c.items():
       if param.name in self.excluded_types:
-        # skip all the type parameters in `excluded_types`
         continue
-      if count == 1 and not (param.constraints or param.bound or
-                             param.covariant or param.contravariant):
-        param.ctx.errorlog.invalid_annotation(
-            stack, param, (f"TypeVar {param.name!r} appears only once in the "
-                           "function signature"))
+      if param.full_name == "typing.Self":
+        if not is_attribute_of_class and not self.name.endswith(".__new__"):
+          ctx.errorlog.invalid_annotation(
+              stack, param, "Cannot use 'typing.Self' outside of a class"
+          )
+        continue
+      if count == 1 and not (
+          param.constraints
+          or param.bound
+          or param.covariant
+          or param.contravariant
+      ):
+        if param.name in bare_aliases:
+          bare_alias_errors.add(bare_aliases[param.name])
+        else:
+          ctx.errorlog.invalid_annotation(
+              stack,
+              param,
+              (
+                  f"TypeVar {param.name!r} appears only once in the "
+                  "function signature"
+              ),
+          )
+      for var_name in bare_alias_errors:
+        annot = src_annots[var_name]
+        params = ", ".join(p.name for p in bare_alias_params[var_name])
+        any_params = ", ".join("Any" for p in bare_alias_params[var_name])
+        msg = (
+            f"In annotation '{var_name}: {annot}', {annot} is a "
+            "generic alias, and needs to be parameterized.\n"
+            f"Use '{var_name}: {annot}[{params}]' instead, or "
+            f"'{var_name}: {annot}[{any_params}]' to match any instance "
+            f"of {annot}"
+        )
+        ctx.errorlog.invalid_annotation(stack, annot, msg)
 
   def drop_first_parameter(self):
     return self._replace(param_names=self.param_names[1:])
 
+  def _make_concatenated_type(
+      self, type1: _base.BaseValue, type2: Optional[_base.BaseValue]
+  ) -> Optional[_base.BaseValue]:
+    """Concatenates type1 and type2 if possible.
+
+    If type2 is a ParamSpec or Concatenate object, creates a new Concatenate
+    object by adding type1 to the front.
+
+    Args:
+      type1: An abstract value.
+      type2: An abstract value or None.
+
+    Returns:
+      A new Concatenate object, or None if type2 cannot be concatenated to.
+    """
+    if _isinstance(type2, "ParamSpec"):
+      new_args = [type1, type2]
+    elif _isinstance(type2, "Concatenate"):
+      # Pytype can't understand this custom isinstance check.
+      type2 = cast(Any, type2)
+      new_args = [type1] + type2.args + [type2.paramspec]
+    else:
+      return None
+    return _make("Concatenate", new_args, type1.ctx)
+
+  def prepend_parameter(self: _SigT, name: str, typ: _base.BaseValue) -> _SigT:
+    """Returns a new signature with a `name: typ` param added to the front."""
+    if self.param_names:
+      param_name = self.param_names[0]
+      param_type = self.annotations.get(param_name)
+      concatenated_type = self._make_concatenated_type(typ, param_type)
+      if concatenated_type:
+        # The existing first parameter actually represents a variable number of
+        # parameters. Ignore `name` and add in the new type.
+        return self._replace(
+            annotations={**self.annotations, param_name: concatenated_type}
+        )
+    param_names = (name,) + self.param_names
+    annots = {**self.annotations, name: typ}
+    return self._replace(param_names=param_names, annotations=annots)
+
   def mandatory_param_count(self):
-    num = len([name
-               for name in self.param_names if name not in self.defaults])
-    num += len([name
-                for name in self.kwonly_params if name not in self.defaults])
+    num = len([name for name in self.param_names if name not in self.defaults])
+    num += len(
+        [name for name in self.kwonly_params if name not in self.defaults]
+    )
     return num
 
   def maximum_param_count(self):
@@ -186,13 +306,17 @@ class Signature:
   @classmethod
   def from_pytd(cls, ctx, name, sig):
     """Construct an abstract signature from a pytd signature."""
-    pytd_annotations = [(p.name, p.type)
-                        for p in sig.params + (sig.starargs, sig.starstarargs)
-                        if p is not None]
+    pytd_annotations = [
+        (p.name, p.type)
+        for p in sig.params + (sig.starargs, sig.starstarargs)
+        if p is not None
+    ]
     pytd_annotations.append(("return", sig.return_type))
+
     def param_to_var(p):
       return ctx.convert.constant_to_var(
-          p.type, subst=datatypes.AliasingDict(), node=ctx.root_node)
+          p.type, subst=datatypes.AliasingDict(), node=ctx.root_node
+      )
 
     param_names = []
     posonly_count = 0
@@ -213,7 +337,8 @@ class Signature:
         defaults={p.name: param_to_var(p) for p in sig.params if p.optional},
         annotations={
             name: ctx.convert.constant_to_value(
-                typ, subst=datatypes.AliasingDict(), node=ctx.root_node)
+                typ, subst=datatypes.AliasingDict(), node=ctx.root_node
+            )
             for name, typ in pytd_annotations
         },
         postprocess_annotations=False,
@@ -221,8 +346,9 @@ class Signature:
 
   @classmethod
   def from_callable(cls, val):
-    annotations = {argname(i): val.formal_type_parameters[i]
-                   for i in range(val.num_args)}
+    annotations = {
+        argname(i): val.formal_type_parameters[i] for i in range(val.num_args)
+    }
     param_names = tuple(sorted(annotations))
     annotations["return"] = val.formal_type_parameters[abstract_utils.RET]
     return cls(
@@ -279,8 +405,11 @@ class Signature:
     )
 
   def has_param(self, name):
-    return name in self.param_names or name in self.kwonly_params or (
-        name == self.varargs_name or name == self.kwargs_name)
+    return (
+        name in self.param_names
+        or name in self.kwonly_params
+        or (name == self.varargs_name or name == self.kwargs_name)
+    )
 
   def insert_varargs_and_kwargs(self, args):
     """Insert varargs and kwargs from args into the signature.
@@ -300,13 +429,16 @@ class Signature:
         varargs_names.append(name)
       else:
         kwargs_names.append(name)
-    new_param_names = (self.param_names + tuple(sorted(varargs_names)) +
-                       tuple(sorted(kwargs_names)))
+    new_param_names = (
+        self.param_names
+        + tuple(sorted(varargs_names))
+        + tuple(sorted(kwargs_names))
+    )
     return self._replace(param_names=new_param_names)
 
-  _ATTRIBUTES = (
-      set(__init__.__code__.co_varnames[:__init__.__code__.co_argcount]) -
-      {"self", "postprocess_annotations"})
+  _ATTRIBUTES = set(
+      __init__.__code__.co_varnames[: __init__.__code__.co_argcount]
+  ) - {"self", "postprocess_annotations"}
 
   def _replace(self, **kwargs):
     """Returns a copy of the signature with the specified values replaced."""
@@ -331,7 +463,7 @@ class Signature:
         yield (argname(i), posarg, None)
     for name in sorted(args.namedargs):
       namedarg = args.namedargs[name]
-      if name in self.param_names[:self.posonly_count]:
+      if name in self.param_names[: self.posonly_count]:
         formal = None
       else:
         formal = self.annotations.get(name)
@@ -341,11 +473,17 @@ class Signature:
           formal = kwargs_type.ctx.convert.get_element_type(kwargs_type)
       yield (name, namedarg, formal)
     if self.varargs_name is not None and args.starargs is not None:
-      yield (self.varargs_name, args.starargs,
-             self.annotations.get(self.varargs_name))
+      yield (
+          self.varargs_name,
+          args.starargs,
+          self.annotations.get(self.varargs_name),
+      )
     if self.kwargs_name is not None and args.starstarargs is not None:
-      yield (self.kwargs_name, args.starstarargs,
-             self.annotations.get(self.kwargs_name))
+      yield (
+          self.kwargs_name,
+          args.starstarargs,
+          self.annotations.get(self.kwargs_name),
+      )
 
   def check_defaults(self, ctx):
     """Raises an error if a non-default param follows a default."""
@@ -354,9 +492,11 @@ class Signature:
       if name in self.defaults:
         has_default = True
       elif has_default:
-        msg = (f"In method {self.name}, non-default argument {name} "
-               "follows default argument")
-        ctx.errorlog.invalid_function_definition(ctx.vm.frames, msg)
+        msg = (
+            f"In method {self.name}, non-default argument {name} "
+            "follows default argument"
+        )
+        ctx.errorlog.invalid_function_definition(ctx.vm.stack(), msg)
         return
 
   def _yield_arguments(self):
@@ -374,7 +514,8 @@ class Signature:
       annot = self._print_annot(base_name)
       default = self._print_default(base_name)
       yield name + (": " + annot if annot else "") + (
-          " = " + default if default else "")
+          " = " + default if default else ""
+      )
 
   def _print_annot(self, name):
     return _print(self.annotations[name]) if name in self.annotations else None
@@ -385,14 +526,14 @@ class Signature:
       if len(values) > 1:
         return "..."
       else:
-        return abstract_utils.show_constant(values[0])
+        return pp.PrettyPrinterBase.show_constant(values[0])
     else:
       return None
 
   def __repr__(self):
     args = list(self._yield_arguments())
     if self.posonly_count:
-      args = args[:self.posonly_count] + ["/"] + args[self.posonly_count:]
+      args = args[: self.posonly_count] + ["/"] + args[self.posonly_count :]
     args = ", ".join(args)
     ret = self._print_annot("return")
     return f"def {self.name}({args}) -> {ret if ret else 'Any'}"
@@ -443,12 +584,12 @@ class Args:
       cfg.Variable.
     starargs: The *args parameter, or None.
     starstarargs: The **kwargs parameter, or None.
-
   """
 
   posargs: Tuple[cfg.Variable, ...]
-  namedargs: Dict[str, cfg.Variable] = attrs.field(converter=_convert_namedargs,
-                                                   default=None)
+  namedargs: Dict[str, cfg.Variable] = attrs.field(
+      converter=_convert_namedargs, default=None
+  )
   starargs: Optional[cfg.Variable] = None
   starstarargs: Optional[cfg.Variable] = None
 
@@ -464,14 +605,16 @@ class Args:
   def starargs_as_tuple(self, node, ctx):
     try:
       args = self.starargs and abstract_utils.get_atomic_python_constant(
-          self.starargs, tuple)
+          self.starargs, tuple
+      )
     except abstract_utils.ConversionError:
       args = None
     if not args:
       return args
     return tuple(
         var if var.bindings else ctx.convert.empty.to_variable(node)
-        for var in args)
+        for var in args
+    )
 
   def starstarargs_as_dict(self):
     """Return **args as a python dict."""
@@ -479,7 +622,7 @@ class Args:
     # could have is_concrete=False.
     if not self.starstarargs or len(self.starstarargs.data) != 1:
       return None
-    kwdict, = self.starstarargs.data
+    (kwdict,) = self.starstarargs.data
     if not _isinstance(kwdict, "Dict"):
       return None
     return kwdict.pyval
@@ -545,7 +688,8 @@ class Args:
       for var in all_args[n_params:]:
         if abstract_utils.is_var_splat(var):
           star.append(
-              abstract_utils.merged_type_parameter(node, var, abstract_utils.T))
+              abstract_utils.merged_type_parameter(node, var, abstract_utils.T)
+          )
         else:
           star.append(var)
       if star:
@@ -599,7 +743,8 @@ class Args:
           # parameterize it now to preserve them.
           params = {
               name: ctx.convert.merge_classes(
-                  kwdict.get_instance_type_parameter(name, node).data)
+                  kwdict.get_instance_type_parameter(name, node).data
+              )
               for name in (abstract_utils.K, abstract_utils.V)
           }
           cls = _make("ParameterizedClass", ctx.convert.dict_type, params, ctx)
@@ -609,11 +754,12 @@ class Args:
     starargs_as_tuple = self.starargs_as_tuple(node, ctx)
     if starargs_as_tuple is not None:
       if match_signature:
-        posargs, starargs = self._unpack_and_match_args(node, ctx,
-                                                        match_signature,
-                                                        starargs_as_tuple)
-      elif (starargs_as_tuple and
-            abstract_utils.is_var_splat(starargs_as_tuple[-1])):
+        posargs, starargs = self._unpack_and_match_args(
+            node, ctx, match_signature, starargs_as_tuple
+        )
+      elif starargs_as_tuple and abstract_utils.is_var_splat(
+          starargs_as_tuple[-1]
+      ):
         # If the last arg is an indefinite iterable keep it in starargs. Convert
         # any other splats to Any.
         # TODO(mdemello): If there are multiple splats should we just fall
@@ -627,9 +773,12 @@ class Args:
         posargs = self.posargs + _splats_to_any(starargs_as_tuple, ctx)
         starargs = None
     simplify = lambda var: abstract_utils.simplify_variable(var, node, ctx)
-    return Args(tuple(simplify(posarg) for posarg in posargs),
-                {k: simplify(namedarg) for k, namedarg in namedargs.items()},
-                simplify(starargs), simplify(starstarargs))
+    return Args(
+        tuple(simplify(posarg) for posarg in posargs),
+        {k: simplify(namedarg) for k, namedarg in namedargs.items()},
+        simplify(starargs),
+        simplify(starstarargs),
+    )
 
   def get_variables(self):
     variables = list(self.posargs) + list(self.namedargs.values())
@@ -640,7 +789,7 @@ class Args:
     return variables
 
   def replace_posarg(self, pos, val):
-    new_posargs = self.posargs[:pos] + (val,) + self.posargs[pos + 1:]
+    new_posargs = self.posargs[:pos] + (val,) + self.posargs[pos + 1 :]
     return self.replace(posargs=new_posargs)
 
   def replace_namedarg(self, name, val):
@@ -656,24 +805,10 @@ class Args:
     return attrs.evolve(self, **kwargs)
 
   def has_opaque_starargs_or_starstarargs(self):
-    return any(arg and not _isinstance(arg, "PythonConstant")
-               for arg in (self.starargs, self.starstarargs))
-
-
-class ReturnValueMixin:
-  """Mixin for exceptions that hold a return node and variable."""
-
-  def __init__(self):
-    super().__init__()
-    self.return_node = None
-    self.return_variable = None
-
-  def set_return(self, node, var):
-    self.return_node = node
-    self.return_variable = var
-
-  def get_return(self, state):
-    return state.change_cfg_node(self.return_node), self.return_variable
+    return any(
+        arg and not _isinstance(arg, "PythonConstant")
+        for arg in (self.starargs, self.starstarargs)
+    )
 
 
 class ParamSpecMatch(_base.BaseValue):
@@ -694,133 +829,6 @@ class ParamSpecMatch(_base.BaseValue):
       return ()
 
 
-# These names are chosen to match pytype error classes.
-# pylint: disable=g-bad-exception-name
-class FailedFunctionCall(Exception, ReturnValueMixin):
-  """Exception for failed function calls."""
-
-  def __init__(self):
-    super().__init__()
-    self.name = "<no name>"
-
-  def __gt__(self, other):
-    return other is None
-
-  def __le__(self, other):
-    return not self.__gt__(other)
-
-
-class NotCallable(FailedFunctionCall):
-  """For objects that don't have __call__."""
-
-  def __init__(self, obj):
-    super().__init__()
-    self.obj = obj
-
-
-class UndefinedParameterError(FailedFunctionCall):
-  """Function called with an undefined variable."""
-
-  def __init__(self, name):
-    super().__init__()
-    self.name = name
-
-
-class DictKeyMissing(Exception, ReturnValueMixin):
-  """When retrieving a key that does not exist in a dict."""
-
-  def __init__(self, name):
-    super().__init__()
-    self.name = name
-
-  def __gt__(self, other):
-    return other is None
-
-  def __le__(self, other):
-    return not self.__gt__(other)
-
-
-@dataclasses.dataclass(eq=True, frozen=True)
-class BadCall:
-  sig: Signature
-  passed_args: Sequence[Tuple[str, _base.BaseValue]]
-  bad_param: Optional[abstract_utils.BadType]
-
-
-@dataclasses.dataclass(eq=True, frozen=True)
-class Arg:
-  name: str
-  value: cfg.Variable
-  typ: _base.BaseValue
-
-
-class InvalidParameters(FailedFunctionCall):
-  """Exception for functions called with an incorrect parameter combination."""
-
-  def __init__(self, sig, passed_args, ctx, bad_param=None):
-    super().__init__()
-    self.name = sig.name
-    passed_args = [(name, ctx.convert.merge_values(arg.data))
-                   for name, arg, _ in sig.iter_args(passed_args)]
-    self.bad_call = BadCall(sig=sig, passed_args=passed_args,
-                            bad_param=bad_param)
-
-
-class WrongArgTypes(InvalidParameters):
-  """For functions that were called with the wrong types."""
-
-  def __init__(self, sig, passed_args, ctx, bad_param):
-    if not sig.has_param(bad_param.name):
-      sig = sig.insert_varargs_and_kwargs(
-          name for name, *_ in sig.iter_args(passed_args))
-    super().__init__(sig, passed_args, ctx, bad_param)
-
-  def __gt__(self, other):
-    if other is None:
-      return True
-    if not isinstance(other, WrongArgTypes):
-      # WrongArgTypes should take precedence over other FailedFunctionCall
-      # subclasses but not over unrelated errors like DictKeyMissing.
-      return isinstance(other, FailedFunctionCall)
-    # The signature that has fewer *args/**kwargs tends to be more precise.
-    def starcount(err):
-      return (bool(err.bad_call.sig.varargs_name) +
-              bool(err.bad_call.sig.kwargs_name))
-    return starcount(self) < starcount(other)
-
-  def __le__(self, other):
-    return not self.__gt__(other)
-
-
-class WrongArgCount(InvalidParameters):
-  """E.g. if a function expecting 4 parameters is called with 3."""
-
-
-class WrongKeywordArgs(InvalidParameters):
-  """E.g. an arg "x" is passed to a function that doesn't have an "x" param."""
-
-  def __init__(self, sig, passed_args, ctx, extra_keywords):
-    super().__init__(sig, passed_args, ctx)
-    self.extra_keywords = tuple(extra_keywords)
-
-
-class DuplicateKeyword(InvalidParameters):
-  """E.g. an arg "x" is passed to a function as both a posarg and a kwarg."""
-
-  def __init__(self, sig, passed_args, ctx, duplicate):
-    super().__init__(sig, passed_args, ctx)
-    self.duplicate = duplicate
-
-
-class MissingParameter(InvalidParameters):
-  """E.g. a function requires parameter 'x' but 'x' isn't passed."""
-
-  def __init__(self, sig, passed_args, ctx, missing_parameter):
-    super().__init__(sig, passed_args, ctx)
-    self.missing_parameter = missing_parameter
-# pylint: enable=g-bad-exception-name
-
-
 @dataclasses.dataclass(frozen=True)
 class Mutation:
   """A type mutation."""
@@ -830,15 +838,18 @@ class Mutation:
   value: cfg.Variable
 
   def __eq__(self, other):
-    return (self.instance == other.instance and
-            self.name == other.name and
-            frozenset(self.value.data) == frozenset(other.value.data))
+    return (
+        self.instance == other.instance
+        and self.name == other.name
+        and frozenset(self.value.data) == frozenset(other.value.data)
+    )
 
   def __hash__(self):
     return hash((self.instance, self.name, frozenset(self.value.data)))
 
 
 class _ReturnType(abc.ABC):
+  """Wrapper for a function return type."""
 
   @property
   @abc.abstractmethod
@@ -849,6 +860,10 @@ class _ReturnType(abc.ABC):
   def instantiate_parameter(self, node, param_name):
     ...
 
+  @abc.abstractmethod
+  def get_parameter(self, node, param_name):
+    ...
+
 
 class AbstractReturnType(_ReturnType):
   """An abstract return type."""
@@ -863,8 +878,10 @@ class AbstractReturnType(_ReturnType):
 
   def instantiate_parameter(self, node, param_name):
     param = self._type.get_formal_type_parameter(param_name)
-    _, param_instance = self._ctx.vm.init_class(node, param)
-    return param_instance
+    return self._ctx.vm.init_class(node, param)
+
+  def get_parameter(self, node, param_name):
+    return self._type.get_formal_type_parameter(param_name)
 
 
 class PyTDReturnType(_ReturnType):
@@ -900,30 +917,44 @@ class PyTDReturnType(_ReturnType):
           abstract_utils.AsReturnValue(self._type),
           self._subst,
           node,
-          source_sets=[self._sources])
+          source_sets=[self._sources],
+      )
     except self._ctx.convert.TypeParameterError:
       # The return type contains a type parameter without a substitution.
       subst = abstract_utils.with_empty_substitutions(
-          self._subst, self._type, node, self._ctx)
+          self._subst, self._type, node, self._ctx
+      )
       return node, self._ctx.convert.constant_to_var(
           abstract_utils.AsReturnValue(self._type),
           subst,
           node,
-          source_sets=[self._sources])
+          source_sets=[self._sources],
+      )
     if not ret.bindings and isinstance(self._type, pytd.TypeParameter):
       ret.AddBinding(self._ctx.convert.empty, [], node)
     return node, ret
 
+  def get_parameter(self, node, param_name):
+    t = self._ctx.convert.constant_to_value(self._type, self._subst, node)
+    return t.get_formal_type_parameter(param_name)
+
 
 def _splats_to_any(seq, ctx):
   return tuple(
       ctx.new_unsolvable(ctx.root_node) if abstract_utils.is_var_splat(v) else v
-      for v in seq)
+      for v in seq
+  )
 
 
 def call_function(
-    ctx, node, func_var, args, fallback_to_unsolvable=True,
-    allow_noreturn=False, strict_filter=True):
+    ctx,
+    node,
+    func_var,
+    args,
+    fallback_to_unsolvable=True,
+    allow_never=False,
+    strict_filter=True,
+):
   """Call a function.
 
   Args:
@@ -932,8 +963,9 @@ def call_function(
     func_var: A variable of the possible functions to call.
     args: The arguments to pass. See function.Args.
     fallback_to_unsolvable: If the function call fails, create an unknown.
-    allow_noreturn: Whether typing.NoReturn is allowed in the return type.
+    allow_never: Whether typing.Never is allowed in the return type.
     strict_filter: Whether function bindings should be strictly filtered.
+
   Returns:
     A tuple (CFGNode, Variable). The Variable is the return value.
   Raises:
@@ -945,25 +977,27 @@ def call_function(
   result = ctx.program.NewVariable()
   nodes = []
   error = None
-  has_noreturn = False
+  has_never = False
   for funcb in func_var.bindings:
     func = funcb.data
     one_result = None
     try:
       new_node, one_result = func.call(node, funcb, args)
-    except (DictKeyMissing, FailedFunctionCall) as e:
-      if e > error and ((not strict_filter and len(func_var.bindings) == 1) or
-                        funcb.IsVisible(node)):
+    except (error_types.DictKeyMissing, error_types.FailedFunctionCall) as e:
+      if e > error and (
+          (not strict_filter and len(func_var.bindings) == 1)
+          or funcb.IsVisible(node)
+      ):
         error = e
     else:
-      if ctx.convert.no_return in one_result.data:
-        if allow_noreturn:
-          # Make sure NoReturn was the only thing returned.
+      if ctx.convert.never in one_result.data:
+        if allow_never:
+          # Make sure Never was the only thing returned.
           assert len(one_result.data) == 1
-          has_noreturn = True
+          has_never = True
         else:
           for b in one_result.bindings:
-            if b.data != ctx.convert.no_return:
+            if b.data != ctx.convert.never:
               result.PasteBinding(b)
       else:
         result.PasteVariable(one_result, new_node, {funcb})
@@ -971,10 +1005,11 @@ def call_function(
   if nodes:
     node = ctx.join_cfg_nodes(nodes)
     if not result.bindings:
-      v = ctx.convert.no_return if has_noreturn else ctx.convert.unsolvable
+      v = ctx.convert.never if has_never else ctx.convert.unsolvable
       result.AddBinding(v, [], node)
-  elif (isinstance(error, FailedFunctionCall) and
-        all(func.name.endswith(".__init__") for func in func_var.data)):
+  elif isinstance(error, error_types.FailedFunctionCall) and all(
+      func.name.endswith(".__init__") for func in func_var.data
+  ):
     # If the function failed with a FailedFunctionCall exception, try calling
     # it again with fake arguments. This allows for calls to __init__ to
     # always succeed, ensuring pytype has a full view of the class and its
@@ -982,24 +1017,28 @@ def call_function(
     # abstract.Unsolvable.
     node, result = ctx.vm.call_with_fake_args(node, func_var)
   elif ctx.options.precise_return and len(func_var.bindings) == 1:
-    funcb, = func_var.bindings
+    (funcb,) = func_var.bindings
     func = funcb.data
     if _isinstance(func, "BoundFunction"):
       func = func.underlying
     if _isinstance(func, "PyTDFunction"):
       node, result = PyTDReturnType(
-          func.signatures[0].pytd_sig.return_type, datatypes.HashableDict(),
-          [funcb], ctx).instantiate(node)
+          func.signatures[0].pytd_sig.return_type,
+          datatypes.HashableDict(),
+          [funcb],
+          ctx,
+      ).instantiate(node)
     elif _isinstance(func, "InterpreterFunction"):
       sig = func.signature_functions()[0].signature
       ret = sig.annotations.get("return", ctx.convert.unsolvable)
-      node, result = ctx.vm.init_class(node, ret)
+      result = ctx.vm.init_class(node, ret)
     else:
       result = ctx.new_unsolvable(node)
   else:
     result = ctx.new_unsolvable(node)
-  ctx.vm.trace_opcode(None, func_var.data[0].name.rpartition(".")[-1],
-                      (func_var, result))
+  ctx.vm.trace_opcode(
+      None, func_var.data[0].name.rpartition(".")[-1], (func_var, result)
+  )
   if (nodes and not ctx.options.strict_parameter_checks) or not error:
     return node, result
   elif fallback_to_unsolvable:
@@ -1025,7 +1064,7 @@ def match_all_args(ctx, node, func, args):
       where new_args = args with all incorrectly typed values set to Any
             errors = a list of [(type mismatch error, arg name, value)]
 
-  Reraises any error that is not function.InvalidParameters
+  Reraises any error that is not InvalidParameters
   """
   positional_names = func.get_positional_names()
   needs_checking = True
@@ -1033,19 +1072,20 @@ def match_all_args(ctx, node, func, args):
   while needs_checking:
     try:
       func.match_args(node, args)
-    except FailedFunctionCall as e:
-      if isinstance(e, WrongKeywordArgs):
+    except error_types.FailedFunctionCall as e:
+      if isinstance(e, error_types.WrongKeywordArgs):
         errors.append((e, e.extra_keywords[0], None))
         for i in e.extra_keywords:
           args = args.delete_namedarg(i)
-      elif isinstance(e, DuplicateKeyword):
+      elif isinstance(e, error_types.DuplicateKeyword):
         errors.append((e, e.duplicate, None))
         args = args.delete_namedarg(e.duplicate)
-      elif isinstance(e, MissingParameter):
+      elif isinstance(e, error_types.MissingParameter):
         errors.append((e, e.missing_parameter, None))
         args = args.replace_namedarg(
-            e.missing_parameter, ctx.new_unsolvable(node))
-      elif isinstance(e, WrongArgTypes):
+            e.missing_parameter, ctx.new_unsolvable(node)
+        )
+      elif isinstance(e, error_types.WrongArgTypes):
         arg_name = e.bad_call.bad_param.name
         for name, value in e.bad_call.passed_args:
           if name != arg_name:
@@ -1084,7 +1124,7 @@ def has_visible_namedarg(node, args, names):
 
 
 def handle_typeguard(node, ret: _ReturnType, first_arg, ctx, func_name=None):
-  """Returns a variable of the return value of a typeguard function.
+  """Returns a variable of the return value of a type guard function.
 
   Args:
     node: The current node.
@@ -1095,18 +1135,39 @@ def handle_typeguard(node, ret: _ReturnType, first_arg, ctx, func_name=None):
   """
   frame = ctx.vm.frame
   if not hasattr(frame, "f_locals"):
-    return None  # no need to apply TypeGuard if we're in a dummy frame
-  if ret.name != "typing.TypeGuard":
+    return None  # no need to apply the type guard if we're in a dummy frame
+  if ret.name == "typing.TypeIs":
+    match_result = ctx.matcher(node).compute_one_match(
+        first_arg, ret.get_parameter(node, abstract_utils.T)
+    )
+    matched = [m.view[first_arg] for m in match_result.good_matches]
+    unmatched = [m.view[first_arg] for m in match_result.bad_matches]
+  elif ret.name == "typing.TypeGuard":
+    matched = []
+    unmatched = first_arg.bindings
+  else:
     return None
+  if matched:
+    # When a TypeIs function is applied to a variable with matching bindings, it
+    # behaves like isinstance(), narrowing in both positive and negative cases.
+    typeis_return = ctx.program.NewVariable()
+    for b in matched:
+      typeis_return.AddBinding(ctx.convert.true, {b}, node)
+    for b in unmatched:
+      typeis_return.AddBinding(ctx.convert.false, {b}, node)
+    return typeis_return
 
-  # Get the local/global variable that first_arg comes from, and add new
-  # bindings for the TypeGuard type.
+  # We have either a TypeIs function that does not match any existing bindings,
+  # or a TypeGuard function. Get the local/global variable that first_arg comes
+  # from, and add new bindings for the type guard type.
   target_name = ctx.vm.get_var_name(first_arg)
   if not target_name:
     desc = f" function {func_name!r}" if func_name else ""
     ctx.errorlog.not_supported_yet(
-        ctx.vm.frames, f"Calling TypeGuard{desc} with an arbitrary expression",
-        "Please assign the expression to a local variable.")
+        ctx.vm.frames,
+        f"Calling {ret.name}{desc} with an arbitrary expression",
+        "Please assign the expression to a local variable.",
+    )
     return None
   target = frame.lookup_name(target_name)
   # Forward all the target's bindings to the current node, so we don't have
@@ -1119,7 +1180,7 @@ def handle_typeguard(node, ret: _ReturnType, first_arg, ctx, func_name=None):
       target.PasteBinding(b, node)
 
   # Create a boolean return variable with True bindings for values that
-  # originate from the TypeGuard type and False for the rest.
+  # originate from the type guard type and False for the rest.
   typeguard_return = ctx.program.NewVariable()
   for b in target.bindings:
     boolvals = {b.data not in old_data} | {b.data in new_instance.data}
@@ -1147,6 +1208,7 @@ def build_paramspec_signature(pspec_match, r_args, return_value, ctx):
   param_names = tuple(ret_posargs) + sig.param_names[l_nargs:]
   # All params need to be in the annotations dict or output.py crashes
   sig.populate_annotation_dict(ann, ctx, param_names)
-  posonly_count = sig.posonly_count + len(r_args) - l_nargs
-  return sig._replace(param_names=param_names, annotations=ann,
-                      posonly_count=posonly_count)
+  posonly_count = max(sig.posonly_count + len(r_args) - l_nargs, 0)
+  return sig._replace(
+      param_names=param_names, annotations=ann, posonly_count=posonly_count
+  )
