@@ -8,6 +8,7 @@ from pytype.abstract import function
 from pytype.pytd import pytd
 from pytype.pytd import pytd_utils
 from pytype.typegraph import cfg
+from pytype.types import types
 
 log = logging.getLogger(__name__)
 _isinstance = abstract_utils._isinstance  # pylint: disable=protected-access
@@ -47,24 +48,28 @@ class MixinMeta(type):
 
     Args:
       method: The method in the mix-in.
+
     Returns:
       The method overloaded by 'method'.
     """
+    method_cls = type(method.__self__)
     # Bound methods have a __self__ attribute, but we don't have a way of
     # annotating `method` as being a bound rather than unbound method.
     # pytype: disable=attribute-error
-    for supercls in type(method.__self__).__mro__:
+    for supercls in method_cls.__mro__:
       # Fetch from __dict__ rather than using getattr() because we only want
       # to consider methods defined on supercls itself (not on a base).
-      if ("__mixin_overloads__" in supercls.__dict__ and
-          supercls.__mixin_overloads__.get(method.__name__) is cls):
+      if (
+          "__mixin_overloads__" in supercls.__dict__
+          and supercls.__mixin_overloads__.get(method.__name__) is cls
+      ):
         method_cls = supercls
         break
     return getattr(super(method_cls, method.__self__), method.__name__)
     # pytype: enable=attribute-error
 
 
-class PythonConstant(metaclass=MixinMeta):
+class PythonConstant(types.PythonConstant, metaclass=MixinMeta):
   """A mix-in for storing actual Python constants, not just their types.
 
   This is used for things that are stored in cfg.Variable, but where we
@@ -85,8 +90,8 @@ class PythonConstant(metaclass=MixinMeta):
     """Get a string representation of this constant.
 
     Args:
-      printer: A BaseValue -> str function that will be used to
-        print abstract values.
+      printer: A BaseValue -> str function that will be used to print abstract
+        values.
 
     Returns:
       A string of self.pyval.
@@ -125,7 +130,8 @@ class HasSlots(metaclass=MixinMeta):
     # can trigger a recursion error, so use only the base class.
     base = self.base_cls if _isinstance(self, "ParameterizedClass") else self
     _, attr = self.ctx.attribute_handler.get_attribute(
-        self.ctx.root_node, base, name, base.to_binding(self.ctx.root_node))
+        self.ctx.root_node, base, name, base.to_binding(self.ctx.root_node)
+    )
     self._super[name] = attr
     self._slots[name] = slot
 
@@ -140,7 +146,8 @@ class HasSlots(metaclass=MixinMeta):
         node,
         self._super[name],
         function.Args(args),
-        fallback_to_unsolvable=False)
+        fallback_to_unsolvable=False,
+    )
 
   def get_special_attribute(self, node, name, valself):
     if name not in self._slots:
@@ -173,6 +180,7 @@ class NestedAnnotation(metaclass=MixinMeta):
   def init_mixin(self):
     self.processed = False
     self._seen_for_formal = False  # for calculating the 'formal' property
+    self._formal = None
 
   @property
   def formal(self):
@@ -180,11 +188,16 @@ class NestedAnnotation(metaclass=MixinMeta):
     # We can't compute self.formal in __init__ because doing so would force
     # evaluation of our type parameters during initialization, possibly
     # leading to an infinite loop.
+    if self._formal is not None:
+      return self._formal
     if self._seen_for_formal:
       return False
     self._seen_for_formal = True
     formal = any(t.formal for _, t in self.get_inner_types())
     self._seen_for_formal = False
+    if self.ctx.vm.late_annotations is None:
+      # Caching 'formal' is safe once all LateAnnotations have been resolved.
+      self._formal = formal
     return formal
 
   def get_inner_types(self):
@@ -230,9 +243,13 @@ class LazyMembers(metaclass=MixinMeta):
     assert isinstance(variable, cfg.Variable)
     # 'subst' can vary between attribute accesses, so it's not safe to store the
     # attribute value in 'members' if it uses any of the subst keys.
-    if store and not (isinstance(member, pytd.Node) and subst and
-                      any(t.full_name in subst
-                          for t in pytd_utils.GetTypeParameters(member))):
+    if store and not (
+        isinstance(member, pytd.Node)
+        and subst
+        and any(
+            t.full_name in subst for t in pytd_utils.GetTypeParameters(member)
+        )
+    ):
       self.members[name] = variable
     return variable
 
@@ -242,6 +259,7 @@ class PythonDict(PythonConstant):
 
   Not all dict methods are implemented, such as methods for modifying the dict.
   """
+
   # This was derived from pytd_utils.WrapsDict, which used `exec` to generate
   # a custom base class. Only the base methods from WrapsDict are implemented
   # here, because those are the only ones that were being used.
