@@ -3,8 +3,7 @@
 import collections
 import dataclasses
 import itertools
-
-from typing import AbstractSet, Any, Dict, Optional, Sequence, Tuple
+from typing import Any, Dict, Optional, Sequence, Set, Tuple
 
 from pytype import state
 from pytype import utils
@@ -30,28 +29,38 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
   def sub_annotations(self, node, annotations, substs, instantiate_unbound):
     """Apply type parameter substitutions to a dictionary of annotations."""
     if substs and all(substs):
-      return {name: self.sub_one_annotation(node, annot, substs,
-                                            instantiate_unbound)
-              for name, annot in annotations.items()}
+      return {
+          name: self.sub_one_annotation(
+              node, annot, substs, instantiate_unbound
+          )
+          for name, annot in annotations.items()
+      }
     return annotations
 
   def _get_type_parameter_subst(
-      self, node: cfg.CFGNode, annot: abstract.TypeParameter,
-      substs: Sequence[Dict[str, cfg.Variable]], instantiate_unbound: bool
+      self,
+      node: cfg.CFGNode,
+      annot: abstract.TypeParameter,
+      substs: Sequence[Dict[str, cfg.Variable]],
+      instantiate_unbound: bool,
   ) -> abstract.BaseValue:
     """Helper for sub_one_annotation."""
     # We use the given substitutions to bind the annotation if
     # (1) every subst provides at least one binding, and
     # (2) none of the bindings are ambiguous, and
     # (3) at least one binding is non-empty.
-    if all(annot.full_name in subst and subst[annot.full_name].bindings
-           for subst in substs):
+    if all(
+        annot.full_name in subst and subst[annot.full_name].bindings
+        for subst in substs
+    ):
       vals = sum((subst[annot.full_name].data for subst in substs), [])
     else:
       vals = None
-    if (vals is None or
-        any(isinstance(v, abstract.AMBIGUOUS) for v in vals) or
-        all(isinstance(v, abstract.Empty) for v in vals)):
+    if (
+        vals is None
+        or any(isinstance(v, abstract.AMBIGUOUS) for v in vals)
+        or all(isinstance(v, abstract.Empty) for v in vals)
+    ):
       if instantiate_unbound:
         vals = annot.instantiate(node).data
       else:
@@ -61,8 +70,9 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
   def sub_one_annotation(self, node, annot, substs, instantiate_unbound=True):
 
     def get_type_parameter_subst(annotation):
-      return self._get_type_parameter_subst(node, annotation, substs,
-                                            instantiate_unbound)
+      return self._get_type_parameter_subst(
+          node, annotation, substs, instantiate_unbound
+      )
 
     return self._do_sub_one_annotation(node, annot, get_type_parameter_subst)
 
@@ -91,7 +101,8 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
             param_strings = []
             for t in utils.unique_list(self.get_type_parameters(cur)):
               s = pytd_utils.Print(
-                  get_type_parameter_subst_fn(t).get_instance_type(node))
+                  get_type_parameter_subst_fn(t).to_pytd_type_of_instance(node)
+              )
               param_strings.append(s)
             expr = f"{cur.expr}[{', '.join(param_strings)}]"
             late_annot = abstract.LateAnnotation(expr, cur.stack, cur.ctx)
@@ -114,10 +125,12 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
             if "[" in late_annot.expr:
               if self.ctx.vm.late_annotations is None:
                 self.ctx.vm.flatten_late_annotation(
-                    node, late_annot, self.ctx.vm.frame.f_globals)
+                    node, late_annot, self.ctx.vm.frame.f_globals
+                )
               else:
                 self.ctx.vm.late_annotations[
-                    late_annot.expr.split("[", 1)[0]].append(late_annot)
+                    late_annot.expr.split("[", 1)[0]
+                ].append(late_annot)
           done.append(done_annot)
       else:
         done.append(get_type_parameter_subst_fn(cur))
@@ -125,8 +138,9 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     return done[0]
 
   def sub_annotations_for_parameterized_class(
-      self, cls: abstract.ParameterizedClass,
-      annotations: Dict[str, abstract.BaseValue]
+      self,
+      cls: abstract.ParameterizedClass,
+      annotations: Dict[str, abstract.BaseValue],
   ) -> Dict[str, abstract.BaseValue]:
     """Apply type parameter substitutions to a dictionary of annotations.
 
@@ -141,7 +155,8 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     formal_type_parameters = cls.get_formal_type_parameters()
 
     def get_type_parameter_subst(
-        annotation: abstract.TypeParameter) -> Optional[abstract.BaseValue]:
+        annotation: abstract.BaseValue,
+    ) -> Optional[abstract.BaseValue]:
       # Normally the type parameter module is set correctly at this point.
       # Except for the case when a method that references this type parameter
       # is inherited in a subclass that does not specialize this parameter:
@@ -155,17 +170,23 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
       # In this case t in A[T].f will be annotated with T with no module set,
       # since we don't know the correct module until T is specialized in
       # B[int].
-      annotation = annotation.with_module(cls.full_name)
+      for name in (
+          f"{cls.full_name}.{annotation.name}",
+          f"{cls.name}.{annotation.name}",
+      ):
+        if name in formal_type_parameters:
+          return formal_type_parameters[name]
       # Method parameter can be annotated with a typevar that doesn't
       # belong to the class template:
       #   class A(Generic[T]):
       #     def f(self, t: U): ...
       # In this case we return it as is.
-      return formal_type_parameters.get(annotation.full_name, annotation)
+      return annotation
 
     return {
-        name: self._do_sub_one_annotation(self.ctx.root_node, annot,
-                                          get_type_parameter_subst)
+        name: self._do_sub_one_annotation(
+            self.ctx.root_node, annot, get_type_parameter_subst
+        )
         for name, annot in annotations.items()
     }
 
@@ -176,7 +197,7 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
       for _, typ in annot.get_inner_types():
         yield from self.get_late_annotations(typ)
 
-  def add_scope(self, annot, types, module, seen=None):
+  def add_scope(self, annot, types, cls, seen=None):
     """Add scope for type parameters.
 
     In original type class, all type parameters that should be added a scope
@@ -185,7 +206,7 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     Args:
       annot: The type class.
       types: A type name list that should be added a scope.
-      module: Module name.
+      cls: The class that type parameters should be scoped to.
       seen: Already seen types.
 
     Returns:
@@ -193,19 +214,24 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     """
     if seen is None:
       seen = {annot}
-    elif annot in seen:
+    elif annot in seen or not annot.formal:
       return annot
     else:
       seen.add(annot)
     if isinstance(annot, abstract.TypeParameter):
       if annot.name in types:
-        new_annot = annot.copy()
-        new_annot.module = module
-        return new_annot
-      return annot
+        return annot.with_scope(cls.full_name)
+      elif annot.full_name == "typing.Self":
+        bound_annot = annot.copy()
+        bound_annot.bound = cls
+        return bound_annot
+      else:
+        return annot
     elif isinstance(annot, mixin.NestedAnnotation):
-      inner_types = [(key, self.add_scope(typ, types, module, seen))
-                     for key, typ in annot.get_inner_types()]
+      inner_types = [
+          (key, self.add_scope(typ, types, cls, seen))
+          for key, typ in annot.get_inner_types()
+      ]
       return annot.replace(inner_types)
     return annot
 
@@ -220,7 +246,7 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
       seen: A seen set.
     """
     seen = seen or set()
-    if annot in seen:
+    if annot in seen or not annot.formal:
       return []
     if isinstance(annot, mixin.NestedAnnotation):
       # We track parameterized classes to avoid recursion errors when a class
@@ -231,12 +257,18 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     elif isinstance(annot, abstract.TupleClass):
       annots = []
       for idx in range(annot.tuple_length):
-        annots.extend(self.get_type_parameters(
-            annot.formal_type_parameters[idx], seen))
+        annots.extend(
+            self.get_type_parameters(annot.formal_type_parameters[idx], seen)
+        )
       return annots
     elif isinstance(annot, mixin.NestedAnnotation):
-      return sum((self.get_type_parameters(t, seen)
-                  for _, t in annot.get_inner_types()), [])
+      return sum(
+          (
+              self.get_type_parameters(t, seen)
+              for _, t in annot.get_inner_types()
+          ),
+          [],
+      )
     return []
 
   def get_callable_type_parameter_names(self, val: abstract.BaseValue):
@@ -246,7 +278,7 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     stack = [val]
     while stack:
       annot = stack.pop()
-      if annot in seen:
+      if annot in seen or not annot.formal:
         continue
       seen.add(annot)
       if annot.full_name == "typing.Callable":
@@ -255,8 +287,11 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
           # pytype represents Callable[[T1, T2], None] as
           # CallableClass({0: T1, 1: T2, ARGS: Union[T1, T2], RET: None}),
           # so we have to fix double-counting of argument type parameters.
-          params -= collections.Counter(self.get_type_parameters(
-              annot.formal_type_parameters[abstract_utils.ARGS]))
+          params -= collections.Counter(
+              self.get_type_parameters(
+                  annot.formal_type_parameters[abstract_utils.ARGS]
+              )
+          )
         # Type parameters that appear only once in a function signature are
         # invalid, so ignore them.
         type_params.update(p.name for p, n in params.items() if n > 1)
@@ -294,8 +329,9 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
       if t is None or abstract_utils.is_ellipsis(t):
         # '...' is an experimental "inferred type"; see b/213607272.
         continue
-      annot = self._process_one_annotation(node, t, name,
-                                           self.ctx.vm.simple_stack())
+      annot = self._process_one_annotation(
+          node, t, name, self.ctx.vm.simple_stack()
+      )
       if annot is not None:
         annotations[name] = annot
     return annotations
@@ -307,23 +343,34 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     for name, t in raw_items:
       # Don't use the parameter name, since it's often something unhelpful
       # like `0`.
-      annot = self._process_one_annotation(node, t, None,
-                                           self.ctx.vm.simple_stack())
+      annot = self._process_one_annotation(
+          node, t, None, self.ctx.vm.simple_stack()
+      )
       annotations[name] = annot or self.ctx.convert.unsolvable
     return annotations
 
   def init_annotation(self, node, name, annot, container=None, extra_key=None):
-    node, value = self.ctx.vm.init_class(
-        node, annot, container=container, extra_key=extra_key)
+    value = self.ctx.vm.init_class(
+        node, annot, container=container, extra_key=extra_key
+    )
     for d in value.data:
       d.from_annotation = name
     return node, value
 
+  def _in_class_frame(self):
+    frame = self.ctx.vm.frame
+    if not frame.func:
+      return False
+    return (
+        isinstance(frame.func.data, abstract.BoundFunction)
+        or frame.func.data.is_attribute_of_class
+    )
+
   def extract_and_init_annotation(self, node, name, var):
     """Extracts an annotation from var and instantiates it."""
     frame = self.ctx.vm.frame
     substs = frame.substs
-    if frame.func and isinstance(frame.func.data, abstract.BoundFunction):
+    if self._in_class_frame():
       self_var = frame.first_arg
       if self_var:
         # self_var is an instance of (a subclass of) the class on which
@@ -333,37 +380,55 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
         type_params = []
         defining_classes = []
         for v in self_var.data:
-          for cls in v.cls.mro:
+          v_cls = v if isinstance(v, abstract.Class) else v.cls
+          for cls in v_cls.mro:
             if cls.name == defining_cls_name:
               # Normalize type parameter names by dropping the scope.
-              type_params.extend(p.with_module(None) for p in cls.template)
+              type_params.extend(p.with_scope(None) for p in cls.template)
               defining_classes.append(cls)
               break
         self_substs = tuple(
             abstract_utils.get_type_parameter_substitutions(cls, type_params)
-            for cls in defining_classes)
+            for cls in defining_classes
+        )
         substs = abstract_utils.combine_substs(substs, self_substs)
     typ = self.extract_annotation(
         node,
         var,
         name,
         self.ctx.vm.simple_stack(),
-        allowed_type_params=set(itertools.chain(*substs)))
-    orig_typ = typ
+        allowed_type_params=set(itertools.chain(*substs)),
+    )
+    if isinstance(typ, typing_overlay.Final):
+      return typ, self.ctx.new_unsolvable(node)
+    return self._sub_and_instantiate(node, name, typ, substs)
+
+  def _sub_and_instantiate(self, node, name, typ, substs):
     if isinstance(typ, abstract.FinalAnnotation):
-      typ = typ.annotation
-    elif isinstance(typ, typing_overlay.Final):
-      typ = self.ctx.convert.unsolvable
+      t, value = self._sub_and_instantiate(node, name, typ.annotation, substs)
+      return abstract.FinalAnnotation(t, self.ctx), value
     if typ.formal:
-      typ = self.sub_one_annotation(node, typ, substs,
-                                    instantiate_unbound=False)
-    _, value = self.init_annotation(node, name, typ)
-    return orig_typ, value
+      substituted_type = self.sub_one_annotation(
+          node, typ, substs, instantiate_unbound=False
+      )
+    else:
+      substituted_type = typ
+    if typ.formal and self._in_class_frame():
+      class_substs = abstract_utils.combine_substs(
+          substs, [{"typing.Self": self.ctx.vm.frame.first_arg}]
+      )
+      type_for_value = self.sub_one_annotation(
+          node, typ, class_substs, instantiate_unbound=False
+      )
+    else:
+      type_for_value = substituted_type
+    _, value = self.init_annotation(node, name, type_for_value)
+    return substituted_type, value
 
   def apply_annotation(self, node, op, name, value):
     """If there is an annotation for the op, return its value."""
     assert op is self.ctx.vm.frame.current_opcode
-    if op.code.co_filename != self.ctx.vm.filename:
+    if op.code.filename != self.ctx.vm.filename:
       return AnnotatedValue(None, value)
     if not op.annotation:
       return AnnotatedValue(None, value)
@@ -374,11 +439,13 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     frame = self.ctx.vm.frame
     stack = self.ctx.vm.simple_stack()
     with self.ctx.vm.generate_late_annotations(stack):
-      var, errorlog = abstract_utils.eval_expr(self.ctx, node, frame.f_globals,
-                                               frame.f_locals, annot)
+      var, errorlog = abstract_utils.eval_expr(
+          self.ctx, node, frame.f_globals, frame.f_locals, annot
+      )
     if errorlog:
       self.ctx.errorlog.invalid_annotation(
-          self.ctx.vm.frames, annot, details=errorlog.details)
+          self.ctx.vm.frames, annot, details=errorlog.details
+      )
     typ, annot_val = self.extract_and_init_annotation(node, name, var)
     if isinstance(typ, typing_overlay.Final):
       # return the original value, we want type inference here.
@@ -393,8 +460,13 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
       return AnnotatedValue(typ, annot_val)
 
   def extract_annotation(
-      self, node, var, name, stack,
-      allowed_type_params: Optional[AbstractSet[str]] = None):
+      self,
+      node,
+      var,
+      name,
+      stack,
+      allowed_type_params: Optional[Set[str]] = None,
+  ):
     """Returns an annotation extracted from 'var'.
 
     Args:
@@ -416,38 +488,52 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     if not typ:
       return self.ctx.convert.unsolvable
     if typ.formal and allowed_type_params is not None:
-      allowed_type_params = (allowed_type_params |
-                             self.get_callable_type_parameter_names(typ))
-      illegal_params = [x.name for x in self.get_type_parameters(typ)
-                        if x.name not in allowed_type_params]
+      allowed_type_params = (
+          allowed_type_params | self.get_callable_type_parameter_names(typ)
+      )
+      if self.ctx.vm.frame.func and (
+          isinstance(self.ctx.vm.frame.func.data, abstract.BoundFunction)
+          or self.ctx.vm.frame.func.data.is_class_builder
+      ):
+        allowed_type_params.add("typing.Self")
+      illegal_params = []
+      for x in self.get_type_parameters(typ):
+        if not allowed_type_params.intersection([x.name, x.full_name]):
+          illegal_params.append(x.name)
       if illegal_params:
-        details = "TypeVar(s) %s not in scope" % ", ".join(
-            repr(p) for p in utils.unique_list(illegal_params))
-        if self.ctx.vm.frame.func:
-          method = self.ctx.vm.frame.func.data
-          if isinstance(method, abstract.BoundFunction):
-            desc = "class"
-            frame_name = method.name.rsplit(".", 1)[0]
-          else:
-            desc = "class" if method.is_class_builder else "method"
-            frame_name = method.name
-          details += f" for {desc} {frame_name!r}"
-        if "AnyStr" in illegal_params:
-          str_type = "Union[str, bytes]"
-          details += (f"\nNote: For all string types, use {str_type}.")
-        self.ctx.errorlog.invalid_annotation(stack, typ, details, name)
+        self._log_illegal_params(illegal_params, stack, typ, name)
         return self.ctx.convert.unsolvable
     return typ
 
+  def _log_illegal_params(self, illegal_params, stack, typ, name):
+    out_of_scope_params = utils.unique_list(illegal_params)
+    details = "TypeVar(s) %s not in scope" % ", ".join(
+        repr(p) for p in out_of_scope_params
+    )
+    if self.ctx.vm.frame.func:
+      method = self.ctx.vm.frame.func.data
+      if isinstance(method, abstract.BoundFunction):
+        desc = "class"
+        frame_name = method.name.rsplit(".", 1)[0]
+      else:
+        desc = "class" if method.is_class_builder else "method"
+        frame_name = method.name
+      details += f" for {desc} {frame_name!r}"
+    if "AnyStr" in out_of_scope_params:
+      str_type = "Union[str, bytes]"
+      details += f"\nNote: For all string types, use {str_type}."
+    self.ctx.errorlog.invalid_annotation(stack, typ, details, name)
+
   def eval_multi_arg_annotation(self, node, func, annot, stack):
     """Evaluate annotation for multiple arguments (from a type comment)."""
     args, errorlog = self._eval_expr_as_tuple(node, annot, stack)
     if errorlog:
       self.ctx.errorlog.invalid_function_type_comment(
-          stack, annot, details=errorlog.details)
+          stack, annot, details=errorlog.details
+      )
     code = func.code
     expected = code.get_arg_count()
-    names = code.co_varnames
+    names = code.varnames
 
     # This is a hack.  Specifying the type of the first arg is optional in
     # class and instance methods.  There is no way to tell at this time
@@ -463,7 +549,8 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
       self.ctx.errorlog.invalid_function_type_comment(
           stack,
           annot,
-          details="Expected %d args, %d given" % (expected, len(args)))
+          details="Expected %d args, %d given" % (expected, len(args)),
+      )
       return
     for name, arg in zip(names, args):
       resolved = self._process_one_annotation(node, arg, name, stack)
@@ -471,36 +558,48 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
         func.signature.set_annotation(name, resolved)
 
   def _process_one_annotation(
-      self, node: cfg.CFGNode, annotation: abstract.BaseValue,
+      self,
+      node: cfg.CFGNode,
+      annotation: abstract.BaseValue,
       # We require `stack` to be a tuple to make sure we pass in a frozen
       # snapshot of the frame stack, rather than the actual stack, since late
       # annotations need to snapshot the stack at time of creation in order to
       # get the right line information for error messages.
-      name: Optional[str], stack: Tuple[state.FrameType, ...]
+      name: Optional[str],
+      stack: Tuple[state.FrameType, ...],
   ) -> Optional[abstract.BaseValue]:
     """Change annotation / record errors where required."""
     if isinstance(annotation, abstract.AnnotationContainer):
       annotation = annotation.base_cls
 
     if isinstance(annotation, typing_overlay.Union):
-      self.ctx.errorlog.invalid_annotation(stack, annotation, "Needs options",
-                                           name)
-      return None
-    elif (name is not None and name != "return" and
-          annotation.full_name == "typing.TypeGuard"):
       self.ctx.errorlog.invalid_annotation(
-          stack, annotation,
-          f"{annotation.name} is only allowed as a return annotation", name)
+          stack, annotation, "Needs options", name
+      )
       return None
-    elif (isinstance(annotation, abstract.Instance) and
-          annotation.cls == self.ctx.convert.str_type):
+    elif (
+        name is not None
+        and name != "return"
+        and annotation.full_name in abstract_utils.TYPE_GUARDS
+    ):
+      self.ctx.errorlog.invalid_annotation(
+          stack,
+          annotation,
+          f"{annotation.name} is only allowed as a return annotation",
+          name,
+      )
+      return None
+    elif (
+        isinstance(annotation, abstract.Instance)
+        and annotation.cls == self.ctx.convert.str_type
+    ):
       # String annotations : Late evaluation
-      if isinstance(annotation, mixin.PythonConstant):
+      if isinstance(annotation, abstract.PythonConstant):
         expr = annotation.pyval
         if not expr:
-          self.ctx.errorlog.invalid_annotation(stack, annotation,
-                                               "Cannot be an empty string",
-                                               name)
+          self.ctx.errorlog.invalid_annotation(
+              stack, annotation, "Cannot be an empty string", name
+          )
           return None
         frame = self.ctx.vm.frame
         # Immediately try to evaluate the reference, generating LateAnnotation
@@ -512,9 +611,9 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
         #   evaluate the union immediately so we don't end up with a complex
         #   LateAnnotation, which can lead to bugs when instantiated.
         with self.ctx.vm.generate_late_annotations(stack):
-          v, errorlog = abstract_utils.eval_expr(self.ctx, node,
-                                                 frame.f_globals,
-                                                 frame.f_locals, expr)
+          v, errorlog = abstract_utils.eval_expr(
+              self.ctx, node, frame.f_globals, frame.f_locals, expr
+          )
         if errorlog:
           self.ctx.errorlog.copy_from(errorlog.errors, stack)
         if len(v.data) == 1:
@@ -529,18 +628,23 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
         return annotation
       annotation.processed = True
       for key, typ in annotation.get_inner_types():
-        if (annotation.full_name == "typing.Callable" and
-            key == abstract_utils.RET):
+        if (
+            annotation.full_name == "typing.Callable"
+            and key == abstract_utils.RET
+        ):
           inner_name = "return"
         else:
           inner_name = name
         processed = self._process_one_annotation(node, typ, inner_name, stack)
         if processed is None:
           return None
-        elif name == inner_name and processed.full_name == "typing.TypeGuard":
+        elif (
+            name == inner_name
+            and processed.full_name in abstract_utils.TYPE_GUARDS
+        ):
           self.ctx.errorlog.invalid_annotation(
-              stack, typ, f"{processed.name} is not allowed as inner type",
-              name)
+              stack, typ, f"{processed.name} is not allowed as inner type", name
+          )
           return None
         annotation.update_inner_type(key, processed)
       return annotation
@@ -557,13 +661,14 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
             abstract.FinalAnnotation,
             function.ParamSpecMatch,
             typing_overlay.Final,
-            typing_overlay.NoReturn,
+            typing_overlay.Never,
         ),
     ):
       return annotation
     else:
-      self.ctx.errorlog.invalid_annotation(stack, annotation, "Not a type",
-                                           name)
+      self.ctx.errorlog.invalid_annotation(
+          stack, annotation, "Not a type", name
+      )
       return None
 
   def _eval_expr_as_tuple(self, node, expr, stack):
@@ -574,14 +679,18 @@ class AnnotationUtils(utils.ContextWeakrefMixin):
     f_globals = self.ctx.vm.frame.f_globals
     f_locals = self.ctx.vm.frame.f_locals
     with self.ctx.vm.generate_late_annotations(stack):
-      result_var, errorlog = abstract_utils.eval_expr(self.ctx, node, f_globals,
-                                                      f_locals, expr)
+      result_var, errorlog = abstract_utils.eval_expr(
+          self.ctx, node, f_globals, f_locals, expr
+      )
     result = abstract_utils.get_atomic_value(result_var)
     # If the result is a tuple, expand it.
-    if (isinstance(result, mixin.PythonConstant) and
-        isinstance(result.pyval, tuple)):
-      return (tuple(abstract_utils.get_atomic_value(x) for x in result.pyval),
-              errorlog)
+    if isinstance(result, abstract.PythonConstant) and isinstance(
+        result.pyval, tuple
+    ):
+      return (
+          tuple(abstract_utils.get_atomic_value(x) for x in result.pyval),
+          errorlog,
+      )
     else:
       return (result,), errorlog
 
