@@ -26,8 +26,8 @@ input/output.
 from typing import Any, Dict, FrozenSet, Tuple
 
 import attrs
-
-from pytype.pyc import loadmarshal
+from pycnite import marshal as pyc_marshal
+from pytype.blocks import blocks
 from pytype.pyc import opcodes
 from pytype.pyc import pyc
 
@@ -72,6 +72,7 @@ class ConstantError(Exception):
 @attrs.define
 class _Constant:
   """A folded python constant."""
+
   typ: Tuple[str, Any]
   value: Any
   elements: Any
@@ -81,10 +82,14 @@ class _Constant:
   def tag(self):
     return self.typ[0]
 
+  def __repr__(self):
+    return repr(self.value)
+
 
 @attrs.define
 class _Collection:
   """A linear collection (e.g. list, tuple, set)."""
+
   types: FrozenSet[Any]
   values: Tuple[Any, ...]
   elements: Tuple[Any, ...]
@@ -93,6 +98,7 @@ class _Collection:
 @attrs.define
 class _Map:
   """A dictionary."""
+
   key_types: FrozenSet[Any]
   keys: Tuple[Any, ...]
   value_types: FrozenSet[Any]
@@ -117,7 +123,8 @@ class _CollectionBuilder:
     return _Collection(
         types=frozenset(self.types),
         values=tuple(reversed(self.values)),
-        elements=tuple(reversed(self.elements)))
+        elements=tuple(reversed(self.elements)),
+    )
 
 
 class _MapBuilder:
@@ -143,7 +150,8 @@ class _MapBuilder:
         keys=tuple(reversed(self.keys)),
         value_types=frozenset(self.value_types),
         values=tuple(reversed(self.values)),
-        elements=self.elements)
+        elements=self.elements,
+    )
 
 
 class _Stack:
@@ -164,8 +172,9 @@ class _Stack:
 
   def _preserve_constant(self, c):
     if c and (
-        not isinstance(c.op, opcodes.LOAD_CONST) or
-        isinstance(c.op, opcodes.BUILD_STRING)):
+        not isinstance(c.op, opcodes.LOAD_CONST)
+        or isinstance(c.op, opcodes.BUILD_STRING)
+    ):
       self.consts[id(c.op)] = c
 
   def clear(self):
@@ -257,10 +266,10 @@ class _FoldedOps:
     return f
 
 
-class _FoldConstants:
+class _FoldConstants(pyc.CodeVisitor):
   """Fold constant literals in pyc code."""
 
-  def visit_code(self, code):
+  def visit_code(self, code: blocks.OrderedCode):
     """Visit code, folding literals."""
 
     def build_tuple(tup):
@@ -277,7 +286,7 @@ class _FoldConstants:
       stack = _Stack()
       for op in block:
         if isinstance(op, opcodes.LOAD_CONST):
-          elt = code.co_consts[op.arg]
+          elt = code.consts[op.arg]
           if isinstance(elt, tuple):
             typ = build_tuple(elt)
             stack.push(_Constant(typ, elt, typ[1], op))
@@ -288,7 +297,7 @@ class _FoldConstants:
         elif isinstance(op, opcodes.BUILD_SET):
           stack.build(set, op)
         elif isinstance(op, opcodes.FORMAT_VALUE):
-          if op.arg & loadmarshal.FVS_MASK:
+          if op.arg & pyc_marshal.Flags.FVS_MASK:
             stack.build_str(2, op)
           else:
             stack.build_str(1, op)
@@ -329,13 +338,22 @@ class _FoldConstants:
             other_tag, other_et = other.typ
             if other_tag == 'tuple':
               # Deconstruct the tuple built in opcodes.LOAD_CONST above
-              other_elts = tuple(_Constant(('prim', e), v, None, other.op)
-                                 for (_, e), v in zip(other_et, other.value))
+              other_elts = tuple(
+                  _Constant(('prim', e), v, None, other.op)
+                  for (_, e), v in zip(other_et, other.value)
+              )
             elif other_tag == 'prim':
-              assert other_et == str
-              other_et = {other.typ}
-              other_elts = tuple(_Constant(('prim', str), v, None, other.op)
-                                 for v in other.value)
+              if other_et == str:
+                other_et = {other.typ}
+                other_elts = tuple(
+                    _Constant(('prim', str), v, None, other.op)
+                    for v in other.value
+                )
+              else:
+                # We have some malformed code, e.g. [*42]
+                name = other_et.__name__
+                msg = f'Value after * must be an iterable, not {name}'
+                raise ConstantError(msg, op)
             else:
               other_elts = other.elements
             typ = (tag, et | set(other_et))
@@ -356,6 +374,11 @@ class _FoldConstants:
           elements = stack.fold_args(2, op)
           if elements:
             map1, map2 = elements.elements
+            if map2.typ[0] != 'map':
+              # We have some malformed code, e.g. {**42}
+              name = map2.typ[1].__name__
+              msg = f'Value after ** must be an mapping, not {name}'
+              raise ConstantError(msg, op)
             tag1, (kt1, vt1) = map1.typ
             tag2, (kt2, vt2) = map2.typ
             assert tag1 == tag2 == 'map'
@@ -379,8 +402,10 @@ class _FoldConstants:
         if id(op) in stack.consts:
           t = stack.consts[id(op)]
           arg = t
-          pretty_arg = t
-          o = opcodes.LOAD_FOLDED_CONST(op.index, op.line, arg, pretty_arg)
+          argval = t
+          o = opcodes.LOAD_FOLDED_CONST(
+              op.index, op.line, op.endline, op.col, op.endcol, arg, argval
+          )
           o.next = op.next
           o.target = op.target
           o.block_target = op.block_target
@@ -412,7 +437,7 @@ def to_literal(typ, always_tuple=False):
   def union(params):
     ret = tuple(sorted(expand(params), key=str))
     if len(ret) == 1 and not always_tuple:
-      ret, = ret  # pylint: disable=self-assigning-variable
+      (ret,) = ret  # pylint: disable=self-assigning-variable
     return ret
 
   tag, params = typ
@@ -453,13 +478,13 @@ def from_literal(tup):
       k, v = vals
       return (tag, (union(k), union(v)))
     else:
-      vals, = vals  # pylint: disable=self-assigning-variable
+      (vals,) = vals  # pylint: disable=self-assigning-variable
       return (tag, union(vals))
   else:
     return tuple(expand(tup))
 
 
-def optimize(code):
+def fold_constants(code: blocks.OrderedCode):
   """Fold all constant literals in the bytecode into LOAD_FOLDED_CONST ops."""
   return pyc.visit(code, _FoldConstants())
 
@@ -515,7 +540,8 @@ def build_folded_type(ctx, state, const):
       # TODO(b/175443170): We should use a smaller MAX_SUBSCRIPT cutoff; this
       # behaviour is unrelated to MAX_VAR_SIZE (which limits the number of
       # distinct bindings for the overall typevar).
-      elts = elements[:MAX_VAR_SIZE] + tuple(typeconst(t) for t in params)
+      n = MAX_VAR_SIZE - len(params) - 1
+      elts = elements[:n] + tuple(typeconst(t) for t in params)
       state, vs = expand(state, elts)
       return state, ctx.convert.build_list(state.node, vs)
 
@@ -530,7 +556,7 @@ def build_folded_type(ctx, state, const):
     # issues down the line.
     str_key = ctx.convert.str_type.instantiate(node)
     if elements is not None and len(elements) < MAX_VAR_SIZE:
-      for (k, v) in elements.items():
+      for k, v in elements.items():
         _, v = build_pyval(state, v)
         k_var = ctx.convert.constant_to_var(k)
         m.setitem(node, k_var, v)
@@ -557,7 +583,7 @@ def build_folded_type(ctx, state, const):
     if const.value:
       return state, ctx.convert.constant_to_var(const.value)
     else:
-      val = ctx.convert.primitive_class_instances[params]
+      val = ctx.convert.primitive_instances[params]
       return state, val.to_variable(state.node)
   elif tag == 'list':
     return collect_list(state, params, const.elements)
