@@ -41,10 +41,14 @@ class TestFolding(test_base.UnitTest):
   def _compile(self, src, mode="exec"):
     exe = (["python" + ".".join(map(str, self.python_version))], [])
     pyc_data = compiler.compile_src_string_to_pyc_string(
-        src, filename="test_input.py", python_version=self.python_version,
-        python_exe=exe, mode=mode)
+        src,
+        filename="test_input.py",
+        python_version=self.python_version,
+        python_exe=exe,
+        mode=mode,
+    )
     code = pyc.parse_pyc_string(pyc_data)
-    code = blocks.process_code(code, self.python_version)
+    code, _ = blocks.process_code(code)
     return code
 
   def _find_load_folded(self, code):
@@ -54,7 +58,7 @@ class TestFolding(test_base.UnitTest):
     return out
 
   def _fold(self, code):
-    code = constant_folding.optimize(code)
+    code = constant_folding.fold_constants(code)
     folded = self._find_load_folded(code)
     actual = [show_op(op) for op in folded]
     return actual
@@ -65,30 +69,45 @@ class TestFolding(test_base.UnitTest):
     actual = self._fold(code)
     return actual
 
+  def check_folding_error(self, src):
+    with self.assertRaises(constant_folding.ConstantError):
+      self._process(src)
+
   def test_basic(self):
     actual = self._process("a = [1, 2, 3]")
-    self.assertCountEqual(actual, [
-        (1, ("list", int), [1, 2, 3], [int, int, int])
-    ])
+    self.assertCountEqual(
+        actual, [(1, ("list", int), [1, 2, 3], [int, int, int])]
+    )
 
   def test_union(self):
     actual = self._process("a = [1, 2, '3']")
-    self.assertCountEqual(actual, [
-        (1, ("list", (int, str)), [1, 2, "3"], [int, int, str])
-    ])
+    self.assertCountEqual(
+        actual, [(1, ("list", (int, str)), [1, 2, "3"], [int, int, str])]
+    )
 
   @test_utils.skipBeforePy((3, 9), "Test for new LIST_EXTEND opcode in 3.9")
   def test_str_to_list(self):
     actual = self._process("a = [*'abc']")
-    self.assertCountEqual(actual, [
-        (1, ("list", str), ["a", "b", "c"], [str, str, str])
-    ])
+    self.assertCountEqual(
+        actual, [(1, ("list", str), ["a", "b", "c"], [str, str, str])]
+    )
+
+  @test_utils.skipBeforePy((3, 9), "Test for new LIST_EXTEND opcode in 3.9")
+  def test_bad_extend(self):
+    with self.assertRaises(constant_folding.ConstantError):
+      self._process("a = [1, 2, *3]")
 
   def test_map(self):
     actual = self._process("a = {'x': 1, 'y': '2'}")
-    self.assertCountEqual(actual, [
-        (1, ("map", str, (int, str)), {"x": 1, "y": "2"}, {"x": int, "y": str})
-    ])
+    self.assertCountEqual(
+        actual,
+        [(
+            1,
+            ("map", str, (int, str)),
+            {"x": 1, "y": "2"},
+            {"x": int, "y": str},
+        )],
+    )
 
   def test_tuple(self):
     actual = self._process("a = (1, '2', True)")
@@ -99,48 +118,24 @@ class TestFolding(test_base.UnitTest):
     actual = self._process("a = [(1, '2', 3), (4, '5', 6)]")
     val = [(1, "2", 3), (4, "5", 6)]
     elements = [("tuple", int, str, int), ("tuple", int, str, int)]
-    self.assertCountEqual(actual, [
-        (1, ("list", ("tuple", int, str, int)), val, elements)
-    ])
+    self.assertCountEqual(
+        actual, [(1, ("list", ("tuple", int, str, int)), val, elements)]
+    )
 
   def test_list_of_varied_tuple(self):
     actual = self._process("a = [(1, '2', 3), ('4', '5', 6)]")
     val = [(1, "2", 3), ("4", "5", 6)]
-    elements = [("tuple", int, str, int),
-                ("tuple", str, str, int)]
-    self.assertCountEqual(actual, [
-        (1, ("list", (
-            ("tuple", int, str, int),
-            ("tuple", str, str, int)
-        )), val, elements)
-    ])
+    elements = [("tuple", int, str, int), ("tuple", str, str, int)]
+    self.assertCountEqual(
+        actual,
+        [(
+            1,
+            ("list", (("tuple", int, str, int), ("tuple", str, str, int))),
+            val,
+            elements,
+        )],
+    )
 
-  @test_utils.skipFromPy((3, 8), "opcode line number changed in 3.8")
-  def test_nested_pre38(self):
-    actual = self._process("""
-      a = {
-        'x': [(1, '2', 3), ('4', '5', 6)],
-        'y': [{'a': 'b'}, {'c': 'd'}],
-        ('p', 'q'): 'r'
-      }
-    """)
-    val = {
-        "x": [(1, "2", 3), ("4", "5", 6)],
-        "y": [{"a": "b"}, {"c": "d"}],
-        ("p", "q"): "r"
-    }
-    x = ("list", (
-        ("tuple", int, str, int),
-        ("tuple", str, str, int)
-    ))
-    y = ("list", ("map", str, str))
-    k = (("tuple", str, str), str)
-    elements = {"x": x, "y": y, ("p", "q"): str}
-    self.assertCountEqual(actual, [
-        (4, ("map", k, (y, x, str)), val, elements)
-    ])
-
-  @test_utils.skipBeforePy((3, 8), "opcode line number changed in 3.8")
   def test_nested(self):
     actual = self._process("""
       a = {
@@ -152,18 +147,13 @@ class TestFolding(test_base.UnitTest):
     val = {
         "x": [(1, "2", 3), ("4", "5", 6)],
         "y": [{"a": "b"}, {"c": "d"}],
-        ("p", "q"): "r"
+        ("p", "q"): "r",
     }
-    x = ("list", (
-        ("tuple", int, str, int),
-        ("tuple", str, str, int)
-    ))
+    x = ("list", (("tuple", int, str, int), ("tuple", str, str, int)))
     y = ("list", ("map", str, str))
     k = (("tuple", str, str), str)
     elements = {"x": x, "y": y, ("p", "q"): str}
-    self.assertCountEqual(actual, [
-        (1, ("map", k, (y, x, str)), val, elements)
-    ])
+    self.assertCountEqual(actual, [(1, ("map", k, (y, x, str)), val, elements)])
 
   def test_partial(self):
     actual = self._process("""
@@ -175,9 +165,9 @@ class TestFolding(test_base.UnitTest):
     """)
     val = [{"a": "b"}, {"c": "d"}]
     map_type = ("map", str, str)
-    self.assertCountEqual(actual, [
-        (4, ("list", map_type), val, [map_type, map_type])
-    ])
+    self.assertCountEqual(
+        actual, [(4, ("list", map_type), val, [map_type, map_type])]
+    )
 
   def test_nested_partial(self):
     # Test that partial expressions get cleaned off the stack properly. The 'if'
@@ -219,9 +209,13 @@ class TestFolding(test_base.UnitTest):
     """)
     self.assertCountEqual(actual, [(x, str, "", None) for x in (4, 6, 7)])
 
-  def test_type_error(self):
-    with self.assertRaises(constant_folding.ConstantError):
-      self._process("x = {[1, 2]}")
+  @test_utils.skipBeforePy((3, 9), "Depends on 3.9+ bytecode")
+  def test_errors(self):
+    # Check that malformed constants raise a ConstantFoldingError rather than
+    # crash pytype.
+    self.check_folding_error("{[1, 2]}")
+    self.check_folding_error("[*42]")
+    self.check_folding_error("{**42}")
 
 
 class TypeBuilderTestBase(test_base.UnitTest):
@@ -233,7 +227,7 @@ class TypeBuilderTestBase(test_base.UnitTest):
     self.ctx = test_utils.make_context(options)
 
   def assertPytd(self, val, expected):
-    pytd_tree = val.to_type()
+    pytd_tree = val.to_pytd_type()
     pytd_tree = pytd_tree.Visit(visitors.CanonicalOrderingVisitor())
     actual = pytd_utils.Print(pytd_tree)
     self.assertEqual(actual, expected)
@@ -250,13 +244,15 @@ class TypeBuilderTest(TypeBuilderTestBase):
     typ = constant_folding.from_literal(typ)
     const = constant_folding._Constant(typ, None, None, None)
     _, var = constant_folding.build_folded_type(self.ctx, self.state, const)
-    val, = var.data
+    (val,) = var.data
     return val
 
   def _is_primitive(self, val, cls):
-    return (isinstance(val, abstract.Instance) and
-            isinstance(val.cls, abstract.PyTDClass) and
-            val.cls.pytd_cls.name == "builtins." + cls)
+    return (
+        isinstance(val, abstract.Instance)
+        and isinstance(val.cls, abstract.PyTDClass)
+        and val.cls.pytd_cls.name == "builtins." + cls
+    )
 
   def test_prim(self):
     val = self._convert(("prim", str))
@@ -327,10 +323,10 @@ class PyvalTest(TypeBuilderTestBase):
     d = defs["d"].data[0]
     t1 = "List[int]"
     t2 = "List[str]"
-    self.assertPytd(a, "List[Any]")
+    self.assertPytd(a, "List[Union[List[int], List[str]]]")
     self.assertPytd(b, t1)
     self.assertPytd(c, t2)
-    self.assertPytd(d, "List[Any]")
+    self.assertPytd(d, "List[Union[List[int], List[str]]]")
 
   def test_long_list_of_tuples(self):
     elts = ["  (1, 2),", "  ('a', False),"] * 82
