@@ -8,15 +8,16 @@ from pytype import annotation_utils
 from pytype import attribute
 from pytype import config
 from pytype import convert
-from pytype import errors
 from pytype import load_pytd
 from pytype import matcher
 from pytype import output
+from pytype import pretty_printer
 from pytype import tracer_vm
 from pytype import vm_utils
 from pytype.abstract import abstract
 from pytype.abstract import abstract_utils
 from pytype.abstract import function
+from pytype.errors import errors
 from pytype.overlays import special_builtins
 from pytype.typegraph import cfg
 from pytype.typegraph import cfg_utils
@@ -31,15 +32,13 @@ class Context:
       self,
       options: config.Options,
       loader: load_pytd.Loader,
-      generate_unknowns: bool = False,
-      store_all_calls: bool = False,
+      src: str,
   ):
     # Inputs
     self.options = options
-    self.python_version: Tuple[int, int] = self.options.python_version
+    self.python_version: Tuple[int, int] = options.python_version
     self.loader = loader
-    self.generate_unknowns = generate_unknowns
-    self.store_all_calls = store_all_calls
+    self.generate_unknowns = not options.check and options.protocols
 
     # Typegraph
     self.program = cfg.Program()
@@ -50,7 +49,7 @@ class Context:
 
     # Helper classes/modules
     self.vm = tracer_vm.CallTracer(self)
-    self.errorlog = errors.ErrorLog()
+    self.errorlog = errors.VmErrorLog(pretty_printer.PrettyPrinter(self), src)
     self.annotation_utils = annotation_utils.AnnotationUtils(self)
     self.attribute_handler = attribute.AbstractAttributeHandler(self)
     self.converter_minimally_initialized = False
@@ -67,9 +66,7 @@ class Context:
         # The object type.
         "object": self.convert.object_type,
         # for more pretty branching tests.
-        "__random__": self.convert.primitive_class_instances[bool],
-        # for debugging
-        "reveal_type": special_builtins.RevealType(self),
+        "__random__": self.convert.primitive_instances[bool],
         # boolean values.
         "True": self.convert.true,
         "False": self.convert.false,
@@ -82,7 +79,7 @@ class Context:
         special_builtins.StaticMethod,
         special_builtins.Type,
     ):
-      instance = cls(self)
+      instance = cls.make(self)
       self.special_builtins[instance.name] = instance
     # builtin functions
     for cls in (
@@ -93,15 +90,23 @@ class Context:
         special_builtins.IsInstance,
         special_builtins.IsSubclass,
         special_builtins.Next,
+        special_builtins.RevealType,
+        special_builtins.Round,
     ):
-      self.special_builtins[cls.name] = cls.make(self)
+      instance = cls.make(self)
+      self.special_builtins[instance.name] = instance
     # If set, allow construction of recursive values, setting the
     # self-referential field to Any
     self.recursion_allowed = False
     # Map from classes to maps from names of the instance methods
     # of the class to their signatures.
-    self.method_signature_map: Dict[abstract.Class,
-                                    Dict[str, function.Signature]] = {}
+    self.method_signature_map: Dict[
+        abstract.Class, Dict[str, function.Signature]
+    ] = {}
+    # Cache used by InterpreterFunction.
+    # There isn't a more appropriate place to put it while ensuring that the
+    # cache doesn't persist between runs.
+    self.function_cache = {}
 
   def matcher(self, node):
     return matcher.AbstractMatcher(node, self)
@@ -148,7 +153,8 @@ class Context:
     return vm_utils.make_class(node, props, self)
 
   def check_annotation_type_mismatch(
-      self, node, name, typ, value, stack, allow_none, details=None):
+      self, node, name, typ, value, stack, allow_none, details=None
+  ):
     """Checks for a mismatch between a variable's annotation and value.
 
     Args:
@@ -162,15 +168,24 @@ class Context:
     """
     if not typ or not value:
       return
-    if (value.data == [self.convert.ellipsis] or
-        allow_none and value.data == [self.convert.none]):
+    if (
+        value.data == [self.convert.ellipsis]
+        or allow_none
+        and value.data == [self.convert.none]
+    ):
       return
     contained_type = abstract_utils.match_type_container(
-        typ, ("typing.ClassVar", "dataclasses.InitVar"))
+        typ, ("typing.ClassVar", "dataclasses.InitVar")
+    )
     if contained_type:
       typ = contained_type
     bad = self.matcher(node).compute_one_match(value, typ).bad_matches
     for match in bad:
       self.errorlog.annotation_type_mismatch(
-          stack, match.expected.typ, match.actual_binding, name,
-          match.error_details, details)
+          stack,
+          match.expected.typ,
+          match.actual_binding,
+          name,
+          match.error_details,
+          details,
+      )
