@@ -2,13 +2,12 @@
 
 import bisect
 import collections
-import dataclasses
 import logging
 import sys
-from typing import AbstractSet, Optional
+
+from typing import Set
 
 from pytype import config
-from pytype.blocks import blocks
 from pytype.directors import annotations
 
 # directors.parser uses the stdlib ast library, which is much faster than
@@ -29,6 +28,8 @@ _ALL_ERRORS = "*"  # Wildcard for disabling all errors.
 
 _ALLOWED_FEATURES = frozenset(x.flag for x in config.FEATURE_FLAGS)
 
+_PRAGMAS = frozenset({"cache-return"})
+
 _FUNCTION_CALL_ERRORS = frozenset((
     # A function call may implicitly access a magic method attribute.
     "attribute-error",
@@ -90,8 +91,8 @@ class _LineSet:
 
     Args:
       line: A line number.
-      membership: If True, lines >= line are included in the set (starting
-        a range), otherwise they are excluded (ending a range).
+      membership: If True, lines >= line are included in the set (starting a
+        range), otherwise they are excluded (ending a range).
 
     Raises:
       ValueError: if line is less than that of a previous call to start_range().
@@ -157,8 +158,9 @@ class _BlockRanges:
       else:
         # Skip nested intervals
         while (
-            1 < i <= num_intervals and
-            self._start_to_end[self._starts[i - 1]] < line):
+            1 < i <= num_intervals
+            and self._start_to_end[self._starts[i - 1]] < line
+        ):
           i -= 1
         start = self._starts[i - 1]
       end = self._start_to_end[start]
@@ -173,86 +175,10 @@ class _BlockRanges:
     self._end_to_start[new_end] = start
 
 
-def _collect_bytecode(ordered_code):
-  bytecode_blocks = []
-  stack = [ordered_code]
-  while stack:
-    code = stack.pop()
-    bytecode_blocks.append(code.original_co_code)
-    for const in code.co_consts:
-      if isinstance(const, blocks.OrderedCode):
-        stack.append(const)
-  return bytecode_blocks
-
-
-def _adjust_line_number(line, allowed_lines, min_line):
-  adjusted_line = line
-  while adjusted_line not in allowed_lines and adjusted_line >= min_line:
-    adjusted_line -= 1
-  return adjusted_line if adjusted_line >= min_line else None
-
-
-def _is_function_call(opcode_name):
-  return opcode_name.startswith("CALL_") or opcode_name in {
-      "BINARY_SUBSCR",
-      "COMPARE_OP",
-      "FOR_ITER",
-  }
-
-
-def _is_load_attribute_op(opcode_name):
-  """Checks whether the opcode loads an attribute."""
-  return (opcode_name.startswith("GET_") or
-          opcode_name.startswith("UNPACK_") or
-          opcode_name in {
-              "LOAD_ATTR",
-              "LOAD_METHOD",
-              "SETUP_WITH",
-          })
-
-
-def _is_return_op(opcode_name):
-  return opcode_name.startswith("YIELD_") or opcode_name == "RETURN_VALUE"
-
-
-@dataclasses.dataclass
-class _OpcodeLines:
-  """Stores opcode line numbers for Director.adjust_line_numbers()."""
-
-  store_lines: AbstractSet[int]
-  make_function_lines: AbstractSet[int]
-  load_attr_lines: AbstractSet[int]
-  return_lines: AbstractSet[int]
-  call_lines: AbstractSet[int]
-
-  @classmethod
-  def from_code(cls, code):
-    """Builds an _OpcodeLines from a code object."""
-    store_lines = set()
-    make_function_lines = set()
-    load_attr_lines = set()
-    return_lines = set()
-    call_lines = set()
-    for block in _collect_bytecode(code):
-      for opcode in block:
-        if opcode.name.startswith("STORE_"):
-          store_lines.add(opcode.line)
-        elif opcode.name == "MAKE_FUNCTION":
-          make_function_lines.add(opcode.line)
-        elif _is_load_attribute_op(opcode.name):
-          load_attr_lines.add(opcode.line)
-        elif _is_return_op(opcode.name):
-          return_lines.add(opcode.line)
-        elif _is_function_call(opcode.name):
-          call_lines.add(opcode.line)
-    return cls(store_lines, make_function_lines, load_attr_lines, return_lines,
-               call_lines)
-
-
 class Director:
   """Holds all of the directive information for a source file."""
 
-  def __init__(self, src_tree, errorlog, filename, disable, code):
+  def __init__(self, src_tree, errorlog, filename, disable):
     """Create a Director for a source file.
 
     Args:
@@ -261,26 +187,24 @@ class Director:
         errorlog.
       filename: The name of the source file.
       disable: List of error messages to always ignore.
-      code: Optionally, bytecode for adjusting line numbers. If provided,
-        directives will be moved to lines at which corresponding opcodes are
-        present. Otherwise, directives will be moved to the starting line of
-        their containing statement.
     """
     self._filename = filename
     self._errorlog = errorlog
     # Collects type comments and variable annotations by line number
     self._variable_annotations = annotations.VariableAnnotations()
+    self._param_annotations = None
     # Lines that have "type: ignore".  These will disable all errors, and in
     # the future may have other impact (such as not attempting an import).
     self._ignore = _LineSet()
     # Map from error name to lines for which that error is disabled.  Note
     # that _ALL_ERRORS is essentially a wildcard name (it matches all names).
     self._disables = collections.defaultdict(_LineSet)
-    # Line numbers of decorators. Since this is used to mark a class or function
-    # as decorated, stacked decorators will record the one closest to the
-    # definition (i.e. the last one). The python bytecode uses this line number
-    # for all the stacked decorator invocations.
-    self._decorators = set()
+    # Map from pragma to lines for which that pragma is set
+    self._pragmas = collections.defaultdict(_LineSet)
+    # Function line number -> decorators map.
+    self._decorators = collections.defaultdict(list)
+    # Decorator line number -> decorated function line number map.
+    self._decorated_functions = {}
     # Apply global disable, from the command line arguments:
     for error_name in disable:
       self._disables[error_name].start_range(0, True)
@@ -290,7 +214,7 @@ class Director:
     self.block_returns = None
     self._function_ranges = _BlockRanges({})
     # Parse the source code for directives.
-    self._parse_src_tree(src_tree, code)
+    self._parse_src_tree(src_tree)
 
   @property
   def type_comments(self):
@@ -300,6 +224,14 @@ class Director:
   def annotations(self):
     return self._variable_annotations.annotations
 
+  @property
+  def param_annotations(self):
+    ret = {}
+    for a in self._param_annotations:
+      for i in range(a.start_line, a.end_line):
+        ret[i] = a.annotations
+    return ret
+
   @property
   def ignore(self):
     return self._ignore
@@ -308,98 +240,79 @@ class Director:
   def decorators(self):
     return self._decorators
 
-  def _parse_src_tree(self, src_tree, code):
+  @property
+  def decorated_functions(self):
+    return self._decorated_functions
+
+  def has_pragma(self, pragma, line):
+    return pragma in self._pragmas and line in self._pragmas[pragma]
+
+  def _parse_src_tree(self, src_tree):
     """Parse a source file, extracting directives from comments."""
     visitor = parser.visit_src_tree(src_tree)
     # TODO(rechen): This check can be removed once parser_libcst is gone.
     if not visitor:
       return
-    if code:
-      opcode_lines = _OpcodeLines.from_code(code)
-    else:
-      opcode_lines = None
 
     self.block_returns = visitor.block_returns
     self.return_lines = visitor.block_returns.all_returns()
     self._function_ranges = _BlockRanges(visitor.function_ranges)
+    self._param_annotations = visitor.param_annotations
     self.matches = visitor.matches
     self.features = set()
 
     for line_range, group in visitor.structured_comment_groups.items():
       for comment in group:
         if comment.tool == "type":
-          self._process_type(comment.line, comment.data, comment.open_ended,
-                             line_range, opcode_lines)
+          self._process_type(
+              comment.line, comment.data, comment.open_ended, line_range
+          )
         else:
           assert comment.tool == "pytype"
           try:
-            self._process_pytype(comment.line, comment.data, comment.open_ended,
-                                 line_range, opcode_lines)
+            self._process_pytype(
+                comment.line, comment.data, comment.open_ended, line_range
+            )
           except _DirectiveError as e:
             self._errorlog.invalid_directive(
-                self._filename, comment.line, str(e))
+                self._filename, comment.line, str(e)
+            )
         # Make sure the function range ends at the last "interesting" line.
-        if (not isinstance(line_range, parser.Call) and
-            self._function_ranges.has_end(line_range.end_line)):
-          if opcode_lines:
-            end = _adjust_line_number(
-                line_range.end_line, opcode_lines.return_lines,
-                line_range.start_line)
-          else:
-            end = line_range.start_line
+        if not isinstance(
+            line_range, parser.Call
+        ) and self._function_ranges.has_end(line_range.end_line):
+          end = line_range.start_line
           self._function_ranges.adjust_end(line_range.end_line, end)
 
     for annot in visitor.variable_annotations:
-      if opcode_lines:
-        final_line = _adjust_line_number(
-            annot.end_line, opcode_lines.store_lines, annot.start_line)
-        if not final_line:
-          log.error("No STORE_* opcode found for annotation %r on line %d",
-                    annot, annot.end_line)
-          continue
-      else:
-        final_line = annot.start_line
       self._variable_annotations.add_annotation(
-          final_line, annot.name, annot.annotation)
+          annot.start_line, annot.name, annot.annotation
+      )
 
-    for decorator in visitor.decorators:
-      # The MAKE_FUNCTION opcode is usually at the 'def' line but pre-3.8 was
-      # sometimes somewhere in the last decorator's line range.
-      if opcode_lines:
-        final_line = _adjust_line_number(
-            decorator.end_line, opcode_lines.make_function_lines,
-            decorator.start_line)
-        if not final_line:
-          log.warning("No MAKE_FUNCTION opcode found for decorator on line %d",
-                      decorator.end_line)
-          continue
-      else:
-        final_line = decorator.end_line
-      self._decorators.add(final_line)
+    for line, decorators in visitor.decorators.items():
+      for decorator_lineno, decorator_name in decorators:
+        self._decorators[line].append(decorator_name)
+        self._decorated_functions[decorator_lineno] = line
 
     if visitor.defs_start is not None:
       disables = list(self._disables.items())
       # Add "# type: ignore" to the list of disables that we check.
       disables.append(("Type checking", self._ignore))
       for name, lineset in disables:
-        lineno = lineset.get_disable_after(visitor.defs_start)
-        if lineno is not None:
-          self._errorlog.late_directive(self._filename, lineno, name)
+        line = lineset.get_disable_after(visitor.defs_start)
+        if line is not None:
+          self._errorlog.late_directive(self._filename, line, name)
 
   def _process_type(
-      self, line: int, data: str, open_ended: bool,
-      line_range: parser.LineRange, opcode_lines: Optional[_OpcodeLines]):
+      self, line: int, data: str, open_ended: bool, line_range: parser.LineRange
+  ):
     """Process a type: comment."""
     is_ignore = parser.IGNORE_RE.match(data)
     if not is_ignore and line != line_range.end_line:
       # Warn and discard type comments placed in the middle of expressions.
       self._errorlog.ignored_type_comment(self._filename, line, data)
       return
-    if opcode_lines:
-      final_line = _adjust_line_number(
-          line, opcode_lines.store_lines, line_range.start_line) or line
-    else:
-      final_line = line_range.start_line
+    final_line = line_range.start_line
     if is_ignore:
       if open_ended:
         self._ignore.start_range(line, True)
@@ -411,12 +324,13 @@ class Director:
         # If we have multiple type comments on the same line, take the last one,
         # but add an error to the log.
         self._errorlog.invalid_directive(
-            self._filename, line, "Multiple type comments on the same line.")
+            self._filename, line, "Multiple type comments on the same line."
+        )
       self._variable_annotations.add_type_comment(final_line, data)
 
   def _process_pytype(
-      self, line: int, data: str, open_ended: bool,
-      line_range: parser.LineRange, opcode_lines: Optional[_OpcodeLines]):
+      self, line: int, data: str, open_ended: bool, line_range: parser.LineRange
+  ):
     """Process a pytype: comment."""
     if not data:
       raise _DirectiveError("Invalid directive syntax.")
@@ -428,81 +342,97 @@ class Director:
       except ValueError as e:
         raise _DirectiveError("Invalid directive syntax.") from e
       # Additional commands may be added in the future.  For now, only
-      # "disable", "enable", and "features" are supported.
+      # "disable", "enable", "pragma", and "features" are supported.
+      values = set(values)
       if command == "disable":
-        disable = True
+        self._process_disable(
+            line, line_range, open_ended, values, disable=True
+        )
       elif command == "enable":
-        disable = False
+        self._process_disable(
+            line, line_range, open_ended, values, disable=False
+        )
+      elif command == "pragma":
+        self._process_pragmas(line, line_range, values)
       elif command == "features":
-        features = set(values)
-        invalid = features - _ALLOWED_FEATURES
-        if invalid:
-          raise _DirectiveError(f"Unknown pytype features: {','.join(invalid)}")
-        self.features |= features
+        self._process_features(values)
         continue
       else:
         raise _DirectiveError(f"Unknown pytype directive: '{command}'")
-      if not values:
-        raise _DirectiveError(
-            "Disable/enable must specify one or more error names.")
 
-      def keep(error_name):
-        if isinstance(line_range, parser.Call):
-          return error_name in _FUNCTION_CALL_ERRORS
-        else:
-          return True
+  def _process_features(self, features: Set[str]):
+    invalid = features - _ALLOWED_FEATURES
+    if invalid:
+      raise _DirectiveError(f"Unknown pytype features: {','.join(invalid)}")
+    self.features |= features
 
-      for error_name in values:
-        if (error_name == _ALL_ERRORS or
-            self._errorlog.is_valid_error_name(error_name)):
-          if not keep(error_name):
-            # Skip the directive if we are in a line range that is irrelevant to
-            # it. (Every directive is also recorded in a base LineRange that is
-            # never skipped.)
-            continue
-          lines = self._disables[error_name]
-          if open_ended:
-            lines.start_range(line, disable)
-          else:
-            final_line = self._adjust_line_number_for_pytype_directive(
-                line, error_name, line_range, opcode_lines)
-            if final_line != line:
-              # Set the disable on the original line so that, even if we mess up
-              # adjusting the line number, silencing an error by adding a
-              # disable to the exact line the error is reported on always works.
-              lines.set_line(line, disable)
-            lines.set_line(final_line, disable)
+  def _process_pragmas(
+      self, line: int, line_range: parser.LineRange, pragmas: Set[str]
+  ):
+    del line_range  # unused
+    invalid = pragmas - _PRAGMAS
+    if invalid:
+      raise _DirectiveError(f"Unknown pytype pragmas: {','.join(invalid)}")
+    for pragma in pragmas:
+      lines = self._pragmas[pragma]
+      lines.set_line(line, True)
+
+  def _process_disable(
+      self,
+      line: int,
+      line_range: parser.LineRange,
+      open_ended: bool,
+      values: Set[str],
+      *,
+      disable: bool,
+  ):
+    """Process enable/disable directives."""
+
+    def keep(error_name):
+      if isinstance(line_range, parser.Call):
+        return error_name in _FUNCTION_CALL_ERRORS
+      else:
+        return True
+
+    if not values:
+      raise _DirectiveError(
+          "Disable/enable must specify one or more error names."
+      )
+
+    for error_name in values:
+      if error_name == _ALL_ERRORS or self._errorlog.is_valid_error_name(
+          error_name
+      ):
+        if not keep(error_name):
+          # Skip the directive if we are in a line range that is irrelevant to
+          # it. (Every directive is also recorded in a base LineRange that is
+          # never skipped.)
+          continue
+        lines = self._disables[error_name]
+        if open_ended:
+          lines.start_range(line, disable)
         else:
-          self._errorlog.invalid_directive(
-              self._filename, line, f"Invalid error name: '{error_name}'")
+          final_line = self._adjust_line_number_for_pytype_directive(
+              line, error_name, line_range
+          )
+          if final_line != line:
+            # Set the disable on the original line so that, even if we mess up
+            # adjusting the line number, silencing an error by adding a
+            # disable to the exact line the error is reported on always works.
+            lines.set_line(line, disable)
+          lines.set_line(final_line, disable)
+      else:
+        self._errorlog.invalid_directive(
+            self._filename, line, f"Invalid error name: '{error_name}'"
+        )
 
   def _adjust_line_number_for_pytype_directive(
-      self, line: int, error_class: str, line_range: parser.LineRange,
-      opcode_lines: Optional[_OpcodeLines]):
+      self, line: int, error_class: str, line_range: parser.LineRange
+  ):
     """Adjusts the line number for a pytype directive."""
     if error_class not in _ALL_ADJUSTABLE_ERRORS:
       return line
-    if not opcode_lines:
-      return line_range.start_line
-    if error_class == "annotation-type-mismatch":
-      allowed_lines = (
-          opcode_lines.store_lines | opcode_lines.make_function_lines)
-    elif error_class == "attribute-error":
-      allowed_lines = opcode_lines.load_attr_lines
-    elif error_class == "bad-return-type":
-      allowed_lines = opcode_lines.return_lines
-    elif error_class == "bad-yield-annotation":
-      allowed_lines = opcode_lines.make_function_lines
-    elif error_class == "invalid-annotation":
-      allowed_lines = opcode_lines.make_function_lines
-    elif error_class == "not-supported-yet":
-      allowed_lines = opcode_lines.store_lines
-    elif error_class == "unsupported-operands":
-      allowed_lines = opcode_lines.store_lines | opcode_lines.call_lines
-    else:
-      allowed_lines = opcode_lines.call_lines
-    return _adjust_line_number(
-        line, allowed_lines, line_range.start_line) or line
+    return line_range.start_line
 
   def filter_error(self, error):
     """Return whether the error should be logged.
@@ -517,19 +447,23 @@ class Director:
     """
     # Always report errors that aren't for this file or do not have a line
     # number.
-    if error.filename != self._filename or error.lineno is None:
+    if error.filename != self._filename or error.line is None:
       return True
-    if (error.name == "bad-return-type" and
-        error.opcode_name == "RETURN_VALUE" and
-        error.lineno not in self.return_lines):
+    if (
+        error.name == "bad-return-type"
+        and error.opcode_name in ("RETURN_VALUE", "RETURN_CONST")
+        and error.line not in self.return_lines
+    ):
       # We have an implicit "return None". Adjust the line number to the last
       # line of the function.
-      _, end = self._function_ranges.find_outermost(error.lineno)
+      _, end = self._function_ranges.find_outermost(error.line)
       if end:
-        error.set_lineno(end)
+        error.set_line(end)
     # Treat line=0 as below the file, so we can filter it.
-    line = error.lineno or sys.maxsize
+    line = error.line or sys.maxsize
     # Report the error if it isn't subject to any ignore or disable.
-    return (line not in self._ignore and
-            line not in self._disables[_ALL_ERRORS] and
-            line not in self._disables[error.name])
+    return (
+        line not in self._ignore
+        and line not in self._disables[_ALL_ERRORS]
+        and line not in self._disables[error.name]
+    )
