@@ -3,6 +3,7 @@
 import contextlib
 import errno
 import os
+import re
 import sys
 
 from pytype.platform_utils import path_utils
@@ -50,6 +51,7 @@ def cd(path):
 
   Arguments:
     path: The directory to change to. If empty, this function is a no-op.
+
   Yields:
     Executes your code, in a changed directory.
   """
@@ -108,16 +110,21 @@ def expand_source_files(filenames, cwd=None):
   Args:
     filenames: A space-separated string of filenames to process.
     cwd: An optional working directory to expand relative paths
+
   Returns:
     A set of full paths to .py files
   """
   out = []
   for f in expand_globpaths(filenames.split(), cwd):
     if path_utils.isdir(f):
-      # If we have a directory, collect all the .py files within it.
+      # If we have a directory, collect all the .py files within it....
       out += recursive_glob(path_utils.join(f, "**", "*.py"))
     elif f.endswith(".py"):
       out.append(f)
+    elif is_file_script(f, cwd):
+      # .....and only process scripts when specfiied by the user.
+      out.append(f)
+
   return set(out)
 
 
@@ -125,7 +132,8 @@ def expand_pythonpath(pythonpath, cwd=None):
   """Expand a/b:c/d into [/path/to/a/b, /path/to/c/d]."""
   if pythonpath:
     return expand_paths(
-        (path.strip() for path in pythonpath.split(os.pathsep)), cwd)
+        (path.strip() for path in pythonpath.split(os.pathsep)), cwd
+    )
   else:
     return []
 
@@ -136,3 +144,16 @@ def replace_separator(path: str):
     return path.replace("/", os.path.sep).replace(":", os.pathsep)
   else:
     return path
+
+
+def is_file_script(filename, directory=None):
+  # This is for python files that do not have the .py extension
+  # of course we assume that they start with a shebang
+  file_path = expand_path(filename, directory)
+  if path_utils.isfile(file_path):
+    with open(file_path, "r") as fi:
+      try:
+        line = fi.readline().rstrip().lower()
+      except UnicodeDecodeError:
+        return False
+      return re.fullmatch(r"#!.+python3?", line) is not None
