@@ -1,8 +1,7 @@
 """Load module type information from the filesystem."""
 
 import logging
-
-from typing import Optional, Tuple
+from typing import Optional, Set, Tuple
 
 from pytype import config
 from pytype import file_utils
@@ -20,8 +19,9 @@ class _PathFinder:
 
   def __init__(self, options: config.Options):
     self.options = options
+    self.accessed_imports_paths: Set[str] = set()
 
-  def find_import(self, module_name: str) -> Tuple[Optional[str], bool]:
+  def find_import(self, module_name: str) -> Optional[Tuple[str, bool]]:
     """Search through pythonpath for a module.
 
     Loops over self.options.pythonpath, taking care of the semantics for
@@ -34,7 +34,7 @@ class _PathFinder:
     Returns:
       - (path, file_exists) if we find a path (file_exists will be false if we
         have found a directory where we need to create an __init__.pyi)
-      - (None, None) if we cannot find a full path
+      - None if we cannot find a full path
     """
     module_name_split = module_name.split(".")
     for searchdir in self.options.pythonpath:
@@ -49,8 +49,9 @@ class _PathFinder:
       elif self.options.imports_map is None and path_utils.isdir(path):
         # We allow directories to not have an __init__ file.
         # The module's empty, but you can still load submodules.
-        log.debug("Created empty module %r with path %r",
-                  module_name, init_path)
+        log.debug(
+            "Created empty module %r with path %r", module_name, init_path
+        )
         full_path = path_utils.join(path, "__init__.pyi")
         return full_path, False
       else:  # Not a directory
@@ -58,10 +59,17 @@ class _PathFinder:
         if full_path is not None:
           log.debug("Found module %r in path %r", module_name, path)
           return full_path, True
-    return None, None
+    return None
 
   def get_pyi_path(self, path: str) -> Optional[str]:
     """Get a pyi file from path if it exists."""
+    path = self._get_pyi_path_no_access_audit(path)
+    if path is not None:
+      self.accessed_imports_paths.add(path)
+    return path
+
+  def _get_pyi_path_no_access_audit(self, path: str) -> Optional[str]:
+    """Get a pyi file, without recording that it was accessed."""
     if self.options.imports_map is not None:
       if path in self.options.imports_map:
         full_path = self.options.imports_map[path]
@@ -87,23 +95,28 @@ class ModuleLoader(base.ModuleLoader):
 
   def find_import(self, module_name: str) -> Optional[base.ModuleInfo]:
     """See if the loader can find a file to import for the module."""
-    full_path, file_exists = self._path_finder.find_import(module_name)
-    if full_path is None:
+    found_import = self._path_finder.find_import(module_name)
+    if found_import is None:
       return None
+    full_path, file_exists = found_import
     return base.ModuleInfo(module_name, full_path, file_exists)
 
   def _load_pyi(self, mod_info: base.ModuleInfo):
     """Load a file and parse it into a pytd AST."""
     with self.options.open_function(mod_info.filename, "r") as f:
       mod_ast = parser.parse_string(
-          f.read(), filename=mod_info.filename, name=mod_info.module_name,
-          options=parser.PyiOptions.from_toplevel_options(self.options))
+          f.read(),
+          filename=mod_info.filename,
+          name=mod_info.module_name,
+          options=parser.PyiOptions.from_toplevel_options(self.options),
+      )
     return mod_ast
 
   def _load_pickle(self, mod_info: base.ModuleInfo):
     """Load and unpickle a serialized pytd AST."""
-    return pickle_utils.LoadPickle(
-        mod_info.filename, open_function=self.options.open_function)
+    return pickle_utils.LoadAst(
+        mod_info.filename, open_function=self.options.open_function
+    )
 
   def load_ast(self, mod_info: base.ModuleInfo):
     if file_utils.is_pickle(mod_info.filename):
@@ -112,7 +125,21 @@ class ModuleLoader(base.ModuleLoader):
       return self._load_pyi(mod_info)
 
   def log_module_not_found(self, module_name: str):
-    log.warning("Couldn't import module %s %r in (path=%r) imports_map: %s",
-                module_name, module_name, self.options.pythonpath,
-                f"{len(self.options.imports_map)} items" if
-                self.options.imports_map is not None else "none")
+    log.warning(
+        "Couldn't import module %s %r in (path=%r) imports_map: %s",
+        module_name,
+        module_name,
+        self.options.pythonpath,
+        f"{len(self.options.imports_map)} items"
+        if self.options.imports_map is not None
+        else "none",
+    )
+
+  def get_unused_imports_map_paths(self) -> Set[str]:
+    if not self.options.imports_map:
+      return set()
+    return {
+        path
+        for path in self.options.imports_map.values()
+        if path not in self._path_finder.accessed_imports_paths
+    }
