@@ -80,17 +80,23 @@ class ImportMapLoaderTest(unittest.TestCase):
   def test_do_not_filter(self):
     with test_utils.Tempdir() as d:
       d.create_file(file_utils.replace_separator("a/b/c.pyi"))
-      imports_info = (f"{file_utils.replace_separator('a/b/c.pyi')} " +
-                      f"{d[file_utils.replace_separator('a/b/c.pyi')]}\n")
+      imports_info = (
+          f"{file_utils.replace_separator('a/b/c.pyi')} "
+          + f"{d[file_utils.replace_separator('a/b/c.pyi')]}\n"
+      )
       d.create_file("imports_info", imports_info)
       imports_map = self.build_imports_map(d["imports_info"])
-      self.assertEqual(imports_map[file_utils.replace_separator("a/b/c")],
-                       d[file_utils.replace_separator("a/b/c.pyi")])
+      self.assertEqual(
+          imports_map[file_utils.replace_separator("a/b/c")],
+          d[file_utils.replace_separator("a/b/c.pyi")],
+      )
 
   def test_invalid_map_entry(self):
     with test_utils.Tempdir() as d:
-      imports_info = (f"{file_utils.replace_separator('a/b/c.pyi')}"
-                      f"{d[file_utils.replace_separator('a/b/c.pyi')]}\n")
+      imports_info = (
+          f"{file_utils.replace_separator('a/b/c.pyi')}"
+          f"{d[file_utils.replace_separator('a/b/c.pyi')]}\n"
+      )
       d.create_file("imports_info", imports_info)
       with self.assertRaises(ValueError):
         self.build_imports_map(d["imports_info"])
