@@ -1,11 +1,10 @@
 """Load and link .pyi files."""
 
+import collections
 import dataclasses
 import logging
 import os
-import pickle
-
-from typing import Dict, Iterable, List, Optional
+from typing import Dict, Iterable, List, Optional, Set
 
 from pytype import file_utils
 from pytype import module_utils
@@ -35,7 +34,8 @@ def create_loader(options, missing_modules=()):
   """Create a pytd loader."""
   if options.precompiled_builtins:
     return PickledPyiLoader.load_from_pickle(
-        options.precompiled_builtins, options, missing_modules)
+        options.precompiled_builtins, options, missing_modules
+    )
   elif options.use_pickled_files:
     return PickledPyiLoader(options, missing_modules=missing_modules)
   else:
@@ -53,6 +53,18 @@ def _is_package(filename):
   return False
 
 
+_ModuleNameType = _AliasNameType = _NameType = str
+
+
+def _merge_aliases(
+    aliases: Dict[_ModuleNameType, Dict[_AliasNameType, _NameType]],
+) -> Dict[_AliasNameType, _NameType]:
+  all_aliases = {}
+  for mod_aliases in aliases.values():
+    all_aliases.update(mod_aliases)
+  return all_aliases
+
+
 @dataclasses.dataclass(eq=True, frozen=True)
 class ResolvedModule:
   module_name: str
@@ -80,14 +92,22 @@ class Module:
   """
 
   # pylint: disable=redefined-outer-name
-  def __init__(self, module_name, filename, ast, metadata=None, pickle=None,
-               has_unresolved_pointers=True):
+  def __init__(
+      self,
+      module_name,
+      filename,
+      ast,
+      metadata=None,
+      pickle=None,
+      has_unresolved_pointers=True,
+  ):
     self.module_name = module_name
     self.filename = filename
     self.ast = ast
     self.pickle = pickle
     self.has_unresolved_pointers = has_unresolved_pointers
     self.metadata = metadata or []
+
   # pylint: enable=redefined-outer-name
 
   def needs_unpickling(self):
@@ -98,8 +118,12 @@ class Module:
 
   @classmethod
   def resolved_internal_stub(cls, name, mod_ast):
-    return cls(name, imports_base.internal_stub_filename(name), mod_ast,
-               has_unresolved_pointers=False)
+    return cls(
+        name,
+        imports_base.internal_stub_filename(name),
+        mod_ast,
+        has_unresolved_pointers=False,
+    )
 
 
 class BadDependencyError(Exception):
@@ -160,8 +184,9 @@ class _ModuleMap:
 
   def get_module_map(self) -> Dict[str, _AST]:
     """Get a {name: ast} map of all modules with a filled-in ast."""
-    return {name: module.ast for name, module in self._modules.items()
-            if module.ast}
+    return {
+        name: module.ast for name, module in self._modules.items() if module.ast
+    }
 
   def get_resolved_modules(self) -> Dict[str, ResolvedModule]:
     """Get a {name: ResolvedModule} map of all resolved modules."""
@@ -169,12 +194,14 @@ class _ModuleMap:
     for name, mod in self._modules.items():
       if not mod.has_unresolved_pointers:
         resolved_modules[name] = ResolvedModule(
-            mod.module_name, mod.filename, mod.ast, mod.metadata)
+            mod.module_name, mod.filename, mod.ast, mod.metadata
+        )
     return resolved_modules
 
   def _base_modules(self):
     bltins, typing = builtin_stubs.GetBuiltinsAndTyping(
-        parser.PyiOptions.from_toplevel_options(self.options))
+        parser.PyiOptions.from_toplevel_options(self.options)
+    )
     return {
         "builtins": Module.resolved_internal_stub("builtins", bltins),
         "typing": Module.resolved_internal_stub("typing", typing),
@@ -195,7 +222,7 @@ class _ModuleMap:
         seen.add(m)
       if not m.pickle:
         continue
-      loaded_ast = pickle.loads(m.pickle)
+      loaded_ast = pickle_utils.DecodeAst(m.pickle)
       deps = [d for d, _ in loaded_ast.dependencies if d != loaded_ast.ast.name]
       loaded_ast = serialize_ast.EnsureAstName(loaded_ast, m.module_name)
       assert m.module_name in self._modules
@@ -216,7 +243,8 @@ class _ModuleMap:
           m.filename = path_utils.join(base, init_file)
         else:
           m.filename = imports_base.internal_stub_filename(
-              path_utils.join(m.module_name, init_file))
+              path_utils.join(m.module_name, init_file)
+          )
       m.pickle = None
     module_map = self.get_module_map()
     for loaded_ast in newly_loaded_asts:
@@ -251,22 +279,44 @@ class _Resolver:
 
   def resolve_builtin_types(self, mod_ast, *, lookup_ast=None):
     bltn_lookup = visitors.LookupBuiltins(
-        self.builtins_ast, full_names=False,
-        allow_singletons=self.allow_singletons)
+        self.builtins_ast,
+        full_names=False,
+        allow_singletons=self.allow_singletons,
+    )
     mod_ast = self._lookup(bltn_lookup, mod_ast, lookup_ast)
     return mod_ast
 
   def resolve_external_types(self, mod_ast, module_map, aliases, *, mod_name):
+    """Resolves external types in mod_ast."""
     name = mod_name or mod_ast.name
     try:
-      mod_ast = mod_ast.Visit(visitors.LookupExternalTypes(
-          module_map, self_name=name, module_alias_map=aliases))
+      return mod_ast.Visit(
+          visitors.LookupExternalTypes(
+              module_map, self_name=name, module_alias_map=aliases[name]
+          )
+      )
     except KeyError as e:
-      raise BadDependencyError(str(e), name) from e
-    return mod_ast
+      key_error = e
+    all_aliases = _merge_aliases(aliases)
+    new_aliases = dict(aliases[name])
+    while isinstance(key_error, visitors.MissingModuleError):
+      if key_error.module not in all_aliases:
+        break
+      new_aliases[key_error.module] = all_aliases[key_error.module]
+      try:
+        return mod_ast.Visit(
+            visitors.LookupExternalTypes(
+                module_map, self_name=name, module_alias_map=new_aliases
+            )
+        )
+      except KeyError as e:
+        key_error = e
+    key = "".join(str(arg) for arg in key_error.args)
+    raise BadDependencyError(key, name) from key_error
 
-  def resolve_module_alias(self, name, *, lookup_ast=None,
-                           lookup_ast_name=None):
+  def resolve_module_alias(
+      self, name, *, lookup_ast=None, lookup_ast_name=None
+  ):
     """Check if a given name is an alias and resolve it if so."""
     # name is bare, but aliases are stored as "ast_name.alias".
     if lookup_ast is None:
@@ -278,7 +328,7 @@ class _Resolver:
       key = f"{ast_name}.{cur_name}"
       value = aliases.get(key)
       if isinstance(value, pytd.Module):
-        return value.module_name + name[len(cur_name):]
+        return value.module_name + name[len(cur_name) :]
       cur_name, _, _ = cur_name.rpartition(".")
     return name
 
@@ -297,12 +347,61 @@ class _Resolver:
     deps = visitors.CollectDependencies()
     mod_ast.Visit(deps)
     if isinstance(mod_ast, (pytd.TypeDeclUnit, pytd.Class)):
-      return {k: v for k, v in deps.dependencies.items()
-              if not isinstance(mod_ast.Get(k), (pytd.Class, pytd.ParamSpec))}
+      return {
+          k: v
+          for k, v in deps.dependencies.items()
+          if not isinstance(mod_ast.Get(k), (pytd.Class, pytd.ParamSpec))
+      }
     else:
       return deps.dependencies
 
 
+class _LateTypeLoader:
+  """Resolves late types, loading modules as needed."""
+
+  def __init__(self, loader: "Loader"):
+    self._loader = loader
+    self._resolved_late_types = {}  # performance cache
+
+  def _load_late_type_module(self, late_type: pytd.LateType):
+    """Load the module of a late type from the qualified name."""
+    parts = late_type.name.split(".")
+    for i in range(len(parts) - 1):
+      module_parts = module_utils.strip_init_suffix(parts[: -(i + 1)])
+      if ast := self._loader.import_name(".".join(module_parts)):
+        return ast, ".".join(parts[-(i + 1) :])
+    return None, late_type.name
+
+  def _load_late_type(self, late_type: pytd.LateType):
+    """Resolve a late type, possibly by loading a module."""
+    if late_type.name not in self._resolved_late_types:
+      if ast := self._loader.import_name(late_type.name):
+        t = pytd.Module(name=late_type.name, module_name=late_type.name)
+      else:
+        ast, attr_name = self._load_late_type_module(late_type)
+        if ast is None:
+          msg = "During dependency resolution, couldn't resolve late type %r"
+          log.error(msg, late_type.name)
+          t = pytd.AnythingType()
+        else:
+          try:
+            cls = pytd.LookupItemRecursive(ast, attr_name)
+          except KeyError:
+            if "__getattr__" not in ast:
+              log.warning("Couldn't resolve %s", late_type.name)
+            t = pytd.AnythingType()
+          else:
+            t = pytd.ToType(cls, allow_functions=True)
+      if isinstance(t, pytd.LateType):
+        t = self._load_late_type(t)
+      self._resolved_late_types[late_type.name] = t
+    return self._resolved_late_types[late_type.name]
+
+  def load_late_type(self, late_type: pytd.LateType) -> pytd.Type:
+    """Convert a pytd.LateType to a pytd type."""
+    return self._load_late_type(late_type)
+
+
 class Loader:
   """A cache for loaded PyTD files.
 
@@ -323,10 +422,12 @@ class Loader:
     pyi_options = parser.PyiOptions.from_toplevel_options(options)
     self._builtin_loader = builtin_stubs.BuiltinLoader(pyi_options)
     self._typeshed_loader = typeshed.TypeshedLoader(
-        pyi_options, missing_modules)
+        pyi_options, missing_modules
+    )
     self._resolver = _Resolver(self.builtins)
+    self._late_type_loader = _LateTypeLoader(self)
     self._import_name_cache = {}  # performance cache
-    self._aliases = {}
+    self._aliases = collections.defaultdict(dict)
     self._prefixes = set()
     # Paranoid verification that pytype.main properly checked the flags:
     if options.imports_map is not None:
@@ -334,41 +435,37 @@ class Loader:
 
   def get_default_ast(self):
     return builtin_stubs.GetDefaultAst(
-        parser.PyiOptions.from_toplevel_options(self.options))
+        parser.PyiOptions.from_toplevel_options(self.options)
+    )
 
   def save_to_pickle(self, filename):
     """Save to a pickle. See PickledPyiLoader.load_from_pickle for reverse."""
     # We assume that the Loader is in a consistent state here. In particular, we
     # assume that for every module in _modules, all the transitive dependencies
     # have been loaded.
-    # pylint: disable=g-complex-comprehension
-    items = tuple(
-        (name, pickle_utils.StoreAst(
-            module.ast, open_function=self.options.open_function,
-            src_path=module.filename))
-        for name, module in sorted(self._modules.items()))
-    # pylint: enable=g-complex-comprehension
+    items = pickle_utils.PrepareModuleBundle(
+        ((name, m.filename, m.ast) for name, m in sorted(self._modules.items()))
+    )
     # Preparing an ast for pickling clears its class pointers, making it
     # unsuitable for reuse, so we have to discard the builtins cache.
     builtin_stubs.InvalidateCache()
-    # Now pickle the pickles. We keep the "inner" modules as pickles as a
-    # performance optimization - unpickling is slow.
-    pickle_utils.SavePickle(items, filename, compress=True,
-                            open_function=self.options.open_function)
+    pickle_utils.Save(
+        items, filename, compress=True, open_function=self.options.open_function
+    )
 
   def _resolve_external_and_local_types(self, mod_ast, lookup_ast=None):
     dependencies = self._resolver.collect_dependencies(mod_ast)
     if dependencies:
       lookup_ast = lookup_ast or mod_ast
       self._load_ast_dependencies(dependencies, lookup_ast)
-      mod_ast = self._resolve_external_types(
-          mod_ast, lookup_ast=lookup_ast)
+      mod_ast = self._resolve_external_types(mod_ast, lookup_ast=lookup_ast)
     mod_ast = self._resolver.resolve_local_types(mod_ast, lookup_ast=lookup_ast)
     return mod_ast
 
   def _create_empty(self, mod_info):
     return self.load_module(
-        mod_info, mod_ast=pytd_utils.CreateModule(mod_info.module_name))
+        mod_info, mod_ast=pytd_utils.CreateModule(mod_info.module_name)
+    )
 
   def load_file(self, module_name, filename, mod_ast=None):
     """Load a module from a filename."""
@@ -385,9 +482,9 @@ class Loader:
       return existing
     if not mod_ast:
       mod_ast = self._module_loader.load_ast(mod_info)
-    return self._process_module(mod_info, mod_ast)
+    return self.process_module(mod_info, mod_ast)
 
-  def _process_module(self, mod_info, mod_ast):
+  def process_module(self, mod_info, mod_ast):
     """Create a module from a loaded ast and save it to the loader cache.
 
     Args:
@@ -429,6 +526,13 @@ class Loader:
       self.add_module_prefixes(module_name)
     return module.ast
 
+  def remove_name(self, module_name: str) -> None:
+    """Removes a module from the cache, if it is present."""
+    if module_name in self._modules:
+      del self._modules[module_name]
+    if module_name in self._import_name_cache:
+      del self._import_name_cache[module_name]
+
   def _try_import_prefix(self, name: str) -> Optional[_AST]:
     """Try importing all prefixes of name, returning the first valid module."""
     prefix = name
@@ -439,16 +543,18 @@ class Loader:
         return ast
     return None
 
-  def _load_ast_dependencies(self, dependencies, lookup_ast,
-                             lookup_ast_name=None):
+  def _load_ast_dependencies(
+      self, dependencies, lookup_ast, lookup_ast_name=None
+  ):
     """Fill in all ClassType.cls pointers and load reexported modules."""
     ast_name = lookup_ast_name or lookup_ast.name
     for dep_name in dependencies:
       name = self._resolver.resolve_module_alias(
-          dep_name, lookup_ast=lookup_ast, lookup_ast_name=lookup_ast_name)
+          dep_name, lookup_ast=lookup_ast, lookup_ast_name=lookup_ast_name
+      )
       if dep_name != name:
         # We have an alias. Store it in the aliases map.
-        self._aliases[dep_name] = name
+        self._aliases[ast_name][dep_name] = name
       if name in self._modules and self._modules[name].ast:
         dep_ast = self._modules[name].ast
       else:
@@ -466,7 +572,8 @@ class Loader:
               pytd.LookupItemRecursive(lookup_ast, name)
             except KeyError as e:
               raise BadDependencyError(
-                  f"Can't find pyi for {name!r}", ast_name) from e
+                  f"Can't find pyi for {name!r}", ast_name
+              ) from e
             # This is a dotted local reference, not an external reference.
             continue
       # If `name` is a package, try to load any base names not defined in
@@ -488,13 +595,16 @@ class Loader:
           # sqlite3.dbapi2 imports local names from the sqlite3 namespace to
           # avoid name collisions.
           maybe_star_import = dep_ast.Get(f"{name}.{ast_name}.*")
-          if (isinstance(maybe_star_import, pytd.Alias) and
-              maybe_star_import.type.name == f"{ast_name}.*"):
+          if (
+              isinstance(maybe_star_import, pytd.Alias)
+              and maybe_star_import.type.name == f"{ast_name}.*"
+          ):
             attr = lookup_ast.Get(f"{ast_name}.{base_name}")
         # 'from . import submodule as submodule' produces
         # Alias(submodule, NamedType(submodule)).
         if attr is None or (
-            isinstance(attr, pytd.Alias) and attr.name == attr.type.name):
+            isinstance(attr, pytd.Alias) and attr.name == attr.type.name
+        ):
           if not self._import_module_by_name(full_name):
             # Add logging to make debugging easier but otherwise ignore the
             # result - resolve_external_types will raise a better error.
@@ -506,7 +616,8 @@ class Loader:
     if mod_name and mod_name not in module_map:
       module_map[mod_name] = lookup_ast
     mod_ast = self._resolver.resolve_external_types(
-        mod_ast, module_map, self._aliases, mod_name=mod_name)
+        mod_ast, module_map, self._aliases, mod_name=mod_name
+    )
     return mod_ast
 
   def _resolve_classtype_pointers(self, mod_ast, *, lookup_ast=None):
@@ -518,9 +629,11 @@ class Loader:
     """Resolve and verify pytd value, using the given ast for local lookup."""
     # NOTE: Modules of dependencies will be loaded into the cache
     pytd_node = self._resolver.resolve_builtin_types(
-        pytd_node, lookup_ast=lookup_ast)
+        pytd_node, lookup_ast=lookup_ast
+    )
     pytd_node = self._resolve_external_and_local_types(
-        pytd_node, lookup_ast=lookup_ast)
+        pytd_node, lookup_ast=lookup_ast
+    )
     self._resolve_classtype_pointers_for_all_modules()
     self._resolve_classtype_pointers(pytd_node, lookup_ast=lookup_ast)
     self._resolver.verify(pytd_node, mod_name=lookup_ast.name)
@@ -581,6 +694,15 @@ class Loader:
     self._import_name_cache[module_name] = mod_ast
     return mod_ast
 
+  def _resolve_module(self, name, aliases):
+    if name in aliases:
+      name = aliases[name]
+    while name not in self._modules:
+      if "." not in name:
+        break
+      name, _ = name.rsplit(".", 1)
+    return name
+
   def finish_and_verify_ast(self, mod_ast):
     """Verify the ast, doing external type resolution first if necessary."""
     if mod_ast:
@@ -593,19 +715,18 @@ class Loader:
         # transitive imports, but lookups are expensive.
         dependencies = self._resolver.collect_dependencies(mod_ast)
         for k in dependencies:
-          if k in self._aliases:
-            k = self._aliases[k]
-          while k not in self._modules:
-            if "." not in k:
-              break
-            k, _ = k.rsplit(".", 1)
+          k = self._resolve_module(k, self._aliases[mod_ast.name])
+          if k not in self._modules:
+            all_aliases = _merge_aliases(self._aliases)
+            k = self._resolve_module(k, all_aliases)
           if k not in self._modules:
             assert mod_ast
             raise (
                 BadDependencyError(f"Can't find pyi for {k!r}", mod_ast.name)
             ) from e
           self._modules[k].ast = self._resolve_external_types(
-              self._modules[k].ast)
+              self._modules[k].ast
+          )
         mod_ast = self._resolve_external_types(mod_ast)
         self._resolver.verify(mod_ast)
     return mod_ast
@@ -658,6 +779,7 @@ class Loader:
 
     # Now try to retrieve an external module from the module loader.
     mod_ast = None
+    default = None
     mod_info = self._module_loader.find_import(module_name)
     if mod_info:
       if mod_info.file_exists:
@@ -703,9 +825,16 @@ class Loader:
     """Gets a name -> ResolvedModule map of the loader's resolved modules."""
     return self._modules.get_resolved_modules()
 
-  def lookup_builtin(self, name):
-    found = self.builtins.Get(name)
-    return found if found is not None else self.typing.Lookup(name)
+  def lookup_pytd(self, module: str, name: str) -> pytd.Node:
+    ast = self.import_name(module)
+    assert ast, f"Module not found: {module}"
+    return ast.Lookup(f"{module}.{name}")
+
+  def load_late_type(self, late_type: pytd.LateType):
+    return self._late_type_loader.load_late_type(late_type)
+
+  def get_unused_imports_map_paths(self) -> Set[str]:
+    return self._module_loader.get_unused_imports_map_paths()
 
 
 class PickledPyiLoader(Loader):
@@ -714,12 +843,18 @@ class PickledPyiLoader(Loader):
   @classmethod
   def load_from_pickle(cls, filename, options, missing_modules=()):
     """Load a pytd module from a pickle file."""
-    items = pickle_utils.LoadPickle(filename, compress=True,
-                                    open_function=options.open_function)
+    items = pickle_utils.LoadBuiltins(
+        filename, compress=True, open_function=options.open_function
+    )
     modules = {
-        name: Module(name, filename=None, ast=None, pickle=pickle,
-                     has_unresolved_pointers=False)
-        for name, pickle in items
+        name: Module(
+            name,
+            filename=None,
+            ast=None,
+            pickle=raw.copy(),
+            has_unresolved_pointers=False,
+        )
+        for name, raw in items
     }
     return cls(options, modules=modules, missing_modules=missing_modules)
 
@@ -734,14 +869,21 @@ class PickledPyiLoader(Loader):
     loaded_ast = self._module_loader.load_ast(mod_info)
     # At this point ast.name and module_name could be different.
     # They are later synced in ProcessAst.
-    dependencies = {d: names for d, names in loaded_ast.dependencies
-                    if d != loaded_ast.ast.name}
+    dependencies = {
+        d: names
+        for d, names in loaded_ast.dependencies
+        if d != loaded_ast.ast.name
+    }
     loaded_ast = serialize_ast.EnsureAstName(loaded_ast, module_name, fix=True)
     self._modules[module_name] = Module(
-        module_name, mod_info.filename, loaded_ast.ast,
-        metadata=loaded_ast.metadata)
-    self._load_ast_dependencies(dependencies, lookup_ast=mod_ast,
-                                lookup_ast_name=module_name)
+        module_name,
+        mod_info.filename,
+        loaded_ast.ast,
+        metadata=loaded_ast.metadata,
+    )
+    self._load_ast_dependencies(
+        dependencies, lookup_ast=mod_ast, lookup_ast_name=module_name
+    )
     try:
       ast = serialize_ast.ProcessAst(loaded_ast, self._modules.get_module_map())
     except serialize_ast.UnrestorableDependencyError as e:
