@@ -28,6 +28,7 @@ AttributeKinds = class_mixin.AttributeKinds
 # Probably should make this an enum.Enum at some point.
 class Ordering:
   """Possible orderings for get_class_locals."""
+
   # Order by each variable's first annotation. For example, for
   #   class Foo:
   #     x: int
@@ -52,7 +53,7 @@ class Decorator(abstract.PyTDFunction, metaclass=abc.ABCMeta):
 
   # Defaults for the args that we support (dataclasses only support 'init',
   # but the others default to false so they should not affect anything).
-  _DEFAULT_ARGS: ClassVar[Dict[str, Any]] = {
+  DEFAULT_ARGS: ClassVar[Dict[str, Any]] = {
       "init": True,
       "kw_only": False,
       "auto_attribs": False,
@@ -74,7 +75,7 @@ class Decorator(abstract.PyTDFunction, metaclass=abc.ABCMeta):
     """Apply the decorator to cls."""
 
   def get_initial_args(self):
-    ret = self._DEFAULT_ARGS.copy()
+    ret = self.DEFAULT_ARGS.copy()
     ret.update(self.partial_args)
     return ret
 
@@ -87,7 +88,8 @@ class Decorator(abstract.PyTDFunction, metaclass=abc.ABCMeta):
           self._current_args[k] = abstract_utils.get_atomic_python_constant(v)
         except abstract_utils.ConversionError:
           self.ctx.errorlog.not_supported_yet(
-              self.ctx.vm.frames, f"Non-constant argument to decorator: {k!r}")
+              self.ctx.vm.frames, f"Non-constant argument to decorator: {k!r}"
+          )
 
   def set_current_args(self, kwargs):
     """Set current_args when constructing a class directly."""
@@ -109,10 +111,7 @@ class Decorator(abstract.PyTDFunction, metaclass=abc.ABCMeta):
       # call self.init_name in case the name differs from the field name -
       # e.g. attrs removes leading underscores from attrib names when
       # generating kwargs for __init__.
-      param = Param(
-          name=self.init_name(attr),
-          typ=typ,
-          default=attr.default)
+      param = Param(name=self.init_name(attr), typ=typ, default=attr.default)
 
       # kw_only=False in a field does not override kw_only=True in the class.
       if all_kwonly or attr.kw_only:
@@ -120,8 +119,9 @@ class Decorator(abstract.PyTDFunction, metaclass=abc.ABCMeta):
       else:
         pos_params.append(param)
 
-    return overlay_utils.make_method(self.ctx, node, init_method_name,
-                                     pos_params, 0, kwonly_params)
+    return overlay_utils.make_method(
+        self.ctx, node, init_method_name, pos_params, 0, kwonly_params
+    )
 
   def call(self, node, func, args, alias_map=None):
     """Construct a decorator, and call it on the class."""
@@ -159,7 +159,7 @@ class Decorator(abstract.PyTDFunction, metaclass=abc.ABCMeta):
 
     cls_var = args.posargs[0]
     # We should only have a single binding here
-    cls, = cls_var.data
+    (cls,) = cls_var.data
 
     if not isinstance(cls, abstract.Class):
       # There are other valid types like abstract.Unsolvable that we don't need
@@ -184,8 +184,9 @@ class FieldConstructor(abstract.PyTDFunction):
     try:
       return abstract_utils.get_atomic_python_constant(args.namedargs[name])
     except abstract_utils.ConversionError:
-      self.ctx.errorlog.not_supported_yet(self.ctx.vm.frames,
-                                          f"Non-constant argument {name!r}")
+      self.ctx.errorlog.not_supported_yet(
+          self.ctx.vm.frames, f"Non-constant argument {name!r}"
+      )
 
   def get_positional_names(self):
     # TODO(mdemello): We currently assume all field constructors are called with
@@ -197,12 +198,15 @@ class FieldConstructor(abstract.PyTDFunction):
 def is_method(var):
   if var is None:
     return False
-  return isinstance(var.data[0], (
-      abstract.INTERPRETER_FUNCTION_TYPES,
-      special_builtins.ClassMethodInstance,
-      special_builtins.PropertyInstance,
-      special_builtins.StaticMethodInstance
-  ))
+  return isinstance(
+      var.data[0],
+      (
+          abstract.INTERPRETER_FUNCTION_TYPES,
+          special_builtins.ClassMethodInstance,
+          special_builtins.PropertyInstance,
+          special_builtins.StaticMethodInstance,
+      ),
+  )
 
 
 def is_dunder(name):
@@ -217,12 +221,15 @@ def add_member(node, cls, name, typ):
   else:
     # See test_attr.TestAttrib.test_repeated_default - keying on the name
     # prevents attributes from sharing the same default object.
-    _, instance = typ.ctx.vm.init_class(node, typ, extra_key=name)
+    instance = typ.ctx.vm.init_class(node, typ, extra_key=name)
   cls.members[name] = instance
 
 
-def is_relevant_class_local(class_local: abstract_utils.Local,
-                            class_local_name: str, allow_methods: bool):
+def is_relevant_class_local(
+    class_local: abstract_utils.Local,
+    class_local_name: str,
+    allow_methods: bool,
+):
   """Tests whether the current class local could be relevant for type checking.
 
   For example, this doesn't match __dunder__ class locals.
@@ -342,26 +349,7 @@ class ClassProperties:
 
 
 def make_annotations_dict(fields, node, ctx):
-  locals_ = {f.name: abstract_utils.Local(node, None, f.typ, None, ctx)
-             for f in fields}
+  locals_ = {
+      f.name: abstract_utils.Local(node, None, f.typ, None, ctx) for f in fields
+  }
   return abstract.AnnotationsDict(locals_, ctx).to_variable(node)
-
-
-def make_interpreter_class(class_type, props, node, ctx):
-  """Make an InterpreterClass from ClassProperties."""
-
-  name_var = ctx.convert.build_string(node, props.name)
-  bases = props.bases
-  members = {f.name: f.typ.instantiate(node) for f in props.fields}
-  annots = make_annotations_dict(props.fields, node, ctx)
-  members["__annotations__"] = annots
-  cls_dict = abstract.Dict(ctx)
-  cls_dict.update(node, members)
-  cls_props = class_mixin.ClassBuilderProperties(
-      name_var=name_var,
-      bases=bases,
-      class_dict_var=cls_dict.to_variable(node),
-      class_type=class_type)
-  node, cls_var = ctx.make_class(node, cls_props)
-  ctx.vm.trace_classdef(cls_var)
-  return node, cls_var
