@@ -22,7 +22,6 @@ call_metaclass_init is called, allowing EnumMetaInit to transform the PyTDClass
 into a proper enum.
 """
 
-
 import collections
 import contextlib
 import logging
@@ -32,47 +31,60 @@ from pytype.abstract import abstract
 from pytype.abstract import abstract_utils
 from pytype.abstract import class_mixin
 from pytype.abstract import function
+from pytype.errors import error_types
 from pytype.overlays import classgen
 from pytype.overlays import overlay
 from pytype.overlays import overlay_utils
 from pytype.overlays import special_builtins
 from pytype.pytd import pytd
 from pytype.pytd import pytd_utils
-from pytype.pytd import visitors
 from pytype.typegraph import cfg
 
 log = logging.getLogger(__name__)
 
 
+# These members have been added in Python 3.11 and are not yet supported.
+_unsupported = (
+    "ReprEnum",
+    "EnumCheck",
+    "FlagBoundary",
+    "verify",
+    "property",
+    "member",
+    "nonmember",
+    "global_enum",
+    "show_flag_values",
+)
+
+
 class EnumOverlay(overlay.Overlay):
   """An overlay for the enum std lib module."""
 
   def __init__(self, ctx):
-    ast = ctx.loader.import_name("enum")
-    if ctx.options.use_enum_overlay:
-      member_map = {
-          "Enum": EnumBuilder,
-          "EnumMeta": EnumMeta,
-          "IntEnum": IntEnumBuilder,
-      }
-      ast = ast.Visit(visitors.RemoveMethods())
-    else:
-      member_map = {}
+    member_map = {
+        "Enum": overlay.add_name("Enum", EnumBuilder),
+        "EnumMeta": EnumMeta,
+        "EnumType": EnumMeta,
+        "IntEnum": overlay.add_name("IntEnum", EnumBuilder),
+        "StrEnum": overlay.add_name("StrEnum", EnumBuilder),
+        **{
+            name: overlay.add_name(name, overlay_utils.not_supported_yet)
+            for name in _unsupported
+        },
+    }
 
-    super().__init__(ctx, "enum", member_map, ast)
+    super().__init__(ctx, "enum", member_map, ctx.loader.import_name("enum"))
 
 
 class EnumBuilder(abstract.PyTDClass):
   """Overlays enum.Enum."""
 
-  def __init__(self, ctx, name="Enum"):
-    enum_ast = ctx.vm.loaded_overlays["enum"].ast
-    pyval = enum_ast.Lookup(f"enum.{name}")
-    super().__init__(name, pyval, ctx)
+  def __init__(self, name, ctx, module):
+    super().__init__(name, ctx.loader.lookup_pytd(module, name), ctx)
 
   def make_class(self, node, props):
     """Check the members for errors, then create the enum class."""
-    # TODO(tsudol): Handle is_decorated: @enum.unique, for example.
+    # TODO(tsudol): Handle decorators: @enum.unique, for example.
 
     # make_class intercepts the class creation for enums in order to check for
     # errors. EnumMeta turns the class into a full enum, but that's too late for
@@ -86,13 +98,17 @@ class EnumBuilder(abstract.PyTDClass):
     props.bases = props.bases or [self.to_variable(node)]
     last_base = props.bases[-1]
     if not any(b.is_enum for b in last_base.data):
-      msg = ("The last base class for an enum must be enum.Enum or a subclass "
-             "of enum.Enum")
+      msg = (
+          "The last base class for an enum must be enum.Enum or a subclass "
+          "of enum.Enum"
+      )
       self.ctx.errorlog.base_class_error(
-          self.ctx.vm.frames, last_base, details=msg)
+          self.ctx.vm.frames, last_base, details=msg
+      )
       return node, self.ctx.new_unsolvable(node)
     props.metaclass_var = props.metaclass_var or (
-        self.ctx.vm.loaded_overlays["enum"].members["EnumMeta"])
+        self.ctx.vm.loaded_overlays["enum"].members["EnumMeta"]
+    )
     props.class_type = EnumInstance
     return self.ctx.make_class(node, props)
 
@@ -109,18 +125,21 @@ class EnumBuilder(abstract.PyTDClass):
     # Note that super().call or _call_new_and_init won't work here, because
     # they don't raise FailedFunctionCall.
     node, pytd_new_var = self.ctx.attribute_handler.get_attribute(
-        node, self, "__new__", self.to_binding(node))
+        node, self, "__new__", self.to_binding(node)
+    )
     pytd_new = abstract_utils.get_atomic_value(pytd_new_var)
     # There are 2 signatures for Enum.__new__. The one with fewer arguments is
     # for looking up values, and the other is for the functional API.
     # I don't think we have a guarantee of ordering for signatures, so choose
     # them based on parameter count.
-    lookup_sig, api_sig = sorted((s.signature for s in pytd_new.signatures),
-                                 key=lambda s: s.maximum_param_count())
+    lookup_sig, api_sig = sorted(
+        (s.signature for s in pytd_new.signatures),
+        key=lambda s: s.maximum_param_count(),
+    )
     lookup_new = abstract.SimpleFunction(lookup_sig, self.ctx)
     try:
       return lookup_new.call(node, None, args, alias_map)
-    except function.FailedFunctionCall as e:
+    except error_types.FailedFunctionCall as e:
       log.info("Called Enum.__new__ as lookup, but failed:\n%s", e)
     api_new = abstract.SimpleFunction(api_sig, self.ctx)
     api_new.call(node, None, args, alias_map)
@@ -144,8 +163,9 @@ class EnumBuilder(abstract.PyTDClass):
     else:
       # List of names, or list of (name, value) pairs.
       try:
-        possible_pairs = [abstract_utils.get_atomic_python_constant(p)
-                          for p in names]
+        possible_pairs = [
+            abstract_utils.get_atomic_python_constant(p) for p in names
+        ]
       except abstract_utils.ConversionError as e:
         log.debug("Failed to unwrap possible enum field pairs:\n  %s", e)
         return node, self.ctx.new_unsolvable(node)
@@ -161,8 +181,7 @@ class EnumBuilder(abstract.PyTDClass):
         # so the values in the tuple still need to be unwrapped.
         try:
           fields = {
-              abstract_utils.get_atomic_python_constant(name):
-                  value
+              abstract_utils.get_atomic_python_constant(name): value
               for name, value in possible_pairs
           }
         except abstract_utils.ConversionError as e:
@@ -179,17 +198,11 @@ class EnumBuilder(abstract.PyTDClass):
         bases=[self.to_variable(node)],
         class_dict_var=cls_dict.to_variable(node),
         metaclass_var=metaclass,
-        class_type=EnumInstance)
+        class_type=EnumInstance,
+    )
     return self.ctx.make_class(node, props)
 
 
-class IntEnumBuilder(EnumBuilder):
-  """Overlays enum.IntEnum using EnumBuilder."""
-
-  def __init__(self, ctx):
-    super().__init__(ctx, name="IntEnum")
-
-
 class EnumInstance(abstract.InterpreterClass):
   """A wrapper for classes that subclass enum.Enum."""
 
@@ -242,8 +255,11 @@ class EnumInstance(abstract.InterpreterClass):
     return True
 
   def get_enum_members(self, qualified=False):
-    ret = {k: v for k, v in self.members.items()
-           if all(d.cls == self for d in v.data)}
+    ret = {
+        k: v
+        for k, v in self.members.items()
+        if all(d.cls == self for d in v.data)
+    }
     if qualified:
       return {f"{self.name}.{k}": v for k, v in ret.items()}
     else:
@@ -252,6 +268,7 @@ class EnumInstance(abstract.InterpreterClass):
 
 class EnumCmpEQ(abstract.SimpleFunction):
   """Implements the functionality of __eq__ for an enum."""
+
   # b/195136939: Enum equality checks could be made more exact/precise by
   # comparing the members' names. However, this causes issues when enums are
   # used in an if statement; see the bug for examples.
@@ -267,10 +284,11 @@ class EnumCmpEQ(abstract.SimpleFunction):
         defaults={},
         annotations={
             "return": ctx.convert.bool_type,
-        })
+        },
+    )
     super().__init__(sig, ctx)
 
-  def call(self, node, unused_f, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     _, argmap = self.match_and_map_args(node, args, alias_map)
     this_var = argmap["self"]
     other_var = argmap["other"]
@@ -291,9 +309,8 @@ class EnumMeta(abstract.PyTDClass):
   enum behavior: EnumMetaInit for modifying enum classes, for example.
   """
 
-  def __init__(self, ctx):
-    enum_ast = ctx.vm.loaded_overlays["enum"].ast
-    pytd_cls = enum_ast.Lookup("enum.EnumMeta")
+  def __init__(self, ctx, module):
+    pytd_cls = ctx.loader.lookup_pytd(module, "EnumMeta")
     super().__init__("EnumMeta", pytd_cls, ctx)
     init = EnumMetaInit(ctx)
     self._member_map["__init__"] = init
@@ -319,16 +336,17 @@ class EnumMetaInit(abstract.SimpleFunction):
         kwonly_params=(),
         kwargs_name=None,
         defaults={},
-        annotations={})
+        annotations={},
+    )
     super().__init__(sig, ctx)
-    self._str_pytd = ctx.loader.lookup_builtin("builtins.str")
+    self._str_pytd = ctx.loader.lookup_pytd("builtins", "str")
 
   def _get_class_locals(self, node, cls_name, cls_dict):
     # First, check if get_class_locals works for this class.
     if cls_name in self.ctx.vm.local_ops:
-      ret = classgen.get_class_locals(cls_name, False,
-                                      classgen.Ordering.LAST_ASSIGN,
-                                      self.ctx).items()
+      ret = classgen.get_class_locals(
+          cls_name, False, classgen.Ordering.LAST_ASSIGN, self.ctx
+      ).items()
       return ret
 
     # If it doesn't work, then it's likely this class was created using the
@@ -344,15 +362,38 @@ class EnumMetaInit(abstract.SimpleFunction):
     # to `unsolvable` if the enum has no members. Technically, `__new__` should
     # not accept any arguments, because it will always fail if the enum has no
     # members. But `unsolvable` is much simpler to implement and use.
+    value_types = [member_type, cls]
+    # If this enum class defines the _missing_ classmethod, then widen the value
+    # type to include the type of the value parameter in _missing_.
+    if "_missing_" in cls:
+      if isinstance(cls, abstract.PyTDClass):
+        missing_sigs = []
+        with self.ctx.allow_recursive_convert():
+          missing_var = cls.load_lazy_attribute("_missing_")
+        for missing in missing_var.data:
+          if isinstance(missing, abstract.PyTDFunction):
+            missing_sigs.extend(sig.signature for sig in missing.signatures)
+      else:
+        missing_sigs = []
+        for val in cls.members["_missing_"].data:
+          if isinstance(val, special_builtins.ClassMethodInstance):
+            for missing in val.func.data:
+              if isinstance(missing, abstract.SignedFunction):
+                missing_sigs.append(missing.signature)
+      for missing_sig in missing_sigs:
+        value_type = missing_sig.annotations.get(
+            "value", self.ctx.convert.unsolvable
+        )
+        value_types.append(value_type)
     return overlay_utils.make_method(
         ctx=self.ctx,
         node=node,
         name="__new__",
         params=[
-            overlay_utils.Param("value",
-                                abstract.Union([member_type, cls], self.ctx))
+            overlay_utils.Param("value", abstract.Union(value_types, self.ctx))
         ],
-        return_type=cls)
+        return_type=cls,
+    )
 
   def _get_base_type(self, bases):
     # Enums may have a data class as one of their bases: class F(str, Enum) will
@@ -379,8 +420,10 @@ class EnumMetaInit(abstract.SimpleFunction):
         # their member type. Their subclasses use the default base type, int.
         # Some enums may have members with actually unsolvable member types, so
         # check if the enum is empty.
-        if (base_type_cls.member_type == self.ctx.convert.unsolvable and
-            base_type_cls.is_empty_enum()):
+        if (
+            base_type_cls.member_type == self.ctx.convert.unsolvable
+            and base_type_cls.is_empty_enum()
+        ):
           return None
         else:
           return base_type_cls.member_type
@@ -401,7 +444,8 @@ class EnumMetaInit(abstract.SimpleFunction):
     # enum_base.__new__ is saved as __new_member__, if it has a custom __new__.
     if enum_base.full_name != "enum.Enum" and "__new_member__" in enum_base:
       node, new = self.ctx.attribute_handler.get_attribute(
-          node, enum_base, "__new_member__")
+          node, enum_base, "__new_member__"
+      )
       new = abstract_utils.get_atomic_value(new)
       if isinstance(new, abstract.BoundFunction):
         new = new.underlying
@@ -427,17 +471,34 @@ class EnumMetaInit(abstract.SimpleFunction):
     # for built-in descriptors (i.e. functions and methods -- NOT all callables)
     # and anything that has __get__, __set__ or __delete__.
     def _check(value):
-      if isinstance(value, (abstract.Function, abstract.BoundFunction,
-                            abstract.ClassMethod, abstract.StaticMethod)):
+      if isinstance(
+          value,
+          (
+              abstract.Function,
+              abstract.BoundFunction,
+              abstract.ClassMethod,
+              abstract.StaticMethod,
+          ),
+      ):
         return True
       for attr_name in ("__get__", "__set__", "__delete__"):
-        _, attr = self.ctx.attribute_handler.get_attribute(
-            node, value, attr_name)
-        if attr is not None:
+        if self._value_definitely_has_attr(node, value, attr_name):
           return True
       return False
+
     return any(_check(value) for value in local.orig.data)
 
+  def _value_definitely_has_attr(self, node, value, attr_name) -> bool:
+    _, attr = self.ctx.attribute_handler.get_attribute(node, value, attr_name)
+    if attr is None:
+      return False
+    maybe_missing_members = getattr(value, "maybe_missing_members", False)
+    if maybe_missing_members and attr.data == [self.ctx.convert.unsolvable]:
+      # The attribute could be a missing member that was given a default value
+      # of Any.
+      return False
+    return True
+
   def _not_valid_member(self, node, name, local) -> bool:
     # Reject a class local if:
     # - its name is invalid.
@@ -456,20 +517,26 @@ class EnumMetaInit(abstract.SimpleFunction):
     except abstract_utils.ConversionError as e:
       log.info("Failed to extract atomic enum value for auto() check: %s", e)
       return False
-    return (isinstance(data, abstract.Instance) and
-            data.cls.full_name == "enum.auto")
+    return (
+        isinstance(data, abstract.Instance)
+        and data.cls.full_name == "enum.auto"
+    )
 
   def _call_generate_next_value(self, node, cls, name):
     node, method = self.ctx.attribute_handler.get_attribute(
-        node, cls, "_generate_next_value_", cls.to_binding(node))
+        node, cls, "_generate_next_value_", cls.to_binding(node)
+    )
     # It's possible we'll get a unsolvable (due to __getattr__, say) for method.
     # We treat that as if the method is undefined instead.
     if method and all(abstract_utils.is_callable(m) for m in method.data):
       args = function.Args(
-          posargs=(self.ctx.convert.build_string(node, name),
-                   self.ctx.convert.build_int(node),
-                   self.ctx.convert.build_int(node),
-                   self.ctx.convert.build_list(node, [])))
+          posargs=(
+              self.ctx.convert.build_string(node, name),
+              self.ctx.convert.build_int(node),
+              self.ctx.convert.build_int(node),
+              self.ctx.convert.build_list(node, []),
+          )
+      )
       return function.call_function(self.ctx, node, method, args)
     else:
       return node, self.ctx.convert.build_int(node)
@@ -486,7 +553,8 @@ class EnumMetaInit(abstract.SimpleFunction):
     # that value. Otherwise, wrap it in a tuple.
     value = abstract_utils.get_atomic_value(value_var)
     if self.ctx.matcher(node).match_from_mro(
-        value.cls, self.ctx.convert.tuple_type):
+        value.cls, self.ctx.convert.tuple_type
+    ):
       args = value_var
     else:
       args = self.ctx.convert.build_tuple(node, [value_var])
@@ -514,8 +582,11 @@ class EnumMetaInit(abstract.SimpleFunction):
         # Interpreter classes don't have "maybe_missing_members" set even if
         # they have _HAS_DYNAMIC_ATTRIBUTES. But for enums, those markers should
         # apply to the whole class.
-        if ((base.cls.full_name != "enum.EnumMeta") or
-            base.maybe_missing_members or base.has_dynamic_attributes()):
+        if (
+            (base.cls.full_name != "enum.EnumMeta")
+            or base.maybe_missing_members
+            or base.has_dynamic_attributes()
+        ):
           cls.maybe_missing_members = True
           return
 
@@ -530,17 +601,20 @@ class EnumMetaInit(abstract.SimpleFunction):
     for name, local in self._get_class_locals(node, cls.name, cls.members):
       if self._not_valid_member(node, name, local):
         continue
-      assert local.orig, ("A local with no assigned value was passed to the "
-                          "enum overlay.")
+      assert (
+          local.orig
+      ), "A local with no assigned value was passed to the enum overlay."
       value = local.orig
       if self._is_orig_auto(value):
         node, value = self._call_generate_next_value(node, cls, name)
       if enum_new:
         new_args = function.Args(
             posargs=(cls.to_variable(node),),
-            starargs=self._value_to_starargs(node, value, base_type))
+            starargs=self._value_to_starargs(node, value, base_type),
+        )
         node, member_var = function.call_function(
-            self.ctx, node, enum_new, new_args, fallback_to_unsolvable=False)
+            self.ctx, node, enum_new, new_args, fallback_to_unsolvable=False
+        )
         # It's possible (but not likely) for member_var to have multiple
         # bindings of the same type. (See test_multiple_value_bindings in
         # test_enums.py.) This isn't an error, but members need to be Instances.
@@ -548,7 +622,8 @@ class EnumMetaInit(abstract.SimpleFunction):
           member = abstract_utils.get_atomic_value(member_var)
         except abstract_utils.ConversionError:
           if member_var.data and all(
-              m.cls == member_var.data[0].cls for m in member_var.data):
+              m.cls == member_var.data[0].cls for m in member_var.data
+          ):
             member = member_var.data[0]
           else:
             member_var = self.ctx.vm.convert.create_new_unknown(node)
@@ -558,42 +633,51 @@ class EnumMetaInit(abstract.SimpleFunction):
         # creating the class -- pytype complains about recursive types.
         member = abstract.Instance(cls, self.ctx)
         member_var = member.to_variable(node)
-      # This makes literal enum equality checks easier. We could check the name
-      # attribute that we set below, but it's easier to compare these strings.
-      # Use the fully qualified name to be consistent with how literal enums
-      # are parsed from type stubs.
-      member.name = f"{cls.full_name}.{name}"
-      if "_value_" not in member.members:
-        if base_type:
-          args = function.Args(
-              posargs=(),
-              starargs=self._value_to_starargs(node, value, base_type))
-          node, value = base_type.call(node, base_type.to_binding(node), args)
-        member.members["_value_"] = value
-      if "__init__" in cls:
-        init_args = function.Args(
-            posargs=(member_var,),
-            starargs=self._value_to_starargs(node, value, base_type))
-        node = cls.call_init(node, cls.to_binding(node), init_args)
-      member.members["value"] = member.members["_value_"]
-      member.members["name"] = self.ctx.convert.build_string(node, name)
-      for attr_name in member.members:
-        if attr_name in ("name", "value"):
-          continue
-        member_attrs[attr_name].extend(member.members[attr_name].data)
+      if isinstance(member, abstract.SimpleValue):
+        # This makes literal enum equality checks easier. We could check the
+        # name attribute that we set below, but it's easier to compare these
+        # strings. Use the fully qualified name to be consistent with how
+        # literal enums are parsed from type stubs.
+        member.name = f"{cls.full_name}.{name}"
+        if "_value_" not in member.members:
+          if base_type:
+            args = function.Args(
+                posargs=(),
+                starargs=self._value_to_starargs(node, value, base_type),
+            )
+            node, value = base_type.call(node, base_type.to_binding(node), args)
+          member.members["_value_"] = value
+        if "__init__" in cls:
+          init_args = function.Args(
+              posargs=(member_var,),
+              starargs=self._value_to_starargs(node, value, base_type),
+          )
+          node, init = self.ctx.attribute_handler.get_attribute(
+              node, cls, "__init__", cls.to_binding(node)
+          )
+          node, _ = function.call_function(self.ctx, node, init, init_args)
+        member.members["value"] = member.members["_value_"]
+        member.members["name"] = self.ctx.convert.build_string(node, name)
+        for attr_name in member.members:
+          if attr_name in ("name", "value"):
+            continue
+          member_attrs[attr_name].extend(member.members[attr_name].data)
       cls.members[name] = member.to_variable(node)
       member_types.extend(value.data)
     # After processing enum members, there's some work to do on the enum itself.
     # If cls has a __new__, save it for later. (See _get_member_new above.)
     # It needs to be marked as a classmethod, or else pytype will try to
     # pass an instance of cls instead of cls when analyzing it.
-    if "__new__" in cls:
+    if "__new__" in cls.members:
       saved_new = cls.members["__new__"]
-      if not any(isinstance(x, special_builtins.ClassMethodInstance)
-                 for x in saved_new.data):
+      if not any(
+          isinstance(x, special_builtins.ClassMethodInstance)
+          for x in saved_new.data
+      ):
         args = function.Args(posargs=(saved_new,))
         node, saved_new = self.ctx.vm.load_special_builtin("classmethod").call(
-            node, None, args)
+            node, None, args
+        )
       cls.members["__new_member__"] = saved_new
     self._mark_dynamic_enum(cls)
     if base_type:
@@ -618,11 +702,13 @@ class EnumMetaInit(abstract.SimpleFunction):
     # However, we skip this if it's already a staticmethod.
     if "_generate_next_value_" in cls.members:
       gnv = cls.members["_generate_next_value_"]
-      if not any(isinstance(x, special_builtins.StaticMethodInstance)
-                 for x in gnv.data):
+      if not any(
+          isinstance(x, special_builtins.StaticMethodInstance) for x in gnv.data
+      ):
         args = function.Args(posargs=(gnv,))
         node, new_gnv = self.ctx.vm.load_special_builtin("staticmethod").call(
-            node, None, args)
+            node, None, args
+        )
         cls.members["_generate_next_value_"] = new_gnv
     return node
 
@@ -640,13 +726,17 @@ class EnumMetaInit(abstract.SimpleFunction):
       # but pytype doesn't differentiate between class and instance attributes
       # for PyTDClass and there's no mechanism to ensure canonical instances
       # have these attributes.
-      if (isinstance(pytd_val.type, pytd.Annotated) and
-          "'property'" in pytd_val.type.annotations):
+      if (
+          isinstance(pytd_val.type, pytd.Annotated)
+          and "'property'" in pytd_val.type.annotations
+      ):
         continue
       # Class-level attributes are marked as ClassVars, and should not be
       # converted to enum instances either.
-      if (isinstance(pytd_val.type, pytd.GenericType) and
-          pytd_val.type.base_type.name == "typing.ClassVar"):
+      if (
+          isinstance(pytd_val.type, pytd.GenericType)
+          and pytd_val.type.base_type.name == "typing.ClassVar"
+      ):
         continue
       # Build instances directly, because you can't call instantiate() when
       # creating the class -- pytype complains about recursive types.
@@ -658,8 +748,10 @@ class EnumMetaInit(abstract.SimpleFunction):
       member.name = f"{cls.full_name}.{pytd_val.name}"
       member.members["name"] = self.ctx.convert.constant_to_var(
           pyval=pytd.Constant(
-              name="name", type=self._str_pytd, value=pytd_val.name),
-          node=node)
+              name="name", type=self._str_pytd, value=pytd_val.name
+          ),
+          node=node,
+      )
       # Some type stubs may use the class type for enum member values, instead
       # of the actual value type. Detect that and use Any.
       if pytd_val.type.name == cls.pytd_cls.name:
@@ -667,7 +759,8 @@ class EnumMetaInit(abstract.SimpleFunction):
       else:
         value_type = pytd_val.type
       member.members["value"] = self.ctx.convert.constant_to_var(
-          pyval=pytd.Constant(name="value", type=value_type), node=node)
+          pyval=pytd.Constant(name="value", type=value_type), node=node
+      )
       member.members["_value_"] = member.members["value"]
       cls._member_map[pytd_val.name] = member  # pylint: disable=protected-access
       cls.members[pytd_val.name] = member.to_variable(node)
@@ -676,8 +769,15 @@ class EnumMetaInit(abstract.SimpleFunction):
     # Of course, this can be moved later once custom __init__ is supported.
     self._mark_dynamic_enum(cls)
     if member_types:
+      # Convert Literals to their underlying types, to avoid raising type errors
+      # in code that passes a non-concrete value to an enum constructor.
+      typ = lambda t: pytd.LateType(f"builtins.{t.value.__class__.__name__}")
+      types = [
+          typ(t) if isinstance(t, pytd.Literal) else t for t in member_types
+      ]
       member_type = self.ctx.convert.constant_to_value(
-          pytd_utils.JoinTypes(member_types))
+          pytd_utils.JoinTypes(types)
+      )
       # Only set the lookup-only __new__ on non-empty enums, since using a
       # non-empty enum for the functional API is a type error.
       # Note that this has to happen AFTER _mark_dynamic_enum.
@@ -693,13 +793,14 @@ class EnumMetaInit(abstract.SimpleFunction):
     # cls is the EnumInstance created by EnumBuilder.make_class, or an
     # abstract.PyTDClass created by convert.py.
     cls_var = argmap["cls"]
-    cls, = cls_var.data
+    (cls,) = cls_var.data
 
     # This function will get called for every class that has enum.EnumMeta as
     # its metaclass, including enum.Enum and other enum module members.
     # We don't have anything to do for those, so return early.
-    if (isinstance(cls, abstract.PyTDClass) and
-        cls.full_name.startswith("enum.")):
+    if isinstance(cls, abstract.PyTDClass) and cls.full_name.startswith(
+        "enum."
+    ):
       return node, ret
 
     if isinstance(cls, abstract.InterpreterClass):
@@ -708,7 +809,8 @@ class EnumMetaInit(abstract.SimpleFunction):
       node = self._setup_pytdclass(node, cls)
     else:
       raise ValueError(
-          f"Expected an InterpreterClass or PyTDClass, but got {type(cls)}")
+          f"Expected an InterpreterClass or PyTDClass, but got {type(cls)}"
+      )
 
     return node, ret
 
@@ -725,7 +827,8 @@ class EnumMetaGetItem(abstract.SimpleFunction):
         kwonly_params=(),
         kwargs_name=None,
         defaults={},
-        annotations={"name": ctx.convert.str_type})
+        annotations={"name": ctx.convert.str_type},
+    )
     super().__init__(sig, ctx)
 
   def _get_member_by_name(
@@ -738,7 +841,7 @@ class EnumMetaGetItem(abstract.SimpleFunction):
         enum.load_lazy_attribute(name)
         return enum.members[name]
 
-  def call(self, node, _, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     _, argmap = self.match_and_map_args(node, args, alias_map)
     cls_var = argmap["cls"]
     name_var = argmap["name"]
@@ -760,6 +863,7 @@ class EnumMetaGetItem(abstract.SimpleFunction):
     if inst:
       return node, inst
     else:
-      self.ctx.errorlog.attribute_error(self.ctx.vm.frames, cls_var.bindings[0],
-                                        name)
+      self.ctx.errorlog.attribute_error(
+          self.ctx.vm.frames, cls_var.bindings[0], name
+      )
       return node, self.ctx.new_unsolvable(node)
