@@ -1,12 +1,13 @@
 """Custom implementations of builtin types."""
 
-import sys
+import contextlib
 
 from pytype.abstract import abstract
 from pytype.abstract import abstract_utils
 from pytype.abstract import class_mixin
 from pytype.abstract import function
 from pytype.abstract import mixin
+from pytype.errors import error_types
 
 
 class TypeNew(abstract.PyTDFunction):
@@ -21,14 +22,18 @@ class TypeNew(abstract.PyTDFunction):
         if not bases:
           bases = [self.ctx.convert.object_type.to_variable(self.ctx.root_node)]
         props = class_mixin.ClassBuilderProperties(
-            name_var, bases, class_dict_var, metaclass_var=cls)
+            name_var, bases, class_dict_var, metaclass_var=cls
+        )
         node, variable = self.ctx.make_class(node, props)
       except abstract_utils.ConversionError:
         pass
       else:
         return node, variable
-    elif (args.posargs and self.ctx.callself_stack and
-          args.posargs[-1].data == self.ctx.callself_stack[-1].data):
+    elif (
+        args.posargs
+        and self.ctx.callself_stack
+        and args.posargs[-1].data == self.ctx.callself_stack[-1].data
+    ):
       # We're calling type(self) in an __init__ method. A common pattern for
       # making a class non-instantiable is:
       #   class Foo:
@@ -42,7 +47,8 @@ class TypeNew(abstract.PyTDFunction):
       self.match_args(node, args)  # May raise FailedFunctionCall.
       return node, self.ctx.new_unsolvable(node)
     elif args.posargs and all(
-        v.full_name == "typing.Protocol" for v in args.posargs[-1].data):
+        v.full_name == "typing.Protocol" for v in args.posargs[-1].data
+    ):
       # type(Protocol) is a _ProtocolMeta class that inherits from abc.ABCMeta.
       # Changing the definition of Protocol in typing.pytd to include this
       # metaclass causes a bunch of weird breakages, so we instead return the
@@ -57,34 +63,31 @@ class TypeNew(abstract.PyTDFunction):
     ret = self.ctx.program.NewVariable()
     for b in raw_ret.bindings:
       value = self.ctx.annotation_utils.deformalize(b.data)
-      ret.AddBinding(value, {b}, node)
+      ret.PasteBindingWithNewData(b, value)
     return node, ret
 
 
 class BuiltinFunction(abstract.PyTDFunction):
   """Implementation of functions in builtins.pytd."""
 
-  name = None
+  _NAME: str = None
 
   @classmethod
   def make(cls, ctx):
-    assert cls.name
-    return super().make(cls.name, ctx, "builtins")
+    assert cls._NAME
+    return super().make(cls._NAME, ctx, "builtins")
 
   @classmethod
-  def make_alias(cls, name, ctx, module_name):
-    """Create an alias to this function."""
-    # See overlays/pytype_extensions_overlay.py
-    self = super().make(name, ctx, module_name)
-    self.module_name = module_name
-    return self
+  def make_alias(cls, name, ctx, module):
+    return super().make(name, ctx, module)
 
   def get_underlying_method(self, node, receiver, method_name):
     """Get the bound method that a built-in function delegates to."""
     results = []
     for b in receiver.bindings:
       node, result = self.ctx.attribute_handler.get_attribute(
-          node, b.data, method_name, valself=b)
+          node, b.data, method_name, valself=b
+      )
       if result is not None:
         results.append(result)
     if results:
@@ -104,9 +107,9 @@ def get_file_mode(sig, args):
 class Abs(BuiltinFunction):
   """Implements abs."""
 
-  name = "abs"
+  _NAME = "abs"
 
-  def call(self, node, _, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     self.match_args(node, args)
     arg = args.posargs[0]
     node, fn = self.get_underlying_method(node, arg, "__abs__")
@@ -119,7 +122,7 @@ class Abs(BuiltinFunction):
 class Next(BuiltinFunction):
   """Implements next."""
 
-  name = "next"
+  _NAME = "next"
 
   def _get_args(self, args):
     arg = args.posargs[0]
@@ -131,7 +134,7 @@ class Next(BuiltinFunction):
       default = self.ctx.program.NewVariable()
     return arg, default
 
-  def call(self, node, _, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     self.match_args(node, args)
     arg, default = self._get_args(args)
     node, fn = self.get_underlying_method(node, arg, "__next__")
@@ -143,6 +146,20 @@ class Next(BuiltinFunction):
       return node, self.ctx.new_unsolvable(node)
 
 
+class Round(BuiltinFunction):
+  """Implements round."""
+
+  _NAME = "round"
+
+  def call(self, node, func, args, alias_map=None):
+    self.match_args(node, args)
+    node, fn = self.get_underlying_method(node, args.posargs[0], "__round__")
+    if fn is None:
+      return super().call(node, func, args, alias_map)
+    new_args = args.replace(posargs=args.posargs[1:])
+    return function.call_function(self.ctx, node, fn, new_args)
+
+
 class ObjectPredicate(BuiltinFunction):
   """The base class for builtin predicates of the form f(obj, ...) -> bool.
 
@@ -153,13 +170,13 @@ class ObjectPredicate(BuiltinFunction):
   def run(self, node, args, result):
     raise NotImplementedError(self.__class__.__name__)
 
-  def call(self, node, _, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     try:
       self.match_args(node, args)
       node = self.ctx.connect_new_cfg_node(node, f"CallPredicate:{self.name}")
       result = self.ctx.program.NewVariable()
       self.run(node, args, result)
-    except function.InvalidParameters as ex:
+    except error_types.InvalidParameters as ex:
       self.ctx.errorlog.invalid_function_call(self.ctx.vm.frames, ex)
       result = self.ctx.new_unsolvable(node)
     return node, result
@@ -179,8 +196,9 @@ class UnaryPredicate(ObjectPredicate):
   def run(self, node, args, result):
     for obj in args.posargs[0].bindings:
       node, pyval = self._call_predicate(node, obj)
-      result.AddBinding(self.ctx.convert.bool_values[pyval],
-                        source_set=(obj,), where=node)
+      result.AddBinding(
+          self.ctx.convert.bool_values[pyval], source_set=(obj,), where=node
+      )
 
 
 class BinaryPredicate(ObjectPredicate):
@@ -195,19 +213,40 @@ class BinaryPredicate(ObjectPredicate):
     raise NotImplementedError(self.__class__.__name__)
 
   def run(self, node, args, result):
-    for left in abstract_utils.expand_type_parameter_instances(
-        args.posargs[0].bindings):
-      for right in abstract_utils.expand_type_parameter_instances(
-          args.posargs[1].bindings):
+    for right in abstract_utils.expand_type_parameter_instances(
+        args.posargs[1].bindings
+    ):
+      one_result = []
+      for left in abstract_utils.expand_type_parameter_instances(
+          args.posargs[0].bindings
+      ):
         node, pyval = self._call_predicate(node, left, right)
-        result.AddBinding(self.ctx.convert.bool_values[pyval],
-                          source_set=(left, right), where=node)
+        one_result.append((left, node, pyval))
+      unsolvable_matches = any(
+          isinstance(left.data, abstract.Unsolvable) and pyval in (None, True)
+          for (left, _, pyval) in one_result
+      )
+      for left, result_node, pyval in one_result:
+        if (
+            unsolvable_matches
+            and not isinstance(left.data, abstract.Unsolvable)
+            and pyval is None
+        ):
+          # If unsolvable (i.e., Any) satisfies the predicate, then we should
+          # ignore non-Any values. See test_splits2:AmbiguousIsInstanceTest for
+          # the reasoning.
+          pyval = False
+        result.AddBinding(
+            self.ctx.convert.bool_values[pyval],
+            source_set=(left, right),
+            where=result_node,
+        )
 
 
 class HasAttr(BinaryPredicate):
   """The hasattr() function."""
 
-  name = "hasattr"
+  _NAME = "hasattr"
 
   def _call_predicate(self, node, left, right):
     return self._has_attr(node, left.data, right.data)
@@ -217,10 +256,8 @@ class HasAttr(BinaryPredicate):
 
     Args:
       node: The given node.
-      obj: A BaseValue, generally the left hand side of a
-          hasattr() call.
-      attr: A BaseValue, generally the right hand side of a
-          hasattr() call.
+      obj: A BaseValue, generally the left hand side of a hasattr() call.
+      attr: A BaseValue, generally the right hand side of a hasattr() call.
 
     Returns:
       (node, result) where result = True if the object has attribute attr, False
@@ -229,8 +266,9 @@ class HasAttr(BinaryPredicate):
     if isinstance(obj, abstract.AMBIGUOUS_OR_EMPTY):
       return node, None
     # If attr is not a literal constant, don't try to resolve it.
-    if (not isinstance(attr, mixin.PythonConstant) or
-        not isinstance(attr.pyval, str)):
+    if not isinstance(attr, abstract.PythonConstant) or not isinstance(
+        attr.pyval, str
+    ):
       return node, None
     node, ret = self.ctx.attribute_handler.get_attribute(node, obj, attr.pyval)
     return node, ret is not None
@@ -239,7 +277,7 @@ class HasAttr(BinaryPredicate):
 class IsInstance(BinaryPredicate):
   """The isinstance() function."""
 
-  name = "isinstance"
+  _NAME = "isinstance"
 
   def _call_predicate(self, node, left, right):
     return node, self._is_instance(left.data, right.data)
@@ -248,18 +286,18 @@ class IsInstance(BinaryPredicate):
     """Check if the object matches a class specification.
 
     Args:
-      obj: A BaseValue, generally the left hand side of an
-          isinstance() call.
-      class_spec: A BaseValue, generally the right hand side of an
-          isinstance() call.
+      obj: A BaseValue, generally the left hand side of an isinstance() call.
+      class_spec: A BaseValue, generally the right hand side of an isinstance()
+        call.
 
     Returns:
       True if the object is derived from a class in the class_spec, False if
       it is not, and None if it is ambiguous whether obj matches class_spec.
     """
     cls = obj.cls
-    if (isinstance(obj, abstract.AMBIGUOUS_OR_EMPTY) or
-        isinstance(cls, abstract.AMBIGUOUS_OR_EMPTY)):
+    if isinstance(obj, abstract.AMBIGUOUS_OR_EMPTY) or isinstance(
+        cls, abstract.AMBIGUOUS_OR_EMPTY
+    ):
       return None
     return abstract_utils.check_against_mro(self.ctx, cls, class_spec)
 
@@ -267,7 +305,7 @@ class IsInstance(BinaryPredicate):
 class IsSubclass(BinaryPredicate):
   """The issubclass() function."""
 
-  name = "issubclass"
+  _NAME = "issubclass"
 
   def _call_predicate(self, node, left, right):
     return node, self._is_subclass(left.data, right.data)
@@ -293,7 +331,7 @@ class IsSubclass(BinaryPredicate):
 class IsCallable(UnaryPredicate):
   """The callable() function."""
 
-  name = "callable"
+  _NAME = "callable"
 
   def _call_predicate(self, node, obj):
     return self._is_callable(node, obj)
@@ -320,7 +358,8 @@ class IsCallable(UnaryPredicate):
       return node, True
     # Otherwise, see if the object has a __call__ method.
     node, ret = self.ctx.attribute_handler.get_attribute(
-        node, val, "__call__", valself=obj)
+        node, val, "__call__", valself=obj
+    )
     return node, ret is not None
 
 
@@ -331,13 +370,21 @@ class BuiltinClass(abstract.PyTDClass):
   module in builtins and inherit the custom behaviour.
   """
 
-  def __init__(self, ctx, name, module="builtins"):
-    if module == "builtins":
-      pytd_cls = ctx.loader.lookup_builtin(f"builtins.{name}")
-    else:
-      ast = ctx.loader.import_name(module)
-      pytd_cls = ast.Lookup(f"{module}.{name}")
-    super().__init__(name, pytd_cls, ctx)
+  _NAME: str = None
+
+  @classmethod
+  def make(cls, ctx):
+    assert cls._NAME
+    return cls(cls._NAME, ctx, "builtins")
+
+  @classmethod
+  def make_alias(cls, name, ctx, module):
+    # Although this method has the same signature as __init__, it makes alias
+    # creation more readable and consistent with BuiltinFunction.
+    return cls(name, ctx, module)
+
+  def __init__(self, name, ctx, module):
+    super().__init__(name, ctx.loader.lookup_pytd(module, name), ctx)
     self.module = module
 
 
@@ -359,7 +406,8 @@ class SuperInstance(abstract.BaseValue):
     ret = []
     for b in obj.bindings:
       _, attr = self.ctx.attribute_handler.get_attribute(
-          node, b.data, "__get__", valself=b)
+          node, b.data, "__get__", valself=b
+      )
       if attr:
         ret.append(attr)
     if ret:
@@ -378,7 +426,7 @@ class SuperInstance(abstract.BaseValue):
     else:
       return super().get_special_attribute(node, name, valself)
 
-  def call(self, node, _, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     self.ctx.errorlog.not_callable(self.ctx.vm.frames, self)
     return node, self.ctx.new_unsolvable(node)
 
@@ -388,30 +436,29 @@ class Super(BuiltinClass):
 
   # Minimal signature, only used for constructing exceptions.
   _SIGNATURE = function.Signature.from_param_names("super", ("cls", "self"))
+  _NAME = "super"
 
-  def __init__(self, ctx):
-    super().__init__(ctx, "super")
-
-  def call(self, node, _, args):
+  def call(self, node, func, args, alias_map=None):
     result = self.ctx.program.NewVariable()
     num_args = len(args.posargs)
     if num_args == 0:
-      # The implicit type argument is available in a freevar named '__class__'.
-      cls_var = None
       # If we are in a list comprehension we want the enclosing frame.
       index = -1
-      while self.ctx.vm.frames[index].f_code.co_name == "<listcomp>":
+      while self.ctx.vm.frames[index].f_code.name == "<listcomp>":
         index -= 1
       frame = self.ctx.vm.frames[index]
-      for i, free_var in enumerate(frame.f_code.co_freevars):
-        if free_var == abstract.BuildClass.CLOSURE_NAME:
-          cls_var = frame.cells[len(frame.f_code.co_cellvars) + i]
-          break
+      # The implicit type argument is available in a freevar named '__class__'.
+      closure_name = abstract.BuildClass.CLOSURE_NAME
+      if closure_name in frame.f_code.freevars:
+        cls_var = frame.get_cell_by_name(closure_name)
+      else:
+        cls_var = None
       if not (cls_var and cls_var.bindings):
         self.ctx.errorlog.invalid_super_call(
             self.ctx.vm.frames,
             message="Missing __class__ closure for super call.",
-            details="Is 'super' being called from a method defined in a class?")
+            details="Is 'super' being called from a method defined in a class?",
+        )
         return node, self.ctx.new_unsolvable(node)
       # The implicit super object argument is the first argument to the function
       # calling 'super'.
@@ -419,38 +466,41 @@ class Super(BuiltinClass):
       if not self_arg:
         self.ctx.errorlog.invalid_super_call(
             self.ctx.vm.frames,
-            message="Missing 'self' argument to 'super' call.")
+            message="Missing 'self' argument to 'super' call.",
+        )
         return node, self.ctx.new_unsolvable(node)
       super_objects = self_arg.bindings
     elif 1 <= num_args <= 2:
       cls_var = args.posargs[0]
       super_objects = args.posargs[1].bindings if num_args == 2 else [None]
     else:
-      raise function.WrongArgCount(self._SIGNATURE, args, self.ctx)
+      raise error_types.WrongArgCount(self._SIGNATURE, args, self.ctx)
     for cls in cls_var.bindings:
       if isinstance(cls.data, (abstract.Class, abstract.AMBIGUOUS_OR_EMPTY)):
         cls_data = cls.data
       elif any(base.full_name == "builtins.type" for base in cls.data.cls.mro):
         cls_data = self.ctx.convert.unsolvable
       else:
-        bad = abstract_utils.BadType(name="cls", typ=self.ctx.convert.type_type)
-        raise function.WrongArgTypes(
-            self._SIGNATURE, args, self.ctx, bad_param=bad)
+        bad = error_types.BadType(name="cls", typ=self.ctx.convert.type_type)
+        raise error_types.WrongArgTypes(
+            self._SIGNATURE, args, self.ctx, bad_param=bad
+        )
       for obj in super_objects:
         if obj:
           result.AddBinding(
-              SuperInstance(cls_data, obj.data, self.ctx), [cls, obj], node)
+              SuperInstance(cls_data, obj.data, self.ctx), [cls, obj], node
+          )
         else:
           result.AddBinding(
-              SuperInstance(cls_data, None, self.ctx), [cls], node)
+              SuperInstance(cls_data, None, self.ctx), [cls], node
+          )
     return node, result
 
 
 class Object(BuiltinClass):
   """Implementation of builtins.object."""
 
-  def __init__(self, ctx):
-    super().__init__(ctx, "object")
+  _NAME = "object"
 
   def is_object_new(self, func):
     """Whether the given function is object.__new__.
@@ -464,8 +514,9 @@ class Object(BuiltinClass):
     """
     self.load_lazy_attribute("__new__")
     self.load_lazy_attribute("__new__extra_args")
-    return ([func] == self.members["__new__"].data or
-            [func] == self.members["__new__extra_args"].data)
+    return [func] == self.members["__new__"].data or [func] == self.members[
+        "__new__extra_args"
+    ].data
 
   def _has_own(self, node, cls, method):
     """Whether a class has its own implementation of a particular method.
@@ -498,20 +549,22 @@ class Object(BuiltinClass):
       if name == "__new__" and self._has_own(node, val, "__init__"):
         self.load_lazy_attribute("__new__extra_args")
         return self.members["__new__extra_args"]
-      elif (name == "__init__" and isinstance(val, abstract.Instance) and
-            self._has_own(node, val.cls, "__new__")):
+      elif (
+          name == "__init__"
+          and isinstance(val, abstract.Instance)
+          and self._has_own(node, val.cls, "__new__")
+      ):
         self.load_lazy_attribute("__init__extra_args")
         return self.members["__init__extra_args"]
     return super().get_special_attribute(node, name, valself)
 
 
-class RevealType(abstract.BaseValue):
+class RevealType(BuiltinFunction):
   """For debugging. reveal_type(x) prints the type of "x"."""
 
-  def __init__(self, ctx):
-    super().__init__("reveal_type", ctx)
+  _NAME = "reveal_type"
 
-  def call(self, node, _, args, alias_map=None):
+  def call(self, node, func, args, alias_map=None):
     for a in args.posargs:
       self.ctx.errorlog.reveal_type(self.ctx.vm.frames, node, a)
     return node, self.ctx.convert.build_none(node)
@@ -522,26 +575,39 @@ class AssertType(BuiltinFunction):
 
   # Minimal signature, only used for constructing exceptions.
   _SIGNATURE = function.Signature.from_param_names(
-      "assert_type", ("variable", "type"))
+      "assert_type", ("variable", "type")
+  )
+  _NAME = "assert_type"
 
-  name = "assert_type"
-
-  def call(self, node, _, args, alias_map=None):
-    if len(args.posargs) == 1:
-      a, = args.posargs
-      t = None
-    elif len(args.posargs) == 2:
-      a, t = args.posargs
+  def call(self, node, func, args, alias_map=None):
+    if len(args.posargs) == 2:
+      var, typ = args.posargs
     else:
-      raise function.WrongArgCount(self._SIGNATURE, args, self.ctx)
-    self.ctx.errorlog.assert_type(self.ctx.vm.frames, node, a, t)
+      raise error_types.WrongArgCount(self._SIGNATURE, args, self.ctx)
+
+    # Convert both args to strings and compare them
+    pp = self.ctx.errorlog.pretty_printer
+    actual = pp.print_var_type(var, node)
+    try:
+      expected = abstract_utils.get_atomic_python_constant(typ, str)
+    except abstract_utils.ConversionError:
+      # NOTE: Converting types to strings is provided as a fallback, but is not
+      # really supported, since there are issues around name resolution.
+      typ = self.ctx.annotation_utils.extract_annotation(
+          node, typ, "assert_type", self.ctx.vm.simple_stack()
+      )
+      node, instance = self.ctx.vm.init_class_and_forward_node(node, typ)
+      expected = pp.print_var_type(instance, node)
+    if actual != expected:
+      self.ctx.errorlog.assert_type(self.ctx.vm.frames, actual, expected)
     return node, self.ctx.convert.build_none(node)
 
 
-class PropertyTemplate(BuiltinClass):
-  """Template for property decorators."""
+class Property(BuiltinClass):
+  """Property decorator."""
 
   _KEYS = ["fget", "fset", "fdel", "doc"]
+  _NAME = "property"
 
   def signature(self):
     # Minimal signature, only used for constructing exceptions.
@@ -551,14 +617,17 @@ class PropertyTemplate(BuiltinClass):
     ret = dict(zip(self._KEYS, args.posargs))
     for k, v in args.namedargs.items():
       if k not in self._KEYS:
-        raise function.WrongKeywordArgs(self.signature(), args, self.ctx, [k])
+        raise error_types.WrongKeywordArgs(
+            self.signature(), args, self.ctx, [k]
+        )
       ret[k] = v
     return ret
 
-  def call(self, node, funcv, args):
+  def call(self, node, func, args, alias_map=None):
     property_args = self._get_args(args)
     return node, PropertyInstance(
-        self.ctx, self.name, self, **property_args).to_variable(node)
+        self.ctx, self.name, self, **property_args
+    ).to_variable(node)
 
 
 def _is_fn_abstract(func_var):
@@ -598,63 +667,64 @@ class PropertyInstance(abstract.Function, mixin.HasSlots):
 
   def fget_slot(self, node, obj, objtype):
     obj_val = abstract_utils.get_atomic_value(
-        obj, default=self.ctx.convert.unsolvable)
-    set_self_annot_managers = []
+        obj, default=self.ctx.convert.unsolvable
+    )
+    # If this property is defined on a generic class, we need to annotate self
+    # with a parameterized type for the property return type to be computed
+    # properly, e.g.:
+    #   @property
+    #   def x(self) -> T: ...
+    # is changed to:
+    #   @property
+    #   def x(self: Foo[T]) -> T: ...
     t = abstract_utils.get_generic_type(obj_val)
-    if t and any(self in member.data for member in t.members.values()):
-      # This property is defined on a generic class. We need to annotate self
-      # with a parameterized type for the property return type to be computed
-      # properly, e.g.:
-      #   @property
-      #   def x(self) -> T: ...
-      # is changed to:
-      #   @property
-      #   def x(self: Foo[T]) -> T: ...
-      for f in self.fget.data:
-        if f.should_replace_self_annot():
-          set_self_annot_managers.append(f.set_self_annot(t))
-    for manager in set_self_annot_managers:
-      manager.__enter__()
-    try:
-      return function.call_function(self.ctx, node, self.fget,
-                                    function.Args((obj,)))
-    finally:
-      exc_info = sys.exc_info()
-      for manager in set_self_annot_managers:
-        manager.__exit__(*exc_info)
+    generic = t and any(self in member.data for member in t.members.values())
+    with contextlib.ExitStack() as stack:
+      if generic:
+        for f in self.fget.data:
+          if f.should_set_self_annot():
+            stack.enter_context(f.set_self_annot(t))
+      return function.call_function(
+          self.ctx, node, self.fget, function.Args((obj,))
+      )
 
   def fset_slot(self, node, obj, value):
-    return function.call_function(self.ctx, node, self.fset,
-                                  function.Args((obj, value)))
+    return function.call_function(
+        self.ctx, node, self.fset, function.Args((obj, value))
+    )
 
   def fdelete_slot(self, node, obj):
-    return function.call_function(self.ctx, node, self.fdel,
-                                  function.Args((obj,)))
+    return function.call_function(
+        self.ctx, node, self.fdel, function.Args((obj,))
+    )
 
   def getter_slot(self, node, fget):
-    prop = PropertyInstance(self.ctx, self.name, self.cls, fget, self.fset,
-                            self.fdel, self.doc)
+    prop = PropertyInstance(
+        self.ctx, self.name, self.cls, fget, self.fset, self.fdel, self.doc
+    )
     result = self.ctx.program.NewVariable([prop], fget.bindings, node)
     return node, result
 
   def setter_slot(self, node, fset):
-    prop = PropertyInstance(self.ctx, self.name, self.cls, self.fget, fset,
-                            self.fdel, self.doc)
+    prop = PropertyInstance(
+        self.ctx, self.name, self.cls, self.fget, fset, self.fdel, self.doc
+    )
     result = self.ctx.program.NewVariable([prop], fset.bindings, node)
     return node, result
 
   def deleter_slot(self, node, fdel):
-    prop = PropertyInstance(self.ctx, self.name, self.cls, self.fget, self.fset,
-                            fdel, self.doc)
+    prop = PropertyInstance(
+        self.ctx, self.name, self.cls, self.fget, self.fset, fdel, self.doc
+    )
     result = self.ctx.program.NewVariable([prop], fdel.bindings, node)
     return node, result
 
-
-class Property(PropertyTemplate):
-  """Property method decorator."""
-
-  def __init__(self, ctx):
-    super().__init__(ctx, "property")
+  def update_signature_scope(self, cls):
+    for fvar in (self.fget, self.fset, self.fdel):
+      if fvar:
+        for f in fvar.data:
+          if isinstance(f, abstract.Function):
+            f.update_signature_scope(cls)
 
 
 def _check_method_decorator_arg(fn_var, name, ctx):
@@ -664,10 +734,8 @@ def _check_method_decorator_arg(fn_var, name, ctx):
       _ = function.get_signatures(d)
     except NotImplementedError:
       # We are wrapping something that is not a function in a method decorator.
-      # TODO(mdemello): The error line is the function definition rather than
-      # the line with `@classmethod` or `@staticmethod`
       details = f"@{name} applied to something that is not a function."
-      ctx.errorlog.not_callable(ctx.vm.frames, d, details)
+      ctx.errorlog.not_callable(ctx.vm.stack(), d, details)
       return False
   return True
 
@@ -689,27 +757,22 @@ class StaticMethodInstance(abstract.Function, mixin.HasSlots):
     return node, self.func
 
 
-class StaticMethodTemplate(BuiltinClass):
+class StaticMethod(BuiltinClass):
   """Static method decorator."""
 
   # Minimal signature, only used for constructing exceptions.
   _SIGNATURE = function.Signature.from_param_names("staticmethod", ("func",))
+  _NAME = "staticmethod"
 
-  def call(self, node, funcv, args):
+  def call(self, node, func, args, alias_map=None):
     if len(args.posargs) != 1:
-      raise function.WrongArgCount(self._SIGNATURE, args, self.ctx)
+      raise error_types.WrongArgCount(self._SIGNATURE, args, self.ctx)
     arg = args.posargs[0]
     if not _check_method_decorator_arg(arg, "staticmethod", self.ctx):
       return node, self.ctx.new_unsolvable(node)
     return node, StaticMethodInstance(self.ctx, self, arg).to_variable(node)
 
 
-class StaticMethod(StaticMethodTemplate):
-
-  def __init__(self, ctx):
-    super().__init__(ctx, "staticmethod")
-
-
 class ClassMethodCallable(abstract.BoundFunction):
   """Tag a ClassMethod bound function so we can dispatch on it."""
 
@@ -731,58 +794,62 @@ class ClassMethodInstance(abstract.Function, mixin.HasSlots):
     results = [ClassMethodCallable(objtype, b.data) for b in self.func.bindings]
     return node, self.ctx.program.NewVariable(results, [], node)
 
+  def update_signature_scope(self, cls):
+    for f in self.func.data:
+      if isinstance(f, abstract.Function):
+        f.update_signature_scope(cls)
 
-class ClassMethodTemplate(BuiltinClass):
+
+class ClassMethod(BuiltinClass):
   """Class method decorator."""
+
   # Minimal signature, only used for constructing exceptions.
   _SIGNATURE = function.Signature.from_param_names("classmethod", ("func",))
+  _NAME = "classmethod"
 
-  def call(self, node, funcv, args):
+  def call(self, node, func, args, alias_map=None):
     if len(args.posargs) != 1:
-      raise function.WrongArgCount(self._SIGNATURE, args, self.ctx)
+      raise error_types.WrongArgCount(self._SIGNATURE, args, self.ctx)
     arg = args.posargs[0]
     if not _check_method_decorator_arg(arg, "classmethod", self.ctx):
       return node, self.ctx.new_unsolvable(node)
+    if any(isinstance(v, ClassMethodInstance) for v in arg.data):
+      return node, arg
     for d in arg.data:
       d.is_classmethod = True
       d.is_attribute_of_class = True
     return node, ClassMethodInstance(self.ctx, self, arg).to_variable(node)
 
 
-class ClassMethod(ClassMethodTemplate):
-
-  def __init__(self, ctx):
-    super().__init__(ctx, "classmethod")
-
-
 class Dict(BuiltinClass):
   """Implementation of builtins.dict."""
 
-  def __init__(self, ctx):
-    super().__init__(ctx, "dict")
+  _NAME = "dict"
 
-  def call(self, node, funcb, args):
+  def call(self, node, func, args, alias_map=None):
     if not args.has_non_namedargs():
       # special-case a dict constructor with explicit k=v args
       d = abstract.Dict(self.ctx)
-      for (k, v) in args.namedargs.items():
+      for k, v in args.namedargs.items():
         d.set_str_item(node, k, v)
       return node, d.to_variable(node)
     else:
-      return super().call(node, funcb, args)
+      return super().call(node, func, args, alias_map)
+
+  def get_special_attribute(self, node, name, valself):
+    # For doing something like getting the __getitem__ attribute to subscript
+    # dict, we want to use the real dict type.
+    return self.ctx.convert.dict_type.get_special_attribute(node, name, valself)
 
 
-class TypeTemplate(BuiltinClass, mixin.HasSlots):
+class Type(BuiltinClass, mixin.HasSlots):
+
+  _NAME = "type"
 
   def __init__(self, *args, **kwargs):
     super().__init__(*args, **kwargs)
     mixin.HasSlots.init_mixin(self)
     slot = self.ctx.convert.convert_pytd_function(
-        self.pytd_cls.Lookup("__new__"), TypeNew)
+        self.pytd_cls.Lookup("__new__"), TypeNew
+    )
     self.set_slot("__new__", slot)
-
-
-class Type(TypeTemplate):
-
-  def __init__(self, ctx):
-    super().__init__(ctx, "type")
