@@ -1,21 +1,25 @@
 """Implementation of TypedDict."""
 
 import dataclasses
-
-from typing import Any, Dict, Optional, Set
+from typing import Dict, Optional, Set
 
 from pytype.abstract import abstract
 from pytype.abstract import abstract_utils
 from pytype.abstract import function
+from pytype.errors import error_types
 from pytype.overlays import classgen
+from pytype.overlays import overlay_utils
 from pytype.pytd import pytd
 
 
-class TypedDictKeyMissing(function.DictKeyMissing):
-
-  def __init__(self, typed_dict: "TypedDict", key: Optional[str]):
-    super().__init__(key)
-    self.typed_dict = typed_dict
+def _is_required(value: abstract.BaseValue) -> Optional[bool]:
+  name = value.full_name
+  if name == "typing.Required":
+    return True
+  elif name == "typing.NotRequired":
+    return False
+  else:
+    return None
 
 
 @dataclasses.dataclass
@@ -23,7 +27,7 @@ class TypedDictProperties:
   """Collection of typed dict properties passed between various stages."""
 
   name: str
-  fields: Dict[str, Any]
+  fields: Dict[str, abstract.BaseValue]
   required: Set[str]
   total: bool
 
@@ -36,8 +40,17 @@ class TypedDictProperties:
     return self.keys - self.required
 
   def add(self, k, v, total):
-    self.fields[k] = v  # pylint: disable=unsupported-assignment-operation
-    if total:
+    """Adds key and value."""
+    req = _is_required(v)
+    if req is None:
+      value = v
+    elif isinstance(v, abstract.ParameterizedClass):
+      value = v.formal_type_parameters[abstract_utils.T]
+    else:
+      value = v.ctx.convert.unsolvable
+    required = total if req is None else req
+    self.fields[k] = value  # pylint: disable=unsupported-assignment-operation
+    if required:
       self.required.add(k)
 
   def check_keys(self, keys):
@@ -51,17 +64,17 @@ class TypedDictBuilder(abstract.PyTDClass):
   """Factory for creating typing.TypedDict classes."""
 
   def __init__(self, ctx):
-    typing_ast = ctx.loader.import_name("typing")
-    pyval = typing_ast.Lookup("typing.TypedDict")
+    pyval = ctx.loader.lookup_pytd("typing", "TypedDict")
     super().__init__("TypedDict", pyval, ctx)
     # Signature for the functional constructor
-    fn = typing_ast.Lookup("typing._TypedDictFunction")
+    fn = ctx.loader.lookup_pytd("typing", "_TypedDictFunction")
     fn = fn.Replace(name="typing.TypedDict")
-    sig, = fn.signatures
+    (sig,) = fn.signatures
     self.fn_sig = function.Signature.from_pytd(
-        self.ctx, "typing.TypedDict", sig)
+        self.ctx, "typing.TypedDict", sig
+    )
 
-  def call(self, node, _, args):
+  def call(self, node, func, args, alias_map=None):
     """Call the functional constructor."""
     props = self._extract_args(args)
     cls = TypedDictClass(props, self, self.ctx)
@@ -69,21 +82,37 @@ class TypedDictBuilder(abstract.PyTDClass):
     return node, cls_var
 
   def _extract_param(self, args, pos, name, pyval_type, typ):
-    var = args.posargs[pos]
+    var = args.namedargs[name] if pos is None else args.posargs[pos]
     try:
       return abstract_utils.get_atomic_python_constant(var, pyval_type)
     except abstract_utils.ConversionError as e:
-      bad = abstract_utils.BadType(name, typ)
-      raise function.WrongArgTypes(self.fn_sig, args, self.ctx, bad) from e
+      bad = error_types.BadType(name, typ)
+      raise error_types.WrongArgTypes(self.fn_sig, args, self.ctx, bad) from e
 
   def _extract_args(self, args):
     if len(args.posargs) != 2:
-      raise function.WrongArgCount(self.fn_sig, args, self.ctx)
+      raise error_types.WrongArgCount(self.fn_sig, args, self.ctx)
     name = self._extract_param(args, 0, "name", str, self.ctx.convert.str_type)
     fields = self._extract_param(
-        args, 1, "fields", dict, self.ctx.convert.dict_type)
+        args, 1, "fields", dict, self.ctx.convert.dict_type
+    )
+    if "total" in args.namedargs:
+      total = self._extract_param(
+          args, None, "total", bool, self.ctx.convert.bool_type
+      )
+    else:
+      total = True
     props = TypedDictProperties(
-        name=name, fields=fields, required=set(fields.keys()), total=True)
+        name=name, fields={}, required=set(), total=total
+    )
+    # Force Required/NotRequired evaluation
+    for k, v in fields.items():
+      try:
+        value = abstract_utils.get_atomic_value(v)
+      except abstract_utils.ConversionError:
+        self.ctx.errorlog.ambiguous_annotation(self.ctx.vm.frames, v.data, k)
+        value = self.ctx.convert.unsolvable
+      props.add(k, value, total)
     return props
 
   def _validate_bases(self, cls_name, bases):
@@ -91,10 +120,12 @@ class TypedDictBuilder(abstract.PyTDClass):
     for base_var in bases:
       for base in base_var.data:
         if not isinstance(base, (TypedDictClass, TypedDictBuilder)):
-          details = (f"TypedDict {cls_name} cannot inherit from "
-                     "a non-TypedDict class.")
+          details = (
+              f"TypedDict {cls_name} cannot inherit from a non-TypedDict class."
+          )
           self.ctx.errorlog.base_class_error(
-              self.ctx.vm.frames, base_var, details)
+              self.ctx.vm.frames, base_var, details
+          )
 
   def _merge_base_class_fields(self, bases, props):
     """Add the merged list of base class fields to the fields dict."""
@@ -109,7 +140,8 @@ class TypedDictBuilder(abstract.PyTDClass):
             classes = f"{base.name} and {provenance[k]}"
             details = f"Duplicate TypedDict key {k} in classes {classes}"
             self.ctx.errorlog.base_class_error(
-                self.ctx.vm.frames, base_var, details)
+                self.ctx.vm.frames, base_var, details
+            )
           else:
             props.add(k, v, base.props.total)
             provenance[k] = base.name
@@ -134,17 +166,25 @@ class TypedDictBuilder(abstract.PyTDClass):
     else:
       total = abstract_utils.get_atomic_python_constant(total, bool)
     props = TypedDictProperties(
-        name=cls_name, fields={}, required=set(), total=total)
+        name=cls_name, fields={}, required=set(), total=total
+    )
 
     # Collect the key types defined in the current class.
     cls_locals = classgen.get_class_locals(
         cls_name,
         allow_methods=False,
         ordering=classgen.Ordering.FIRST_ANNOTATE,
-        ctx=self.ctx)
+        ctx=self.ctx,
+    )
     for k, local in cls_locals.items():
-      assert local.typ
-      props.add(k, local.typ, total)
+      var = local.typ
+      assert var
+      try:
+        typ = abstract_utils.get_atomic_value(var)
+      except abstract_utils.ConversionError:
+        self.ctx.errorlog.ambiguous_annotation(self.ctx.vm.frames, var.data, k)
+        typ = self.ctx.convert.unsolvable
+      props.add(k, typ, total)
 
     # Process base classes and generate the __init__ signature.
     self._validate_bases(cls_name, bases)
@@ -154,22 +194,26 @@ class TypedDictBuilder(abstract.PyTDClass):
     cls_var = cls.to_variable(node)
     return node, cls_var
 
-  def make_class_from_pyi(self, cls_name, pytd_cls, total=True):
+  def make_class_from_pyi(self, cls_name, pytd_cls):
     """Make a TypedDictClass from a pyi class."""
     # NOTE: Returns the abstract class, not a variable.
     name = pytd_cls.name or cls_name
-    if total is None:
+    for k, v in pytd_cls.keywords:
+      if k == "total":
+        total = v.value
+        break
+    else:
       total = True
     props = TypedDictProperties(
-        name=name, fields={}, required=set(), total=total)
+        name=name, fields={}, required=set(), total=total
+    )
 
     for c in pytd_cls.constants:
-      typ = self.ctx.convert.constant_to_var(c.type)
+      typ = self.ctx.convert.constant_to_value(c.type)
       props.add(c.name, typ, total)
 
     # Process base classes and generate the __init__ signature.
-    bases = [self.ctx.convert.constant_to_var(x)
-             for x in pytd_cls.bases]
+    bases = [self.ctx.convert.constant_to_var(x) for x in pytd_cls.bases]
     self._validate_bases(cls_name, bases)
     self._merge_base_class_fields(bases, props)
 
@@ -192,20 +236,22 @@ class TypedDictClass(abstract.PyTDClass):
   def _make_init(self, props):
     # __init__ method for type checking signatures.
     # We construct this here and pass it to TypedDictClass because we need
-    # access to abstract.SignedFunction.
+    # access to abstract.SimpleFunction.
     sig = function.Signature.from_param_names(
-        f"{props.name}.__init__", props.fields.keys(),
-        kind=pytd.ParameterKind.KWONLY)
-    sig.annotations = {k: abstract_utils.get_atomic_value(v)
-                       for k, v in props.fields.items()}
-    sig.defaults = {k: self.ctx.new_unsolvable(self.ctx.root_node)
-                    for k in props.optional}
-    return abstract.SignedFunction(sig, self.ctx)
+        f"{props.name}.__init__",
+        props.fields.keys(),
+        kind=pytd.ParameterKind.KWONLY,
+    )
+    sig.annotations = dict(props.fields)
+    sig.defaults = {
+        k: self.ctx.new_unsolvable(self.ctx.root_node) for k in props.optional
+    }
+    return abstract.SimpleFunction(sig, self.ctx)
 
   def _new_instance(self, container, node, args):
     self.init_method.match_and_map_args(node, args, None)
     ret = TypedDict(self.props, self.ctx)
-    for (k, v) in args.namedargs.items():
+    for k, v in args.namedargs.items():
       ret.set_str_item(node, k, v)
     ret.cls = self
     return ret
@@ -213,8 +259,7 @@ class TypedDictClass(abstract.PyTDClass):
   def instantiate_value(self, node, container):
     args = function.Args(())
     for name, typ in self.props.fields.items():
-      args.namedargs[name] = self.ctx.join_variables(
-          node, [t.instantiate(node) for t in typ.data])
+      args.namedargs[name] = typ.instantiate(node)
     return self._new_instance(container, node, args)
 
   def instantiate(self, node, container=None):
@@ -253,17 +298,21 @@ class TypedDict(abstract.Dict):
 
   def _check_str_key(self, name):
     if name not in self.fields:
-      raise TypedDictKeyMissing(self, name)
+      raise error_types.TypedDictKeyMissing(self, name)
     return name
 
   def _check_str_key_value(self, node, name, value_var):
     self._check_str_key(name)
-    typ = abstract_utils.get_atomic_value(self.fields[name])
+    typ = self.fields[name]
     bad = self.ctx.matcher(node).compute_one_match(value_var, typ).bad_matches
     for match in bad:
       self.ctx.errorlog.annotation_type_mismatch(
-          self.ctx.vm.frames, match.expected.typ, match.actual_binding, name,
-          match.error_details, typed_dict=self
+          self.ctx.vm.frames,
+          match.expected.typ,
+          match.actual_binding,
+          name,
+          match.error_details,
+          typed_dict=self,
       )
     return name, value_var
 
@@ -272,7 +321,7 @@ class TypedDict(abstract.Dict):
     try:
       name = abstract_utils.get_atomic_python_constant(name_var, str)
     except abstract_utils.ConversionError as e:
-      raise TypedDictKeyMissing(self, None) from e
+      raise error_types.TypedDictKeyMissing(self, None) from e
     return self._check_str_key(name)
 
   def _check_value(self, node, name_var, value_var):
@@ -308,9 +357,15 @@ class TypedDict(abstract.Dict):
   def get_slot(self, node, key_var, default_var=None):
     try:
       str_key = self._check_key(key_var)
-    except TypedDictKeyMissing:
+    except error_types.TypedDictKeyMissing:
+      return node, default_var or self.ctx.convert.none.to_variable(node)
+    if str_key in self.pyval:
+      return node, self.pyval[str_key]
+    else:
+      # If total=False the key might not be in self.pyval.
+      # TODO(mdemello): Should we return `self.props[key].typ | default | None`
+      # here, or just `default | None`?
       return node, default_var or self.ctx.convert.none.to_variable(node)
-    return node, self.pyval[str_key]
 
   def merge_instance_type_parameter(self, node, name, value):
     _, _, short_name = name.rpartition(".")
@@ -324,3 +379,59 @@ class TypedDict(abstract.Dict):
       # them once fully set.
       return
     super().merge_instance_type_parameter(node, name, value)
+
+
+def _is_typeddict(val: abstract.BaseValue):
+  if isinstance(val, abstract.Union):
+    return all(_is_typeddict(v) for v in val.options)
+  return isinstance(val, TypedDictClass)
+
+
+class IsTypedDict(abstract.PyTDFunction):
+  """Implementation of typing.is_typeddict."""
+
+  def call(self, node, func, args, alias_map=None):
+    self.match_args(node, args)
+    if args.posargs:
+      tp = args.posargs[0]
+    elif "tp" in args.namedargs:
+      tp = args.namedargs["tp"]
+    else:
+      return node, self.ctx.convert.bool_values[None].to_variable(node)
+    is_typeddict = [_is_typeddict(v) for v in tp.data]
+    if all(is_typeddict):
+      boolval = True
+    elif not any(is_typeddict):
+      boolval = False
+    else:
+      boolval = None
+    return node, self.ctx.convert.bool_values[boolval].to_variable(node)
+
+
+class _TypedDictItemRequiredness(overlay_utils.TypingContainer):
+  """typing.(Not)Required."""
+
+  _REQUIREDNESS = None
+
+  def _get_value_info(self, inner, ellipses, allowed_ellipses=frozenset()):
+    template, processed_inner, abstract_class = super()._get_value_info(
+        inner, ellipses, allowed_ellipses
+    )
+    for annotation in processed_inner:
+      req = _is_required(annotation)
+      if req not in (None, self._REQUIREDNESS):
+        error = "Cannot mark a TypedDict item as both Required and NotRequired"
+        self.ctx.errorlog.invalid_annotation(
+            stack=self.ctx.vm.frames, annot=self.name, details=error
+        )
+    return template, processed_inner, abstract_class
+
+
+class Required(_TypedDictItemRequiredness):
+
+  _REQUIREDNESS = True
+
+
+class NotRequired(_TypedDictItemRequiredness):
+
+  _REQUIREDNESS = False
