@@ -3,7 +3,7 @@
 import dataclasses
 import enum
 import logging
-from typing import Any, Mapping
+from typing import Any, Mapping, Optional
 
 from pytype.abstract import abstract
 from pytype.abstract import abstract_utils
@@ -20,6 +20,7 @@ _SignatureMapType = Mapping[str, function.Signature]
 @enum.unique
 class SignatureErrorType(enum.Enum):
   """Constants representing various signature mismatch errors."""
+
   NO_ERROR = enum.auto()
   DEFAULT_PARAMETER_MISMATCH = enum.auto()
   DEFAULT_VALUE_MISMATCH = enum.auto()
@@ -32,16 +33,6 @@ class SignatureErrorType(enum.Enum):
   RETURN_TYPE_MISMATCH = enum.auto()
 
 
-SIGNATURE_ERROR_TYPE_TO_OPTION_NAME = {
-    SignatureErrorType.DEFAULT_PARAMETER_MISMATCH:
-        "overriding_parameter_count_checks",
-    SignatureErrorType.KWONLY_PARAMETER_COUNT_MISMATCH:
-        "overriding_parameter_count_checks",
-    SignatureErrorType.POSITIONAL_PARAMETER_COUNT_MISMATCH:
-        "overriding_parameter_count_checks",
-}
-
-
 @dataclasses.dataclass
 class SignatureError:
   error_code: SignatureErrorType = SignatureErrorType.NO_ERROR
@@ -59,8 +50,9 @@ def _get_varargs_annotation_type(param_type):
   return param_type.get_formal_type_parameter(abstract_utils.T)
 
 
-def _check_positional_parameter_annotations(method_signature, base_signature,
-                                            is_subtype):
+def _check_positional_parameter_annotations(
+    method_signature, base_signature, is_subtype
+):
   """Checks type annotations for positional parameters of the overriding method.
 
   Args:
@@ -71,8 +63,9 @@ def _check_positional_parameter_annotations(method_signature, base_signature,
   Returns:
     SignatureError if a mismatch is detected. Otherwise returns None.
   """
-  for param_index in range(max(len(base_signature.param_names),
-                               len(method_signature.param_names))):
+  for param_index in range(
+      max(len(base_signature.param_names), len(method_signature.param_names))
+  ):
     if param_index == 0:
       # No type checks for 'self' parameter.
       continue
@@ -99,8 +92,10 @@ def _check_positional_parameter_annotations(method_signature, base_signature,
       # Positional-only parameters should match by position.
       method_param_name = method_signature.param_names[param_index]
     elif param_index < len(method_signature.param_names):
-      if (base_param_name == "_" or
-          method_signature.param_names[param_index] == "_"):
+      if (
+          base_param_name == "_"
+          or method_signature.param_names[param_index] == "_"
+      ):
         # Underscore parameters should match by position.
         method_param_name = method_signature.param_names[param_index]
       else:
@@ -128,23 +123,27 @@ def _check_positional_parameter_annotations(method_signature, base_signature,
     if not is_subtype(base_param_type, method_param_type):
       return SignatureError(
           SignatureErrorType.POSITIONAL_PARAMETER_TYPE_MISMATCH,
-          f"Type mismatch for parameter '{method_param_name}'.")
+          f"Type mismatch for parameter '{method_param_name}'.",
+      )
 
   return None
 
 
-def _check_positional_parameters(method_signature, base_signature, is_subtype):
+def _check_positional_parameters(
+    method_signature, base_signature, is_subtype, ctx
+):
   """Checks that the positional parameters of the overriding method match.
 
   Args:
     method_signature: signature of the overriding method.
     base_signature: signature of the overridden method.
     is_subtype: a binary function to compare types.
+    ctx: Context
 
   Returns:
     SignatureError if a mismatch is detected. Otherwise returns None.
   """
-
+  check_types = True
   # Check mappings of positional-or-keyword parameters of the overridden method.
   for base_param_pos, base_param_name in enumerate(base_signature.param_names):
     # Skip positional-only parameters - those that cannot be passed by name.
@@ -157,7 +156,8 @@ def _check_positional_parameters(method_signature, base_signature, is_subtype):
     if base_param_pos < method_signature.posonly_count:
       return SignatureError(
           SignatureErrorType.POSITIONAL_PARAMETER_COUNT_MISMATCH,
-          "Too many positional-only parameters in overriding method.")
+          "Too many positional-only parameters in overriding method.",
+      )
     elif base_param_pos < len(method_signature.param_names):
       method_param_name = method_signature.param_names[base_param_pos]
     else:
@@ -167,7 +167,8 @@ def _check_positional_parameters(method_signature, base_signature, is_subtype):
         break
       return SignatureError(
           SignatureErrorType.POSITIONAL_PARAMETER_COUNT_MISMATCH,
-          "Not enough positional parameters in overriding method.")
+          "Not enough positional parameters in overriding method.",
+      )
 
     # Positional-or-keyword parameters must have the same name or underscore.
     method_param_name = method_signature.param_names[base_param_pos]
@@ -175,14 +176,20 @@ def _check_positional_parameters(method_signature, base_signature, is_subtype):
       # We don't report it as an error, as this is a very common practice
       # in the absence of positional-only parameters.
       # TODO(sinopalnikov): clean it up and start flagging the error.
-      log.warning("Name mismatch for parameter '%r'.", base_param_name)
-      return None
-
+      log.warning("Name mismatch for parameter %r.", base_param_name)
+      # We match positional parameter type annotations by name, not position,
+      # later on, so if we have a name mismatch here we should disable
+      # annotation checking and just check param count.
+      if not ctx.options.overriding_renamed_parameter_count_checks:
+        return None
+      check_types = False
   # Check mappings of remaining positional parameters of the overriding method
   # that don't map to any positional parameters of the overridden method.
   remaining_method_params = (
-      method_signature.param_names[len(base_signature.param_names):]
-      if not base_signature.varargs_name else [])
+      method_signature.param_names[len(base_signature.param_names) :]
+      if not base_signature.varargs_name
+      else []
+  )
   for method_param_name in remaining_method_params:
 
     # Keyword-only can map to remaining positional.
@@ -193,14 +200,19 @@ def _check_positional_parameters(method_signature, base_signature, is_subtype):
     if method_param_name not in method_signature.defaults:
       return SignatureError(
           SignatureErrorType.DEFAULT_PARAMETER_MISMATCH,
-          f"Parameter '{method_param_name}' must have a default value.")
+          f"Parameter '{method_param_name}' must have a default value.",
+      )
 
-  return _check_positional_parameter_annotations(method_signature,
-                                                 base_signature, is_subtype)
+  if not check_types:
+    return None
+  return _check_positional_parameter_annotations(
+      method_signature, base_signature, is_subtype
+  )
 
 
-def _check_keyword_only_parameters(method_signature, base_signature,
-                                   is_subtype):
+def _check_keyword_only_parameters(
+    method_signature, base_signature, is_subtype
+):
   """Checks that the keyword-only parameters of the overriding method match.
 
   Args:
@@ -220,10 +232,12 @@ def _check_keyword_only_parameters(method_signature, base_signature,
     # Keyword-only parameters of the overriding method that don't match any
     # keyword-only parameter of the overridden method must have a default value.
     for method_param_name in method_kwonly_params.difference(
-        base_kwonly_params).difference(method_defaults):
+        base_kwonly_params
+    ).difference(method_defaults):
       return SignatureError(
           SignatureErrorType.DEFAULT_PARAMETER_MISMATCH,
-          f"Parameter '{method_param_name}' must have a default value.")
+          f"Parameter '{method_param_name}' must have a default value.",
+      )
 
   # A keyword-only parameter of the overridden method cannot have the same name
   # as a positional-only parameter of the overriding method.
@@ -234,14 +248,18 @@ def _check_keyword_only_parameters(method_signature, base_signature,
       if not method_signature.kwargs_name:
         return SignatureError(
             SignatureErrorType.KWONLY_PARAMETER_NAME_MISMATCH,
-            f"Parameter '{base_param_name}' not found in overriding method.")
+            f"Parameter '{base_param_name}' not found in overriding method.",
+        )
     else:
       if method_param_index < method_signature.posonly_count:
         return SignatureError(
             SignatureErrorType.KWONLY_PARAMETER_NAME_MISMATCH,
-            (f"Keyword-only parameter '{base_param_name}' of the overridden "
-             "method has the same name as a positional-only parameter"
-             "of the overriding method."))
+            (
+                f"Keyword-only parameter '{base_param_name}' of the overridden "
+                "method has the same name as a positional-only parameter"
+                "of the overriding method."
+            ),
+        )
 
   # Check annotations of keyword-only parameters.
   for base_param_name in base_signature.kwonly_params:
@@ -251,8 +269,10 @@ def _check_keyword_only_parameters(method_signature, base_signature,
       # Parameter not annotated in the overridden method.
       continue
 
-    if (base_param_name in method_kwonly_params or
-        base_param_name in method_signature.param_names):
+    if (
+        base_param_name in method_kwonly_params
+        or base_param_name in method_signature.param_names
+    ):
       method_param_name = base_param_name
     elif method_signature.kwargs_name:
       method_param_name = method_signature.kwargs_name
@@ -270,7 +290,8 @@ def _check_keyword_only_parameters(method_signature, base_signature,
         # If kwargs are annotated with type T in the function definition,
         # the annotation in the signature will be Dict[str, T].
         method_param_type = method_param_type.get_formal_type_parameter(
-            abstract_utils.V)
+            abstract_utils.V
+        )
       else:
         # If the kwargs type is a plain dict, there's nothing to check.
         continue
@@ -280,7 +301,8 @@ def _check_keyword_only_parameters(method_signature, base_signature,
     if not is_subtype(base_param_type, method_param_type):
       return SignatureError(
           SignatureErrorType.KWONLY_PARAMETER_TYPE_MISMATCH,
-          f"Type mismatch for parameter '{base_param_name}'.")
+          f"Type mismatch for parameter '{base_param_name}'.",
+      )
 
   return None
 
@@ -299,8 +321,10 @@ def _check_default_values(method_signature, base_signature):
   for base_param_name, base_default_value in base_signature.defaults.items():
     if base_param_name in base_signature.kwonly_params:
       # The parameter is keyword-only, check match by name.
-      if (base_param_name not in method_signature.kwonly_params and
-          base_param_name not in method_signature.param_names):
+      if (
+          base_param_name not in method_signature.kwonly_params
+          and base_param_name not in method_signature.param_names
+      ):
         continue
       method_param_name = base_param_name
     else:
@@ -315,22 +339,25 @@ def _check_default_values(method_signature, base_signature):
     except KeyError:
       return SignatureError(
           SignatureErrorType.DEFAULT_PARAMETER_MISMATCH,
-          f"Parameter '{method_param_name}' must have a default value.")
+          f"Parameter '{method_param_name}' must have a default value.",
+      )
 
     # Only concrete values can be compared for an exact match.
     try:
       base_default = abstract_utils.get_atomic_python_constant(
-          base_default_value)
+          base_default_value
+      )
       method_default = abstract_utils.get_atomic_python_constant(
-          method_default_value)
+          method_default_value
+      )
     except abstract_utils.ConversionError:
       continue
 
     if base_default != method_default:
       return SignatureError(
           SignatureErrorType.DEFAULT_VALUE_MISMATCH,
-          (f"Parameter '{base_param_name}' must have the same default "
-           "value."))
+          f"Parameter '{base_param_name}' must have the same default value.",
+      )
 
   return None
 
@@ -344,21 +371,24 @@ def _check_return_types(method_signature, base_signature, is_subtype):
     # Return type not annotated in either of the two methods.
     return None
 
-  if (isinstance(base_return_type, abstract.AMBIGUOUS_OR_EMPTY) or
-      isinstance(method_return_type, abstract.AMBIGUOUS_OR_EMPTY)):
+  if isinstance(base_return_type, abstract.AMBIGUOUS_OR_EMPTY) or isinstance(
+      method_return_type, abstract.AMBIGUOUS_OR_EMPTY
+  ):
     return None
 
   # Return type of the overriding method must be a subtype of the
   # return type of the overridden method.
   if not is_subtype(method_return_type, base_return_type):
-    return SignatureError(SignatureErrorType.RETURN_TYPE_MISMATCH,
-                          "Return type mismatch.")
+    return SignatureError(
+        SignatureErrorType.RETURN_TYPE_MISMATCH, "Return type mismatch."
+    )
 
   return None
 
 
-def _check_signature_compatible(method_signature, base_signature,
-                                stack, matcher, ctx):
+def _check_signature_compatible(
+    method_signature, base_signature, stack, matcher, ctx
+):
   """Checks if the signatures match for the overridden and overriding methods.
 
   Adds the first error found to the context's error log.
@@ -413,29 +443,33 @@ def _check_signature_compatible(method_signature, base_signature,
 
   def is_subtype(this_type, that_type):
     """Return True iff this_type is a subclass of that_type."""
-    this_type_instance = this_type.instantiate(ctx.root_node, None)
+    if this_type == ctx.convert.never:
+      return True  # Never is the bottom type, so it matches everything
+    this_type_instance = this_type.instantiate(
+        ctx.root_node, container=abstract_utils.DUMMY_CONTAINER
+    )
     return matcher.compute_one_match(this_type_instance, that_type).success
 
   check_result = (
-      _check_positional_parameters(method_signature, base_signature, is_subtype)
-      or _check_keyword_only_parameters(method_signature, base_signature,
-                                        is_subtype) or
-      _check_default_values(method_signature, base_signature) or
-      _check_return_types(method_signature, base_signature, is_subtype))
+      _check_positional_parameters(
+          method_signature, base_signature, is_subtype, ctx
+      )
+      or _check_keyword_only_parameters(
+          method_signature, base_signature, is_subtype
+      )
+      or _check_default_values(method_signature, base_signature)
+      or _check_return_types(method_signature, base_signature, is_subtype)
+  )
 
   if check_result:
-    if check_result.error_code in SIGNATURE_ERROR_TYPE_TO_OPTION_NAME:
-      # Check if the error is disabled by an option.
-      option_name = SIGNATURE_ERROR_TYPE_TO_OPTION_NAME[check_result.error_code]
-      if not getattr(ctx.options, option_name):
-        log.warning(check_result.message)
-        return
     ctx.errorlog.overriding_signature_mismatch(
-        stack, base_signature, method_signature, details=check_result.message)
+        stack, base_signature, method_signature, details=check_result.message
+    )
 
 
 def _get_pytd_class_signature_map(
-    cls: abstract.PyTDClass, ctx: _ContextType) -> _SignatureMapType:
+    cls: abstract.PyTDClass, ctx: _ContextType
+) -> _SignatureMapType:
   """Returns a map from method names to their signatures for a PyTDClass."""
   if cls in ctx.method_signature_map:
     return ctx.method_signature_map[cls]
@@ -460,7 +494,8 @@ def _get_pytd_class_signature_map(
 
 
 def _get_parameterized_class_signature_map(
-    cls: abstract.ParameterizedClass, ctx: _ContextType) -> _SignatureMapType:
+    cls: abstract.ParameterizedClass, ctx: _ContextType
+) -> _SignatureMapType:
   """Returns a map from method names to signatures for a ParameterizedClass."""
   if cls in ctx.method_signature_map:
     return ctx.method_signature_map[cls]
@@ -477,9 +512,11 @@ def _get_parameterized_class_signature_map(
   for base_method_name, base_method_signature in base_signature_map.items():
     # Replace formal type parameters with their values.
     annotations = ctx.annotation_utils.sub_annotations_for_parameterized_class(
-        cls, base_method_signature.annotations)
+        cls, base_method_signature.annotations
+    )
     method_signature_map[base_method_name] = base_method_signature._replace(
-        annotations=annotations)
+        annotations=annotations
+    )
 
   ctx.method_signature_map[cls] = method_signature_map
   return method_signature_map
@@ -493,7 +530,8 @@ def check_overriding_members(cls, bases, members, matcher, ctx):
   for member_name, member_value in members.items():
     try:
       atomic_value = abstract_utils.get_atomic_value(
-          member_value, constant_type=abstract.InterpreterFunction)
+          member_value, constant_type=abstract.InterpreterFunction
+      )
     except abstract_utils.ConversionError:
       continue
     method = atomic_value
@@ -507,29 +545,30 @@ def check_overriding_members(cls, bases, members, matcher, ctx):
 
   class_signature_map = {}
   for method_name, method in class_method_map.items():
-    if method.is_coroutine():
+    if method.is_unannotated_coroutine():
       annotations = dict(method.signature.annotations)
-      coroutine_params = {
-          abstract_utils.T: ctx.convert.unsolvable,
-          abstract_utils.T2: ctx.convert.unsolvable,
-          abstract_utils.V: annotations.get("return", ctx.convert.unsolvable),
-      }
-      annotations["return"] = abstract.ParameterizedClass(
-          ctx.convert.coroutine_type, coroutine_params, ctx)
+      annotations["return"] = ctx.convert.coroutine_type
       signature = method.signature._replace(annotations=annotations)
     else:
       signature = method.signature
     class_signature_map[method_name] = signature
+
+  base_classes = []
   for base in bases:
     try:
       base_class = abstract_utils.get_atomic_value(base)
     except abstract_utils.ConversionError:
       continue
+    base_classes.append(base_class)
+  mro = [mro_class.full_name for mro_class in cls.mro]
+
+  for i, base_class in enumerate(base_classes):
     if isinstance(base_class, abstract.InterpreterClass):
       base_signature_map = ctx.method_signature_map[base_class]
     elif isinstance(base_class, abstract.ParameterizedClass):
       base_signature_map = _get_parameterized_class_signature_map(
-          base_class, ctx)
+          base_class, ctx
+      )
     elif isinstance(base_class, abstract.PyTDClass):
       base_signature_map = _get_pytd_class_signature_map(base_class, ctx)
     else:
@@ -544,13 +583,46 @@ def check_overriding_members(cls, bases, members, matcher, ctx):
       # class definition.
       method_def_opcode = (
           class_method_map[base_method_name].def_opcode
-          if base_method_name in class_method_map else None)
+          if base_method_name in class_method_map
+          else None
+      )
       stack = ctx.vm.simple_stack(method_def_opcode)
-      _check_signature_compatible(class_method_signature, base_method_signature,
-                                  stack, matcher, ctx)
+      _check_signature_compatible(
+          class_method_signature, base_method_signature, stack, matcher, ctx
+      )
 
+    # We filter out any methods inherited from a base class that comes after the
+    # next direct base class in the MRO, to avoid checking signature
+    # compatibility in the wrong direction.
+    if i < len(base_classes) - 1:
+      next_index = mro.index(base_classes[i + 1].full_name)
+      filtered_base_map = {}
+      for base_method_name, base_method_signature in base_signature_map.items():
+        defining_class = _get_defining_class(base_method_signature)
+        if defining_class:
+          try:
+            defining_index = mro.index(defining_class)
+          except ValueError:
+            include = True
+          else:
+            include = defining_index < next_index
+        else:
+          include = True
+        if include:
+          filtered_base_map[base_method_name] = base_method_signature
+    else:
+      filtered_base_map = base_signature_map
     # Methods defined in this class take precedence.
-    class_signature_map = {**base_signature_map, **class_signature_map}
+    class_signature_map = {**filtered_base_map, **class_signature_map}
 
   assert cls not in ctx.method_signature_map
   ctx.method_signature_map[cls] = class_signature_map
+
+
+def _get_defining_class(sig: function.Signature) -> Optional[str]:
+  if "self" in sig.annotations:
+    return sig.annotations["self"].full_name
+  elif "." in sig.name:
+    return sig.name.rsplit(".", 1)[0]
+  else:
+    return None
