@@ -1,38 +1,36 @@
 """Process conditional blocks in pyi files."""
 
-import sys
+import ast as astlib
 from typing import Optional, Tuple, Union
 
 from pytype.ast import visitor as ast_visitor
-from pytype.pyi.types import ParseError  # pylint: disable=g-importing-member
+from pytype.pyi import types
 from pytype.pytd import slots as cmp_slots
 
-# pylint: disable=g-import-not-at-top
-if sys.version_info >= (3, 8):
-  import ast as ast3
-else:
-  from typed_ast import ast3
-# pylint: enable=g-import-not-at-top
+_ParseError = types.ParseError
 
 
 class ConditionEvaluator(ast_visitor.BaseVisitor):
   """Evaluates if statements in pyi files."""
 
   def __init__(self, options):
-    super().__init__(ast=ast3)
+    super().__init__(ast=astlib)
     self._compares = {
-        ast3.Eq: cmp_slots.EQ,
-        ast3.Gt: cmp_slots.GT,
-        ast3.Lt: cmp_slots.LT,
-        ast3.GtE: cmp_slots.GE,
-        ast3.LtE: cmp_slots.LE,
-        ast3.NotEq: cmp_slots.NE
+        astlib.Eq: cmp_slots.EQ,
+        astlib.Gt: cmp_slots.GT,
+        astlib.Lt: cmp_slots.LT,
+        astlib.GtE: cmp_slots.GE,
+        astlib.LtE: cmp_slots.LE,
+        astlib.NotEq: cmp_slots.NE,
     }
     self._options = options
 
   def _eval_comparison(
-      self, ident: Tuple[str, Optional[Union[int, slice]]], op: str,
-      value: Union[str, int, Tuple[int, ...]]) -> bool:
+      self,
+      ident: Tuple[str, Optional[Union[int, slice]]],
+      op: str,
+      value: Union[str, int, Tuple[int, ...]],
+  ) -> bool:
     """Evaluate a comparison and return a bool.
 
     Args:
@@ -51,28 +49,31 @@ class ConditionEvaluator(ast_visitor.BaseVisitor):
       if key is None:
         key = slice(None, None, None)
       if isinstance(key, int) and not isinstance(value, int):
-        raise ParseError(
-            "an element of sys.version_info must be compared to an integer")
+        raise _ParseError(
+            "an element of sys.version_info must be compared to an integer"
+        )
       if isinstance(key, slice) and not _is_int_tuple(value):
-        raise ParseError(
-            "sys.version_info must be compared to a tuple of integers")
+        raise _ParseError(
+            "sys.version_info must be compared to a tuple of integers"
+        )
       try:
         actual = self._options.python_version[key]
       except IndexError as e:
-        raise ParseError(str(e)) from e
+        raise _ParseError(str(e)) from e
       if isinstance(key, slice):
         actual = _three_tuple(actual)
         value = _three_tuple(value)
     elif name == "sys.platform":
       if not isinstance(value, str):
-        raise ParseError("sys.platform must be compared to a string")
+        raise _ParseError("sys.platform must be compared to a string")
       valid_cmps = (cmp_slots.EQ, cmp_slots.NE)
       if op not in valid_cmps:
-        raise ParseError(
-            "sys.platform must be compared using %s or %s" % valid_cmps)
+        raise _ParseError(
+            "sys.platform must be compared using %s or %s" % valid_cmps
+        )
       actual = self._options.platform
     else:
-      raise ParseError(f"Unsupported condition: {name!r}.")
+      raise _ParseError(f"Unsupported condition: {name!r}.")
     return cmp_slots.COMPARES[op](actual, value)
 
   def fail(self, name=None):
@@ -81,10 +82,10 @@ class ConditionEvaluator(ast_visitor.BaseVisitor):
     else:
       msg = "Unsupported condition. "
     msg += "Supported checks are sys.platform and sys.version_info"
-    raise ParseError(msg)
+    raise _ParseError(msg)
 
   def visit_Attribute(self, node):
-    if not isinstance(node.value, ast3.Name):
+    if not isinstance(node.value, astlib.Name):
       self.fail()
     name = f"{node.value.id}.{node.attr}"
     if node.value.id == "sys":
@@ -100,15 +101,9 @@ class ConditionEvaluator(ast_visitor.BaseVisitor):
   def visit_Index(self, node):
     return node.value
 
-  def visit_Constant(self, node):
+  def visit_Pyval(self, node):
     return node.value
 
-  def visit_Num(self, node):
-    return node.n
-
-  def visit_Str(self, node):
-    return node.s
-
   def visit_Subscript(self, node):
     return (node.value, node.slice)
 
@@ -116,18 +111,18 @@ class ConditionEvaluator(ast_visitor.BaseVisitor):
     return tuple(node.elts)
 
   def visit_BoolOp(self, node):
-    if isinstance(node.op, ast3.Or):
+    if isinstance(node.op, astlib.Or):
       return any(node.values)
-    elif isinstance(node.op, ast3.And):
+    elif isinstance(node.op, astlib.And):
       return all(node.values)
     else:
-      raise ParseError(f"Unexpected boolean operator: {node.op}")
+      raise _ParseError(f"Unexpected boolean operator: {node.op}")
 
   def visit_UnaryOp(self, node):
-    if isinstance(node.op, ast3.USub) and isinstance(node.operand, int):
+    if isinstance(node.op, astlib.USub) and isinstance(node.operand, int):
       return -node.operand
     else:
-      raise ParseError(f"Unexpected unary operator: {node.op}")
+      raise _ParseError(f"Unexpected unary operator: {node.op}")
 
   def visit_Compare(self, node):
     if isinstance(node.left, tuple):
@@ -139,7 +134,7 @@ class ConditionEvaluator(ast_visitor.BaseVisitor):
     return self._eval_comparison(ident, op, right)
 
 
-def evaluate(test: ast3.AST, options) -> bool:
+def evaluate(test: astlib.AST, options) -> bool:
   return ConditionEvaluator(options).visit(test)
 
 
