@@ -1,22 +1,17 @@
 """Function definitions in pyi files."""
 
-import sys
+import ast as astlib
+import dataclasses
 import textwrap
-
-from typing import Any, List, Optional
+from typing import Any, List, Optional, cast
 
 from pytype.pyi import types
-from pytype.pyi.types import ParseError  # pylint: disable=g-importing-member
 from pytype.pytd import pytd
 from pytype.pytd import visitors
 from pytype.pytd.codegen import function as pytd_function
+from pytype.pytd.parse import parser_constants
 
-# pylint: disable=g-import-not-at-top
-if sys.version_info >= (3, 8):
-  import ast as ast3
-else:
-  from typed_ast import ast3
-# pylint: enable=g-import-not-at-top
+_ParseError = types.ParseError
 
 
 class Mutator(visitors.Visitor):
@@ -42,7 +37,8 @@ class Mutator(visitors.Visitor):
       self.successful = True
       if p.optional:
         raise NotImplementedError(
-            f"Argument {p.name!r} cannot be both mutable and optional")
+            f"Argument {p.name!r} cannot be both mutable and optional"
+        )
       return p.Replace(mutated_type=self.new_type)
     else:
       return p
@@ -57,7 +53,7 @@ class Param(pytd_function.Param):
   """Internal representation of function parameters."""
 
   @classmethod
-  def from_arg(cls, arg: ast3.arg, kind: pytd.ParameterKind) -> "Param":
+  def from_arg(cls, arg: astlib.arg, kind: pytd.ParameterKind) -> "Param":
     """Constructor from an ast.argument node."""
     p = cls(arg.arg)
     if arg.annotation:
@@ -66,31 +62,32 @@ class Param(pytd_function.Param):
     return p
 
 
+@dataclasses.dataclass
+class SigProperties:
+  abstract: bool
+  coroutine: bool
+  final: bool
+  overload: bool
+  is_async: bool
+
+
 class NameAndSig(pytd_function.NameAndSig):
   """Internal representation of function signatures."""
 
   @classmethod
   def from_function(
-      cls, function: ast3.FunctionDef, is_async: bool) -> "NameAndSig":
+      cls, function: astlib.FunctionDef, props: SigProperties
+  ) -> "NameAndSig":
     """Constructor from an ast.FunctionDef node."""
     name = function.name
 
-    # decorators
-    decorators = set(function.decorator_list)
-    abstracts = {"abstractmethod", "abc.abstractmethod"}
-    coroutines = {"coroutine", "asyncio.coroutine", "coroutines.coroutine"}
-    overload = {"overload"}
-    final = {"final"}
-    ignored = {"type_check_only"}
-    is_abstract = bool(decorators & abstracts)
-    is_coroutine = bool(decorators & coroutines)
-    is_overload = bool(decorators & overload)
-    is_final = bool(decorators & final)
-    decorators -= (abstracts | coroutines | overload | final | ignored)
-    # TODO(mdemello): do we need this limitation?
-    if len(decorators) > 1:
-      raise ParseError(f"Too many decorators for {name}")
-    decorator, = decorators if decorators else (None,)
+    decorators = cast(List[pytd.Alias], function.decorator_list)
+    mutually_exclusive = {"property", "staticmethod", "classmethod"}
+    if len({d.type.name for d in decorators} & mutually_exclusive) > 1:
+      raise _ParseError(
+          f"'{name}' can be decorated with at most one of "
+          "property, staticmethod, and classmethod"
+      )
 
     exceptions = []
     mutators = []
@@ -101,9 +98,12 @@ class NameAndSig(pytd_function.NameAndSig):
         mutators.append(x)
       elif isinstance(x, types.Ellipsis):
         pass
-      elif (isinstance(x, ast3.Expr) and
-            isinstance(x.value, ast3.Str) and
-            i == 0):
+      elif (
+          isinstance(x, astlib.Expr)
+          and isinstance(x.value, astlib.Constant)
+          and isinstance(x.value.value, str)
+          and i == 0
+      ):
         # docstring
         pass
       else:
@@ -111,46 +111,60 @@ class NameAndSig(pytd_function.NameAndSig):
             Unexpected statement in function body.
             Only `raise` statements and type mutations are valid
         """).lstrip()
-        if isinstance(x, ast3.AST):
-          raise ParseError(msg).at(x)
+        if isinstance(x, astlib.AST):
+          raise _ParseError(msg).at(x)
         else:
-          raise ParseError(msg)
+          raise _ParseError(msg)
 
     # exceptions
-    sig = _pytd_signature(function, is_async, exceptions=exceptions)
+    sig = _pytd_signature(function, props.is_async, exceptions=exceptions)
 
     # mutators
     # If `self` is generic, a type parameter is being mutated.
-    if (sig.params and sig.params[0].name == "self" and
-        isinstance(sig.params[0].type, pytd.GenericType)):
+    if (
+        sig.params
+        and sig.params[0].name == "self"
+        and isinstance(sig.params[0].type, pytd.GenericType)
+    ):
       mutators.append(Mutator("self", sig.params[0].type))
     for mutator in mutators:
       try:
         sig = sig.Visit(mutator)
       except NotImplementedError as e:
-        raise ParseError(str(e)) from e
+        raise _ParseError(str(e)) from e
       if not mutator.successful:
-        raise ParseError(f"No parameter named {mutator.name!r}")
+        raise _ParseError(f"No parameter named {mutator.name!r}")
 
-    return cls(name, sig, decorator, is_abstract, is_coroutine, is_final,
-               is_overload)
+    return cls(
+        name,
+        sig,
+        tuple(decorators),
+        props.abstract,
+        props.coroutine,
+        props.final,
+        props.overload,
+    )
 
 
 def _pytd_signature(
-    function: ast3.FunctionDef,
+    function: astlib.FunctionDef,
     is_async: bool,
-    exceptions: Optional[List[pytd.Type]] = None
+    exceptions: Optional[List[pytd.Type]] = None,
 ) -> pytd.Signature:
   """Construct a pytd signature from an ast.FunctionDef node."""
   name = function.name
   args = function.args
   # Positional-only parameters are new in Python 3.8.
-  posonly_params = [Param.from_arg(a, pytd.ParameterKind.POSONLY)
-                    for a in getattr(args, "posonlyargs", ())]
-  pos_params = [Param.from_arg(a, pytd.ParameterKind.REGULAR)
-                for a in args.args]
-  kwonly_params = [Param.from_arg(a, pytd.ParameterKind.KWONLY)
-                   for a in args.kwonlyargs]
+  posonly_params = [
+      Param.from_arg(a, pytd.ParameterKind.POSONLY)
+      for a in getattr(args, "posonlyargs", ())
+  ]
+  pos_params = [
+      Param.from_arg(a, pytd.ParameterKind.REGULAR) for a in args.args
+  ]
+  kwonly_params = [
+      Param.from_arg(a, pytd.ParameterKind.KWONLY) for a in args.kwonlyargs
+  ]
   _apply_defaults(posonly_params + pos_params, args.defaults)
   _apply_defaults(kwonly_params, args.kw_defaults)
   all_params = posonly_params + pos_params + kwonly_params
@@ -159,26 +173,35 @@ def _pytd_signature(
   starstarargs = _pytd_starstar_param(args.kwarg)
   ret = pytd_function.pytd_return_type(name, function.returns, is_async)
   exceptions = exceptions or []
-  return pytd.Signature(params=params,
-                        return_type=ret,
-                        starargs=starargs,
-                        starstarargs=starstarargs,
-                        exceptions=tuple(exceptions), template=())
+  return pytd.Signature(
+      params=params,
+      return_type=ret,
+      starargs=starargs,
+      starstarargs=starstarargs,
+      exceptions=tuple(exceptions),
+      template=(),
+  )
 
 
-def _pytd_star_param(arg: ast3.arg) -> Optional[pytd.Parameter]:
+def _pytd_star_param(arg: astlib.arg) -> Optional[pytd.Parameter]:
   """Return a pytd.Parameter for a *args argument."""
   if not arg:
     return None
-  return pytd_function.pytd_star_param(arg.arg, arg.annotation)
+  # Temporary hack: until pytype supports Unpack, treat it as Any.
+  unpack = parser_constants.EXTERNAL_NAME_PREFIX + "typing_extensions.Unpack"
+  if (
+      isinstance(arg.annotation, pytd.GenericType)
+      and arg.annotation.base_type.name == unpack
+  ):
+    arg.annotation = pytd.AnythingType()
+  return pytd_function.pytd_star_param(arg.arg, arg.annotation)  # pytype: disable=wrong-arg-types
 
 
-def _pytd_starstar_param(
-    arg: Optional[ast3.arg]) -> Optional[pytd.Parameter]:
+def _pytd_starstar_param(arg: Optional[astlib.arg]) -> Optional[pytd.Parameter]:
   """Return a pytd.Parameter for a **kwargs argument."""
   if not arg:
     return None
-  return pytd_function.pytd_starstar_param(arg.arg, arg.annotation)
+  return pytd_function.pytd_starstar_param(arg.arg, arg.annotation)  # pytype: disable=wrong-arg-types
 
 
 def _apply_defaults(params: List[Param], defaults: List[Any]) -> None:
