@@ -1,10 +1,9 @@
 """Printer to output pytd trees in pyi format."""
 
 import collections
-import copy
-import keyword
 import logging
 import re
+from typing import Dict, Optional
 
 from pytype import utils
 from pytype.pytd import base_visitor
@@ -12,38 +11,169 @@ from pytype.pytd import pep484
 from pytype.pytd import pytd
 from pytype.pytd.parse import parser_constants
 
+# Aliases for readability:
+_NameType = _AliasType = str
+
+
+class _TypingImports:
+  """Imports from the `typing` module."""
+
+  def __init__(self):
+    # Typing members that are imported via `from typing import ...`.
+    self._members: Dict[_AliasType, _NameType] = {}
+    # The number of times that each typing member is used.
+    self._counts: Dict[_NameType, int] = collections.defaultdict(int)
+
+  @property
+  def members(self):
+    # Note that when a typing member has multiple aliases, this keeps only one.
+    return {name: alias for alias, name in self._members.items()}
+
+  def add(self, name: str, alias: str):
+    self._counts[name] += 1
+    self._members[alias] = name
+
+  def decrement_count(self, name: str):
+    self._counts[name] -= 1
+
+  def to_import_statements(self):
+    targets = []
+    for alias, name in self._members.items():
+      if not self._counts[name]:
+        continue
+      targets.append(f"{name} as {alias}" if alias != name else name)
+    if targets:
+      return ["from typing import " + ", ".join(sorted(targets))]
+    else:
+      return []
+
+
+class _Imports:
+  """Imports tracker."""
+
+  def __init__(self):
+    self.track_imports = True
+    self._typing = _TypingImports()
+    self._direct_imports: Dict[_AliasType, _NameType] = {}
+    self._from_imports: Dict[_NameType, Dict[_AliasType, _NameType]] = {}
+    # Map from fully qualified import name to alias
+    self._reverse_alias_map: Dict[_NameType, _AliasType] = {}
+
+  @property
+  def typing_members(self):
+    return self._typing.members
+
+  def add(self, full_name: str, alias: Optional[str] = None):
+    """Adds an import.
+
+    Examples:
+    -------------------------------------------------------
+    Import Statement           | Method Call
+    -------------------------------------------------------
+    import abc                 | add('abc')
+    import abc as xyz          | add('abc', 'xyz')
+    import foo.bar             | add('foo.bar')
+    from foo import bar        | add('foo.bar', 'bar')
+    from foo import bar as baz | add('foo.bar', 'baz')
+
+    Args:
+      full_name: The full name of the thing being imported.
+      alias: The name that the imported thing is assigned to.
+    """
+    if not self.track_imports:
+      return
+    alias = alias or full_name
+    if "." not in full_name or full_name == alias and not alias.endswith(".*"):
+      self._direct_imports[alias] = full_name
+    else:
+      module, name = full_name.rsplit(".", 1)
+      if name == "*":
+        alias = "*"
+      if module == "typing":
+        self._typing.add(name, alias)
+      else:
+        self._from_imports.setdefault(module, {})[alias] = name
+    self._reverse_alias_map[full_name] = alias
+
+  def decrement_typing_count(self, member: str):
+    self._typing.decrement_count(member)
+
+  def get_alias(self, name: str):
+    if name.startswith("typing."):
+      return self._typing.members.get(utils.strip_prefix(name, "typing."))
+    return self._reverse_alias_map.get(name)
+
+  def to_import_statements(self):
+    """Converts self to import statements."""
+    imports = self._typing.to_import_statements()
+    for alias, module in self._direct_imports.items():
+      if alias == module:
+        imports.append(f"import {module}")
+      else:
+        imports.append(f"import {module} as {alias}")
+    for module, members in self._from_imports.items():
+      targets = ", ".join(
+          sorted(
+              f"{name} as {alias}" if alias != name else name
+              for alias, name in members.items()
+          )
+      )
+      imports.append(f"from {module} import {targets}")
+    # Sort import lines lexicographically and ensure import statements come
+    # before from-import statements.
+    return sorted(imports, key=lambda s: (s.startswith("from "), s))
+
 
 class PrintVisitor(base_visitor.Visitor):
   """Visitor for converting ASTs back to pytd source code."""
+
   visits_all_node_types = True
   unchecked_node_names = base_visitor.ALL_NODE_NAMES
 
   INDENT = " " * 4
-  _RESERVED = frozenset(parser_constants.RESERVED +
-                        parser_constants.RESERVED_PYTHON)
+  _RESERVED = frozenset(
+      parser_constants.RESERVED + parser_constants.RESERVED_PYTHON
+  )
 
   def __init__(self, multiline_args=False):
     super().__init__()
     self.class_names = []  # allow nested classes
-    self.imports = collections.defaultdict(set)
     self.in_alias = False
     self.in_parameter = False
     self.in_literal = False
     self.in_constant = False
     self.in_signature = False
+    self.in_function = False
     self.multiline_args = multiline_args
 
     self._unit = None
-    self._local_names = {}
+    self._local_names = set()
     self._class_members = set()
-    self._typing_import_counts = collections.defaultdict(int)
-    self._module_aliases = {}
-    self._alias_imports = collections.defaultdict(set)
     self._paramspec_names = set()
-    self._maybe_from_typing = set()
+    self._imports = _Imports()
+
+  @property
+  def typing_imports(self):
+    return self._imports.typing_members
+
+  def copy(self):
+    # Note that copy.deepcopy is too slow to use here.
+    copy = PrintVisitor(self.multiline_args)
+    copy.in_alias = self.in_alias
+    copy.in_parameter = self.in_parameter
+    copy.in_literal = self.in_literal
+    copy.in_constant = self.in_constant
+    copy.in_signature = self.in_signature
+    copy.in_function = self.in_function
+    # pylint: disable=protected-access
+    copy._local_names = set(self._local_names)
+    copy._imports._typing._members = dict(self._imports._typing._members)
+    copy._imports._reverse_alias_map = dict(self._imports._reverse_alias_map)
+    # pylint: enable=protected-access
+    return copy
 
   def Print(self, node):
-    return node.Visit(copy.deepcopy(self))
+    return node.Visit(self.copy())
 
   def _IsEmptyTuple(self, t: pytd.GenericType) -> bool:
     """Check if it is an empty tuple."""
@@ -59,66 +189,38 @@ class PrintVisitor(base_visitor.Visitor):
     """Check if it is typing.Callable type."""
     return t.name == "typing.Callable"
 
-  def _RequireImport(self, module, name=None):
-    """Register that we're using name from module.
-
-    Args:
-      module: string identifier.
-      name: if None, means we want 'import module'. Otherwise string identifier
-       that we want to import.
-    """
-    self.imports[module].add(name)
-
-  def _ProcessTypingImports(self, imports):
-    # If we have imported something from typing_extensions do not try to also
-    # import it from typing
-    names = imports["typing"]
-    self._maybe_from_typing -= imports["typing_extensions"]
-    names |= self._maybe_from_typing
-    need_typing = False
-    for (name, count) in self._typing_import_counts.items():
-      if count:
-        need_typing = True
-      else:
-        names.discard(name)
-    if not need_typing:
-      names.discard(None)
-
-  def _GenerateImportStrings(self):
-    """Generate import statements needed by the nodes we've visited so far.
-
-    Returns:
-      List of strings.
-    """
-    ret = []
-    imports = self.imports.copy()
-    for k in self._alias_imports:
-      imports[k] = imports[k] | self._alias_imports[k]
-    self._ProcessTypingImports(imports)
-    for module, names in sorted(imports.items()):
-      if None in names:
-        ret.append(f"import {module}")
-        names.remove(None)
-      if names:
-        name_str = ", ".join(sorted(names))
-        ret.append(f"from {module} import {name_str}")
-
-    return ret
-
   def _IsBuiltin(self, module):
     return module == "builtins"
 
+  def _LookupTypingMember(self, name):
+    prefixes = ("typing", "typing_extensions")
+    for prefix in prefixes:
+      alias = self._imports.get_alias(f"{prefix}.{name}")
+      if alias:
+        return alias
+    for prefix in prefixes:
+      prefix_alias = self._imports.get_alias(prefix)
+      if prefix_alias:
+        return f"{prefix_alias}.{name}"
+    raise AssertionError("This should never happen.")
+
   def _FormatTypeParams(self, type_params):
     formatted_type_params = []
     for t in type_params:
+      if t.full_name == "typing.Self":
+        continue
       args = [f"'{t.name}'"]
       args += [self.Print(c) for c in t.constraints]
       if t.bound:
         args.append(f"bound={self.Print(t.bound)}")
+      if isinstance(t.default, tuple):
+        args.append(f"default=[{', '.join(self.Print(d) for d in t.default)}]")
+      elif t.default:
+        args.append(f"default={self.Print(t.default)}")
       if isinstance(t, pytd.ParamSpec):
-        typename = "ParamSpec"
+        typename = self._LookupTypingMember("ParamSpec")
       else:
-        typename = "TypeVar"
+        typename = self._LookupTypingMember("TypeVar")
       formatted_type_params.append(f"{t.name} = {typename}({', '.join(args)})")
     return sorted(formatted_type_params)
 
@@ -126,33 +228,22 @@ class PrintVisitor(base_visitor.Visitor):
 
     def name_in(members):
       return name in members or (
-          self._unit and f"{self._unit.name}.{name}" in members)
+          self._unit and f"{self._unit.name}.{name}" in members
+      )
 
     return name_in(self._class_members) or name_in(self._local_names)
 
   def _FromTyping(self, name):
-    self._typing_import_counts[name] += 1
+    extensions_name = self._imports.get_alias(f"typing_extensions.{name}")
+    if extensions_name:
+      return extensions_name
+    full_name = f"typing.{name}"
     if self._NameCollision(name):
-      self._RequireImport("typing")
-      return f"typing.{name}"
-    else:
-      self._RequireImport("typing", name)
-      return name
-
-  def _ImportTypingExtension(self, name):
-    if self._unit and self._unit.name:
-      full_name = f"{self._unit.name}.{name}"
-    else:
-      full_name = name
-    # `name` is a typing construct that is not supported in all Python versions.
-    if (self._local_names.get(name) == "alias" or
-        self._local_names.get(full_name) == "alias"):
-      # A typing_extensions import is parsed as Alias(X, typing_extensions.X).
-      # If we see an alias to `name`, assume it's been explicitly imported from
-      # typing_extensions due to the current Python version not supporting it.
-      return name
-    else:
-      return self._FromTyping(name)
+      self._imports.add("typing")
+      return full_name
+    alias = self._imports.get_alias(full_name) or name
+    self._imports.add(full_name, alias)
+    return alias
 
   def _StripUnitPrefix(self, name):
     if self._unit:
@@ -160,55 +251,72 @@ class PrintVisitor(base_visitor.Visitor):
     else:
       return name
 
+  def _IsAliasImport(self, node):
+    if not self._unit or self.in_constant or self.in_signature:
+      return False
+    elif isinstance(node.type, pytd.Module):
+      return True
+    # Modules and classes can be imported. Modules are represented as NamedTypes
+    # in partially resolved asts and sometimes as LateTypes in pickled asts.
+    return (
+        isinstance(node.type, (pytd.NamedType, pytd.ClassType, pytd.LateType))
+        and "." in node.type.name
+    )
+
+  def _ProcessDecorators(self, node):
+    # Our handling of class and function decorators is a bit hacky (see
+    # output.py); this makes sure that typing classes read in directly from a
+    # pyi file and then reemitted (e.g. in assertTypesMatchPytd) have their
+    # required module imports handled correctly.
+    decorators = []
+    for d in node.decorators:
+      decorators.append("@" + self.VisitNamedType(d))
+      if d.type.name.startswith("typing."):
+        self.VisitNamedType(d.type)
+    return utils.unique_list(decorators)
+
   def EnterTypeDeclUnit(self, unit):
     self._unit = unit
-    for definitions, label in [(unit.classes, "class"),
-                               (unit.functions, "function"),
-                               (unit.constants, "constant"),
-                               (unit.type_params, "type_param"),
-                               (unit.aliases, "alias")]:
-      for defn in definitions:
-        self._local_names[defn.name] = label
+    definitions = (
+        unit.classes
+        + unit.functions
+        + unit.constants
+        + unit.type_params
+        + unit.aliases
+    )
+    self._local_names = {c.name for c in definitions}
     for alias in unit.aliases:
-      # Modules are represented as NamedTypes in partially resolved asts and
-      # sometimes as LateTypes in asts modified for pickling.
-      if isinstance(alias.type, pytd.Module):
-        module_name = alias.type.module_name
-      elif isinstance(alias.type, (pytd.NamedType, pytd.LateType)):
-        module_name = alias.type.name
-      else:
+      if not self._IsAliasImport(alias):
         continue
-      name = self._StripUnitPrefix(alias.name)
-      self._module_aliases[module_name] = name
-    self._paramspec_names = {x.name for x in unit.type_params
-                             if isinstance(x, pytd.ParamSpec)}
+      if isinstance(alias.type, pytd.Module):
+        name = alias.type.module_name
+        alias_name = alias.type.name
+      else:
+        name = alias.type.name
+        alias_name = self._StripUnitPrefix(alias.name)
+      self._imports.add(name, alias_name)
+    self._paramspec_names = {
+        x.name for x in unit.type_params if isinstance(x, pytd.ParamSpec)
+    }
 
   def LeaveTypeDeclUnit(self, _):
     self._unit = None
-    self._local_names = {}
+    self._local_names = set()
 
   def VisitTypeDeclUnit(self, node):
     """Convert the AST for an entire module back to a string."""
     for t in self.old_node.type_params:
       if isinstance(t, pytd.ParamSpec):
-        self._maybe_from_typing.add("ParamSpec")
+        self._FromTyping("ParamSpec")
+      elif t.full_name == "typing.Self":
+        self._imports.add("typing.Self", "Self")
       else:
         self._FromTyping("TypeVar")
+    imports = self._imports.to_import_statements()
 
-    aliases = []
-    imports = set(self._GenerateImportStrings())
-    for alias in filter(None, node.aliases):
-      if alias.startswith(("from ", "import ")):
-        imports.add(alias)
-      else:
-        aliases.append(alias)
-
-    # Sort import lines lexicographically and ensure import statements come
-    # before from-import statements.
-    imports = sorted(imports, key=lambda s: (s.startswith("from "), s))
-
-    # Remove deleted constants
-    constants = [c for c in node.constants if "<deleted>" not in c]
+    # Remove deleted nodes
+    aliases = list(filter(None, node.aliases))
+    constants = list(filter(None, node.constants))
 
     sections = [
         imports,
@@ -221,9 +329,11 @@ class PrintVisitor(base_visitor.Visitor):
 
     # We put one blank line after every class,so we need to strip the blank line
     # after the last class.
-    sections_as_string = ("\n".join(section_suite).rstrip()
-                          for section_suite in sections
-                          if section_suite)
+    sections_as_string = (
+        "\n".join(section_suite).rstrip()
+        for section_suite in sections
+        if section_suite
+    )
     return "\n\n".join(sections_as_string)
 
   def EnterConstant(self, node):
@@ -240,10 +350,19 @@ class PrintVisitor(base_visitor.Visitor):
     #   from typing import Foo
     if self.class_names or node.value:
       return False
-    if node.type == f"Type[typing.{node.name}]":
-      self._RequireImport("typing", node.name)
-      self._typing_import_counts["Type"] -= 1
-      del self._local_names[node.name]
+    full_typing_name = f"typing.{node.name}"
+    # TODO(b/315507078): This is only necessary while these three classes
+    # are aliases of typing members.
+    if full_typing_name in (
+        "typing.ChainMap",
+        "typing.Counter",
+        "typing.OrderedDict",
+    ):
+      return False
+    if node.type == f"Type[{full_typing_name}]":
+      self._imports.add(full_typing_name, node.name)
+      self._imports.decrement_typing_count("Type")
+      self._local_names.remove(node.name)
       return True
 
   def VisitConstant(self, node):
@@ -257,35 +376,26 @@ class PrintVisitor(base_visitor.Visitor):
       else:
         return node.name
     # Decrement Any, since the actual value is never printed.
-    if node.value == "Any":
-      self._typing_import_counts["Any"] -= 1
+    if isinstance(self.old_node.value, pytd.AnythingType):
+      self._imports.decrement_typing_count("Any")
     if self._DropTypingConstant(node):
-      return "<deleted>"
+      return None
     # Whether the constant has a default value is important for fields in
     # generated classes like namedtuples.
     suffix = " = ..." if node.value else ""
     return f"{node.name}: {node.type}{suffix}"
 
-  def EnterAlias(self, _):
-    self.old_imports = self.imports.copy()
+  def EnterAlias(self, node):
+    if self.in_function or self._IsAliasImport(node):
+      self._imports.track_imports = False
+
+  def LeaveAlias(self, _):
+    self._imports.track_imports = True
 
   def VisitAlias(self, node):
     """Convert an import or alias to a string (or None if handled elsewhere)."""
-    if (isinstance(self.old_node.type,
-                   (pytd.NamedType, pytd.ClassType, pytd.LateType)) and
-        not self.in_constant and not self.in_signature):
-      full_name = self.old_node.type.name
-      suffix = ""
-      module, _, name = full_name.rpartition(".")
-      if module:
-        alias_name = self._StripUnitPrefix(self.old_node.name)
-        if name not in ("*", alias_name):
-          suffix += f" as {alias_name}"
-        self.imports = self.old_imports  # undo unnecessary imports change
-        self._alias_imports[module].add(f"{name}{suffix}")
-        # Return None here since we do not want to emit the import statement
-        # from both self._alias_imports and unit.aliases
-        return None
+    if self._IsAliasImport(self.old_node):
+      return None  # the import statement has already been processed
     elif isinstance(self.old_node.type, (pytd.Constant, pytd.Function)):
       return self.Print(self.old_node.type.Replace(name=node.name))
     elif isinstance(self.old_node.type, pytd.Module):
@@ -318,33 +428,34 @@ class PrintVisitor(base_visitor.Visitor):
       for c in node.constants:
         name, typ = c.split(": ")
         constants[name] = typ
-      if any(keyword.iskeyword(name) for name in constants):
+      if any(not utils.is_valid_name(name) for name in constants):
         # We output the TypedDict in functional form, since using the class form
         # would produce a parse error when the pyi file is ingested.
-        fields = "{%s}" % ", ".join(f"{name!r}: {typ}"
-                                    for name, typ in constants.items())
+        fields = "{%s}" % ", ".join(
+            f"{name!r}: {typ}" for name, typ in constants.items()
+        )
         return f"{node.name} = TypedDict('{node.name}', {fields})"
     # If object is the only base, we don't need to list any bases.
     if bases == ("object",):
       bases = ()
-    if node.metaclass is not None:
-      bases += ("metaclass=" + node.metaclass,)
+    keywords = []
+    for k, v in node.keywords:
+      vmatch = re.fullmatch(r"Literal\[(.+)\]", v)
+      if vmatch:
+        self._imports.decrement_typing_count("Literal")
+        vprint = vmatch.group(1)
+      else:
+        vprint = v
+      keywords.append(f"{k}={vprint}")
+    bases += tuple(keywords)
     bases_str = f"({', '.join(bases)})" if bases else ""
     header = [f"class {node.name}{bases_str}:"]
     if node.slots is not None:
-      slots_str = ", ".join(f"\"{s}\"" for s in node.slots)
+      slots_str = ", ".join(f'"{s}"' for s in node.slots)
       slots = [self.INDENT + f"__slots__ = [{slots_str}]"]
     else:
       slots = []
-    decorators = ["@" + self.VisitNamedType(d)
-                  for d in self.old_node.decorators]
-    # Our handling of class decorators is a bit hacky (see output.py); this
-    # makes sure that typing classes read in directly from a pyi file and then
-    # reemitted (e.g. in assertTypesMatchPytd) have their required module
-    # imports handled correctly.
-    for d in self.old_node.decorators:
-      if d.type.name.startswith("typing."):
-        self.VisitNamedType(d.type)
+    decorators = self._ProcessDecorators(self.old_node)
     if node.classes or node.methods or node.constants or slots:
       # We have multiple methods, and every method has multiple signatures
       # (i.e., the method string will have multiple lines). Combine this into
@@ -362,17 +473,28 @@ class PrintVisitor(base_visitor.Visitor):
     lines = decorators + header + slots + classes + constants + methods
     return "\n".join(lines) + "\n"
 
+  def EnterFunction(self, node):
+    self.in_function = True
+
+  def LeaveFunction(self, node):
+    self.in_function = False
+
   def VisitFunction(self, node):
     """Visit function, producing multi-line string (one for each signature)."""
     function_name = node.name
-    decorators = ""
+    if self.old_node.decorators:
+      decorators = self._ProcessDecorators(self.old_node)
+      decorators = "\n".join(decorators) + "\n"
+    else:
+      decorators = ""
     if node.is_final:
       decorators += "@" + self._FromTyping("final") + "\n"
-    if (node.kind == pytd.MethodKind.STATICMETHOD and
-        function_name != "__new__"):
+    if node.kind == pytd.MethodKind.STATICMETHOD and function_name != "__new__":
       decorators += "@staticmethod\n"
-    elif (node.kind == pytd.MethodKind.CLASSMETHOD and
-          function_name != "__init_subclass__"):
+    elif (
+        node.kind == pytd.MethodKind.CLASSMETHOD
+        and function_name != "__init_subclass__"
+    ):
       decorators += "@classmethod\n"
     elif node.kind == pytd.MethodKind.PROPERTY:
       decorators += "@property\n"
@@ -382,8 +504,9 @@ class PrintVisitor(base_visitor.Visitor):
       decorators += "@coroutine\n"
     if len(node.signatures) > 1:
       decorators += "@" + self._FromTyping("overload") + "\n"
-    signatures = "\n".join(decorators + "def " + function_name + sig
-                           for sig in node.signatures)
+    signatures = "\n".join(
+        decorators + "def " + function_name + sig for sig in node.signatures
+    )
     return signatures
 
   def _FormatContainerContents(self, node: pytd.Parameter) -> str:
@@ -391,14 +514,15 @@ class PrintVisitor(base_visitor.Visitor):
     if isinstance(node.type, pytd.GenericType):
       container_name = node.type.name.rpartition(".")[2]
       assert container_name in ("tuple", "dict")
-      self._typing_import_counts[container_name.capitalize()] -= 1
+      self._imports.decrement_typing_count(container_name.capitalize())
       # If the type is "Any", e.g. `**kwargs: Any`, decrement Any to avoid an
       # extraneous import of typing.Any. Any was visited before this function
       # transformed **kwargs, so it was incremented at least once already.
       if isinstance(node.type.parameters[-1], pytd.AnythingType):
-        self._typing_import_counts["Any"] -= 1
+        self._imports.decrement_typing_count("Any")
       return self.Print(
-          node.Replace(type=node.type.parameters[-1], optional=False))
+          node.Replace(type=node.type.parameters[-1], optional=False)
+      )
     else:
       return self.Print(node.Replace(type=pytd.AnythingType(), optional=False))
 
@@ -411,8 +535,7 @@ class PrintVisitor(base_visitor.Visitor):
   def VisitSignature(self, node):
     """Visit a signature, producing a string."""
     if node.return_type == "nothing":
-      return_type = "NoReturn"  # a prettier alias for nothing
-      self._FromTyping(return_type)
+      return_type = self._FromTyping("Never")  # a prettier alias for nothing
     else:
       return_type = node.return_type
     ret = f" -> {return_type}"
@@ -428,15 +551,18 @@ class PrintVisitor(base_visitor.Visitor):
     params = []
     for i, p in enumerate(node.params):
       if self.old_node.params[i].kind == pytd.ParameterKind.KWONLY:
-        assert all(p.kind == pytd.ParameterKind.KWONLY
-                   for p in self.old_node.params[i:])
+        assert all(
+            p.kind == pytd.ParameterKind.KWONLY
+            for p in self.old_node.params[i:]
+        )
         params.append("*" + starargs)
         params.extend(node.params[i:])
         break
       params.append(p)
-      if (self.old_node.params[i].kind == pytd.ParameterKind.POSONLY and
-          (i == len(node.params)-1 or
-           self.old_node.params[i+1].kind != pytd.ParameterKind.POSONLY)):
+      if self.old_node.params[i].kind == pytd.ParameterKind.POSONLY and (
+          i == len(node.params) - 1
+          or self.old_node.params[i + 1].kind != pytd.ParameterKind.POSONLY
+      ):
         params.append("/")
     else:
       if starargs:
@@ -449,8 +575,11 @@ class PrintVisitor(base_visitor.Visitor):
     # Handle Mutable parameters
     # pylint: disable=no-member
     # (old_node is set in parse/node.py)
-    mutable_params = [(p.name, p.mutated_type) for p in self.old_node.params
-                      if p.mutated_type is not None]
+    mutable_params = [
+        (p.name, p.mutated_type)
+        for p in self.old_node.params
+        if p.mutated_type is not None
+    ]
     # pylint: enable=no-member
     for name, new_type in mutable_params:
       body.append(f"\n{self.INDENT}{name} = {self.Print(new_type)}")
@@ -475,26 +604,32 @@ class PrintVisitor(base_visitor.Visitor):
     assert self.in_parameter
     self.in_parameter = False
 
+  def _DecrementParameterImports(self, name):
+    if "[" not in name:
+      return
+    param = name.split("[", 1)[-1]
+    for k in self._imports.typing_members:
+      if re.search(r"\b%s\b" % k, param):
+        self._imports.decrement_typing_count(k)
+
   def VisitParameter(self, node):
     """Convert a function parameter to a string."""
     suffix = " = ..." if node.optional else ""
-    if node.type == "Any":
-      # Abbreviated form. "Any" is the default.
-      self._typing_import_counts["Any"] -= 1
-      return node.name + suffix
     # For parameterized class, for example: ClsName[T, V].
     # Its name is `ClsName` before `[`.
-    elif node.name == "self" and self.class_names and (
-        self.class_names[-1].split("[")[0] == node.type.split("[")[0]):
-      if "[" in node.type:
-        elided = node.type.split("[", 1)[-1]
-        for k in self._typing_import_counts:
-          if re.search(r"(^|\W)%s($|\W)" % k, elided):
-            self._typing_import_counts[k] -= 1
+    class_name = self.class_names[-1].split("[")[0] if self.class_names else ""
+    if isinstance(self.old_node.type, pytd.AnythingType):
+      # Abbreviated form. "Any" is the default.
+      self._imports.decrement_typing_count("Any")
       return node.name + suffix
-    elif node.name == "cls" and self.class_names and (
-        node.type == f"Type[{self.class_names[-1]}]"):
-      self._typing_import_counts["Type"] -= 1
+    elif node.name == "self" and class_name == node.type.split("[")[0]:
+      self._DecrementParameterImports(node.type)
+      return node.name + suffix
+    elif node.name == "cls" and re.fullmatch(
+        rf"Type\[{class_name}(\[.+\])?\]", node.type
+    ):
+      self._imports.decrement_typing_count("Type")
+      self._DecrementParameterImports(node.type[5:-1])
       return node.name + suffix
     elif node.type is None:
       logging.warning("node.type is None")
@@ -509,8 +644,9 @@ class PrintVisitor(base_visitor.Visitor):
   def _UseExistingModuleAlias(self, name):
     prefix, suffix = name.rsplit(".", 1)
     while prefix:
-      if prefix in self._module_aliases:
-        return f"{self._module_aliases[prefix]}.{suffix}"
+      prefix_alias = self._imports.get_alias(prefix)
+      if prefix_alias:
+        return f"{prefix_alias}.{suffix}"
       prefix, _, remainder = prefix.rpartition(".")
       suffix = f"{remainder}.{suffix}"
     return None
@@ -535,7 +671,8 @@ class PrintVisitor(base_visitor.Visitor):
     elif prefix == "typing":
       node_name = self._FromTyping(suffix)
     elif prefix == "typing_extensions":
-      node_name = self._ImportTypingExtension(suffix)
+      self._imports.add(node.name, suffix)
+      node_name = self._FromTyping(suffix)
     elif "." not in node.name:
       node_name = node.name
     else:
@@ -551,11 +688,10 @@ class PrintVisitor(base_visitor.Visitor):
             module_alias = module
             while self._NameCollision(module_alias):
               module_alias = f"_{module_alias}"
+            self._imports.add(module, module_alias)
             if module_alias == module:
-              self._RequireImport(module)
               node_name = node.name
             else:
-              self._RequireImport(f"{module} as {module_alias}")
               node_name = ".".join(filter(bool, (module_alias, rest, suffix)))
         else:
           node_name = node.name
@@ -599,20 +735,7 @@ class PrintVisitor(base_visitor.Visitor):
     return f"{node.name}.kwargs"
 
   def VisitModule(self, node):
-    if self.in_constant or self.in_signature:
-      return "module"
-    elif not node.is_aliased:
-      return f"import {node.module_name}"
-    elif "." in node.module_name:
-      # `import x.y as z` and `from x import y as z` are equivalent, but the
-      # latter is a bit prettier.
-      prefix, suffix = node.module_name.rsplit(".", 1)
-      imp = f"from {prefix} import {suffix}"
-      if node.name != suffix:
-        imp += f" as {node.name}"
-      return imp
-    else:
-      return f"import {node.module_name} as {node.name}"
+    return "module"
 
   def MaybeCapitalize(self, name):
     """Capitalize a generic type, if necessary."""
@@ -632,12 +755,16 @@ class PrintVisitor(base_visitor.Visitor):
       param = self.old_node.parameters[0]
       # Callable[Any, X] is rewritten to Callable[..., X].
       if isinstance(param, pytd.AnythingType):
-        self._typing_import_counts["Any"] -= 1
+        self._imports.decrement_typing_count("Any")
       else:
         assert isinstance(param, (pytd.NothingType, pytd.TypeParameter)), param
       parameters = ("...",) + parameters[1:]
-    return (self.MaybeCapitalize(node.base_type) +
-            "[" + ", ".join(parameters) + "]")
+    return (
+        self.MaybeCapitalize(node.base_type)
+        + "["
+        + ", ".join(str(p) for p in parameters)
+        + "]"
+    )
 
   def VisitCallableType(self, node):
     typ = self.MaybeCapitalize(node.base_type)
@@ -651,7 +778,7 @@ class PrintVisitor(base_visitor.Visitor):
       return f"{typ}[[{args}], {node.ret}]"
 
   def VisitConcatenate(self, node):
-    base = self._ImportTypingExtension("Concatenate")
+    base = self._FromTyping("Concatenate")
     parameters = ", ".join(node.parameters)
     return f"{base}[{parameters}]"
 
@@ -672,7 +799,7 @@ class PrintVisitor(base_visitor.Visitor):
     """Form list of types within a set type."""
     type_list = dict.fromkeys(node.type_list)
     if self.in_parameter:
-      for compat, name in pep484.COMPAT_ITEMS:
+      for compat, name in pep484.get_compat_items():
         # name can replace compat.
         if compat in type_list and name in type_list:
           del type_list[compat]
@@ -703,8 +830,12 @@ class PrintVisitor(base_visitor.Visitor):
     if len(new_type_list) == 1:
       return new_type_list[0]
     elif "None" in new_type_list:
-      return (self._FromTyping("Optional") + "[" +
-              self._BuildUnion(t for t in new_type_list if t != "None") + "]")
+      return (
+          self._FromTyping("Optional")
+          + "["
+          + self._BuildUnion(t for t in new_type_list if t != "None")
+          + "]"
+      )
     else:
       return self._FromTyping("Union") + "[" + ", ".join(new_type_list) + "]"
 
@@ -733,10 +864,10 @@ class PrintVisitor(base_visitor.Visitor):
     self.in_literal = False
 
   def VisitLiteral(self, node):
-    base = self._ImportTypingExtension("Literal")
+    base = self._FromTyping("Literal")
     return f"{base}[{node.value}]"
 
   def VisitAnnotated(self, node):
-    base = self._ImportTypingExtension("Annotated")
+    base = self._FromTyping("Annotated")
     annotations = ", ".join(node.annotations)
     return f"{base}[{node.base_type}, {annotations}]"
