@@ -1,57 +1,36 @@
 """Internal representation of pytd nodes.
 
-All pytd nodes should be frozen attrs inheriting from node.Node (aliased
-to Node in this module). Nodes representing types should also inherit from Type.
-
-Since we use frozen classes, setting attributes in __post_init__ needs to be
-done via
-  object.__setattr__(self, 'attr', value)
+All pytd nodes should be frozen msgspec.Structs inheriting from node.Node
+(aliased to Node in this module). Nodes representing types should inherit from
+Type.
 
 NOTE: The way we introspect on the types of the fields requires forward
 references to be simple classes, hence use
   x: Union['Foo', 'Bar']
 rather than
   x: 'Union[Foo, Bar]'
+
+pytd_test.py tests the construction and runtime usage of these classes.
+To test how these classes serialize, see serialize_ast_test.py.
 """
 
+from __future__ import annotations
+
 import enum
 import itertools
-
-import typing
-from typing import Any, Optional, Tuple, TypeVar, Union
-
-import attrs
+from typing import Any, Dict, Generator, Optional, Tuple, Union
 
 from pytype.pytd.parse import node
 
 # Alias node.Node for convenience.
 Node = node.Node
 
-_TypeT = TypeVar('_TypeT', bound='Type')
+
+class Type(Node):
+  """Each type class below should inherit from this marker."""
 
 
-class Type:
-  """Each type class below should inherit from this mixin."""
-  name: Optional[str]
-
-  # We type-annotate many things as pytd.Type when we'd really want them to be
-  # Intersection[pytd.Type, pytd.parse.node.Node], so we need to copy some type
-  # signatures from Node here.
-  if typing.TYPE_CHECKING:
-
-    def Replace(self: _TypeT, **kwargs) -> _TypeT:
-      del kwargs  # unused
-      return self
-
-    def Visit(self: _TypeT, visitor, *args, **kwargs) -> _TypeT:
-      del visitor, args, kwargs  # unused
-      return self
-
-  __slots__ = ()
-
-
-@attrs.frozen(slots=False, eq=False)
-class TypeDeclUnit(Node):
+class TypeDeclUnit(Node, eq=False):
   """Module node. Holds module contents (constants / classes / functions).
 
   Attributes:
@@ -62,18 +41,23 @@ class TypeDeclUnit(Node):
     classes: Iterable of classes defined in this type decl unit.
     aliases: Iterable of aliases (or imports) for types in other modules.
   """
-  name: Optional[str]
-  constants: Tuple['Constant', ...]
-  type_params: Tuple['TypeParameter', ...]
-  classes: Tuple['Class', ...]
-  functions: Tuple['Function', ...]
-  aliases: Tuple['Alias', ...]
+
+  name: str
+  constants: Tuple[Constant, ...]
+  type_params: Tuple[TypeParameterU, ...]
+  classes: Tuple[Class, ...]
+  functions: Tuple[Function, ...]
+  aliases: Tuple[Alias, ...]
+  # _name2item is the lookup cache. It should not be treated as a child or used
+  # in equality or hash operations.
+  _name2item: Dict[str, Any] = {}
 
   def _InitCache(self):
     # TODO(b/159053187): Put constants, functions, classes and aliases into a
     # combined dict.
-    self.PopulateLookupCache(
-        self.constants, self.functions, self.classes, self.aliases)
+    for x in (self.constants, self.functions, self.classes, self.aliases):
+      for item in x:
+        self._name2item[item.name] = item
     for x in self.type_params:
       self._name2item[x.full_name] = x
 
@@ -91,43 +75,47 @@ class TypeDeclUnit(Node):
     Raises:
       KeyError: if this identifier doesn't exist.
     """
-    try:
-      return self._name2item[name]
-    except AttributeError:
+    if not self._name2item:
       self._InitCache()
-      return self._name2item[name]
+    return self._name2item[name]
 
   def Get(self, name):
     """Version of Lookup that returns None instead of raising."""
-    try:
-      return self._name2item.get(name)
-    except AttributeError:
+    if not self._name2item:
       self._InitCache()
-      return self._name2item.get(name)
+    return self._name2item.get(name)
 
   def __contains__(self, name):
     return bool(self.Get(name))
 
+  def IterChildren(self) -> Generator[Tuple[str, Any | None], None, None]:
+    for name, child in super().IterChildren():
+      if name == '_name2item':
+        continue
+      yield name, child
+
+  def Replace(self, **kwargs):
+    if '_name2item' not in kwargs:
+      kwargs['_name2item'] = {}
+    return super().Replace(**kwargs)
+
   # The hash/eq/ne values are used for caching and speed things up quite a bit.
 
   def __hash__(self):
     return id(self)
 
-  def __eq__(self, other):
-    return id(self) == id(other)
 
-  def __ne__(self, other):
-    return id(self) != id(other)
-
-
-@attrs.frozen(cache_hash=True)
 class Constant(Node):
   name: str
-  type: Type
-  value: Any = None
+  type: TypeU
+  # Value may be any value allowed in a Literal (int, str, bool, None),
+  # a tuple[str, ...] (used to hold `__all__`) or Any (i.e. AnythingType).
+  # (bytes is excluded because it serializes the same as str.)
+  # We can't use just `value: Any` because msgspec isn't able to decode
+  # AnythingType in that case, since it's indistinguishable from a dict.
+  value: Optional[Union[AnythingType, int, str, bool, Tuple[str, ...]]] = None
 
 
-@attrs.frozen(cache_hash=True)
 class Alias(Node):
   """An alias (symbolic link) for a class implemented in some other module.
 
@@ -135,22 +123,18 @@ class Alias(Node):
   type. It's generated, among others, from imports - e.g. "from x import y as z"
   will create a local alias "z" for "x.y".
   """
+
   name: str
-  type: Union[Type, Constant, 'Function', 'Module']
+  type: Union[TypeU, Constant, Function, Module]
 
 
-@attrs.frozen(cache_hash=True)
 class Module(Node):
   """A module imported into the current module, possibly with an alias."""
+
   name: str
   module_name: str
 
-  @property
-  def is_aliased(self):
-    return self.name != self.module_name
 
-
-@attrs.frozen(slots=False, cache_hash=True)
 class Class(Node):
   """Represents a class declaration.
 
@@ -159,27 +143,33 @@ class Class(Node):
   Attributes:
     name: Class name (string)
     bases: The super classes of this class (instances of pytd.Type).
-    methods: Tuple of methods, classmethods, staticmethods
-      (instances of pytd.Function).
+    methods: Tuple of methods, classmethods, staticmethods (instances of
+      pytd.Function).
     constants: Tuple of constant class attributes (instances of pytd.Constant).
     classes: Tuple of nested classes.
     slots: A.k.a. __slots__, declaring which instance attributes are writable.
     template: Tuple of pytd.TemplateItem instances.
   """
+
   name: str
-  metaclass: Union[None, Type]
-  bases: Tuple[Union['Class', Type], ...]
-  methods: Tuple['Function', ...]
+  keywords: Tuple[Tuple[str, TypeU], ...]
+  bases: Tuple[Union[Class, TypeU], ...]
+  methods: Tuple[Function, ...]
   constants: Tuple[Constant, ...]
-  classes: Tuple['Class', ...]
+  classes: Tuple[Class, ...]
   decorators: Tuple[Alias, ...]
   slots: Optional[Tuple[str, ...]]
-  template: Tuple['TemplateItem', ...]
+  template: Tuple[TemplateItem, ...]
+  # _name2item is the lookup cache. It should not be treated as a child or used
+  # in equality or hash operations.
+  _name2item: Dict[str, Any] = {}
 
   def _InitCache(self):
     # TODO(b/159053187): Put constants, functions, classes and aliases into a
     # combined dict.
-    self.PopulateLookupCache(self.methods, self.constants, self.classes)
+    for x in (self.methods, self.constants, self.classes):
+      for item in x:
+        self._name2item[item.name] = item
 
   def Lookup(self, name):
     """Convenience function: Look up a given name in the class namespace.
@@ -196,23 +186,45 @@ class Class(Node):
       KeyError: if this identifier doesn't exist in this class.
     """
     # TODO(b/159053187): Remove this. Make methods and constants dictionaries.
-    try:
-      return self._name2item[name]
-    except AttributeError:
+    if not self._name2item:
       self._InitCache()
-      return self._name2item[name]
+    return self._name2item[name]
 
   def Get(self, name):
     """Version of Lookup that returns None instead of raising."""
-    try:
-      return self._name2item.get(name)
-    except AttributeError:
+    if not self._name2item:
       self._InitCache()
-      return self._name2item.get(name)
+    return self._name2item.get(name)
 
   def __contains__(self, name):
     return bool(self.Get(name))
 
+  def __hash__(self):
+    # _name2item is a dict, so it can't be hashed. This worked in the previous
+    # version by pretending that _name2item didn't exist.
+    # We could also delete the cache on self, but making a new instance should
+    # be cheaper than recomputing the cache.
+    nohash = self.Replace(_name2item=None)
+    return super(Class, nohash).__hash__()
+
+  def IterChildren(self) -> Generator[Tuple[str, Any | None], None, None]:
+    for name, child in super().IterChildren():
+      if name == '_name2item':
+        continue
+      yield name, child
+
+  def Replace(self, **kwargs):
+    if '_name2item' not in kwargs:
+      kwargs['_name2item'] = {}
+    return super().Replace(**kwargs)
+
+  @property
+  def metaclass(self):
+    for key, val in self.keywords:
+      if key == 'metaclass':
+        return val
+    return None
+
 
 class MethodKind(enum.Enum):
   METHOD = 'method'
@@ -233,7 +245,6 @@ class MethodFlag(enum.Flag):
     return cls.ABSTRACT if is_abstract else cls.NONE
 
 
-@attrs.frozen(cache_hash=True)
 class Function(Node):
   """A function or a method, defined by one or more PyTD signatures.
 
@@ -243,10 +254,12 @@ class Function(Node):
     kind: The kind of function (e.g., MethodKind.STATICMETHOD).
     flags: A bitfield of flags like is_abstract
   """
+
   name: str
-  signatures: Tuple['Signature', ...]
+  signatures: Tuple[Signature, ...]
   kind: MethodKind
   flags: MethodFlag = MethodFlag.NONE
+  decorators: Tuple[Alias, ...] = ()
 
   @property
   def is_abstract(self):
@@ -266,7 +279,6 @@ class Function(Node):
     return self.Replace(flags=new_flags)
 
 
-@attrs.frozen(cache_hash=True)
 class Signature(Node):
   """Represents an individual signature of a function.
 
@@ -282,16 +294,13 @@ class Signature(Node):
     exceptions: List of exceptions for this function definition.
     template: names for bindings for bounded types in params/return_type
   """
-  params: Tuple['Parameter', ...]
-  starargs: Optional['Parameter']
-  starstarargs: Optional['Parameter']
-  return_type: Type
-  exceptions: Tuple[Type, ...]
-  template: Tuple['TemplateItem', ...]
 
-  @property
-  def name(self):
-    return None
+  params: Tuple[Parameter, ...]
+  starargs: Optional[Parameter]
+  starstarargs: Optional[Parameter]
+  return_type: TypeU
+  exceptions: Tuple[TypeU, ...]
+  template: Tuple[TemplateItem, ...]
 
   @property
   def has_optional(self):
@@ -304,7 +313,6 @@ class ParameterKind(enum.Enum):
   KWONLY = 'kwonly'
 
 
-@attrs.frozen(cache_hash=True)
 class Parameter(Node):
   """Represents a parameter of a function definition.
 
@@ -316,15 +324,15 @@ class Parameter(Node):
     mutated_type: The type the parameter will have after the function is called
       if the type is mutated, None otherwise.
   """
+
   name: str
-  type: Type
+  type: TypeU
   kind: ParameterKind
   optional: bool
-  mutated_type: Optional[Type]
+  mutated_type: Optional[TypeU]
 
 
-@attrs.frozen(cache_hash=True)
-class TypeParameter(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class TypeParameter(Type):
   """Represents a type parameter.
 
   A type parameter is a bound variable in the context of a function or class
@@ -334,12 +342,16 @@ class TypeParameter(Node, Type):  # pytype: disable=signature-mismatch  # overri
 
   Attributes:
     name: Name of the parameter. E.g. "T".
-    scope: Fully-qualified name of the class or function this parameter is
-      bound to. E.g. "mymodule.MyClass.mymethod", or None.
+    constraints: The valid types for the TypeParameter. Exclusive with bound.
+    bound: The upper bound for the TypeParameter. Exclusive with constraints.
+    scope: Fully-qualified name of the class or function this parameter is bound
+      to. E.g. "mymodule.MyClass.mymethod", or None.
   """
+
   name: str
-  constraints: Tuple[Type, ...] = attrs.field(factory=tuple)
-  bound: Optional[Type] = None
+  constraints: Tuple[TypeU, ...] = ()
+  bound: Optional[TypeU] = None
+  default: Optional[Union[TypeU, Tuple[TypeU, ...]]] = None
   scope: Optional[str] = None
 
   def __lt__(self, other):
@@ -368,7 +380,6 @@ class TypeParameter(Node, Type):  # pytype: disable=signature-mismatch  # overri
       return AnythingType()
 
 
-@attrs.frozen(cache_hash=True)
 class ParamSpec(TypeParameter):
   """ParamSpec is a specific case of TypeParameter."""
 
@@ -382,19 +393,18 @@ class ParamSpec(TypeParameter):
       return None
 
 
-@attrs.frozen(cache_hash=True)
 class ParamSpecArgs(Node):
   """ParamSpec.args special form."""
+
   name: str
 
 
-@attrs.frozen(cache_hash=True)
 class ParamSpecKwargs(Node):
   """ParamSpec.kwargs special form."""
+
   name: str
 
 
-@attrs.frozen(cache_hash=True)
 class TemplateItem(Node):
   """Represents template name for generic types.
 
@@ -410,7 +420,8 @@ class TemplateItem(Node):
     type_param: the TypeParameter instance used. This is the actual instance
       that's used wherever this type parameter appears, e.g. within a class.
   """
-  type_param: TypeParameter
+
+  type_param: TypeParameterU
 
   @property
   def name(self):
@@ -438,39 +449,42 @@ class TemplateItem(Node):
 # corresponding AST representations.
 
 
-@attrs.frozen(cache_hash=True)
-class NamedType(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class NamedType(Type):
   """A type specified by name and, optionally, the module it is in."""
+
   name: str
 
   def __str__(self):
     return self.name
 
 
-@attrs.mutable(init=False, slots=False, eq=False)
-class ClassType(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class ClassType(Type, frozen=False, eq=False):
   """A type specified through an existing class node."""
+
   # This type is different from normal nodes:
   # (a) It's mutable, and there are functions
   #     (parse/visitors.py:FillInLocalPointers) that modify a tree in place.
-  # (b) The cls pointer is not treated as a regular attr field.
-  # (c) Visitors will not process the "children" of this node. Since we point
+  # (b) Visitors will not process the "children" of this node. Since we point
   #     to classes that are back at the top of the tree, that would generate
   #     cycles.
 
-  name: str = attrs.field()
+  name: str
+  # This should be Type, or even Class. Unfortunately, doing so breaks
+  # base_visitor:_GetAncestorMap by making ClassType an ancestor of all the
+  # Type subclasses. This is undesirable, because the cls pointer should be
+  # treated as if it doesn't exist.
+  cls: Optional[Any] = None
 
-  # We do not want cls to be a child node, but we do want it to be an optional
-  # arg to __init__ and accessible via self.cls
-  cls = None
-
-  def __init__(self, name, cls=None):
-    self.name = name
-    self.cls = cls
+  def IterChildren(self) -> Generator[Tuple[str, Any | None], None, None]:
+    # It is very important that visitors do not follow the cls pointer. To avoid
+    # this, we claim that `name` is the only child.
+    yield 'name', self.name
 
   def __eq__(self, other):
-    return (self.__class__ == other.__class__ and
-            self.name == other.name)
+    return self.__class__ == other.__class__ and self.name == other.name
+
+  def __ne__(self, other):
+    return not self == other
 
   def __hash__(self):
     return hash((self.__class__.__name__, self.name))
@@ -480,13 +494,15 @@ class ClassType(Node, Type):  # pytype: disable=signature-mismatch  # overriding
 
   def __repr__(self):
     return '{type}{cls}({name})'.format(
-        type=type(self).__name__, name=self.name,
-        cls='<unresolved>' if self.cls is None else '')
+        type=type(self).__name__,
+        name=self.name,
+        cls='<unresolved>' if self.cls is None else '',
+    )
 
 
-@attrs.frozen(cache_hash=True)
-class LateType(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class LateType(Type):
   """A type we have yet to resolve."""
+
   name: str
   recursive: bool = False
 
@@ -494,30 +510,20 @@ class LateType(Node, Type):  # pytype: disable=signature-mismatch  # overriding-
     return self.name
 
 
-@attrs.frozen(cache_hash=True)
-class AnythingType(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class AnythingType(Type):
   """A type we know nothing about yet (? in pytd)."""
 
-  @property
-  def name(self):
-    return None
-
   def __bool__(self):
     return True
 
 
-@attrs.frozen(cache_hash=True)
-class NothingType(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class NothingType(Type):
   """An "impossible" type, with no instances (nothing in pytd).
 
   Also known as the "uninhabited" type, or, in type systems, the "bottom" type.
   For representing empty lists, and functions that never return.
   """
 
-  @property
-  def name(self):
-    return None
-
   def __bool__(self):
     return True
 
@@ -526,29 +532,26 @@ def _FlattenTypes(type_list) -> Tuple[Type, ...]:
   """Helper function for _SetOfTypes initialization."""
   assert type_list  # Disallow empty sets. Use NothingType for these.
   flattened = itertools.chain.from_iterable(
-      t.type_list if isinstance(t, _SetOfTypes) else [t]
-      for t in type_list)
+      t.type_list if isinstance(t, _SetOfTypes) else [t] for t in type_list
+  )
   # Remove duplicates, preserving order
   unique = tuple(dict.fromkeys(flattened))
   return unique
 
 
-@attrs.frozen(eq=False)
-class _SetOfTypes(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class _SetOfTypes(Type, frozen=False, eq=False):
   """Super class for shared behavior of UnionType and IntersectionType."""
+
+  # NOTE: This class is not frozen so that we can flatten types after
+  # initialization. It should still be treated as a frozen type.
   # NOTE: type_list is kept as a tuple, to preserve the original order
   #       even though in most respects it acts like a frozenset.
   #       It also flattens the input, such that printing without
   #       parentheses gives the same result.
-  type_list: Tuple[Type, ...] = attrs.field(converter=_FlattenTypes)
+  type_list: Tuple[TypeU, ...] = ()
 
-  @property
-  def name(self):
-    return None
-
-  def __hash__(self):
-    # See __eq__ - order doesn't matter, so use frozenset
-    return hash(frozenset(self.type_list))
+  def __post_init__(self):
+    self.type_list = _FlattenTypes(self.type_list)
 
   def __eq__(self, other):
     if self is other:
@@ -561,19 +564,19 @@ class _SetOfTypes(Node, Type):  # pytype: disable=signature-mismatch  # overridi
   def __ne__(self, other):
     return not self == other
 
+  def __hash__(self):
+    return hash(self.type_list)
+
 
 class UnionType(_SetOfTypes):
   """A union type that contains all types in self.type_list."""
-  __slots__ = ()
 
 
 class IntersectionType(_SetOfTypes):
   """An intersection type."""
-  __slots__ = ()
 
 
-@attrs.frozen(cache_hash=True)
-class GenericType(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
+class GenericType(Type):
   """Generic type. Takes a base type and type parameters.
 
   This is used for homogeneous tuples, lists, dictionaries, user classes, etc.
@@ -582,8 +585,9 @@ class GenericType(Node, Type):  # pytype: disable=signature-mismatch  # overridi
     base_type: The base type. Instance of Type.
     parameters: Type parameters. Tuple of instances of Type.
   """
+
   base_type: Union[NamedType, ClassType, LateType]
-  parameters: Tuple[Type, ...]
+  parameters: Tuple[TypeU, ...]
 
   @property
   def name(self):
@@ -592,7 +596,7 @@ class GenericType(Node, Type):  # pytype: disable=signature-mismatch  # overridi
   @property
   def element_type(self):
     """Type of the contained type, assuming we only have one type parameter."""
-    element_type, = self.parameters
+    (element_type,) = self.parameters
     return element_type
 
 
@@ -602,7 +606,6 @@ class TupleType(GenericType):
   A tuple with length len(self.parameters), whose item type is specified at
   each index.
   """
-  __slots__ = ()
 
 
 class CallableType(GenericType):
@@ -612,7 +615,6 @@ class CallableType(GenericType):
   the individual argument types, in the order of the arguments, and the last
   parameter is the return type.
   """
-  __slots__ = ()
 
   @property
   def args(self):
@@ -628,7 +630,6 @@ class CallableType(GenericType):
 
 class Concatenate(GenericType):
   """Concatenate params and ParamSpec."""
-  __slots__ = ()
 
   @property
   def args(self):
@@ -639,29 +640,61 @@ class Concatenate(GenericType):
     return self.parameters[-1]
 
 
-@attrs.frozen(cache_hash=True)
-class Literal(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
-  value: Union[int, str, Type, Constant]
-
-  @property
-  def name(self):
-    return None
+class Literal(Type):
+  value: Union[int, str, bool, TypeU, Constant]
 
 
-@attrs.frozen(cache_hash=True)
-class Annotated(Node, Type):  # pytype: disable=signature-mismatch  # overriding-return-type-checks
-  base_type: Type
+class Annotated(Type):
+  base_type: TypeU
   annotations: Tuple[str, ...]
 
-  @property
-  def name(self):
-    return None
-
 
 # Types that can be a base type of GenericType:
 GENERIC_BASE_TYPE = (NamedType, ClassType)
 
 
+# msgspec will not deserialize subclasses. That is, for a class like:
+#   class Example(msgspec.Struct):
+#     typ: Type
+# then **ONLY** pytd.Type is allowed in that field, not any of its subclasses.
+# Instead, a union of all possible types must be used instead.
+# For convenience, we define those unions below.
+# If you add a new subclass of one of the base classes, you MUST add it to the
+# appropriate Union.
+
+# All subclasses of TypeParameter.
+TypeParameterU = Union[TypeParameter, ParamSpec]
+
+# All subclasses of _SetOfType.
+SetOfTypesU = Union[UnionType, IntersectionType]
+
+# All subclasses of GenericType.
+GenericTypeU = Union[GenericType, TupleType, CallableType, Concatenate]
+
+# All subclasses of Type.
+TypeU = Union[
+    NamedType,
+    ClassType,
+    LateType,
+    AnythingType,
+    NothingType,
+    Literal,
+    Annotated,
+    # These are all Type subclasses too.
+    TypeParameterU,
+    SetOfTypesU,
+    GenericTypeU,
+    # TODO(tsudol): There's something weird going on here where msgspec cannot
+    # see that Parameter.type can be ParamSpecArg or ParamSpecKwargs when it's
+    # in the annotation, e.g.:
+    #   `type: Union[ParamSpecArgs, ParamSpecKwargs, TypeU]`
+    # So for now, add it to the TypeU union. This was uncovered by :main_test,
+    # so a dedicated test case would be good to write. Perhaps hypothesis/fuzz?
+    ParamSpecArgs,
+    ParamSpecKwargs,
+]
+
+
 def IsContainer(t: Class) -> bool:
   """Checks whether class t is a container."""
   if t.name in ('typing.Generic', 'typing.Protocol'):
@@ -671,8 +704,11 @@ def IsContainer(t: Class) -> bool:
       base = p.base_type
       # We consider a LateType a container, since we don't have enough
       # information to determine whether it is one or not.
-      if (isinstance(base, ClassType) and IsContainer(base.cls) or
-          isinstance(base, LateType)):
+      if (
+          isinstance(base, ClassType)
+          and IsContainer(base.cls)
+          or isinstance(base, LateType)
+      ):
         return True
   return False
 
@@ -682,8 +718,9 @@ def IsContainer(t: Class) -> bool:
 SINGLETON_TYPES = frozenset({'Ellipsis', 'builtins.Ellipsis'})
 
 
-def ToType(item, allow_constants=False, allow_functions=False,
-           allow_singletons=False):
+def ToType(
+    item, allow_constants=False, allow_functions=False, allow_singletons=False
+):
   """Convert a pytd AST item into a type.
 
   Takes an AST item representing the definition of a type and returns an item
@@ -723,8 +760,10 @@ def ToType(item, allow_constants=False, allow_functions=False,
         return item.type.parameters[0]
       else:
         return AnythingType()
-    elif (isinstance(item.type, AnythingType) or
-          item.name == 'typing_extensions.TypedDict'):
+    elif (
+        isinstance(item.type, AnythingType)
+        or item.name == 'typing_extensions.TypedDict'
+    ):
       # A constant with type Any may be a type, and TypedDict is a class that
       # looks like a constant:
       #   https://github.com/python/typeshed/blob/8cad322a8ccf4b104cafbac2c798413edaa4f327/third_party/2and3/typing_extensions.pyi#L68
@@ -761,21 +800,29 @@ def AliasMethod(func, from_constant):
   # first parameter ('self' or 'cls') is dropped.
   new_func = func.Replace(kind=MethodKind.METHOD)
   if func.kind == MethodKind.STATICMETHOD or (
-      func.kind == MethodKind.METHOD and not from_constant):
+      func.kind == MethodKind.METHOD and not from_constant
+  ):
     return new_func
-  return new_func.Replace(signatures=tuple(
-      s.Replace(params=s.params[1:]) for s in new_func.signatures))
+  return new_func.Replace(
+      signatures=tuple(
+          s.Replace(params=s.params[1:]) for s in new_func.signatures
+      )
+  )
 
 
-def LookupItemRecursive(module, name):
+def LookupItemRecursive(module: TypeDeclUnit, name: str) -> Node:
   """Recursively look up name in module."""
 
-  def ExtractClass(t):
+  def ExtractClass(t, prev_item):
     if isinstance(t, ClassType) and t.cls:
       return t.cls
-    t = module.Get(t.name)
-    if isinstance(t, Class):
-      return t
+    top_level_t = module.Get(t.name)
+    if isinstance(top_level_t, Class):
+      return top_level_t
+    elif isinstance(prev_item, Class):
+      nested_t = prev_item.Get(f'{prev_item.name}.{t.name}')
+      if isinstance(nested_t, Class):
+        return nested_t
     return None
 
   def Lookup(item, *names):
@@ -791,11 +838,11 @@ def LookupItemRecursive(module, name):
   item = module
 
   for part in parts:
-    prev_item = item
+    next_prev_item = item
     # Check the type of item and give up if we encounter a type we don't know
     # how to handle.
     if isinstance(item, Constant):
-      found = ExtractClass(item.type)
+      found = ExtractClass(item.type, prev_item)
       if not found:
         raise KeyError(item.type.name)
       item = found
@@ -807,12 +854,24 @@ def LookupItemRecursive(module, name):
     # we try lookup for both naming conventions.
     found = Lookup(item, lookup_name, part)
     if found:
+      seen = {found}
+      while (
+          isinstance(found, Alias)
+          and isinstance(found.type, NamedType)
+          and found.type.name.startswith(f'{item.name}.')
+      ):
+        resolved = Lookup(item, found.type.name)
+        if resolved and resolved not in seen:
+          found = resolved
+          seen.add(resolved)
+        else:
+          break
       item = found
     else:
       if not isinstance(item, Class):
         raise KeyError(item)
       for base in item.bases:
-        base_cls = ExtractClass(base)
+        base_cls = ExtractClass(base, prev_item)
         if base_cls is None:
           raise KeyError(item)
         found = Lookup(base_cls, lookup_name, part)
@@ -821,10 +880,11 @@ def LookupItemRecursive(module, name):
           break  # name found!
       else:
         raise KeyError(item)  # unresolved
-    if isinstance(item, Constant):
+    if isinstance(item, (Constant, Class)):
       partial_name += '.' + item.name.rsplit('.', 1)[-1]
     else:
       partial_name = lookup_name
+    prev_item = next_prev_item
   if isinstance(item, Function):
     return AliasMethod(item, from_constant=isinstance(prev_item, Constant))
   else:
