@@ -13,8 +13,6 @@ signatures against new inference results.
 import logging
 from typing import Dict, Optional, Union
 
-import attrs
-
 from pytype import utils
 from pytype.pytd import booleq
 from pytype.pytd import escape
@@ -57,23 +55,29 @@ def get_all_subclasses(asts):
   hierarchy = {}
   for ast in asts:
     hierarchy.update(ast.Visit(visitors.ExtractSuperClasses()))
+
   def filter_superclasses(superclasses):
-    return [superclass for superclass in superclasses
-            if is_complete(superclass)]
-  hierarchy = {cls: filter_superclasses(superclasses)
-               for cls, superclasses in hierarchy.items() if is_complete(cls)}
+    return [
+        superclass for superclass in superclasses if is_complete(superclass)
+    ]
+
+  hierarchy = {
+      cls: filter_superclasses(superclasses)
+      for cls, superclasses in hierarchy.items()
+      if is_complete(cls)
+  }
   # typically this is a fairly short list, e.g.:
   #  [ClassType(basestring), ClassType(int), ClassType(object)]
   return utils.invert_dict(hierarchy)
 
 
-@attrs.frozen(slots=True, cache_hash=True)
 class StrictType(node.Node):
   """A type that doesn't allow sub- or superclasses to match.
 
   For example, "int" is considered a valid argument for a function that accepts
   "object", but StrictType("int") is not.
   """
+
   name: str
 
   def __str__(self):
@@ -89,9 +93,8 @@ class TypeMatch(pytd_utils.TypeMatcher):
     Args:
       direct_subclasses: A dictionary, mapping pytd.Type to lists of pytd.Type.
       any_also_is_bottom: Whether we should, (if True) consider
-        pytd.AnythingType() to also be at the bottom of the type hierarchy,
-        thus making it a subclass of everything, or (if False) to be only
-        at the top.
+        pytd.AnythingType() to also be at the bottom of the type hierarchy, thus
+        making it a subclass of everything, or (if False) to be only at the top.
     """
     self.direct_subclasses = direct_subclasses or {}
     self.any_also_is_bottom = any_also_is_bottom
@@ -101,13 +104,15 @@ class TypeMatch(pytd_utils.TypeMatcher):
   def default_match(self, t1, t2, *unused_args, **unused_kwargs):
     # Don't allow pytd_utils.TypeMatcher to do default matching.
     raise AssertionError(
-        f"Can't compare {type(t1).__name__} and {type(t2).__name__}")
+        f"Can't compare {type(t1).__name__} and {type(t2).__name__}"
+    )
 
   def get_superclasses(self, t):
     """Get all base classes of this type.
 
     Args:
         t: A pytd.Type
+
     Returns:
         A list of pytd.Type.
     """
@@ -127,19 +132,25 @@ class TypeMatch(pytd_utils.TypeMatcher):
 
     Args:
         t: A pytd.Type
+
     Returns:
         A list of pytd.Type.
     """
     if isinstance(t, pytd.ClassType):
       subclasses = self.direct_subclasses.get(t, [])
-      return sum((self.get_subclasses(pytd.ClassType(c.name, c))
-                  for c in subclasses), [t])
+      return sum(
+          (self.get_subclasses(pytd.ClassType(c.name, c)) for c in subclasses),
+          [t],
+      )
     else:
       raise NotImplementedError(f"Can't extract subclasses from {type(t)}")
 
   def type_parameter(
-      self, unknown: _UnknownType, base_class: pytd.Class,
-      item: pytd.TemplateItem) -> StrictType:
+      self,
+      unknown: _UnknownType,
+      base_class: pytd.Class,
+      item: pytd.TemplateItem,
+  ) -> StrictType:
     """This generates the type parameter when matching against a generic type.
 
     For example, when we match ~unknown1 against list[T], we need an additional
@@ -152,6 +163,7 @@ class TypeMatch(pytd_utils.TypeMatcher):
       base_class: The base class of the generic we're matching the unknown
         against. E.g. "list".
       item: The actual type parameter. ("T" in the examples above).
+
     Returns:
       A type (pytd.Node) to represent this type parameter.
     """
@@ -176,17 +188,22 @@ class TypeMatch(pytd_utils.TypeMatcher):
       return (t1.element_type,) * len(t2.parameters), t2.parameters
     elif isinstance(t1, pytd.TupleType):
       return (pytd_utils.JoinTypes(t1.parameters),), t2.parameters
-    elif (isinstance(t1, pytd.CallableType) and
-          isinstance(t2, pytd.CallableType)):
+    elif isinstance(t1, pytd.CallableType) and isinstance(
+        t2, pytd.CallableType
+    ):
       # Flip the arguments, since argument types are contravariant.
       return t2.args + (t1.ret,), t1.args + (t2.ret,)
-    elif (t1.base_type.cls.name == "builtins.type" and
-          t2.base_type.cls.name == "typing.Callable"):
+    elif (
+        t1.base_type.cls.name == "builtins.type"
+        and t2.base_type.cls.name == "typing.Callable"
+    ):
       # We'll only check the return type, since getting the argument types for
       # initializing a class is tricky.
       return t1.parameters, (t2.parameters[-1],)
-    elif (t1.base_type.cls.name == "typing.Callable" and
-          t2.base_type.cls.name == "builtins.type"):
+    elif (
+        t1.base_type.cls.name == "typing.Callable"
+        and t2.base_type.cls.name == "builtins.type"
+    ):
       return (t1.parameters[-1],), t2.parameters
     elif isinstance(t1, pytd.CallableType):
       # We're matching against GenericType(Callable, (Any, _RET)), so we don't
@@ -197,13 +214,18 @@ class TypeMatch(pytd_utils.TypeMatcher):
     else:
       num_extra_params = len(t1.parameters) - len(t2.parameters)
       # Matching, e.g., Dict[str, int] against Iterable[K] is legitimate.
-      assert num_extra_params >= 0, (t1.base_type.cls.name,
-                                     t2.base_type.cls.name)
+      assert num_extra_params >= 0, (
+          t1.base_type.cls.name,
+          t2.base_type.cls.name,
+      )
       t2_parameters = t2.parameters + (pytd.AnythingType(),) * num_extra_params
       return t1.parameters, t2_parameters
 
   def match_Generic_against_Generic(  # pylint: disable=invalid-name
-      self, t1: pytd.GenericType, t2: pytd.GenericType, subst: _SubstType,
+      self,
+      t1: pytd.GenericType,
+      t2: pytd.GenericType,
+      subst: _SubstType,
   ) -> booleq.BooleanTerm:
     """Match a pytd.GenericType against another pytd.GenericType."""
     assert isinstance(t1.base_type, pytd.ClassType), type(t1.base_type)
@@ -218,8 +240,10 @@ class TypeMatch(pytd_utils.TypeMatcher):
       return booleq.FALSE
     # Type parameters are covariant:
     # E.g. passing list[int] as argument for list[object] succeeds.
-    param_cmp = [self.match_type_against_type(p1, p2, subst)
-                 for p1, p2 in zip(t1_parameters, t2_parameters)]
+    param_cmp = [
+        self.match_type_against_type(p1, p2, subst)
+        for p1, p2 in zip(t1_parameters, t2_parameters)
+    ]
     return booleq.And([base_type_cmp] + param_cmp)
 
   def match_Unknown_against_Generic(  # pylint: disable=invalid-name
@@ -229,16 +253,20 @@ class TypeMatch(pytd_utils.TypeMatcher):
     # No inheritance for base classes - you can only inherit from an
     # instantiated template, but not from a template itself.
     base_match = booleq.Eq(t1.name, t2.base_type.cls.name)
-    type_params = [self.type_parameter(t1, t2.base_type.cls, item)
-                   for item in t2.base_type.cls.template]
+    type_params = [
+        self.type_parameter(t1, t2.base_type.cls, item)
+        for item in t2.base_type.cls.template
+    ]
     for type_param in type_params:
       self.solver.register_variable(type_param.name)
     if isinstance(t2, pytd.TupleType):
       t2_parameters = (pytd_utils.JoinTypes(t2.parameters),)
     else:
       t2_parameters = t2.parameters
-    params = [self.match_type_against_type(p1, p2, subst)
-              for p1, p2 in zip(type_params, t2_parameters)]
+    params = [
+        self.match_type_against_type(p1, p2, subst)
+        for p1, p2 in zip(type_params, t2_parameters)
+    ]
     return booleq.And([base_match] + params)
 
   def match_Generic_against_Unknown(self, t1, t2, subst):  # pylint: disable=invalid-name
@@ -280,7 +308,8 @@ class TypeMatch(pytd_utils.TypeMatcher):
     if types in self._implications:
       return self._implications[types]
     implication = self._implications[types] = self._match_type_against_type(
-        t1, t2, subst)
+        t1, t2, subst
+    )
     return implication
 
   def _full_name(self, t):
@@ -309,30 +338,42 @@ class TypeMatch(pytd_utils.TypeMatcher):
       # nothing itself, above).
       return booleq.FALSE
     elif isinstance(t1, pytd.UnionType):
-      return booleq.And(self.match_type_against_type(u, t2, subst)
-                        for u in t1.type_list)
+      return booleq.And(
+          self.match_type_against_type(u, t2, subst) for u in t1.type_list
+      )
     elif isinstance(t2, pytd.UnionType):
-      return booleq.Or(self.match_type_against_type(t1, u, subst)
-                       for u in t2.type_list)
-    elif (isinstance(t1, pytd.ClassType) and isinstance(t2, StrictType) or
-          isinstance(t1, StrictType) and isinstance(t2, pytd.ClassType)):
+      return booleq.Or(
+          self.match_type_against_type(t1, u, subst) for u in t2.type_list
+      )
+    elif (
+        isinstance(t1, pytd.ClassType)
+        and isinstance(t2, StrictType)
+        or isinstance(t1, StrictType)
+        and isinstance(t2, pytd.ClassType)
+    ):
       # For strict types, avoid subclasses of the left side.
       return booleq.Eq(self._full_name(t1), self._full_name(t2))
     elif isinstance(t1, pytd.ClassType) and t2.name == "builtins.object":
       return booleq.TRUE
-    elif (t1.name in ("builtins.type", "typing.Callable") and
-          t2.name in ("builtins.type", "typing.Callable")):
+    elif t1.name in ("builtins.type", "typing.Callable") and t2.name in (
+        "builtins.type",
+        "typing.Callable",
+    ):
       return booleq.TRUE
     elif isinstance(t1, pytd.ClassType):
       # ClassTypes are similar to Unions, except they're disjunctions: We can
       # match the type or any of its base classes against the formal parameter.
-      return booleq.Or(self.match_type_against_type(t, t2, subst)
-                       for t in self.expand_superclasses(t1))
+      return booleq.Or(
+          self.match_type_against_type(t, t2, subst)
+          for t in self.expand_superclasses(t1)
+      )
     elif isinstance(t2, pytd.ClassType):
       # ClassTypes on the right are exactly like Unions: We can match against
       # this type or any of its subclasses.
-      return booleq.Or(self.match_type_against_type(t1, t, subst)
-                       for t in self.expand_subclasses(t2))
+      return booleq.Or(
+          self.match_type_against_type(t1, t, subst)
+          for t in self.expand_subclasses(t2)
+      )
     assert not isinstance(t1, pytd.ClassType)
     assert not isinstance(t2, pytd.ClassType)
     if is_unknown(t1) and isinstance(t2, pytd.GenericType):
@@ -352,8 +393,9 @@ class TypeMatch(pytd_utils.TypeMatcher):
         return booleq.FALSE
     elif is_unknown(t1) and is_unknown(t2):
       return booleq.Eq(t1.name, t2.name)
-    elif (isinstance(t1, (pytd.NamedType, StrictType)) and
-          isinstance(t2, (pytd.NamedType, StrictType))):
+    elif isinstance(t1, (pytd.NamedType, StrictType)) and isinstance(
+        t2, (pytd.NamedType, StrictType)
+    ):
       if is_complete(t1) and is_complete(t2) and t1.name != t2.name:
         # Optimization: If we know these two can never be equal, just return
         # false right away.
@@ -368,12 +410,14 @@ class TypeMatch(pytd_utils.TypeMatcher):
     elif isinstance(t1, pytd.Literal) and isinstance(t2, pytd.Literal):
       return booleq.TRUE if t1.value == t2.value else booleq.FALSE
     else:
-      raise AssertionError(f"Don't know how to match {type(t1)} against "
-                           f"{type(t2)}")
+      raise AssertionError(
+          f"Don't know how to match {type(t1)} against {type(t2)}"
+      )
 
   # pylint: disable=invalid-name
-  def match_Signature_against_Signature(self, sig1, sig2, subst,
-                                        skip_self=False):
+  def match_Signature_against_Signature(
+      self, sig1, sig2, subst, skip_self=False
+  ):
     """Match a pytd.Signature against another pytd.Signature.
 
     Args:
@@ -382,6 +426,7 @@ class TypeMatch(pytd_utils.TypeMatcher):
       subst: Current type parameters.
       skip_self: If True, doesn't compare the first parameter, which is
         considered (and verified) to be "self".
+
     Returns:
       An instance of booleq.BooleanTerm, i.e. a boolean formula.
     """
@@ -400,11 +445,17 @@ class TypeMatch(pytd_utils.TypeMatcher):
     if len(params1) > len(params2) and not sig2.has_optional:
       return booleq.FALSE  # extra parameters
     if sig1.starargs is not None and sig2.starargs is not None:
-      equalities.append(self.match_type_against_type(
-          sig1.starargs.type, sig2.starargs.type, subst))
+      equalities.append(
+          self.match_type_against_type(
+              sig1.starargs.type, sig2.starargs.type, subst
+          )
+      )
     if sig1.starstarargs is not None and sig2.starstarargs is not None:
-      equalities.append(self.match_type_against_type(
-          sig1.starstarargs.type, sig2.starstarargs.type, subst))
+      equalities.append(
+          self.match_type_against_type(
+              sig1.starstarargs.type, sig2.starstarargs.type, subst
+          )
+      )
     # TODO(b/159058933): Handle kwonly parameters (on either side). Presumably,
     # a kwonly on the left side means that it was a keyword param.
     for p1, p2 in zip(params1, params2):
@@ -419,26 +470,30 @@ class TypeMatch(pytd_utils.TypeMatcher):
       else:
         p1 = params1[i]
         if p1.name != p2.name and not (
-            pytd_utils.ANON_PARAM.match(p1.name) or
-            pytd_utils.ANON_PARAM.match(p2.name)):
+            pytd_utils.ANON_PARAM.match(p1.name)
+            or pytd_utils.ANON_PARAM.match(p2.name)
+        ):
           return booleq.FALSE
         equalities.append(self.match_type_against_type(p1.type, p2.type, subst))
     equalities.append(
-        self.match_type_against_type(
-            sig1.return_type, sig2.return_type, subst))
+        self.match_type_against_type(sig1.return_type, sig2.return_type, subst)
+    )
     return booleq.And(equalities)
 
   def match_Signature_against_Function(self, sig, f, subst, skip_self=False):  # pylint: disable=invalid-name
     def make_or(inner):
       return booleq.Or(
           self.match_Signature_against_Signature(inner, s, subst, skip_self)
-          for s in f.signatures)
+          for s in f.signatures
+      )
+
     return booleq.And(make_or(inner) for inner in visitors.ExpandSignature(sig))
 
   def match_Function_against_Function(self, f1, f2, subst, skip_self=False):  # pylint: disable=invalid-name
     return booleq.And(
         self.match_Signature_against_Function(s1, f2, subst, skip_self)
-        for s1 in f1.signatures)
+        for s1 in f1.signatures
+    )
 
   def match_Function_against_Class(self, f1, cls2, subst, cache):
     cls2_methods = cache.get(id(cls2))
@@ -474,25 +529,23 @@ class TypeMatch(pytd_utils.TypeMatcher):
         else:
           # Funky types like UnionType are hard to match against (and shouldn't
           # appear as a base class) so we treat them as catch-all.
-          log.warning("Assuming that %s has method %s",
-                      pytd_utils.Print(base), f1.name)
+          log.warning(
+              "Assuming that %s has method %s", pytd_utils.Print(base), f1.name
+          )
           return booleq.TRUE
       return booleq.FALSE
     else:
       f2 = cls2_methods[f1.name]
-      return self.match_Function_against_Function(
-          f1, f2, subst, skip_self=True)
+      return self.match_Function_against_Function(f1, f2, subst, skip_self=True)
 
   def match_Class_against_Class(self, cls1, cls2, subst):  # pylint: disable=invalid-name
     """Match a pytd.Class against another pytd.Class."""
-    return self.match_Functions_against_Class(
-        cls1.methods, cls2, subst)
+    return self.match_Functions_against_Class(cls1.methods, cls2, subst)
 
   def match_Protocol_against_Unknown(self, protocol, unknown, subst):  # pylint: disable=invalid-name
     """Match a typing.Protocol against an unknown class."""
     filtered_methods = [f for f in protocol.methods if f.is_abstract]
-    return self.match_Functions_against_Class(
-        filtered_methods, unknown, subst)
+    return self.match_Functions_against_Class(filtered_methods, unknown, subst)
 
   def match_Functions_against_Class(self, methods, cls2, subst):
     implications = []
