@@ -3,13 +3,13 @@
 import collections
 import itertools
 import logging
-from typing import Any, Collection, Dict, Optional, Tuple, Union
+from typing import Any, Collection, Dict, List, Optional, Tuple, Union
 
 from pytype import compare
 from pytype import metrics
+from pytype import module_utils
 from pytype import utils
 from pytype.abstract import abstract
-from pytype.abstract import mixin
 from pytype.blocks import blocks
 from pytype.typegraph import cfg
 
@@ -19,6 +19,9 @@ log = logging.getLogger(__name__)
 # condition cannot be satisfied with any known bindings.
 UNSATISFIABLE = object()
 
+# Represents the `not None` condition for restrict_condition().
+NOT_NONE = object()
+
 FrameType = Union["SimpleFrame", "Frame"]
 # This should be context.Context, which can't be imported due to a circular dep.
 _ContextType = Any
@@ -45,12 +48,24 @@ class FrameState(utils.ContextWeakrefMixin):
     raise AttributeError("States are immutable.")
 
   def set_why(self, why):
-    return FrameState(self.data_stack, self.block_stack, self.node, self.ctx,
-                      self.exception, why)
+    return FrameState(
+        self.data_stack,
+        self.block_stack,
+        self.node,
+        self.ctx,
+        self.exception,
+        why,
+    )
 
   def set_stack(self, new_stack):
-    return FrameState(new_stack, self.block_stack, self.node, self.ctx,
-                      self.exception, self.why)
+    return FrameState(
+        new_stack,
+        self.block_stack,
+        self.node,
+        self.ctx,
+        self.exception,
+        self.why,
+    )
 
   def push(self, *values):
     """Push value(s) onto the value stack."""
@@ -86,8 +101,10 @@ class FrameState(utils.ContextWeakrefMixin):
       # Not an error: E.g. function calls with no parameters pop zero items
       return self, ()
     if len(self.data_stack) < n:
-      raise IndexError("Trying to pop %d values from stack of size %d" %
-                       (n, len(self.data_stack)))
+      raise IndexError(
+          "Trying to pop %d values from stack of size %d"
+          % (n, len(self.data_stack))
+      )
     values = self.data_stack[-n:]
     return self.set_stack(self.data_stack[:-n]), values
 
@@ -102,28 +119,64 @@ class FrameState(utils.ContextWeakrefMixin):
   def rotn(self, n):
     """Rotate the top n values by one."""
     if len(self.data_stack) < n:
-      raise IndexError("Trying to rotate %d values from stack of size %d" %
-                       (n, len(self.data_stack)))
+      raise IndexError(
+          "Trying to rotate %d values from stack of size %d"
+          % (n, len(self.data_stack))
+      )
     top = self.data_stack[-1]
     rot = self.data_stack[-n:-1]
     return self.set_stack(self.data_stack[:-n] + (top,) + rot)
 
+  def swap(self, n):
+    """Swap the top of the data stack with the value in position n."""
+    if len(self.data_stack) < n:
+      raise IndexError(
+          "Trying to swap value %d in stack of size %d"
+          % (n, len(self.data_stack))
+      )
+    top = self.data_stack[-1]
+    nth = self.data_stack[-n]
+    in_between = self.data_stack[(-n + 1) : -1]
+    rest = self.data_stack[:-n]
+    return self.set_stack(rest + (top,) + in_between + (nth,))
+
   def push_block(self, block):
     """Push a block on to the block stack."""
-    return FrameState(self.data_stack, self.block_stack + (block,), self.node,
-                      self.ctx, self.exception, self.why)
+    return FrameState(
+        self.data_stack,
+        self.block_stack + (block,),
+        self.node,
+        self.ctx,
+        self.exception,
+        self.why,
+    )
 
   def pop_block(self):
     """Pop a block from the block stack."""
     block = self.block_stack[-1]
-    return FrameState(self.data_stack, self.block_stack[:-1], self.node,
-                      self.ctx, self.exception, self.why), block
+    return (
+        FrameState(
+            self.data_stack,
+            self.block_stack[:-1],
+            self.node,
+            self.ctx,
+            self.exception,
+            self.why,
+        ),
+        block,
+    )
 
   def change_cfg_node(self, node: cfg.CFGNode) -> "FrameState":
     if self.node is node:
       return self
-    return FrameState(self.data_stack, self.block_stack, node, self.ctx,
-                      self.exception, self.why)
+    return FrameState(
+        self.data_stack,
+        self.block_stack,
+        node,
+        self.ctx,
+        self.exception,
+        self.why,
+    )
 
   def connect_to_cfg_node(self, node):
     self.node.ConnectTo(node)
@@ -148,23 +201,39 @@ class FrameState(utils.ContextWeakrefMixin):
     """Merge with another state."""
     if other is None:
       return self
-    assert len(self.data_stack) == len(other.data_stack)
-    assert len(self.block_stack) == len(other.block_stack)
+    assert len(self.data_stack) == len(other.data_stack), (
+        self.data_stack,
+        other.data_stack,
+    )
+    assert len(self.block_stack) == len(other.block_stack), (
+        self.block_stack,
+        other.block_stack,
+    )
     both = list(zip(self.data_stack, other.data_stack))
     if any(v1 is not v2 for v1, v2 in both):
       for v, o in both:
         o.PasteVariable(v, None)
     if self.node is not other.node:
       self.node.ConnectTo(other.node)
-      return FrameState(other.data_stack, self.block_stack, other.node,
-                        self.ctx, self.exception, self.why)
+      return FrameState(
+          other.data_stack,
+          self.block_stack,
+          other.node,
+          self.ctx,
+          self.exception,
+          self.why,
+      )
     return self
 
   def set_exception(self):
     return FrameState(
-        self.data_stack, self.block_stack,
-        self.ctx.connect_new_cfg_node(self.node, "SetException"), self.ctx,
-        True, self.why)
+        self.data_stack,
+        self.block_stack,
+        self.ctx.connect_new_cfg_node(self.node, "SetException"),
+        self.ctx,
+        True,
+        self.why,
+    )
 
 
 class SimpleFrame:
@@ -217,20 +286,27 @@ class Frame(utils.ContextWeakrefMixin):
   """
 
   def __init__(
-      self, node: cfg.CFGNode, ctx: _ContextType, f_code: blocks.OrderedCode,
-      f_globals: abstract.LazyConcreteDict, f_locals: abstract.LazyConcreteDict,
-      f_back: FrameType, callargs: Dict[str, cfg.Variable],
-      closure: Optional[Tuple[cfg.Variable, ...]], func: Optional[cfg.Binding],
+      self,
+      node: cfg.CFGNode,
+      ctx: _ContextType,
+      f_code: blocks.OrderedCode,
+      f_globals: abstract.LazyConcreteDict,
+      f_locals: abstract.LazyConcreteDict,
+      f_back: FrameType,
+      callargs: Dict[str, cfg.Variable],
+      closure: Optional[Tuple[cfg.Variable, ...]],
+      func: Optional[cfg.Binding],
       first_arg: Optional[cfg.Variable],
-      substs: Collection[Dict[str, cfg.Variable]]):
+      substs: Collection[Dict[str, cfg.Variable]],
+  ):
     """Initialize a special frame as needed by TypegraphVirtualMachine.
 
     Args:
       node: The current CFG graph node.
       ctx: The owning abstract context.
       f_code: The code object to execute in this frame.
-      f_globals: The global context to execute in as a SimpleValue as
-        used by TypegraphVirtualMachine.
+      f_globals: The global context to execute in as a SimpleValue as used by
+        TypegraphVirtualMachine.
       f_locals: Local variables. Will be modified if callargs is passed.
       f_back: The frame above this one on the stack.
       callargs: Additional function arguments to store in f_locals.
@@ -239,6 +315,7 @@ class Frame(utils.ContextWeakrefMixin):
       first_arg: First argument to the function.
       substs: Maps from type parameter names in scope for this frame to their
         possible values.
+
     Raises:
       NameError: If we can't resolve any references into the outer frame.
     """
@@ -254,10 +331,11 @@ class Frame(utils.ContextWeakrefMixin):
       self.f_builtins = f_back.f_builtins
     else:
       _, bltin = self.ctx.attribute_handler.get_attribute(
-          self.ctx.root_node, f_globals, "__builtins__")
-      builtins_pu, = bltin.bindings
+          self.ctx.root_node, f_globals, "__builtins__"
+      )
+      (builtins_pu,) = bltin.bindings
       self.f_builtins = builtins_pu.data
-    self.f_lineno = f_code.co_firstlineno
+    self.f_lineno = f_code.firstlineno
     # The first argument is used to make Python 3 super calls when super is not
     # passed any arguments.
     self.first_arg = first_arg
@@ -285,14 +363,21 @@ class Frame(utils.ContextWeakrefMixin):
     # if they are also parameters of f (because co_varnames[0:co_argcount] are
     # always the parameters), but won't otherwise.
     # Cells 0 .. num(cellvars)-1 : cellvar; num(cellvars) .. end : freevar
-    assert len(f_code.co_freevars) == len(closure or [])
-    self.cells = [self.ctx.program.NewVariable() for _ in f_code.co_cellvars]
-    self.cells.extend(closure or [])
-
+    self.closure = closure
+    freevars = closure or []
+    assert len(f_code.freevars) == len(freevars)
+    if self.ctx.python_version < (3, 11):
+      cell_names = f_code.cellvars
+    elif freevars:
+      cell_names = f_code.localsplus[: -len(freevars)]
+    else:
+      cell_names = f_code.localsplus
+    self.cells = [self.ctx.program.NewVariable() for _ in cell_names]
+    self.cells.extend(freevars)
     if callargs:
       for name, value in sorted(callargs.items()):
-        if name in f_code.co_cellvars:
-          i = f_code.co_cellvars.index(name)
+        if name in f_code.cellvars:
+          i = cell_names.index(name)
           self.cells[i].PasteVariable(value, node)
         else:
           self.ctx.attribute_handler.set_attribute(node, f_locals, name, value)
@@ -308,29 +393,49 @@ class Frame(utils.ContextWeakrefMixin):
     self.class_closure_var = None
     if func and isinstance(func.data, abstract.InterpreterFunction):
       closure_name = abstract.BuildClass.CLOSURE_NAME
-      if func.data.is_class_builder and closure_name in f_code.co_cellvars:
-        i = f_code.co_cellvars.index(closure_name)
-        self.class_closure_var = self.cells[i]
+      if func.data.is_class_builder and closure_name in f_code.cellvars:
+        self.class_closure_var = self.get_cell_by_name(closure_name)
     self.func = func
     self.substs = substs
     # Do not add to error tracebacks
     self.skip_in_tracebacks = False
 
-  def __repr__(self):     # pragma: no cover
+    # Set the module name (used in logging)
+    if f_code.filename:
+      self.module_name = module_utils.path_to_module_name(f_code.filename)
+    else:
+      self.module_name = ""
+
+    # All InterpreterFunction objects created while this frame was at the top of
+    # the frame stack.
+    self.functions_created_in_frame: Dict[
+        str, List[abstract.InterpreterFunction]
+    ] = collections.defaultdict(list)
+
+  def __repr__(self):  # pragma: no cover
     return "<Frame at 0x%08x: %r @ %d>" % (
-        id(self), self.f_code.co_filename, self.f_lineno
+        id(self),
+        self.f_code.filename,
+        self.f_lineno,
     )
 
+  def copy_free_vars(self, n):
+    offset = len(self.cells) - len(self.f_code.freevars)
+    for i in range(n):
+      self.cells[i + offset] = self.closure[i]
+
   @property
   def type_params(self):
     return set(itertools.chain.from_iterable(self.substs))
 
   def lookup_name(self, target_name):
     for store in (self.f_locals, self.f_globals, self.f_builtins):
-      if target_name in store.members:
+      if store is not None and target_name in store.members:
         return store.members[target_name]
-    i = (self.f_code.co_cellvars + self.f_code.co_freevars).index(target_name)
-    return self.cells[i]
+    return self.get_cell_by_name(target_name)
+
+  def get_cell_by_name(self, name):
+    return self.cells[self.f_code.get_cell_index(name)]
 
 
 class Condition:
@@ -358,41 +463,38 @@ class Condition:
 _restrict_counter = metrics.MapCounter("state_restrict")
 
 
-def split_conditions(node, var):
-  """Return a pair of conditions for the value being true and false."""
-  return (_restrict_condition(node, var.bindings, True),
-          _restrict_condition(node, var.bindings, False))
+def _match_condition(value, condition):
+  if isinstance(condition, bool):
+    return compare.compatible_with(value, condition)
+  elif condition is None:
+    return compare.compatible_with_none(value)
+  else:
+    assert condition is NOT_NONE
+    return value.full_name != "builtins.NoneType"
 
 
-def _restrict_condition(node, bindings, logical_value):
+def restrict_condition(node, var, condition):
   """Return a restricted condition based on filtered bindings.
 
   Args:
     node: The CFGNode.
-    bindings: A sequence of bindings.
-    logical_value: Either True or False.
+    var: A variable.
+    condition: A value that we will check each binding for compatibility with.
 
   Returns:
-    A Condition or None.  Each binding is checked for compatibility with
-    logical_value.  If either no bindings match, or all bindings match, then
-    None is returned.  Otherwise a new Condition is built from the specified,
-    compatible, bindings.
+    A Condition or None. Each binding of the variable is checked for
+    compatibility with the condition. If either no bindings match, or all
+    bindings match, then None is returned.  Otherwise a new Condition is built
+    from the specified, compatible, bindings.
   """
   dnf = []
   restricted = False
-  for b in bindings:
-    match = compare.compatible_with(b.data, logical_value)
-    if match is True:  # pylint: disable=g-bool-id-comparison
-      dnf.append([b])
-    elif match is False:  # pylint: disable=g-bool-id-comparison
-      restricted = True
+  for b in var.bindings:
+    match_result = _match_condition(b.data, condition)
+    if match_result:
+      dnf.append([b])  # the binding may match the condition
     else:
-      dnf.extend(match)
-      # In theory, the value could have returned [[b]] as its DNF, in which
-      # case this isn't really a restriction.  However in practice this is
-      # very unlikely to occur, and treating it as a restriction will not
-      # cause any problems.
-      restricted = True
+      restricted = True  # the binding cannot match the condition
   if not dnf:
     _restrict_counter.inc("unsatisfiable")
     return UNSATISFIABLE
@@ -406,20 +508,21 @@ def _restrict_condition(node, bindings, logical_value):
 
 def _is_or_is_not_cmp(left, right, is_not=False):
   """Implementation of 'left is right' amd 'left is not right'."""
-  if (isinstance(left, mixin.PythonConstant) and
-      isinstance(right, mixin.PythonConstant)):
+  if isinstance(left, abstract.PythonConstant) and isinstance(
+      right, abstract.PythonConstant
+  ):
     if left.cls != right.cls:
       return is_not
     return is_not ^ (left.pyval == right.pyval)
-  elif (isinstance(left, abstract.Instance) and
-        isinstance(right, abstract.Instance)):
+  elif isinstance(left, abstract.Instance) and isinstance(
+      right, abstract.Instance
+  ):
     if left.cls != right.cls:
       # If those were the same they could be the same but we can't be sure from
       # comparing types.
       return is_not
     return None
-  elif (isinstance(left, abstract.Class) and
-        isinstance(right, abstract.Class)):
+  elif isinstance(left, abstract.Class) and isinstance(right, abstract.Class):
     # types are singletons. We use the name so that, e.g., two different
     # TupleClass instances compare as identical.
     return is_not ^ (left.full_name == right.full_name)
