@@ -4,6 +4,7 @@ import collections
 import importlib
 import itertools
 import logging
+import re
 import subprocess
 import sys
 from typing import Iterable, Sequence, Tuple
@@ -52,7 +53,7 @@ def _get_executable(binary, module=None):
     return [sys.executable, '-m', module or binary]
   else:
     return [binary]
-PYTYPE_SINGLE = _get_executable('pytype-single', 'pytype.single')
+PYTYPE_SINGLE = _get_executable('pytype-single', 'pytype.main')
 
 
 def resolved_file_to_module(f):
@@ -146,19 +147,18 @@ def _module_to_output_path(mod):
 
 
 def escape_ninja_path(path: str):
-  """escape `:` in absolute path on windows."""
-  if sys.platform == 'win32':
-    new_path = ''
-    last_char = None
-    for ch in path:
-      if last_char != '$' and ch == ':':
-        new_path += '$:'
-      else:
-        new_path += ch
-      last_char = ch
-    return new_path
-  else:
-    return path
+  """Returns the path with special characters escaped.
+
+  Escape new line, space, colon, and dollar sign, for ninja
+  (as described in https://ninja-build.org/manual.html#ref_lexer).
+  This function should be called on an unescaped path string to turn
+  it into a ninja path string. (If you call it on a ninja path string, it will
+  render the ninja variables inert.)
+
+  Args:
+    path: The path.
+  """
+  return re.sub(r'(?P<char>[\n :$])', r'$\g<char>', path)
 
 
 def get_imports_map(deps, module_to_imports_map, module_to_output):
@@ -194,6 +194,7 @@ class PytypeRunner:
       if not report_errors and dest in config.REPORT_ERRORS_ITEMS:
         continue
       arg_info = config.get_pytype_single_item(dest).arg_info
+      assert arg_info is not None
       if arg_info.to_command_line:
         value = arg_info.to_command_line(value)
       if isinstance(value, bool):
@@ -287,8 +288,7 @@ class PytypeRunner:
         if action:
           modules.append((module, action))
       if len(modules) == 1:
-        # TODO(b/73562531): Remove the pytype disable once the bug is fixed.
-        yield modules[0] + (deps, Stage.SINGLE_PASS)  # pytype: disable=bad-return-type
+        yield modules[0] + (deps, Stage.SINGLE_PASS)
       else:
         # If we have a cycle we run pytype over the files twice. So that we
         # don't fail on missing dependencies, we'll ignore errors the first
@@ -338,7 +338,7 @@ class PytypeRunner:
     logging.info('%s %s\n  imports: %s\n  deps: %s\n  output: %s',
                  action, module.name, imports, deps, output)
     if deps:
-      deps = ' | ' +  escape_ninja_path(' '.join(deps))
+      deps = ' | ' + ' '.join(escape_ninja_path(dep) for dep in deps)
     else:
       deps = ''
     with open(self.ninja_file, 'a') as f:
