@@ -3,7 +3,7 @@
 import collections
 import dataclasses
 import re
-import sys
+from typing import Sequence
 
 from pytype import config as pytype_config
 from pytype import file_utils
@@ -37,7 +37,7 @@ class ExpectedBuildStatement:
   output: str
   action: str
   input: str
-  deps: str
+  deps: Sequence[str]
   imports: str
   module: str
 
@@ -465,13 +465,7 @@ class TestNinjaPathEscape(TestBase):
 
   def test_escape(self):
     escaped = pytype_runner.escape_ninja_path('C:/xyz')
-    if sys.platform == 'win32':
-      self.assertEqual(escaped, 'C$:/xyz')
-    else:
-      self.assertEqual(escaped, 'C:/xyz')
-
-  def test_already_escaped(self):
-    self.assertEqual(pytype_runner.escape_ninja_path('C$:/xyz'), 'C$:/xyz')
+    self.assertEqual(escaped, 'C$:/xyz')
 
 
 class TestNinjaPreamble(TestBase):
@@ -592,12 +586,17 @@ class TestNinjaBody(TestBase):
     self.conf = self.parser.config_from_defaults()
 
   def assertBuildStatementMatches(self, build_statement, expected):
+    if expected.deps:
+      deps = ' | ' + ' '.join(pytype_runner.escape_ninja_path(d)
+                              for d in expected.deps)
+    else:
+      deps = ''
     self.assertEqual(
         build_statement[0], 'build {output}: {action} {input}{deps}'.format(
             output=pytype_runner.escape_ninja_path(expected.output),
             action=expected.action,
             input=pytype_runner.escape_ninja_path(expected.input),
-            deps=pytype_runner.escape_ninja_path(expected.deps)))
+            deps=deps))
     self.assertEqual(
         set(build_statement[1:]), {
             f'  imports = {pytype_runner.escape_ninja_path(expected.imports)}',
@@ -620,7 +619,7 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'bar.pyi'),
             action=Action.INFER,
             input='bar.py',
-            deps='',
+            deps=[],
             imports=path_utils.join(runner.imports_dir, 'bar.imports'),
             module='bar'))
     self.assertBuildStatementMatches(
@@ -629,7 +628,7 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'foo.pyi'),
             action=Action.CHECK,
             input='foo.py',
-            deps=' | ' + path_utils.join(runner.pyi_dir, 'bar.pyi'),
+            deps=[path_utils.join(runner.pyi_dir, 'bar.pyi')],
             imports=path_utils.join(runner.imports_dir, 'foo.imports'),
             module='foo'))
 
@@ -651,7 +650,7 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'foo.pyi'),
             action=Action.CHECK,
             input='foo.py',
-            deps='',
+            deps=[],
             imports=path_utils.join(runner.imports_dir, 'foo.imports'),
             module='foo'))
     short_bar_path, bar_path = imports_info.split(' ')
@@ -675,7 +674,7 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'bar.pyi-1'),
             action=Action.INFER,
             input='bar.py',
-            deps='',
+            deps=[],
             imports=path_utils.join(runner.imports_dir, 'bar.imports-1'),
             module='bar'))
     self.assertBuildStatementMatches(
@@ -684,7 +683,7 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'foo.pyi-1'),
             action=Action.INFER,
             input='foo.py',
-            deps='',
+            deps=[],
             imports=path_utils.join(runner.imports_dir, 'foo.imports-1'),
             module='foo'))
     self.assertBuildStatementMatches(
@@ -693,9 +692,9 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'bar.pyi'),
             action=Action.INFER,
             input='bar.py',
-            deps=' | {} {}'.format(
+            deps=[
                 path_utils.join(runner.pyi_dir, 'bar.pyi-1'),
-                path_utils.join(runner.pyi_dir, 'foo.pyi-1')),
+                path_utils.join(runner.pyi_dir, 'foo.pyi-1')],
             imports=path_utils.join(runner.imports_dir, 'bar.imports'),
             module='bar'))
     self.assertBuildStatementMatches(
@@ -704,9 +703,9 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'foo.pyi'),
             action=Action.CHECK,
             input='foo.py',
-            deps=' | {} {}'.format(
+            deps=[
                 path_utils.join(runner.pyi_dir, 'bar.pyi'),
-                path_utils.join(runner.pyi_dir, 'foo.pyi-1')),
+                path_utils.join(runner.pyi_dir, 'foo.pyi-1')],
             imports=path_utils.join(runner.imports_dir, 'foo.imports'),
             module='foo'))
 
@@ -727,7 +726,7 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'foo.pyi-1'),
             action=Action.INFER,
             input='foo.py',
-            deps='',
+            deps=[],
             imports=path_utils.join(runner.imports_dir, 'foo.imports-1'),
             module='foo'))
     self.assertBuildStatementMatches(
@@ -736,7 +735,7 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'bar.pyi-1'),
             action=Action.INFER,
             input='bar.py',
-            deps='',
+            deps=[],
             imports=path_utils.join(runner.imports_dir, 'bar.imports-1'),
             module='bar'))
     self.assertBuildStatementMatches(
@@ -745,9 +744,9 @@ class TestNinjaBody(TestBase):
             output=path_utils.join(runner.pyi_dir, 'foo.pyi'),
             action=Action.CHECK,
             input='foo.py',
-            deps=' | {} {}'.format(
+            deps=[
                 path_utils.join(runner.pyi_dir, 'foo.pyi-1'),
-                path_utils.join(runner.pyi_dir, 'bar.pyi-1')),
+                path_utils.join(runner.pyi_dir, 'bar.pyi-1')],
             imports=path_utils.join(runner.imports_dir, 'foo.imports'),
             module='foo'))
 
