@@ -2,6 +2,7 @@
 
 import ast
 import collections
+import sys
 import textwrap
 from pytype import config
 from pytype.pytd import pytd
@@ -10,6 +11,13 @@ from pytype.tests import test_utils
 from pytype.tools.traces import traces
 import unittest
 
+_PYVER = sys.version_info[:2]
+
+_BINMOD_OP = "BINARY_OP" if _PYVER >= (3, 11) else "BINARY_MODULO"
+_CALLFUNC_OP = "CALL" if _PYVER >= (3, 11) else "CALL_FUNCTION"
+_CALLMETH_OP = "CALL" if _PYVER >= (3, 11) else "CALL_METHOD"
+_FORMAT_OP = "FORMAT_VALUE" if _PYVER >= (3, 11) else "BINARY_MODULO"
+
 
 class _NotImplementedVisitor(traces.MatchAstVisitor):
 
@@ -36,8 +44,10 @@ class TraceTest(unittest.TestCase):
 
   def test_traces(self):
     src = traces.trace("")
-    trace, = src.traces[1]
-    self.assertEqual(trace.op, "LOAD_CONST")
+    trace, = src.traces[0 if _PYVER >= (3, 11) else 1]
+    self.assertEqual(
+        trace.op, "RETURN_CONST" if _PYVER >= (3, 12) else "LOAD_CONST"
+    )
     self.assertIsNone(trace.symbol)
     pyval, = trace.types
     self.assertEqual(pyval.name, "builtins.NoneType")
@@ -116,14 +126,14 @@ class MatchAstVisitorTest(MatchAstTestCase):
   def test_import(self):
     matches = self._get_traces("import os, sys as tzt", ast.Import)
     self.assertTracesEqual(matches, [
-        ((1, 7), "IMPORT_NAME", "os", ("import os",)),
-        ((1, 18), "STORE_NAME", "tzt", ("import sys",))])
+        ((1, 7), "IMPORT_NAME", "os", ("module",)),
+        ((1, 18), "STORE_NAME", "tzt", ("module",))])
 
   def test_import_from(self):
     matches = self._get_traces(
         "from os import path as p, environ", ast.ImportFrom)
     self.assertTracesEqual(matches, [
-        ((1, 23), "STORE_NAME", "p", ("import os.path",)),
+        ((1, 23), "STORE_NAME", "p", ("module",)),
         ((1, 26), "STORE_NAME", "environ", ("os._Environ[str]",))])
 
 
@@ -199,7 +209,7 @@ class MatchCallTest(MatchAstTestCase):
       f(42)
     """, ast.Call)
     self.assertTracesEqual(matches, [
-        ((3, 0), "CALL_FUNCTION", "f", ("Callable[[Any], Any]", "float"))])
+        ((3, 0), _CALLFUNC_OP, "f", ("Callable[[Any], Any]", "float"))])
 
   def test_chain(self):
     matches = self._get_traces("""
@@ -209,8 +219,8 @@ class MatchCallTest(MatchAstTestCase):
       Foo().f(42)
     """, ast.Call)
     self.assertTracesEqual(matches, [
-        ((4, 0), "CALL_FUNCTION", "Foo", ("Type[Foo]", "Foo")),
-        ((4, 0), "CALL_METHOD", "f", ("Callable[[Any], Any]", "int"))])
+        ((4, 0), _CALLFUNC_OP, "Foo", ("Type[Foo]", "Foo")),
+        ((4, 0), _CALLMETH_OP, "f", ("Callable[[Any], Any]", "int"))])
 
   def test_multiple_bindings(self):
     matches = self._get_traces("""
@@ -226,7 +236,7 @@ class MatchCallTest(MatchAstTestCase):
       f(42)
     """, ast.Call)
     self.assertTracesEqual(matches, [
-        ((10, 0), "CALL_FUNCTION", "f",
+        ((10, 0), _CALLFUNC_OP, "f",
          ("Callable[[Any], Any]", "Union[int, float]"))])
 
   def test_bad_call(self):
@@ -235,12 +245,12 @@ class MatchCallTest(MatchAstTestCase):
       f(42)
     """, ast.Call)
     self.assertTracesEqual(
-        matches, [((2, 0), "CALL_FUNCTION", "f", ("Callable[[], Any]", "Any"))])
+        matches, [((2, 0), _CALLFUNC_OP, "f", ("Callable[[], Any]", "Any"))])
 
   def test_literal(self):
     matches = self._get_traces("''.upper()", ast.Call)
     self.assertTracesEqual(matches, [
-        ((1, 0), "CALL_METHOD", "upper", ("Callable[[], str]", "str"))])
+        ((1, 0), _CALLMETH_OP, "upper", ("Callable[[], str]", "str"))])
 
   def test_lookahead(self):
     matches = self._get_traces("""
@@ -253,84 +263,34 @@ class MatchCallTest(MatchAstTestCase):
       )
     """, ast.Call)
     self.assertTracesEqual(matches, [
-        ((3, 0), "CALL_FUNCTION", "f",
+        ((3, 0), _CALLFUNC_OP, "f",
          ("Callable[[Any, Any, Any], Any]", "int"))])
 
 
 class MatchConstantTest(MatchAstTestCase):
 
-  @test_utils.skipFromPy((3, 8),
-                         reason="numbers are ast.Constant in 3.8+")
   def test_num(self):
-    matches = self._get_traces("v = 42", ast.Num)
-    self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", 42, ("int",))])
-
-  @test_utils.skipFromPy((3, 8),
-                         reason="strings are ast.Constant in 3.8+")
-  def test_str(self):
-    matches = self._get_traces("v = 'hello'", ast.Str)
-    self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", "hello", ("str",))])
-
-  @test_utils.skipFromPy((3, 8),
-                         reason="unicode strings are ast.Constant in 3.8+")
-  def test_unicode(self):
-    matches = self._get_traces("v = u'hello'", ast.Str)
-    self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", "hello", ("str",))])
-
-  @test_utils.skipFromPy((3, 8),
-                         reason="byte strings are ast.Constant in 3.8+")
-  def test_bytes(self):
-    matches = self._get_traces("v = b'hello'", ast.Bytes)
-    self.assertTracesEqual(
-        matches, [((1, 4), "LOAD_CONST", b"hello", ("bytes",))])
-
-  @test_utils.skipFromPy((3, 8),
-                         reason="bools are ast.Constant in 3.8+")
-  def test_bool(self):
-    matches = self._get_traces("v = True", ast.NameConstant)
-    self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", True, ("bool",))])
-
-  @test_utils.skipFromPy((3, 8),
-                         reason="ellipsis are ast.Constant in 3.8+")
-  def test_ellipsis(self):
-    matches = self._get_traces("v = ...", ast.Ellipsis)
-    self.assertTracesEqual(
-        matches, [((1, 4), "LOAD_CONST", Ellipsis, ("ellipsis",))])
-
-  @test_utils.skipBeforePy((3, 8),
-                           reason="numbers are ast.Constant in 3.8+")
-  def test_num_38(self):
     matches = self._get_traces("v = 42", ast.Constant)
     self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", 42, ("int",))])
 
-  @test_utils.skipBeforePy((3, 8),
-                           reason="strings are ast.Constant in 3.8+")
-  def test_str_38(self):
+  def test_str(self):
     matches = self._get_traces("v = 'hello'", ast.Constant)
     self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", "hello", ("str",))])
 
-  @test_utils.skipBeforePy((3, 8),
-                           reason="unicode strings are ast.Constant in 3.8+")
-  def test_unicode_38(self):
+  def test_unicode(self):
     matches = self._get_traces("v = u'hello'", ast.Constant)
     self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", "hello", ("str",))])
 
-  @test_utils.skipBeforePy((3, 8),
-                           reason="byte strings are ast.Constant in 3.8+")
-  def test_bytes_38(self):
+  def test_bytes(self):
     matches = self._get_traces("v = b'hello'", ast.Constant)
     self.assertTracesEqual(
         matches, [((1, 4), "LOAD_CONST", b"hello", ("bytes",))])
 
-  @test_utils.skipBeforePy((3, 8),
-                           reason="bools are ast.Constant in 3.8+")
-  def test_bool_38(self):
+  def test_bool(self):
     matches = self._get_traces("v = True", ast.Constant)
     self.assertTracesEqual(matches, [((1, 4), "LOAD_CONST", True, ("bool",))])
 
-  @test_utils.skipBeforePy((3, 8),
-                           reason="ellipsis are ast.Constant in 3.8+")
-  def test_ellipsis_38(self):
+  def test_ellipsis(self):
     matches = self._get_traces("v = ...", ast.Constant)
     self.assertTracesEqual(
         matches, [((1, 4), "LOAD_CONST", Ellipsis, ("ellipsis",))])
@@ -351,8 +311,16 @@ class MatchSubscriptTest(MatchAstTestCase):
       v = "hello"
       print(v[:-1])
     """, ast.Subscript)
-    self.assertTracesEqual(
-        matches, [((2, 6), "BINARY_SUBSCR", "__getitem__", ("str",))])
+    if _PYVER >= (3, 12):
+      expected = [(
+          (2, 6),
+          "BINARY_SLICE",
+          "__getitem__",
+          ("Callable[[Union[int, slice]], str]", "str"),
+      )]
+    else:
+      expected = [((2, 6), "BINARY_SUBSCR", "__getitem__", ("str",))]
+    self.assertTracesEqual(matches, expected)
 
   def test_complex_slice(self):
     matches = self._get_traces("""
@@ -370,10 +338,18 @@ class MatchBinOpTest(MatchAstTestCase):
       v = "hello %s"
       print(v % "world")
     """, ast.BinOp)
-    self.assertTracesEqual(
-        matches, [((2, 6), "BINARY_MODULO", "__mod__", ("str",))])
+    self.assertTracesEqual(matches, [((2, 6), _BINMOD_OP, "__mod__", ("str",))])
 
   def test_modulo_multiline_string(self):
+    matches = self._get_traces("""
+      ('%s'
+       '%s' %
+       ('hello',
+        'world'))
+    """, ast.BinOp)
+    self.assertTracesEqual(matches, [((1, 1), _BINMOD_OP, "__mod__", ("str",))])
+
+  def test_format_multiline_string(self):
     matches = self._get_traces("""
       ('%s'
        '%s' %
@@ -381,7 +357,7 @@ class MatchBinOpTest(MatchAstTestCase):
         __any_object__))
     """, ast.BinOp)
     self.assertTracesEqual(
-        matches, [((1, 1), "BINARY_MODULO", "__mod__", ("str",))])
+        matches, [((1, 1), _FORMAT_OP, "__mod__", ("str",))])
 
 
 class MatchLambdaTest(MatchAstTestCase):
