@@ -1,16 +1,15 @@
 """Generate cross references from a project."""
 
+import ast as astlib
 import collections
 import dataclasses
 import re
-import sys
 import textwrap
 import types
 from typing import Any, List, Optional, Type, TypeVar
 
 from pytype import analyze
 from pytype import config
-from pytype import context
 from pytype import io
 from pytype import load_pytd
 from pytype import module_utils
@@ -26,14 +25,6 @@ from pytype.tools.xref import callgraph
 from pytype.tools.xref import utils as xref_utils
 from pytype.tools.xref import node_utils
 
-# pylint: disable=g-import-not-at-top
-if sys.version_info >= (3, 8):
-  import ast as ast3
-else:
-  from typed_ast import ast3
-# pylint: enable=g-import-not-at-top
-
-
 # A mapping of offsets between a node's start position and the symbol being
 # defined. e.g. in the declaration "class X" the X is at +6 from the start.
 DEF_OFFSETS = {
@@ -67,13 +58,8 @@ def get_location(node):
 
 
 def get_end_location(node):
-  # These attributes are new in Python 3.8.
-  if sys.version_info[:2] >= (3, 8):
-    end_lineno = node.end_lineno
-    end_col_offset = node.end_col_offset
-  else:
-    end_lineno = node.lineno
-    end_col_offset = node.col_offset
+  end_lineno = node.end_lineno
+  end_col_offset = node.end_col_offset
   return source.Location(end_lineno, end_col_offset)
 
 
@@ -201,9 +187,10 @@ class DocString:
     assert isinstance(node, (ast.ClassDef, ast.FunctionDef))
     if (node.body and
         isinstance(node.body[0], ast.Expr) and
-        isinstance(node.body[0].value, ast.Str)):
+        isinstance(node.body[0].value, ast.Constant) and
+        isinstance(node.body[0].value.value, str)):
       doc_node = node.body[0]
-      doc = doc_node.value.s
+      doc = doc_node.value.value
       length = len(doc)  # we want to preserve the byte length
       # strip indentation from multiline docstrings
       if "\n" in doc:
@@ -491,7 +478,7 @@ class ScopedVisitor(ast_visitor.BaseVisitor):
     elif c == self._ast.Module:
       return self.module_name
     else:
-      raise Exception(f"Unexpected scope: {node!r}")
+      raise Exception(f"Unexpected scope: {node!r}")  # pylint: disable=broad-exception-raised
 
   def iprint(self, x):
     """Print messages indented by scope level, for debugging."""
@@ -718,6 +705,7 @@ class IndexVisitor(ScopedVisitor, traces.MatchAstVisitor):
     ])
     # pytype sometimes analyses this twice, leading to duplicate opcode
     # traces. We only want the first two in the list.
+    d = data = None
     if (len(ops) >= 2 and
         ops[0][0] == "LOAD_BUILD_CLASS" and
         ops[1][0] in (
@@ -892,6 +880,7 @@ class IndexVisitor(ScopedVisitor, traces.MatchAstVisitor):
         c, _ = m.span()
         loc = source.Location(loc.line, c + 1)
 
+      imported = None
       try:
         [imported] = _unwrap(data)
       except (TypeError, ValueError):
@@ -1246,7 +1235,7 @@ class Indexer:
   def get_pytd(self, datum):
     if not datum:
       return pytd.AnythingType()
-    t = pytd_utils.JoinTypes(v.to_type() for v in datum).Visit(
+    t = pytd_utils.JoinTypes(v.to_pytd_type() for v in datum).Visit(
         visitors.RemoveUnknownClasses())
     return self.loader.resolve_pytd(t, self.pytd_module)
 
@@ -1322,39 +1311,32 @@ def process_file(options, source_text=None, generate_callgraphs=False,
   with config.verbosity_from(options):
     loader = load_pytd.create_loader(options)
     src = source_text or io.read_source_file(options.input)
-    ctx = context.Context(
-        options=options,
-        generate_unknowns=options.protocols,
-        store_all_calls=True,
-        loader=loader)
     with io.wrap_pytype_exceptions(PytypeError, filename=options.input):
       ret = analyze.infer_types(
           src=src,
-          filename=options.input,
           options=options,
-          loader=loader,
-          ctx=ctx)
+          loader=loader)
       pytd_module = ret.ast
   # pylint: disable=unexpected-keyword-arg
-  ast_root_node = ast3.parse(src, options.input,
-                             feature_version=options.python_version[1])
+  ast_root_node = astlib.parse(src, options.input,
+                               feature_version=options.python_version[1])
   # pylint: enable=unexpected-keyword-arg
 
   # TODO(mdemello): Get from args
   module_name = "module"
   src_code = source.Code(
-      src, ctx.vm.opcode_traces, VmTrace, filename=options.input)
+      src, ret.context.vm.opcode_traces, VmTrace, filename=options.input)
   ix = Indexer(
-      ast=ast3,
+      ast=astlib,
       src=src_code,
-      loader=ctx.loader,
+      loader=loader,
       module_name=module_name,
       pytd_module=pytd_module)
   ix.index(ast_root_node)
   ix.finalize()
 
   # Make the vm available via indexer.vm for post-finalize() functions.
-  ix.vm = ctx.vm
+  ix.vm = ret.context.vm
 
   # Use the indexer as a single object to hold data for calling processes.
   if generate_callgraphs:
