@@ -8,7 +8,6 @@ import re
 from typing import Any, Dict, Sequence, Tuple, Union
 
 import attrs
-
 from pytype import state as frame_state
 from pytype import vm
 from pytype.abstract import abstract
@@ -30,8 +29,14 @@ log = logging.getLogger(__name__)
 # have names like "<listcomp>" and "<genexpr>".
 _SKIP_FUNCTION_RE = re.compile(r"<(?!lambda)\w+>$")
 
-_InstanceCacheType = Dict[abstract.InterpreterClass,
-                          Dict[Any, Union["_InitClassState", cfg.Variable]]]
+_InstanceCacheType = Dict[
+    abstract.InterpreterClass, Dict[Any, Union["_InitClassState", cfg.Variable]]
+]
+
+_METHOD_TYPES = abstract.INTERPRETER_FUNCTION_TYPES + (
+    special_builtins.StaticMethodInstance,
+    special_builtins.ClassMethodInstance,
+)
 
 
 @dataclasses.dataclass(eq=True, frozen=True)
@@ -72,11 +77,12 @@ class CallTracer(vm.VirtualMachine):
   def create_varargs(self, node):
     value = abstract.Instance(self.ctx.convert.tuple_type, self.ctx)
     value.merge_instance_type_parameter(
-        node, abstract_utils.T, self.ctx.convert.create_new_unknown(node))
+        node, abstract_utils.T, self.ctx.convert.create_new_unknown(node)
+    )
     return value.to_variable(node)
 
   def create_kwargs(self, node):
-    key_type = self.ctx.convert.primitive_class_instances[str].to_variable(node)
+    key_type = self.ctx.convert.primitive_instances[str].to_variable(node)
     value_type = self.ctx.convert.create_new_unknown(node)
     kwargs = abstract.Instance(self.ctx.convert.dict_type, self.ctx)
     kwargs.merge_instance_type_parameter(node, abstract_utils.K, key_type)
@@ -116,13 +122,16 @@ class CallTracer(vm.VirtualMachine):
         kws[key] = method.kw_defaults[key]
       else:
         kws[key] = self.ctx.convert.create_new_unknown(
-            node, force=not use_defaults)
+            node, force=not use_defaults
+        )
     starargs = self.create_varargs(node) if method.has_varargs() else None
     starstarargs = self.create_kwargs(node) if method.has_kwargs() else None
-    return node, function.Args(posargs=tuple(args),
-                               namedargs=kws,
-                               starargs=starargs,
-                               starstarargs=starstarargs)
+    return node, function.Args(
+        posargs=tuple(args),
+        namedargs=kws,
+        starargs=starargs,
+        starstarargs=starstarargs,
+    )
 
   def call_function_with_args(self, node, val, args):
     """Call a function.
@@ -138,11 +147,13 @@ class CallTracer(vm.VirtualMachine):
     assert isinstance(val.data, abstract.INTERPRETER_FUNCTION_TYPES)
     with val.data.record_calls():
       new_node, ret = self._call_function_in_frame(  # pylint: disable=no-value-for-parameter
-          node, val, *attrs.astuple(args, recurse=False))
+          node, val, *attrs.astuple(args, recurse=False)
+      )
     return new_node, ret
 
-  def _call_function_in_frame(self, node, val, args, kwargs,
-                              starargs, starstarargs):
+  def _call_function_in_frame(
+      self, node, val, args, kwargs, starargs, starstarargs
+  ):
     # Try to get the function opcode with position information to construct the
     # frame, so that we have the data for any error messages raised before we
     # get to func.call().
@@ -156,27 +167,39 @@ class CallTracer(vm.VirtualMachine):
       assert isinstance(fn, abstract.InterpreterFunction)
       opcode = fn.def_opcode
       f_globals = fn.f_globals
+    log.info("Analyzing function: %r", fn.name)
+    state = frame_state.FrameState.init(node, self.ctx)
     frame = frame_state.SimpleFrame(
-        node=node, opcode=opcode, f_globals=f_globals)
+        node=node, opcode=opcode, f_globals=f_globals
+    )
     # We only want this frame to show up in errors if it's at the top of the
     # stack (i.e. we haven't started analysing the actual function yet)
     frame.skip_in_tracebacks = True
     self.push_frame(frame)
-    log.info("Analyzing function: %r", fn.name)
-    state = frame_state.FrameState.init(node, self.ctx)
-    state, ret = self.call_function_with_state(
-        state, val.AssignToNewVariable(node), args, kwargs, starargs,
-        starstarargs)
-    self.pop_frame(frame)
+    try:
+      state, ret = self.call_function_with_state(
+          state,
+          val.AssignToNewVariable(node),
+          args,
+          kwargs,
+          starargs,
+          starstarargs,
+      )
+    finally:
+      self.pop_frame(frame)
     return state.node, ret
 
   def _maybe_fix_classmethod_cls_arg(self, node, cls, func, args):
     sig = func.signature
-    if (args.posargs and sig.param_names and
-        (sig.param_names[0] not in sig.annotations)):
+    if (
+        args.posargs
+        and sig.param_names
+        and (sig.param_names[0] not in sig.annotations)
+    ):
       # fix "cls" parameter
       return args.replace(
-          posargs=(cls.AssignToNewVariable(node),) + args.posargs[1:])
+          posargs=(cls.AssignToNewVariable(node),) + args.posargs[1:]
+      )
     else:
       return args
 
@@ -185,9 +208,11 @@ class CallTracer(vm.VirtualMachine):
     fname = val.data.name
     if isinstance(method, abstract.INTERPRETER_FUNCTION_TYPES):
       self._analyzed_functions.add(method.get_first_opcode())
-      if (not self.ctx.options.analyze_annotated and
-          (method.signature.has_return_annotation or method.has_overloads) and
-          fname.rsplit(".", 1)[-1] not in self._CONSTRUCTORS):
+      if (
+          not self.ctx.options.analyze_annotated
+          and (method.signature.has_return_annotation or method.has_overloads)
+          and fname.rsplit(".", 1)[-1] not in self._CONSTRUCTORS
+      ):
         log.info("%r has annotations, not analyzing further.", fname)
       else:
         for f in method.iter_signature_functions():
@@ -224,7 +249,8 @@ class CallTracer(vm.VirtualMachine):
             node1,
             funcb.AssignToNewVariable(),
             args,
-            fallback_to_unsolvable=False)
+            fallback_to_unsolvable=False,
+        )
         nodes.append(node2)
         rets.append(ret)
 
@@ -240,8 +266,9 @@ class CallTracer(vm.VirtualMachine):
 
   def analyze_method_var(self, node0, name, var, cls):
     full_name = f"{cls.data.full_name}.{name}"
-    if any(isinstance(v, abstract.INTERPRETER_FUNCTION_TYPES)
-           for v in var.data):
+    if any(
+        isinstance(v, abstract.INTERPRETER_FUNCTION_TYPES) for v in var.data
+    ):
       log.info("Analyzing method: %r", full_name)
       node1 = self.ctx.connect_new_cfg_node(node0, f"Method:{full_name}")
     else:
@@ -251,28 +278,31 @@ class CallTracer(vm.VirtualMachine):
       node2.ConnectTo(node0)
     return node0
 
-  def bind_method(self, node, methodvar, instance_var):
+  def _bind_method(self, node, methodvar, instance_var):
     bound = self.ctx.program.NewVariable()
     for m in methodvar.Data(node):
       if isinstance(m, special_builtins.ClassMethodInstance):
         m = m.func.data[0]
         is_cls = True
       else:
-        is_cls = (isinstance(m, abstract.InterpreterFunction) and
-                  m.is_classmethod)
+        is_cls = (
+            isinstance(m, abstract.InterpreterFunction) and m.is_classmethod
+        )
       bound.AddBinding(m.property_get(instance_var, is_cls), [], node)
     return bound
 
-  def _maybe_instantiate_binding_directly(self, node0, cls, container,
-                                          instantiate_directly):
+  def _maybe_instantiate_binding_directly(
+      self, node0, cls, container, instantiate_directly
+  ):
     node1, new = cls.data.get_own_new(node0, cls)
     if not new:
       instantiate_directly = True
     elif not instantiate_directly:
       # This assumes that any inherited __new__ method defined in a pyi file
       # returns an instance of the current class.
-      instantiate_directly = any(not isinstance(f, abstract.InterpreterFunction)
-                                 for f in new.data)
+      instantiate_directly = any(
+          not isinstance(f, abstract.InterpreterFunction) for f in new.data
+      )
     if instantiate_directly:
       instance = cls.data.instantiate(node0, container=container)
     else:
@@ -282,7 +312,8 @@ class CallTracer(vm.VirtualMachine):
   def _instantiate_binding(self, node0, cls, container, instantiate_directly):
     """Instantiate a class binding."""
     node1, new, maybe_instance = self._maybe_instantiate_binding_directly(
-        node0, cls, container, instantiate_directly)
+        node0, cls, container, instantiate_directly
+    )
     if maybe_instance:
       return node0, maybe_instance
     instance = self.ctx.program.NewVariable()
@@ -292,7 +323,8 @@ class CallTracer(vm.VirtualMachine):
       node2, args = self.create_method_arguments(node1, b.data)
       args = self._maybe_fix_classmethod_cls_arg(node0, cls, b.data, args)
       node3 = self.ctx.connect_new_cfg_node(
-          node2, f"Call:{cls.data.name}.__new__")
+          node2, f"Call:{cls.data.name}.__new__"
+      )
       node4, ret = self.call_function_with_args(node3, b, args)
       instance.PasteVariable(ret)
       nodes.append(node4)
@@ -301,9 +333,10 @@ class CallTracer(vm.VirtualMachine):
   def _instantiate_var(self, node, clsv, container, instantiate_directly):
     """Build an (dummy) instance from a class, for analyzing it."""
     n = self.ctx.program.NewVariable()
-    for cls in clsv.Bindings(node, strict=False):
+    for cls in clsv.Bindings(node):
       node, var = self._instantiate_binding(
-          node, cls, container, instantiate_directly)
+          node, cls, container, instantiate_directly
+      )
       n.PasteVariable(var)
     return node, n
 
@@ -326,7 +359,9 @@ class CallTracer(vm.VirtualMachine):
           for child in v.instance_type_parameters.values():
             values.extend(child.data)
 
-  def init_class(self, node, cls, container=None, extra_key=None):
+  def init_class_and_forward_node(
+      self, node, cls, container=None, extra_key=None
+  ):
     """Instantiate a class, and also call __init__.
 
     Calling __init__ can be expensive, so this method caches its created
@@ -357,13 +392,14 @@ class CallTracer(vm.VirtualMachine):
       # For some reason, checking only `status` leads to attributes set in
       # __init__ not being visible later, so we have to check every instance of
       # this class, ignoring `key`.
-      instantiate_directly = any(v is _InitClassState.INSTANTIATING
-                                 for v in cache.values())
+      instantiate_directly = any(
+          v is _InitClassState.INSTANTIATING for v in cache.values()
+      )
       cache[key] = _InitClassState.INSTANTIATING
       node, instance = self._instantiate_var(
-          node, clsvar, container, instantiate_directly)
-      if (instantiate_directly or
-          status is _InitClassState.INITIALIZING):
+          node, clsvar, container, instantiate_directly
+      )
+      if instantiate_directly or status is _InitClassState.INITIALIZING:
         # We've encountered a recursive pattern in such as
         #   class A:
         #     def __new__(cls) -> "A": ...
@@ -383,15 +419,44 @@ class CallTracer(vm.VirtualMachine):
       cache[key] = instance
     return node, instance
 
-  def _call_method(self, node, binding, method_name):
+  def init_class(self, node, cls, container=None, extra_key=None):
+    return self.init_class_and_forward_node(node, cls, container, extra_key)[-1]
+
+  def get_bound_method(self, node, obj, method_name, valself):
+
+    def bind(cur_node, m):
+      return self._bind_method(cur_node, m, valself.AssignToNewVariable())
+
     node, method = self.ctx.attribute_handler.get_attribute(
-        node, binding.data.cls, method_name, binding)
-    if method:
-      bound_method = self.bind_method(
-          node, method, binding.AssignToNewVariable())
-      node = self.analyze_method_var(
-          node, method_name, bound_method, binding.data.cls.to_binding(node))
-    return node
+        node, obj, method_name, valself
+    )
+    if not method:
+      return node, None
+    cls = valself.data.cls
+    bound_method = bind(node, method) if obj == cls else method
+    if not isinstance(cls, abstract.InterpreterClass) or any(
+        isinstance(m, _METHOD_TYPES) for m in bound_method.data
+    ):
+      return node, bound_method
+    # If the method is not something that pytype recognizes as a function -
+    # which can happen if the method is decorated, for example - then we look up
+    # the method before any decorators were applied and use that instead.
+    undecorated_method = cls.get_undecorated_method(method_name, node)
+    if undecorated_method:
+      return node, bind(node, undecorated_method)
+    else:
+      return node, bound_method
+
+  def _call_method(self, node, binding, method_name):
+    node, bound_method = self.get_bound_method(
+        node, binding.data.cls, method_name, binding
+    )
+    if bound_method:
+      return self.analyze_method_var(
+          node, method_name, bound_method, binding.data.cls.to_binding(node)
+      )
+    else:
+      return node
 
   def _call_init_on_binding(self, node, b):
     if isinstance(b.data, abstract.SimpleValue):
@@ -422,7 +487,7 @@ class CallTracer(vm.VirtualMachine):
     cls = val.data
     log.info("Analyzing class: %r", cls.full_name)
     self._analyzed_classes.add(cls)
-    node, instance = self.init_class(node, cls)
+    node, instance = self.init_class_and_forward_node(node, cls)
     good_instances = [b for b in instance.bindings if cls == b.data.cls]
     if not good_instances:
       # __new__ returned something that's not an instance of our class.
@@ -443,13 +508,27 @@ class CallTracer(vm.VirtualMachine):
           for m in (v.fget, v.fset, v.fdel):
             if m:
               methods.insert(0, (name, m))
-        if (not cls.is_abstract and isinstance(v, abstract.Function) and
-            v.is_abstract):
+        if (
+            not cls.is_abstract
+            and isinstance(v, abstract.Function)
+            and v.is_abstract
+        ):
           unwrapped = abstract_utils.maybe_unwrap_decorated_function(v)
           name = unwrapped.data[0].name if unwrapped else v.name
           self.ctx.errorlog.ignored_abstractmethod(
-              self.ctx.vm.simple_stack(cls.get_first_opcode()), cls.name, name)
-      b = self.bind_method(node, methodvar, instance)
+              self.ctx.vm.simple_stack(cls.get_first_opcode()), cls.name, name
+          )
+      is_method_or_nested_class = any(
+          isinstance(m, (_METHOD_TYPES, abstract.InterpreterClass))
+          for m in methodvar.data
+      )
+      if (
+          self.ctx.options.bind_decorated_methods
+          and not is_method_or_nested_class
+          and (undecorated_method := cls.get_undecorated_method(name, node))
+      ):
+        methodvar = undecorated_method
+      b = self._bind_method(node, methodvar, instance)
       node = self.analyze_method_var(node, name, b, val)
     return node
 
@@ -468,11 +547,13 @@ class CallTracer(vm.VirtualMachine):
     # two ways of recording are equivalent except for closures, which are
     # re-generated when the variables they close over change, but we don't want
     # to re-analyze them.
-    return (isinstance(data, abstract.InterpreterFunction) and
-            not data.is_overload and
-            not data.is_class_builder and
-            data.get_first_opcode() not in self._analyzed_functions and
-            not _SKIP_FUNCTION_RE.search(data.name))
+    return (
+        isinstance(data, abstract.InterpreterFunction)
+        and not data.is_overload
+        and not data.is_class_builder
+        and data.get_first_opcode() not in self._analyzed_functions
+        and not _SKIP_FUNCTION_RE.search(data.name)
+    )
 
   def analyze_toplevel(self, node, defs):
     for name, var in sorted(defs.items()):  # sort, for determinicity
@@ -480,8 +561,10 @@ class CallTracer(vm.VirtualMachine):
         for value in var.bindings:
           if isinstance(value.data, abstract.InterpreterClass):
             new_node = self.analyze_class(node, value)
-          elif (isinstance(value.data, abstract.INTERPRETER_FUNCTION_TYPES) and
-                not value.data.is_overload):
+          elif (
+              isinstance(value.data, abstract.INTERPRETER_FUNCTION_TYPES)
+              and not value.data.is_overload
+          ):
             new_node = self.analyze_function(node, value)
           else:
             continue
@@ -492,15 +575,19 @@ class CallTracer(vm.VirtualMachine):
     # be set for all functions in class.
     for c in self._interpreter_classes:
       for value in c.bindings:
-        if (isinstance(value.data, abstract.InterpreterClass) and
-            value.data not in self._analyzed_classes):
+        if (
+            isinstance(value.data, abstract.InterpreterClass)
+            and value.data not in self._analyzed_classes
+        ):
           node = self.analyze_class(node, value)
     for f in self._interpreter_functions:
       for value in f.bindings:
         if self._should_analyze_as_interpreter_function(value.data):
           node = self.analyze_function(node, value)
     for func, opcode in self.functions_type_params_check:
-      func.signature.check_type_parameter_count(self.simple_stack(opcode))
+      func.signature.check_type_parameters(
+          self.simple_stack(opcode), opcode, func.is_attribute_of_class
+      )
     return node
 
   def analyze(self, node, defs, maximum_depth):
@@ -524,8 +611,9 @@ class CallTracer(vm.VirtualMachine):
       namedargs: The keyword arguments, a dict mapping str to cfg.Variable.
       result: A Variable of the possible result values.
     """
-    log.debug("Logging call to %r with %d args, return %r",
-              func, len(posargs), result)
+    log.debug(
+        "Logging call to %r with %d args, return %r", func, len(posargs), result
+    )
     args = tuple(posargs)
     kwargs = tuple((namedargs or {}).items())
     record = _CallRecord(node, func, sigs, args, kwargs, result)
@@ -549,9 +637,11 @@ class CallTracer(vm.VirtualMachine):
     return classes
 
   def _skip_definition_export(self, name, var):
-    return (name in abstract_utils.TOP_LEVEL_IGNORE or
-            self._is_typing_member(name, var) or
-            self._is_future_feature(name, var))
+    return (
+        name in abstract_utils.TOP_LEVEL_IGNORE
+        or self._is_typing_member(name, var)
+        or self._is_future_feature(var)
+    )
 
   def pytd_for_types(self, defs):
     # If a variable is annotated, we'll always output that type.
@@ -559,39 +649,53 @@ class CallTracer(vm.VirtualMachine):
     data = []
     annots = abstract_utils.get_annotations_dict(defs)
     for name, t in self.ctx.pytd_convert.annotations_to_instance_types(
-        self.ctx.exitpoint, annots):
+        self.ctx.exitpoint, annots
+    ):
       annotated_names.add(name)
       data.append(pytd.Constant(name, t))
     for name, var in defs.items():
       if name in annotated_names or self._skip_definition_export(name, var):
         continue
       log.info("Generating pytd type for top-level definition: %r", name)
-      if any(v == self.ctx.convert.unsolvable
-             for v in var.Data(self.ctx.exitpoint)):
+      if any(
+          v == self.ctx.convert.unsolvable for v in var.Data(self.ctx.exitpoint)
+      ):
         options = [self.ctx.convert.unsolvable]
       else:
-        options = var.FilteredData(self.ctx.exitpoint, strict=False)
-      if (len(options) > 1 and
-          not all(isinstance(o, abstract.FUNCTION_TYPES) for o in options)):
+        all_options = var.FilteredData(self.ctx.exitpoint, strict=False)
+        options = [
+            o for o in all_options if not isinstance(o, abstract.Deleted)
+        ]
+        if all_options and not options:
+          # All bindings are deleted. Don't emit anything.
+          continue
+      if len(options) > 1 and not all(
+          isinstance(o, abstract.FUNCTION_TYPES) for o in options
+      ):
         if all(isinstance(o, abstract.TypeParameter) for o in options):
           pytd_def = pytd_utils.JoinTypes(
-              t.to_pytd_def(self.ctx.exitpoint, name) for t in options)
+              t.to_pytd_def(self.ctx.exitpoint, name) for t in options
+          )
           if isinstance(pytd_def, pytd.TypeParameter):
             data.append(pytd_def)
           else:
             # We have multiple definitions for the same TypeVar name. There's no
             # good way to handle this.
             data.append(pytd.Constant(name, pytd.AnythingType()))
-        elif all(isinstance(o, (abstract.ParameterizedClass, abstract.Union))
-                 for o in options):  # type alias
+        elif all(
+            isinstance(o, (abstract.ParameterizedClass, abstract.Union))
+            for o in options
+        ):  # type alias
           pytd_def = pytd_utils.JoinTypes(
-              t.to_pytd_def(self.ctx.exitpoint, name).type for t in options)
+              t.to_pytd_def(self.ctx.exitpoint, name).type for t in options
+          )
           data.append(pytd.Alias(name, pytd_def))
         else:
           # It's ambiguous whether this is a type, a function or something
           # else, so encode it as a constant.
           combined_types = pytd_utils.JoinTypes(
-              t.to_type(self.ctx.exitpoint) for t in options)
+              t.to_pytd_type(self.ctx.exitpoint) for t in options
+          )
           data.append(pytd.Constant(name, combined_types))
       elif options:
         for option in options:
@@ -604,12 +708,12 @@ class CallTracer(vm.VirtualMachine):
               d = option.to_pytd_def(self.ctx.exitpoint, name)
           except NotImplementedError:
             with self.ctx.pytd_convert.optimize_literals():
-              d = option.to_type(self.ctx.exitpoint)  # Type only
+              d = option.to_pytd_type(self.ctx.exitpoint)  # Type only
             if isinstance(d, pytd.NothingType):
               if isinstance(option, abstract.Empty):
                 d = pytd.AnythingType()
               else:
-                assert isinstance(option, typing_overlay.NoReturn)
+                assert isinstance(option, typing_overlay.Never)
           if isinstance(d, pytd.Type) and not isinstance(d, pytd.TypeParameter):
             data.append(pytd.Constant(name, d))
           else:
@@ -623,40 +727,55 @@ class CallTracer(vm.VirtualMachine):
   def _call_traces_to_function(call_traces, name_transform=lambda x: x):
     funcs = collections.defaultdict(pytd_utils.OrderedSet)
 
-    def to_type(node, arg):
-      return pytd_utils.JoinTypes(a.to_type(node) for a in arg.data)
+    def to_pytd_type(node, arg):
+      return pytd_utils.JoinTypes(a.to_pytd_type(node) for a in arg.data)
 
     for ct in call_traces:
-      log.info("Generating pytd function for call trace: %r",
-               ct.function.data.name)
+      log.info(
+          "Generating pytd function for call trace: %r", ct.function.data.name
+      )
       # The lengths may be different in the presence of optional and kw args.
-      arg_names = max((sig.get_positional_names() for sig in ct.signatures),
-                      key=len)
+      arg_names = max(
+          (sig.get_positional_names() for sig in ct.signatures), key=len
+      )
       for i in range(len(arg_names)):
         if not isinstance(ct.function.data, abstract.BoundFunction) or i > 0:
           arg_names[i] = function.argname(i)
       arg_types = []
       for arg in ct.positional_arguments:
-        arg_types.append(to_type(ct.node, arg))
+        arg_types.append(to_pytd_type(ct.node, arg))
       kw_types = []
       for name, arg in ct.keyword_arguments:
-        kw_types.append((name, to_type(ct.node, arg)))
-      ret = pytd_utils.JoinTypes(t.to_type(ct.node)
-                                 for t in ct.return_value.data)
+        kw_types.append((name, to_pytd_type(ct.node, arg)))
+      ret = pytd_utils.JoinTypes(
+          t.to_pytd_type(ct.node) for t in ct.return_value.data
+      )
       starargs = None
       starstarargs = None
-      funcs[ct.function.data.name].add(pytd.Signature(
-          tuple(pytd.Parameter(n, t, pytd.ParameterKind.REGULAR, False, None)
-                for n, t in zip(arg_names, arg_types)) +
-          tuple(pytd.Parameter(n, t,
-                               pytd.ParameterKind.REGULAR, False, None)
-                for n, t in kw_types),
-          starargs, starstarargs,
-          ret, exceptions=(), template=()))
+      funcs[ct.function.data.name].add(
+          pytd.Signature(
+              tuple(
+                  pytd.Parameter(n, t, pytd.ParameterKind.REGULAR, False, None)
+                  for n, t in zip(arg_names, arg_types)
+              )
+              + tuple(
+                  pytd.Parameter(n, t, pytd.ParameterKind.REGULAR, False, None)
+                  for n, t in kw_types
+              ),
+              starargs,
+              starstarargs,
+              ret,
+              exceptions=(),
+              template=(),
+          )
+      )
     functions = []
     for name, signatures in funcs.items():
-      functions.append(pytd.Function(name_transform(name), tuple(signatures),
-                                     pytd.MethodKind.METHOD))
+      functions.append(
+          pytd.Function(
+              name_transform(name), tuple(signatures), pytd.MethodKind.METHOD
+          )
+      )
     return functions
 
   def _is_typing_member(self, name, var):
@@ -670,7 +789,7 @@ class CallTracer(vm.VirtualMachine):
           return True
     return False
 
-  def _is_future_feature(self, name, var):
+  def _is_future_feature(self, var):
     for v in var.data:
       if isinstance(v, abstract.Instance) and v.cls.module == "__future__":
         return True
@@ -699,44 +818,52 @@ class CallTracer(vm.VirtualMachine):
     for cls, call_records in class_to_records.items():
       full_name = cls.module + "." + cls.name if cls.module else cls.name
       log.info("Generating pytd class for call trace: %r", full_name)
-      classes.append(pytd.Class(
-          name=escape.pack_partial(full_name),
-          metaclass=None,
-          bases=(pytd.NamedType("builtins.object"),),  # not used in solver
-          methods=tuple(self._call_traces_to_function(call_records)),
-          constants=(),
-          classes=(),
-          decorators=(),
-          slots=None,
-          template=(),
-      ))
+      classes.append(
+          pytd.Class(
+              name=escape.pack_partial(full_name),
+              keywords=(),
+              bases=(pytd.NamedType("builtins.object"),),  # not used in solver
+              methods=tuple(self._call_traces_to_function(call_records)),
+              constants=(),
+              classes=(),
+              decorators=(),
+              slots=None,
+              template=(),
+          )
+      )
     return classes
 
   def compute_types(self, defs):
-    classes = (tuple(self.pytd_classes_for_unknowns()) +
-               tuple(self.pytd_classes_for_call_traces()))
+    classes = tuple(self.pytd_classes_for_unknowns()) + tuple(
+        self.pytd_classes_for_call_traces()
+    )
     functions = tuple(self.pytd_functions_for_call_traces())
     aliases = ()  # aliases are instead recorded as constants
     ty = pytd_utils.Concat(
         self.pytd_for_types(defs),
-        pytd_utils.CreateModule("unknowns", classes=classes,
-                                functions=functions, aliases=aliases))
+        pytd_utils.CreateModule(
+            "unknowns", classes=classes, functions=functions, aliases=aliases
+        ),
+    )
     ty = ty.Visit(optimize.CombineReturnsAndExceptions())
     ty = ty.Visit(optimize.PullInMethodClasses())
     ty = ty.Visit(
-        visitors.DefaceUnresolved([ty, self.ctx.loader.concat_all()],
-                                  escape.UNKNOWN))
+        visitors.DefaceUnresolved(
+            [ty, self.ctx.loader.concat_all()], escape.UNKNOWN
+        )
+    )
     return ty.Visit(visitors.AdjustTypeParameters())
 
   def _check_return(self, node, actual, formal):
     if not self.ctx.options.report_errors:
       return True
-    if formal.full_name == "typing.TypeGuard":
+    if formal.full_name in abstract_utils.TYPE_GUARDS:
       expected = self.ctx.convert.bool_type
     else:
       expected = formal
     match_result = self.ctx.matcher(node).compute_one_match(actual, expected)
     if not match_result.success:
       self.ctx.errorlog.bad_return_type(
-          self.frames, node, match_result.bad_matches)
+          self.frames, node, match_result.bad_matches
+      )
     return match_result.success
